#!/bin/bash

RED='[0;31m'
GREEN='[0;32m'
NORM='[0;00m'

##########################################################################
#               N O T   F O R   P U B L I C   U S E
##########################################################################
#
#  This bash script is used by the Virgil developers to commit changes
#  to CVS and is not intended to be used by users. It checks multiple
#  correctness criteria, runs the tests, and increments the build number.
#
##########################################################################

if [ "$1" = "" ]; then
    echo "${RED}Usage${NORM}: ${GREEN}commit.bash <log message>${NORM}"
    exit
fi

if [ "$2" = "release" ]; then
    echo "  -> ${GREEN}Release${NORM} commit selected."
    RELEASE=1
else
    RELEASE=0
fi

TEST_FILES="`find test -name '*.java'` `find test -name '*.tst'`"
JAVA_FILES=`find src -name '*.java'`
JJ_FILES=`find src -name '*.jj'`

MODULES='cck vpc'
ROOTPATH=`pwd`
ARCHPATH='/project/www/html/virgil/archives/vpc'
JAVA_vpc='java5'
JAR_NEEDED=0

report() {
    if [ "$3" = "" ]; then
	echo "  -> $1$2${NORM}"
    else
	echo "  -> $1$2${NORM}: $3"
    fi
}

reportError() {
    report "$RED" "$1" "$2"
}

reportSuccess() {
    report "$GREEN" "$1" "$2"
}

removeOldVersions() {
	for m in $MODULES; do
	    rm -f /tmp/${m}Version.java
	done
}

restoreOldVersions() {
	for m in $MODULES; do
	    if `test -e /tmp/${m}Version.java`; then
		cp /tmp/${m}Version.java $ROOTPATH/src/$m/Version.java
	    fi
	done
}

# routine to check for successful CVS commit conditions
checkSuccess() {

    if [ "$?" = 0 ]; then
	if [ ! "$1" = "" ]; then
	    reportSuccess "$1"
	fi
    else
	reportError "Commit error" "$2"
	$3
	cat /tmp/commit.reason
	# replace all old version files
	restoreOldVersions
	removeOldVersions
	exit 1
    fi
}

assembleCommitErrors() {
    cp /tmp/commit.log /tmp/commit.reason
}

assembleMissing() {
    echo '*** The following files are not in CVS: ***' > /tmp/commit.reason
    echo `grep cvs\ log:\ nothing\ known\ about /tmp/commit.new | awk '{ print $6 }'` >> /tmp/commit.reason
}

assembleCompileErrors() {
    cp /tmp/commit.log /tmp/commit.reason
}

assembleTestErrors() {
    cp /tmp/commit.log /tmp/commit.reason
}

assembleCheckinList() {
    echo '*** The following files are in CVS but missing here: ***'
    grep cvs\ diff:\ cannot\ find /tmp/commit.log | awk '{ print $5}' >> /tmp/commit.reason
}

echo > /tmp/commit.reason

# ====================================================================
#  Perform consistency checks for added and deleted files
# ====================================================================

echo 'Checking for bash mode "vpc"'
test "$BASH_MODE" = "vpc"
checkSuccess 'OK' 'Please convert shell to vpc mode first.' 'echo'

echo 'Checking that all Java files are added to CVS...'
cvs log $JAVA_FILES &> /tmp/commit.new
checkSuccess 'All Java files here are in CVS.' 'There are Java files missing from CVS.' 'assembleMissing'

echo 'Checking that all test files are added to CVS...'
cvs log $TEST_FILES &> /tmp/commit.new
checkSuccess 'All test files here are in CVS.' 'There are test files missing from CVS.' 'assembleMissing'

echo 'Detecting changes to repository...'
cvs diff &> /tmp/commit.log
if `test "$?" = 0`; then
    reportSuccess "No changes to commit."
    exit 0
else
    reportSuccess "Changes detected."

    echo 'Searching for missing files...'
    test `grep cvs\ diff:\ cannot\ find /tmp/commit.log | wc -l` = 0
    checkSuccess 'No files are missing from CVS.' 'Files are missing here that are in CVS.' 'assembleCheckinList'
fi


# ====================================================================
#  Make clean
# ====================================================================
echo "Removing all class files with \"make clean\""
make clean &> /tmp/commit.log
checkSuccess 'make clean successful.' 'make clean failed.' 'assembleCommitErrors'

# ====================================================================
#  Increment build number and compile each module
# ====================================================================
for m in $MODULES; do
    echo "Checking module $m {"
    MODULE_FILES=`find src/$m -name '*.java'`

    cvs diff src/$m &> /tmp/commit.log
    if `test "$?" = 0`; then
	reportSuccess "No changes to commit."
    else

	VERSION_JAVA=src/$m/Version.java

	echo "  -> Changes detected."
	test -e $VERSION_JAVA
	checkSuccess '' 'Version.java does not exist' 'echo'

	cp $VERSION_JAVA /tmp/${m}Version.java
	if [ "$RELEASE" = 1 ]; then
	    awk '{ if ( $1 == "public" && $4 == "int" && $5 == "commit" ) printf("    public static final int commit = 0;\n"); else print }' /tmp/${m}Version.java > $VERSION_JAVA
	else
	    awk '{ if ( $1 == "public" && $4 == "int" && $5 == "commit" ) printf("    public static final int commit = %d;\n",($7+1)); else print }' /tmp/${m}Version.java > $VERSION_JAVA
	    JAR_NEEDED=1
	fi

	checkSuccess 'Version updated.' 'Version not updated correctly' 'echo'
    fi

    echo "  -> Making $m..."
    make $m &> /tmp/commit.log
    checkSuccess 'Compiled successfully.' 'There were compilation errors building the project.' 'assembleCompileErrors'
    echo "}"
done
   
# ====================================================================
#  Run all tests
# ====================================================================
TESTS='parser seman types init sched'
./runtests.bash $TESTS
checkSuccess 'All tests passed.' 'There were test case failures.' 'assembleTestErrors'

# ====================================================================
#  Attempt to make and archive new jar, if necessary
# ====================================================================
if [ "$JAR_NEEDED" = 1 ]; then

    echo "Checking archive path..."
    if [ ! -d $ARCHPATH ]; then
	reportError "Archive path $ARCHPATH does not exist."
	exit 1
    fi

    JARFILE=`./makejar.bash`
    checkSuccess "JAR built successfully" 'JAR build failed' assembleCommitErrors

    echo "Checking for previous jar..."
    if [ -e "$ARCHPATH/$JARFILE" ]; then
	reportError "JAR file $ARCHPATH/$JARFILE already exists."
	restoreOldVersions
	removeOldVersions
	exit 1
    fi
    
fi

if [ "$2" = test ]; then
    echo "Commit aborted by user."
    restoreOldVersions
    removeOldVersions
    exit 0
fi

# ====================================================================
#  Attempt to commit code
# ====================================================================
echo Attempting to commit to CVS...
cvs commit -m "$1" &> /tmp/commit.log
checkSuccess 'Commit completed successfully.' 'There were errors committing to CVS.' 'assembleCommitErrors'

if [ "$JAR_NEEDED" = 1 ]; then
    cp jars/$JARFILE $ARCHPATH
    checkSuccess "$JARFILE archived successfully" 'JAR copy failed' echo
fi

cat /tmp/commit.log

removeOldVersions
