/*
 * The ht_link class implements a linked list of entries which
 * are collisions in the hashtable.
 */
class ht_link {
    field key: keytype;
    field value: valuetype;
    field next: ht_link;

    constructor(k: keytype, v: valuetype, n: ht_link) {
        key = k;
        value = v;
        next = n;
    }
}

/*
 * The ht_impl class implements the main infrastructure of the hashtable.
 * It exists as a fixed size throughout its life. The wrapper class,
 * hashtable, uses instances of ht_impl internally.
 */
class ht_impl {
    field size: int;
    field entries: int;
    field hash_key: function(keytype): int;
    field compare_keys: function(keytype, keytype): boolean;
    field key_array: keytype[];
    field value_array: valuetype[];
    field link_array: ht_link[];

    constructor(s: int,
                f: function(keytype): int,
                e: function(keytype, keytype): boolean) {

        key_array = new keytype[s];
        value_array = new valuetype[s];
        link_array = new ht_link[s];

        size = s;
        hash_key = f;
        compare_keys = e;
    }

    method apply(user_func: function(keytype, valuetype)) {
        local cntr = 0;
        for ( ; cntr < size; cntr++ ) {
            if ( key_array[cntr] != null )
                user_func(key_array[cntr], value_array[cntr]);
            local link = link_array[cntr];
            while ( link != null ) {
                user_func(link.key, link.value);
                link = link.next;
            }
        }
    }

    method put(k: keytype, v: valuetype) {
        local hash = hash_key(k) % size;
        local k_entry = key_array[hash];

        if ( k_entry == null ) {
            key_array[hash] = k;
            value_array[hash] = v;
            entries++;
        } else if ( compare_keys(k_entry, k) ) {
            value_array[hash] = v;
        } else {
            local link = findlink(k, link_array[hash]);
            if ( link != null ) {
                link.value = v;
            } else {
                link_array[hash] = new ht_link(k, v, link_array[hash]);
                entries++;
            }
        }
    }

    method get(k: keytype): valuetype {
        local hash = hash_key(k) % size;
        local k_entry = key_array[hash];

        if ( k_entry == null ) return null;
        if ( compare_keys(k_entry, k) ) return value_array[hash];

        local link = findlink(k, link_array[hash]);

        return (link == null) ? null : link.value;
    }

    private method findlink(k:keytype, link: ht_link): ht_link {
        while ( link != null ) {
            if ( compare_keys(k, link.key) ) break;
            link = link.next;
        }
        return link;
    }

}

/*
 * The hashtable class implements the main interface to the hashtable
 * functionality. It uses ht_impl instances internally to accomplish
 * the storage and manipulation of data.
 */
class hashtable {

    field data: ht_impl;

    constructor(s: int) {
        data = new ht_impl(s, this.hash, this.compare);
    }

    method apply(user_func: function(keytype, valuetype)) {
        data.apply(user_func);
    }

    method put(k: keytype, v: valuetype) {
        data.put(k, v);
        if ( data.entries >= data.size * 4 )
            resize(data.size * 4);
    }

    method get(k: keytype): valuetype {
        return data.get(k);
    }

    method count(): int {
        return data.entries;
    }

    private method resize(size: int) {
        local old_data = data;
        data = new ht_impl(size, this.hash, this.compare);
        old_data.apply(this.put);
    }

    private method hash(k: keytype): int {
        return k.hashCode();
    }

    private method compare(k1: keytype, k2: keytype): boolean {
        return k1.equals(k2);
    }
}
