// imports require an explicit statement about what kind of
// program element is being imported.
import component virgil.system.io;
import primitive virgil.primitives.int32 as int;
import primitive virgil.primitives.string as string;
import component fool.io as myio;

// ??? actual structure of standard libraries?


// locals > params > fields > methods > super fields > super methods > imports > components

// ==================================================================
// Virgil does not have any of the following Java constructs:
// ==================================================================

// * interfaces
//   Interfaces are difficult to implement efficiently and are often a
//   kludge. A real unit system will remove the need for them.

// * static methods, fields, or initializers
//   Static state is the enemy of object orientation. Components
//   are provided in Virgil for collecting static state and methods,
//   for situations where procedural programming is desirable.

// * exceptions
//   Exceptions cause non-local control flow and require the compiler to
//   support stack walking. Optimization and analysis in their presence
//   is tricky.

// * inner classes
//   Inner classes are generally used for scoping purposes, but inner
//   classes can be tricky when inheritance is involved.

// * anonymous classes
//   Anonymous classes are a hack to achieve closures. First class
//   functions will achieve real closures.

// * dynamic casts
//   Dynamic casts are almost always the result of loss of type information
//   from container classes or poor design.

// * java.lang.Object
//   Object was originally designed to allow containers. However java Object
//   is a wormhole into the VM, and defines lots of methods that are too
//   heavyweight.

// * classloaders
//   Who really understands classloaders? They are a dynamic mechanism to
//   deal with dynamic linking on a fine-grained level. One of those,
//   "feels like a library, but its really a wormwhole into the language
//   runtime" classes.

// * serialization
//   Too complicated to implement. Feels like a library, but again, a
//   wormhole into the language runtime.

// * reflection
//   Too dynamic. Demands large amounts of rich runtime information and
//   inhibits optimization.

// * public and protected keywords
//   These are due to a lack of powerful module mechanism.

// * arrays as objects
//   Container hack.

// * array subtyping
//   Admitted hole in the type system.

// * method overloading
//   Common cause of tricky bugs, complicated rules for method selection.
//   Messy to support in the presence of method extraction.

// * field hiding
//   Almost never a good idea. Sometimes causes surprising bugs, especially
//   during refactoring. A powerful unit system should fix this.

// ==================================================================
// ...but Virgil has the following constructs that are not in Java:
// ==================================================================

// * components
//   Both you and the compiler know where all your static state and
//   methods are. Very convenient. Bye bye singletons.

// * first class functions
//   Allows lightweight semi-functional programming. Eliminates many
//   needs for interfaces. May improve performance in some narrow
//   situations.

// * method extraction
//   Allows convenient tie-in to semi-functional programming. Eliminates
//   many needs for interfaces. Can enhance performance for virtual
//   dispatch, double dispatch, allow implementation of dynamic dispatch,
//   and allow local shorthand names for methods.

// * type inference for local variables
//   Allows more pleasant syntax and less tedium.
//   Optional: an explicit type can be supplied.

// * (TO COME) import as
//   Eliminates need for fully-qualified names in programs and reduces
//   name clashes, especially in adapter pattern situations.

// * (TO COME) generics
//   Allows more powerful containers and eliminates almost all need for
//   java.lang.Object

// * (TO COME) partial evaluation
//   Allows more convenient semi-functional programming.

// * (TO COME) structs
//   Precise control over data structure size and layout for
//   low-level programming.

// * (TO COME) a real module system
//   Allows for more powerful access control and scoping. Allows
//   for metaprogramming.

// * (ON THE TABLE) extensible primitives
//   Allows more convenient numerical programming.
//   Allows more precise targetting of the device/machine.

// ==================================================================
// ...and it has cleaner syntax:
// ==================================================================

// * names of array, function types
//
//   Java              Virgil
// -----------------------------------
//   int[] array;      array: int[];
//   int array[];      array: int[];
//   ...               func: function(int): int;

// * declaration of members, parameters, and locals
//
//   Java                        Virgil
// --------------------------------------------------------
//   int my_field;               field my_field: int;
//   int my_method(int a);       method my_method(a: int): int;
//   void my_method(int a);      method my_method(a: int);
//   int my_local;               local my_local: int;
//   int my_local = 0;           local my_local = 0;


// * switch statements
//
//   Java                        Virgil
// --------------------------------------------------------
//   switch(val) {               switch(val) {
//       case 0:                     case(0) {
//           ...                         ...
//           break;                  }
//       case 1:                     case(1, 2) {
//       case 2:                         ...
//           ...                     }
//           break;
//       default:                    default {
//           ...                         ...
//                                   }
//   }                           }

// ==================================================================
// The compiler and toolchain will be cool too
// ==================================================================

// * designed for native compilation with minimal language runtime
// * requires no external tools (C compiler, assembler) besides JVM
// * will have built in assembler and assembly verifier
// * good connectivity between source language and assembly
// * easily retargettable
// * good performance with minimal space usage

// Verbose Notation       Compact notation
// ==================================================================
// Array<T>             = T[]
// Function<T..., T>    = function(T...): T
// Maybe<T>             = T?
// Raw<K>               = K

component fib {
  
  // ??? Constants can be declared as properties
  property MAX : int = 200;

  // ??? enumeration types?

  // Fields can have a default value in a component.
  // The compiler will generate a constructor.
  field title : string = "Fibonacci program written in the Virgil Programming Language";

  // I like this explicit syntax for members better.
  // It also simplifies parsing some.
  method main(arguments: string[]) {

    // notice new syntax for locals fits with methods and fields well.
    local x1: int = 1, 
          x2: int = 2;

    // there is also type inference for explicitly initialized locals.
    local cntr = 0;

    // declaring a function reference
    local func: function(string[]): int;

    // we can use the component we imported
    io.print(title);

    // the && and || operators are replaced with "and" and "or"
    if ( true and false )
       io.print("Unsound universe. Results may be skewed.");

    // the ! operator is retained, however
    if ( !true )
        io.print("what do I do?");

    // Loops are still like C.
    // ??? is there anything that can be done to make loops nicer?
    for ( cntr = 0; cntr < MAX; cntr++ ) {
       int next = x1 + x2;

       // Note that we can use a component we aliased by our alias.
       myio.print3("The next Fibonacci number is: ", next, "\n");

       x1 = x2;
       x2 = next;
    }

    // collection types
    local a: Object;  // exactly one
    local b: Object*; // zero or more => multicast
    local c: Object+; // one or more => multicast
    local d: Object?; // zero or one

    local e: Object|Exception; // union type

    // tuple types
    local f: (Key, Value);

    // generics
    local g: List<Object>;

    // instanceof                         cast                  convert (adapt)
    //----------------------------------------------------------------------------
    if ( e <: Type ) reset();

    local v: Type = e :: Type;

    // tuple expressions
    (a, b) = f(0);                   // lvalue assignment?
    f = (a, b);                      // tuple creation
    f = (a, b).1;                    // access of tuple element
    f = ((a, b)::(l: int, r:int)).l; // casting to named tuple 
    f = (a, b) :: complex;           // casting to named tuple type
    f = ((a, b), c);                 // nesting

    local t = (a, b); f t;           // application denoted by space?

    // Use cases for bit operators:

    // (1) mask low k bits on or off (suppose k = 3)
    i & ~7             i | 7          // traditional
    i & (-1 # 0b000)   i | 0x7        // binary mask
    i & 0b11111111000  i | 0b111      // explicit binary mask
    i & -0b1000                       // sign extension operator
    i & ~7                            // bitwise complement on integer
    i >> 3 << 3                       // use shift operators
    v[2:0] = 0b000     v[2:0] = 0b111 // range update operator

    // (2) round integer up to multiple of 2^k
    (i + 7) % 8      (i + 7) & ~7     // traditionally
                                      // ?

    // (3) flip byte order
    (i >> 8) | (i << 8)               // traditionally
    i[7:0] # i[15:8]                  // concat operator

    // (4) get low k bits
    i & 7                             // traditionally
    i & 0x7                           // and operator
    i[2:0]                            // bit range access operator

    // (5) get high k bits
    i >> 30                           // traditionally
    i[31:30]                          // bit range access operator

    // (6) is multiple of 2^k ?
    i & 7 == 0
    i[2:0] == 0

    // (7) multiply by 2^k
    i << k
    i << k
    i # 0b000

    // (8) divide by 2^k
    i >> k
    i >> k
    i[31:3]

    // (9) get bits m...n
    (i >> n) & Cm
    i[m:n]

    // (10) is signed?
    i < 0
    i < 0

    // (11) update bits m...n
    i & Cm | v << n
    i[m:n] = v

    // (12) flip bit k
    i ^ (1 << k)
    i[k] = ~i[k]
    i[k] ^= 0b1;

    // (13) update bit k
    i & ~Cm | v & Cm
    i[k] = v;

    // Notice the new syntax for switch statements. Cases don't fall
    // through. Multiple case values are allowed for one body, however.
    switch ( cntr ) {
        case(0) return;
        case(1,2) {
            io.print("1 or 2");
        }
        default return;
    }

  }
}

// Experiments with non-null types.
class surething {

    field f: +object;

    constructor(o: +object) {
        f = o;
        // OK: compiler verify field is initialized to non-null
    }

    // +(reference type) is a non-null type constructor
    method my_method(o: object): +object {
        // the + operator for reference types is a null check.
        return +o;
    }

    method caster(o: object): object {
        return tester(o);
    }

    method tester(o: object): +object {
        // compiler will check nullity of locals.
        local o1: object = o;

        if ( o1 == null ) o1 = o;
        else return o1; /* OK: compiler determines o1 != null l*/

        return o; /* not OK: compiler does not know o1 != null */
    }

    method arrays() {
        local a: +object[] = new object[2];
    }
}

// Experiments with generics.
class Hashtable<`key, `value> extends Set<`value> {

    method get(k: `key): `value;
    method set(k: `key, v: `value);
    method app(f: function(`key, `value));

    // ??? parametric types for methods?
    method map<`value2>(f: function(`key, `value): `value2):
         Hashtable<`key, `value2>;
}

class List<`value> {
    method head(): `value;
    method tail(): `value;
    method length(): int;
    method addFirst(v: `value);
    method addLast(v: `value);
    method concat(l: List<`value>);
    method app(f: function(`value));
    method map<`nv>(f: function(`value): `nv): List<`nv>;
}

// ??? standard libraries, high performance libraries
// ??? strings, characters, etc, what is the end goal?
import class virgil.util.highperf.string as string;
import component virgil.util.textio.stdout as out;

class A<`typeA> {

	method accept(f: function(`typeA):int) {
	}
}

// type variables are bound by a scope.
// Other generic types must be bound when they are outside of
// their declaring scope.
class B<`typeB> {

	field testf1: A<`typeA>; // ERROR: unknown type variable `typeA
	field testf2: A<`typeB>; // OK: `typeB is bound here
	field testf3: A<string>; // also OK: string is a concrete type

	method test(a: `typeB): function(`typeB):int {
		// this.ext: function(`typeB):int
		return this.ext;
	}

	method ext(a: `typeB): int {
		out.print("!");
		return 0;
	}

	method generic(o: Other<`typeB>) {
		o.accept(this.ext);
	}

}

// ??? first attempt at what a module definition might look like.
module MyLibrary {

    // ??? what can be contained in a module?
	contains {
		module miniModule;
		class A, B, C;
		component D, E, F;
		primitive G, H;
	}

    // ??? how do importants work?
	imports {
		class a, b, c;
		module virgil;
	}

    // ??? how to exports work? signatures?
	exports {
		class A as MyA, B as MyB;
		component D as MyD;
	}

	
}

module virgil {

	contains {
		primitive boolean, char16, int32, long64, string;
		class string, list, hashtable;
		component textio;
	}

	exports {
		class string, list, hashtable;
		primitive boolean, char16 as char, int32 as int;
		primitive long64 as long, string;
	}

    // ??? compiler option section? how standardized?
	options {
		vpc.compile-only = true;
		vpc.post-link-optimize = true;
		vpc.cross-module-inlining = true;
		sealed = true;
		debug = false;
		reflect = false;
	}
}

concept Number {

   type num;

   operator +(a: num, b: num): num;
   operator -(a: num, b: num): num;
   operator *(a: num, b: num): num;
   operator /(a: num, b: num): num | !DivideByZeroException;
   operator >(a: num, b: num): boolean;
   operator <(a: num, b: num): boolean;
   operator >=(a: num, b: num): boolean;
   operator <=(a: num, b: num): boolean;
   operator ==(a: num, b: num): boolean;
   operator !=(a: num, b: num): boolean;

}

concept Field {

   type num;
   value one: num;
   value zero: num;

}