/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc;

import cck.text.StringUtil;
import cck.text.Terminal;
import cck.util.*;
import vpc.core.Program;
import vpc.help.MainHelp;
import vpc.hil.DeviceDecl;
import vpc.hil.parser.HILParser;
import vpc.model.*;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.HashSet;
import java.util.List;

/**
 * The <code>Compiler</code> class is the entrypoint the vpc compiler.
 * It parses the arguments, sets up the options, and controls the
 * compilation process.
 *
 * @author Ben L. Titzer
 */
public class Compiler {

    public static final Options options = new Options();

    public static long startMillis;

    public static final Option.Bool HELP = options.newOption("help", false,
            "Displays this help message.");
    static final Option.List CONFIG = options.newOptionList("config", "",
            "This option can be used to specify a list of configuration files that contain " +
            "compiler options and switches. This is useful to reuse a large number " +
            "of command line options without cluttering the invocation of the compiler. " +
            "The configuration files are loaded in the order that they are specified, and " +
            "additional command line options specified override the options specified " +
            "in the configuration file(s).");
    static final Option.Bool COLORS = options.newOption("colors", true,
            "This option enables or disables the printing of control characters " +
            "that color text output to the terminal.");
    public static final Option.Bool VERSION = options.newOption("version", false,
            "Display the detailed compiler version, copyright and license text.");
    static final Option.Bool STATISTICS = options.newOption("statistics", false,
            "This option enables reporting of internal compiler statistics for " +
            "each phase of compilation that includes diagnostic information " +
            "about the program being compiled and the operation of the compiler itself.");
    static final Option.List STAGES = options.newOptionList("stages", "virgil,init,tir-opt0,tir-emit-c",
            "This option specifies a list of the compilation stages (or phases) to " +
            "apply to the program. Each phase produces a representation of the " +
            "program that is fed into the next stage.");
    static final Option.Str INPUT = options.newOption("input", "vsc",
            "This option allows the user to specify the input format of the program. The " +
            "input format includes the source language and its representation. For example, " +
            "the default, \"vsc\" corresponds to Virgil source code.");
    static final Option.Str OUTPUT = options.newOption("output", "",
            "This option selects the output format of the compiler. For example, specify the " +
            "\"csc\" value for this option selects C source code as the output format; " +
            "the compiler will transform the input program through a series of passes " +
            "and optimizations and produce C source code.");
    static final Option.List OPTS = options.newOptionList("opts", "",
            "This option allows the user to specify a list of optional passes that should be " +
            "performed by the compiler. The compiler stage scheduler will attempt to find " +
            "a compilation path from the source representation to the output representation " +
            "that includes the specified optional paths.");
    static final Option.Str DEVICE = options.newOption("device", "",
            "This option specifies the target hardware device for the program. The target device " +
            "description includes the memory configuration of the device as well as interrupts, " +
            "hardware registers, and other hardware capabilities that are exposed to the program " +
            "through the hardware interface language. The device specified in this option must be " +
            "compatible with the target specified in the \"target\" option.");

    public static void main(String[] args) {

        startMillis = System.currentTimeMillis();

        try {
            // parse the command line options
            args = parseOptions(args);

            // if the user specified the -version option, print it and exit
            if ( VERSION.get() ) {
                MainHelp.displayVersionAndCopyright();
                return;
            }

            // no files specified on command line, print help
            if (args.length == 0 || HELP.get() ) {
                MainHelp.printHelp(args);
                return;
            }

            // if all the files exist, begin compilation
            if (Util.verifyFilesExist(args)) {
                Stage[] stages = getStages();
                runCompileChain("output", stages, args);
            }

        } catch (Util.Error e) {
            e.report();
            System.exit(1);
        } catch (Throwable t) {
            Terminal.print(Terminal.ERROR_COLOR, "Unexpected exception");
            Terminal.println(": " + t.getClass());
            t.printStackTrace();
            System.exit(2);
        }
    }

    private static Stage[] getStages() {
        Stage[] stages;
        if ( !INPUT.isBlank() && !OUTPUT.isBlank()) {
            String input = INPUT.get();
            String output = OUTPUT.get();
            HashSet<String> opts = new HashSet<String>();
            opts.addAll((List<String>)OPTS.get());
            stages = new Scheduler(null, opts).findPath(input, output);
            if ( stages == null )
                Util.userError("cannot build compilation path from "+StringUtil.quote(input)+" to "+StringUtil.quote(output));
        } else {
            stages = new Scheduler().getPath(STAGES.toArray());
            if ( stages == null )
                Util.userError("cannot build compilation path for stages "+StringUtil.quote(STAGES.stringValue()));
        }
        return stages;
    }

    private static DeviceDecl getDevice() throws Exception {
        if (DEVICE.isBlank()) return null;
        String fname = DEVICE.get();
        Util.verifyFileExists(fname);
        HILParser hp = new HILParser(new FileInputStream(fname));
        return hp.Device();
    }

    private static void runCompileChain(String programname, Stage[] stages, String[] files) throws Exception {
        DeviceDecl d = getDevice();
        Compilation c = newCompilation(programname, d, stages, files);

        try {
            c.run();
        } finally {
            if (STATISTICS.get()) c.reportStatistics();
        }
    }

    private static Compilation newCompilation(String programname, DeviceDecl d, Stage[] stages, String[] files) {
        Program program = new Program(programname);
        program.setDevice(d);
        Compilation c = new Compilation(program, stages, options);
        for (String f: files) c.addFile(f);
        return c;
    }

    public static String[] parseOptions(String[] args) throws IOException {
        options.parseCommandLine(args);
        List configs = CONFIG.get();
        if (!configs.isEmpty()) {
            // if the config-file option is specified, load config file options
            for (Object o : configs) {
                String fname = (String) o;
                Util.verifyFileExists(fname);
                options.loadFile(fname);
            }
            options.parseCommandLine(args);
        }
        Terminal.useColors = COLORS.get();
        return options.getArguments();
    }

}
