/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Feb 15, 2007
 */

package vpc;

import cck.util.*;
import cck.text.Terminal;
import cck.text.StringUtil;
import vpc.help.MainHelp;
import vpc.model.*;
import vpc.core.*;
import vpc.core.virgil.VirgilArray;
import vpc.core.concept.*;
import vpc.tir.TIRInterpreter;

import java.util.List;
import java.io.IOException;

/**
 * The <code>Interpreter</code> class implements a simple command-line interpreter
 * for Virgil programs.
 *
 * @author Ben L. Titzer
 */
public class Interpreter {
    public static final Options options = new Options();

    public static long startMillis;

    public static final Option.Bool HELP = options.newOption("help", false,
            "Displays this help message.");
    static final Option.List CONFIG = options.newOptionList("config", "",
            "This option can be used to specify a list of configuration files that contain " +
            "compiler options and switches. This is useful to reuse a large number " +
            "of command line options without cluttering the invocation of the compiler. " +
            "The configuration files are loaded in the order that they are specified, and " +
            "additional command line options specified override the options specified " +
            "in the configuration file(s).");
    static final Option.Bool COLORS = options.newOption("colors", true,
            "This option enables or disables the printing of control characters " +
            "that color text output to the terminal.");
    static final Option.Bool TRACE = options.newOption("trace", false,
            "This option enables or disables tracing of the execution of the program " +
                    "within the interpreter.");
    public static final Option.Bool VERSION = options.newOption("version", false,
            "Display the detailed compiler version, copyright and license text.");
    static final Option.List STAGES = options.newOptionList("stages", "virgil,init",
            "This option specifies a list of the compilation stages (or phases) to " +
            "apply to the program. Each phase produces a representation of the " +
            "program that is fed into the next stage.");

    protected static String[] programFiles;
    protected static String[] programArgs;

    protected static VirgilArray.IType charArray;
    protected static VirgilArray.IType charArrayArray;

    public static void main(String[] argv) {

        startMillis = System.currentTimeMillis();

        try {
            // parse the command line options
            parseOptions(argv);

            // if the user specified the -version option, print it and exit
            if ( VERSION.get() ) {
                MainHelp.displayVersionAndCopyright();
                return;
            }

            // no files specified on command line, print help
            if (programFiles.length == 0) {
                Util.userError("no input files");
                return;
            }

            // if all the files exist, begin compilation
            Util.verifyFilesExist(programFiles);

            // if all the files exist, begin compilation
            Program p = runCompileChain("temp", getStages());

            // if compilation worked, interpret the program
            System.exit(runProgram(p));

        } catch (Util.Error e) {
            e.report();
            System.exit(1);
        } catch (Throwable t) {
            Terminal.print(Terminal.ERROR_COLOR, "Unexpected exception");
            Terminal.println(": " + t.getClass());
            t.printStackTrace();
            System.exit(2);
        }
    }

    private static Stage[] getStages() {
        Stage[] stages = new Scheduler().getPath(STAGES.toArray());
        if ( stages == null )
            Util.userError("cannot build compilation path for stages "+StringUtil.quote(STAGES.stringValue()));
        return stages;
    }

    private static Program runCompileChain(String name, Stage[] stages) throws Exception {
        Program program = new Program(name);
        Compilation comp = new Compilation(program, stages, options);
        for (String f: programFiles) comp.addFile(f);
        comp.run();
        return program;
    }

    public static void parseOptions(String[] args) throws IOException {
        options.parseCommandLine(args);

        // if there are configuration files, parse them
        if (!CONFIG.get().isEmpty()) loadConfigs(CONFIG.get(), args);

        // set the terminal colors
        Terminal.useColors = COLORS.get();

        parseProgramFiles(options.getArguments());
    }

    private static int runProgram(Program p) {
        TIRInterpreter interp = new TIRInterpreter(p, TRACE.get());
        Method main = ProgramDecl.resolveEntryPoint("main", p);
        Value v = interp.invokeComponentMethod(main, packageArgs(p, programArgs));
        return PrimInt32.fromValue(v);
    }

    private static Value[] packageArgs(Program p, String[] jargv) {
        charArray = VirgilArray.getArrayType(p.typeCache, PrimChar.TYPE);
        charArrayArray = VirgilArray.getArrayType(p.typeCache, charArray);

        return new Value[] { Reference.toValue(copyJavaStringArray(p, jargv)) };
    }

    private static Heap.Record copyJavaStringArray(Program p, String[] jargv) {
        Heap.Record argv = VirgilArray.allocArray(p.heap, charArrayArray, jargv.length);
        for ( int argn = 0; argn < jargv.length; argn++ ) {
            Heap.Record str = copyJavaString(p, jargv[argn]);
            argv.setValue(argn, Reference.toValue(str));
        }
        return argv;
    }

    public static Heap.Record copyJavaString(Program p, String javaStr) {
        Heap.Record str = VirgilArray.allocArray(p.heap, charArray, javaStr.length());
        for ( int cntr = 0; cntr < javaStr.length(); cntr++ ) {
            str.setValue(cntr, PrimChar.toValue(javaStr.charAt(cntr)));
        }
        return str;
    }

    private static void parseProgramFiles(String[] files) {
        int filen = files.length;
        int argn = 0;

        for ( int cntr = 0; cntr < files.length; cntr++ ) {
            String fn = files[cntr];
            if ( !fn.endsWith(".v") ) {
                filen = cntr;
                argn = files.length - filen;
                break;
            }
        }
        programFiles = new String[filen];
        programArgs = new String[argn];

        // copy the files and arguments into separate arrays
        System.arraycopy(files, 0, programFiles, 0, filen);
        System.arraycopy(files, filen, programArgs, 0, argn);
    }

    private static void loadConfigs(List configs, String[] args) throws IOException {
        // if the config-file option is specified, load config file options
        for (Object o : configs) {
            String fname = (String) o;
            Util.verifyFileExists(fname);
            options.loadFile(fname);
        }
        options.parseCommandLine(args);
    }

}
