/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc.core;

import cck.parser.AbstractToken;
import cck.parser.SourcePoint;
import vpc.core.concept.*;
import vpc.types.TypeName;
import vpc.util.HashList;
import vpc.vst.parser.Token;

/**
 * The <code>CompoundDecl</code> class is an abstract class that represents
 * both class declarations and component declarations in the program. This class
 * contains members, which include fields, methods, and a constructor.
 *
 * @author Ben L. Titzer
 */
public abstract class CompoundDecl extends TypeDecl {
    public final HashList<String, Method> methods;
    public final HashList<String, Field> fields;
    protected Constructor constructor;
    public final NameSpace externalNameSpace;
    public final NameSpace internalNameSpace;

    protected CompoundDecl(AbstractToken tok, TypeName tn) {
        super(tok, tn);
        methods = new HashList<String, Method>();
        fields = new HashList<String, Field>();
        externalNameSpace = new NameSpace();
        internalNameSpace = new NameSpace(externalNameSpace);
    }

    public Method getLocalMethod(String name) {
        return methods.get(name);
    }

    public Field getLocalField(String name) {
        return fields.get(name);
    }

    public Member getLocalMember(String name) {
        Field f = getLocalField(name);
        if (f != null) return f;
        else return getLocalMethod(name);
    }

    public Method resolveMethod(String name, Program p) {
        return methods.get(name);
    }

    public Field resolveField(String name, Program p) {
        return fields.get(name);
    }

    public Member resolveMember(String name, Program p) {
        Field f = resolveField(name, p);
        if (f != null) return f;
        else return resolveMethod(name, p);
    }

    public Field newField(AbstractToken name, TypeName type, boolean priv) {
        Field f;
        if (priv) {
            f = new Field(this, name, this.name + "_" + name.image, type);
            internalNameSpace.addDecl(f.getName(), f);
        } else {
            f = new Field(this, name, name.image, type);
            externalNameSpace.addDecl(f.getName(), f);
        }
        fields.add(name.image, f);
        return f;
    }

    public Method newMethod(AbstractToken name, Function.ITypeName tnf, boolean priv) {
        Method m = new Method(this, name, tnf);
        methods.add(name.image, m);
        if (priv) internalNameSpace.addDecl(m.getName(), m);
        else externalNameSpace.addDecl(m.getName(), m);
        return m;
    }

    public Constructor newConstructor(Function.ITypeName typeName) {
        return constructor = new Constructor(this, typeName);
    }

    public Constructor getConstructor() {
        return constructor;
    }

    public Iterable<Method> getMethods() {
        return methods;
    }

    public Iterable<Field> getFields() {
        return fields;
    }

    public int hashCode() {
        return typename.hashCode();
    }

    public boolean equals(Object o) {
        return this == o;
    }

    public static AbstractToken getDefaultToken(CompoundDecl d) {
        Token tok = new Token("<constructor>");
        SourcePoint pt = d.getSourcePoint();
        tok.beginColumn = pt.beginColumn;
        tok.endColumn = pt.endColumn;
        tok.beginLine = pt.beginLine;
        tok.endLine = pt.endLine;
        return tok;
    }
}
