/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Dec 26, 2005
 */
package vpc.core;

import cck.util.Util;
import vpc.core.concept.Field;
import vpc.core.concept.Reference;
import vpc.core.virgil.VirgilDelegate;
import vpc.types.Type;
import java.util.*;

/**
 * The <code>Heap</code> class represents the heap for a particular program
 * that is used during the initialization phase and then outputted as part
 * of the program binary. It contains records which can be used to represent
 * objects, method tables, and other data structures in the program.
 *
 * @author Ben L. Titzer
 */
public class Heap {

    protected HashSet<Record> roots;
    protected int maxUid;
    public final Program program;

    /**
     * The <code>Layout</code> class represents the layout of a record within the heap.
     * It stores a list of fields that have names and types; the layout is used to track
     * the indices of fields and can also be used to migrate a record from one layout to another.
     */
    public class Layout {
        public Type type;
        protected final String name;
        protected ArrayList<Cell> cells;

        public Layout(String n) {
            name = n;
            cells = new ArrayList<Cell>();
        }

        public Cell findCell(Field f) {
            for (Cell c : cells) {
                if (c.field == f) return c;
            }
            return null;
        }

        public Cell getCell(int i) {
            return cells.get(i);
        }

        public Cell addCell(Field f, Value v) {
            Cell c = new Cell(f, v);
            c.index = cells.size();
            cells.add(c);
            return c;
        }

        public Record newInstance() {
            return initialize(newRecordWithSize(cells.size()));
        }

        public Record newInstanceWithUID(int uid) {
            return initialize(newRecordWithUID(cells.size(), uid));
        }

        private Record initialize(Record record) {
            record.layout = this;
            for (Cell f : cells)
                record.values[f.index] = f.defvalue;
            return record;
        }

        public Layout copy(Layout nl) {
            for (Cell f : cells)
                nl.addCell(f.field, f.defvalue);
            return nl;
        }

        public void migrate(Record r) {
            if (r.layout != this) {
                Value[] nv = new Value[size()];
                copyCells(r, nv);
                r.layout = this;
                r.values = nv;
            }
        }

        private void copyCells(Record r, Value[] nv) {
            for (Cell c : cells) {
                Cell prev = r.layout.findCell(c.field);
                if (prev != null) nv[c.index] = r.getValue(prev.index);
                else nv[c.index] = c.defvalue;
            }
        }

        public List<Cell> getCells() {
            return cells;
        }

        public int size() {
            return cells.size();
        }

        public String toString() {
            return name;
        }
    }

    /**
     * The <code>Record</code> class encapsulates a record within the heap; it is a
     * collection of named fields, with each field corresponding to a <code>java.lang.Object</code>
     * that represents a value. A value could be an integer, a string literal, or a reference
     * to another <code>Record</code>.
     */
    public class Record {
        public final int uid;
        public Layout layout;
        protected Value[] values;

        protected Record(int u, int sz) {
            uid = u;
            values = new Value[sz];
        }

        public Value getValue(int index) {
            return values[index];
        }

        public void setValue(int index, Value v) {
            values[index] = v;
        }

        public int getSize() {
            return values.length;
        }

        public String toString() {
            return "#" + uid + ":" + layout;
        }

        public void setRoot(boolean root) {
            if (root) roots.add(this);
            else roots.remove(this);
        }

        public Program getProgram() {
            return program;
        }
    }

    /**
     * The <code>Cell</code> class represents a cell within a layout. A field is a cell,
     * typed, storage location that has a default value. A cell instance can be used to
     * extract a value from a record by using its index.
     */
    public static class Cell implements Variable {
        public final Field field;
        public final Value defvalue;
        protected int index;

        protected Cell(Field f, Value v) {
            field = f;
            defvalue = v;
        }

        public Value getValue(Record r) {
            return r.values[index];
        }

        public Type getType() {
            return field.getType();
        }

        public String getName() {
            return field.getName();
        }

        public int getIndex() {
            return index;
        }
    }

    /**
     * The constructor for the <code>Heap</code> class creates a new with the given number of
     * maximum fields (in all records). The maximum field number is used to determine when to
     * automatically trigger garbage collection, since some program's initialization phase
     * may allocate a large amount of temporary memory.
     *
     */
    public Heap(Program p) {
        program = p;
        roots = new HashSet<Record>();
    }

    /**
     * The <code>newRecordWithSize()</code> method creates a new record within this heap and returns
     * a reference. This version of the method gives the new record the next highest UID, so
     * that the UID of each record corresponds to its allocation order (i.e. 0 being first).
     *
     * @return a reference to a new, empty, record within this heap
     */
    public Record newRecordWithSize(int sz) {
        return new Record(maxUid++, sz);
    }

    /**
     * The <code>newRecordWithUID()</code> method creates a new record within this heap and returns
     * a reference. This version of the method creates a record with specific UID; this is used
     * when loading a heap from a file so that the UID can be used to compare records across
     * different heaps (e.g. to test initialization code).
     *
     * @param sz  the specified size of the new record
     * @param uid the specified id of the new record
     * @return a reference to a new, empty, record within this heap
     */
    public Record newRecordWithUID(int sz, int uid) {
        return new Record(uid, sz);
    }

    /**
     * The <code>getRecords()</code> method returns a reference to an iterable collection of
     * the records in this heap.
     *
     * @return a list of the records in this heap that can be iterated over
     */
    public Iterable<Record> getRecords() {
        return new RecordIterator(roots);
    }

    public Layout newLayout(String n, Type t) {
        Layout layout = new Layout(n);
        layout.type = t;
        return layout;
    }

    private static void addIfUnmarked(Record ref, HashSet<Record> marked, LinkedList<Record> queue) {
        if (ref != null && !marked.contains(ref)) {
            marked.add(ref);
            queue.offer(ref);
        }
    }

    public static String recordName(Record record) {
        return record == null ? "#null" : record.toString();
    }

    public static boolean compareRecords(Record r1, Record r2) {
        // both null? or identical record in same heap?
        if (r1 == r2) return true;
        // one null? (implies not equal)
        if (r1 == null || r2 == null) return false;
        // compare UIDs (records from different heaps)
        return r1.uid == r2.uid;
    }

    /**
     * The <code>RecordIterator</code> class implements an iterator over a record graph.
     * This iterator will return all reachable records in the graph by iterating over
     * the references between records in a depth-first manner.
     */
    public static class RecordIterator implements Iterator<Record>, Iterable<Record> {

        private final LinkedList<Record> queue;
        private final HashSet<Record> marked = new HashSet<Record>();

        public RecordIterator(Collection<Record> roots) {
            queue = new LinkedList<Record>();
            queue.addAll(roots);
            marked.addAll(roots);
        }

        public Record next() {
            Heap.Record r = queue.poll();
            if ( r == null ) throw new NoSuchElementException();
            scan(r);
            return r;
        }

        public void remove() {
            throw Util.unimplemented();
        }

        public boolean hasNext() {
            return !queue.isEmpty();
        }

        private void scan(Heap.Record r) {
            for (Value v : r.values) {
                if (v instanceof Reference.Val)
                    addIfUnmarked(Reference.fromValue(v), marked, queue);
                else if (v instanceof VirgilDelegate.Val)
                    addIfUnmarked(((VirgilDelegate.Val) v).record, marked, queue);
            }
        }

        public Iterator<Record> iterator() {
            return this;
        }
    }

}
