/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Feb 14, 2006
 */

package vpc.core;

import vpc.util.Delegator;
import java.util.HashMap;

/**
 * The <code>NameSpace</code> class represents a collection of declarations
 * indexed by a string name. Name spaces can be nested, so that if resolution
 * of a particular name fails in an inner scope, it can be resolved in outer
 * scopes.
 *
 * @author Ben L. Titzer
 */
public class NameSpace extends Delegator<Decl, String> {

    protected HashMap<String, Decl> decls;

    /**
     * The default constructor for the <code>NameSpace</code> class creates a new
     * name space with the specified parent name space.
     *
     * @param p the parent(s) of this name space
     */
    public NameSpace(NameSpace... p) {
        for (NameSpace ns : p) add(ns);
    }

    /**
     * The <code>addDecl()</code> method adds a declaration to this name space.
     * The declaration is then available in this name space as given by the name
     * specified.
     *
     * @param n the string name of the declaration
     * @param d the declaration to add to this name space
     */
    public void addDecl(String n, Decl d) {
        if (decls == null) decls = new HashMap<String, Decl>();
        decls.put(n, d);
    }

    /**
     * The <code>resolveDecl()</code> method resolves a declaration given its name
     * in this namespace and its parents.
     *
     * @param n the string name of the declaration
     * @return a reference to the declaration for the specified name if it exists in
     *         this name space or the parent namespace; null otherwise
     */
    public Decl resolveDecl(String n) {
        return invoke(n);
    }

    /**
     * The <code>invoke()</code> method overrides the delegation process. It will
     * first attempt to resolve the string in this namespace, and if it fails,
     * resort to delegation using the <code>Delegator</code> inherited functionality.
     * @param n the name to resolve
     * @return a reference to the declaration for the specified name if it exists in this
     * namesapce or any parent namespaces
     */
    public Decl invoke(String n) {
        Decl d = getDecl(n);
        if ( d == null ) return super.invoke(n);
        else return d;
    }

    /**
     * The <code>getDecl()</code> method resolves a declaration given its name
     * in this namespace. This method does not search the parent namespace.
     *
     * @param n the string name of the declaration
     * @return a reference to the declaration for the specified name if it exists in
     *         this name space; null otherwise
     */
    public Decl getDecl(String n) {
        if (decls == null) return null;
        return decls.get(n);
    }

    /**
     * The <code>getFirstParent()</code> method returns the parent name space of this
     * name space.
     *
     * @return a reference to the first namespace that is a parent of this
     */
    public NameSpace getFirstParent() {
        if (delegates.isEmpty()) return null;
        return (NameSpace)delegates.getFirst();
    }
}
