/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc.core;

import vpc.core.concept.*;
import vpc.core.virgil.*;
import vpc.core.csr.CSRProgram;
import vpc.hil.DeviceDecl;
import vpc.types.*;
import java.io.File;
import java.util.LinkedList;
import cck.parser.SourcePoint;

/**
 * The <code>Program</code> interface represents the compiler's highest level
 * view of the program. Instances of this interface will encapsulate the types,
 * classes, components, and methods that comprise an application and will serve
 * as a means by which the compiler can manage and track the compilation
 * process.
 *
 * @author Ben L. Titzer
 */
public class Program {

    public final String name;
    public final VirgilProgram virgil;
    public CSRProgram csr;
    protected final LinkedList<File> files;
    public final Heap heap;
    public DeviceDecl targetDevice;
    public ProgramDecl programDecl;
    public final NameSpace namespace;
    public final TypeCache typeCache;
    public final TypeEnv typeEnv;
    public final TypeSystem typeSystem;
    public Closure closure;

    public Program(String n) {
        name = n;
        namespace = new NameSpace();
        typeSystem = buildTypeSystem();
        typeEnv = buildTypeEnv();
        typeCache = buildTypeCache();
        virgil = new VirgilProgram(this);
        files = new LinkedList<File>();
        heap = new Heap(this);
        closure = new Closure(this);
    }

    private TypeSystem buildTypeSystem() {
        return new VirgilTypeSystem();
    }

    protected TypeEnv buildTypeEnv() {
        /*
        TypeEnv te = new TypeEnv(null);
        te.bindGlobal("int", PrimInt32.TYPE);
        te.bindGlobal("char", PrimChar.TYPE);
        te.bindGlobal("bool", PrimBool.TYPE);
        te.bindGlobal("void", PrimVoid.TYPE);
        te.bindGlobal("null", PrimNull.TYPE);
        for ( int cntr = 1; cntr <= 64; cntr++ ) {
            te.bindGlobal(cntr+"", PrimRaw.getType(cntr));
        }
        */
        return null;
    }

    protected TypeCache buildTypeCache() {
        TypeCache ts = new TypeCache();
        ts.getTypeName(PrimInt32.TYPE);
        ts.getTypeName(PrimBool.TYPE);
        ts.getTypeName(PrimChar.TYPE);
        ts.getTypeName(PrimVoid.TYPE);
        ts.getTypeName(Reference.NULL_TYPE);
        for ( int cntr = 1; cntr <= 64; cntr++ ) {
            ts.getTypeName(PrimRaw.getType(cntr));
        }
        return ts;
    }

    public void addFile(File f) {
        files.addLast(f);
    }

    public Iterable<File> getFiles() {
        return files;
    }

    public TypeDecl getTypeDecl(String s) {
        return getTypeDecl(typeCache.getTypeName(s));
    }

    public TypeDecl getTypeDecl(TypeName tn) {
        TypeDecl d;
        if ((d = virgil.getClassDecl(tn.toString())) != null) return d;
        return virgil.getComponentDecl(tn.toString());
    }

    public void setDevice(DeviceDecl d) {
        if (d == null) return;
        targetDevice = d;
        namespace.addDecl("device", d);
    }

    public Method resolveMethod(Type t, Member m) {
        VirgilClass.IType ct = (VirgilClass.IType)t;
        return ct.getDecl().resolveMethod(m.getName(), this);
    }

    public File getFirstFile() {
        return files.get(0);
    }

    public SourcePoint getDefaultSourcePoint() {
        return new SourcePoint(getFirstFile().getName(), 1, 1, 1);
    }
}
