/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Feb 8, 2006
 */

package vpc.core;

import cck.parser.AbstractToken;
import vpc.core.concept.Method;
import vpc.core.virgil.VirgilComponent;
import vpc.core.virgil.VirgilError;
import vpc.util.HashList;

/**
 * The <code>ProgramDecl</code> class represents a program declaration, which
 * includes a collection of the modules involved as well as entrypoints.
 *
 * @author Ben L. Titzer
 */
public class ProgramDecl extends BaseDecl {

    public static Method resolveEntryPoint(String epn, Program p) {
        ProgramDecl decl = p.programDecl;
        if (decl == null)
            throw VirgilError.MissingProgramDecl(p);
        EntryPoint entry = decl.entryPoints.get(epn);
        if (entry == null)
            throw VirgilError.CannotResolveEntrypoint.gen(decl.getSourcePoint(), epn);
        return entry.resolve(p);
    }

    /**
     * The <code>EntryPoint</code> class represents a declaration of an entrypoibnt
     * into the program. The entrypoint declarations are used to connect interrupt
     * handlers and the main entrypoint of the program to declared methods of
     * components.
     */
    public static class EntryPoint extends BaseDecl {
        public final AbstractToken component;
        public final AbstractToken method;
        protected Method target;

        public EntryPoint(AbstractToken n, AbstractToken c, AbstractToken m) {
            super(n);
            component = c;
            method = m;
        }

        public Method resolve(Program p) {
            VirgilComponent pcd = p.virgil.getComponentDecl(component.image);
            if (pcd == null)
                throw VirgilError.UnresolvedIdentifier.gen(component.getSourcePoint(), component.image);
            Method md = pcd.resolveMethod(method.image, p);
            if (md == null)
                throw VirgilError.UnresolvedMember.gen(component.getSourcePoint(), method.image, pcd.getName());
            return target = md;
        }


    }

    /**
     * The <code>ComponentRef</code> class represents a reference to a component
     * that is declared to be part of the program.
     */
    public static class ComponentRef extends BaseDecl {
        public ComponentRef(AbstractToken tok) {
            super(tok);
        }
    }

    public final HashList<String, EntryPoint> entryPoints;
    public final HashList<String, ComponentRef> components;

    /**
     * The default constructor for the <code>ProgramDecl</code> class accepts
     * a token that represents the name of the program. The program name is used
     * to create output file(s) for the results of the compilation.
     *
     * @param tok the name of the program as a token
     */
    public ProgramDecl(AbstractToken tok) {
        super(tok);

        entryPoints = new HashList<String, EntryPoint>();
        components = new HashList<String, ComponentRef>();
    }

    public void addEntryPoint(EntryPoint p) {
        entryPoints.add(p.getName(), p);
    }

    public void addComponent(ComponentRef r) {
        components.add(r.getName(), r);
    }
}
