/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc.core;

import cck.parser.AbstractToken;
import cck.text.Printer;
import vpc.core.concept.*;
import vpc.core.virgil.VirgilClass;
import vpc.core.virgil.VirgilComponent;
import java.io.PrintStream;

/**
 * The <code>ProgramPrinter</code> class implements a framework for
 * outputting the program in various forms.
 *
 * @author Ben L. Titzer
 */
public abstract class ProgramPrinter {
    public final Printer printer;
    private final String repName;

    protected ProgramPrinter(String r, Printer p) {
        printer = p;
        repName = r;
    }

    protected ProgramPrinter(String r, PrintStream o) {
        printer = new Printer(o);
        repName = r;
    }

    public void visit(VirgilClass pcd) {
        StringBuffer b = new StringBuffer("class ");
        b.append(pcd.getTypeName());
        b.append(" ");
        AbstractToken pn = pcd.getParent();
        if (pn != null) b.append("extends ").append(pn.image);
        printer.startblock(b.toString());
        visitFields(pcd);
        printer.println("");
        visit(pcd.getConstructor());
        printer.println("");
        visitMethods(pcd);
        printer.endblock();
        printer.println("");
    }

    private void visitMethods(CompoundDecl pcd) {
        for (Method m : pcd.getMethods()) visit(m);
    }

    private void visitFields(CompoundDecl pcd) {
        for (Field f : pcd.getFields()) visit(f);
    }

    public void visit(VirgilComponent pcd) {
        String b = "component " + pcd.getTypeName();
        printer.startblock(b);
        visitFields(pcd);
        printer.println("");
        visit(pcd.getConstructor());
        printer.println("");
        visitMethods(pcd);
        printer.endblock();
        printer.println("");
    }

    public void visit(Method m) {
        printer.print(m.toString());
        Method.MethodRep rep = m.getMethodRep(repName);
        if (rep != null) {
            printer.print(" ");
            printMethodRep(m, rep);
        } else {
            printer.println(";");
        }
    }

    public void visit(Constructor c) {
        if (c == null) return;
        printer.print(c.toString());
        Constructor.MethodRep rep = c.getMethodRep(repName);
        if (rep != null) {
            printer.print(" ");
            printConstructorRep(c, rep);
        } else {
            printer.println(";");
        }
    }

    public void visit(Field f) {
        printer.println(f.toString() + ";");
    }

    public void emitHeap(Heap h) {
        printer.startblock("heap");
        for (Heap.Record r : h.getRecords()) {
            printer.startblock("record #" + r.uid + ":" + r.getSize() + ":" + r.layout);
            for (Heap.Cell f : r.layout.getCells()) {
                printField(r, f);
            }
            printer.endblock();
        }
        printer.endblock();
    }

    public void printField(Heap.Record r, Heap.Cell f) {
        StringBuffer buf = new StringBuffer("field ");
        buf.append(f.getName());
        buf.append(": ");
        buf.append(f.getType());
        buf.append(" = ");
        buf.append(f.getValue(r));
        buf.append(";");
        printer.println(buf.toString());
    }

    public abstract void printMethodRep(Method pm, Method.MethodRep r);

    public abstract void printConstructorRep(Constructor pc, Constructor.MethodRep r);

}
