/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Mar 25, 2006
 */
package vpc.core.concept;

import cck.text.StringUtil;
import vpc.core.Value;
import vpc.types.*;
import vpc.util.ArrayUtil;

/**
 * The <code>Function</code> class encapsulates concepts related to functions,
 * including typenames for functions, type constructors, and actual types. It
 * also contains an implementation of <code>Value</code> that represents
 * a reference to a method.
 *
 * @author Ben L. Titzer
 */
public class Function {
    public static final TypeName[] NOARGS = new TypeName[0];

    public static class ITypeCon extends TypeCon {
        public ITypeCon() {
            super("function", UNBOUNDED);
        }

        public Type newType(TypeCache cache, Type... t) {
            assert t.length > 0;
            Type[] at = ArrayUtil.trunc(t, t.length - 1);
            return new IType(t[t.length - 1], at);
        }

        public String render(TypeRef... p) {
            assert p.length > 0;
            TypeRef[] at = ArrayUtil.trunc(p, p.length - 1);
            return buildFuncName(p[p.length -1], at);
        }
    }

    public static class IType extends Type implements Callable {
        protected final Type returnType;
        protected final Type[] argumentTypes;

        public IType(ITypeName f) {
            super(f.toString(), FUNCTION, REFERENCE, EQUALITY);
            returnType = f.returnTypeName.getType();
            argumentTypes = new Type[f.argumentTypeNames.length];
            for ( int cntr = 0; cntr < f.argumentTypeNames.length; cntr++ )
                argumentTypes[cntr] = f.argumentTypeNames[cntr].getType();
        }

        public IType(Type rt, Type[] p) {
            super(buildFuncName(rt, p), FUNCTION, REFERENCE, EQUALITY);
            returnType = rt;
            argumentTypes = p;
        }

        public Type getReturnType() {
            return returnType;
        }

        public Type[] getArgumentTypes() {
            return argumentTypes;
        }

        public boolean canBeComparedTo(Type t) {
            return t == this || t == Reference.NULL_TYPE;
        }
    }

    public static class ITypeName extends TypeName {
        public final TypeName returnTypeName;
        public final TypeName[] argumentTypeNames;

        public ITypeName(TypeName rt, TypeName[] args) {
            super(buildFuncName(rt, args = nonNullArgs(args)));
            returnTypeName = rt;
            argumentTypeNames = args.clone();
        }

        public boolean isResolved() {
            for (TypeName tn : argumentTypeNames)
                if (!tn.isResolved()) return false;
            return returnTypeName.isResolved();
        }

        public Type getType() {
            if (type == null) {
                type = new IType(this);
            }
            return type;
        }
    }

    public static ITypeName getFuncTypeName(TypeCache typeCache, TypeName rt, TypeName[] args) {
        String fullname = buildFuncName(rt, args);
        if ( typeCache.hasTypeName(fullname)) {
            return (ITypeName) typeCache.getTypeName(fullname);
        } else {
            return typeCache.addTypeName(new ITypeName(rt, args));
        }
    }

    public static <T extends TypeToken> String buildFuncName(T rt, T[] args) {
        return buildCallableName("function", args, rt);
    }

    public static <T extends TypeToken> String buildCallableName(String tycon, T[] args, T rt) {
        StringBuffer buf = new StringBuffer(tycon);
        buf.append("(");
        StringUtil.commalist(args, buf);
        buf.append(")");
        appendReturnType(buf, rt);
        return buf.toString();
    }

    public static TypeName[] nonNullArgs(TypeName[] args) {
        if (args == null) return NOARGS;
        else return args;
    }

    public static void appendReturnType(StringBuffer buf, TypeToken t) {
        if (t != null && !"void".equals(t.toString())) {
            buf.append(": ");
            buf.append(t);
        }
    }

    public static String asReturnType(TypeToken t) {
        if (t != null && !"void".equals(t.toString())) {
            return ": "+ t;
        }
        return "";
    }

    /**
     * The <code>Val</code> class represents a value that is simply a reference
     * to a function. Unlike a delegate, a reference to a function does not include
     * a bound object.
     */
    public static class Val extends Value {
        public final Method method;

        public Val(Method m) {
            super(m.getTypeName().getType());
            method = m;
        }

        public boolean equals(Object o) {
            if (!(o instanceof Val)) return false;
            Val meth = (Val) o;
            return meth.method == method;
        }

        public String toString() {
            return "&" + method.getFullName();
        }
    }
}
