/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Apr 9, 2006
 */
package vpc.core.concept;

import cck.parser.AbstractToken;
import vpc.core.*;
import vpc.types.TypeName;
import vpc.types.Type;
import vpc.util.Hierarchy;
import vpc.util.HashList;
import java.util.HashMap;

/**
 * The <code>Method</code> class represents the VPC compiler's view of a
 * method. A method consists of a method contents, its return type, and the
 * types of its arguments. The two inner classes, <code>Virtual</code> and
 * <code>Static</code> represent the two different types of methods that
 * can be present in Virgil programs.
 * Methods might have several different code representations throughout
 * the compilation process. The functionality associated with a method
 * representation is encapsulated in the <code>MethodRep</code> class.
 *
 * @author Ben L. Titzer
 */
public class Method extends Member {
    protected Function.ITypeName methodTypeName;
    protected HashMap<String, MethodRep> codereps;
    public Method.Family family;

    public Method(CompoundDecl decl, AbstractToken tok, Function.ITypeName tnf) {
        super(decl, tok, tnf);
        methodTypeName = tnf;
        codereps = new HashMap<String, MethodRep>();
    }

    public MethodRep getMethodRep(String name) {
        return codereps.get(name);
    }

    public void addMethodRep(String name, MethodRep rep) {
        codereps.put(name, rep);
    }

    public String getFullName() {
        return container.getTypeName() + ":" + name + "()";
    }

    public TypeName[] getArgumentTypes() {
        return methodTypeName.argumentTypeNames;
    }

    public TypeName getReturnType() {
        return methodTypeName.returnTypeName;
    }

    public String toString() {
        return Function.buildCallableName("method",
                methodTypeName.argumentTypeNames,
                methodTypeName.returnTypeName);
    }

    public interface MethodRep {
    }

    public static class BaseMethodRep<Rep, T extends Type> implements MethodRep {

        protected final HashList<String, Temporary<T>> temps;
        protected final HashList<String, Temporary<T>> params;
        protected Rep body;
        protected int numTemps;

        public BaseMethodRep() {
            temps = new HashList<String, Temporary<T>>();
            params = new HashList<String, Temporary<T>>();
        }

        public void setBody(Rep b) {
            body = b;
        }

        public Rep getBody() {
            return body;
        }

        public Temporary<T> newTemporary(T t) {
            return newVariable("__t" + numTemps, t);
        }

        public Temporary<T> newVariable(String name, T t) {
            String unique = uniquify(name);
            Temporary<T> tmp = new Temporary<T>(unique, t, numTemps++);
            temps.add(unique, tmp);
            return tmp;
        }

        private String uniquify(String name) {
            String uniq = name;
            for ( int cntr = 0; temps.get(uniq) != null; cntr++) {
                uniq = name + cntr;
            }
            return uniq;
        }

        public Temporary<T> newParam(String name, T t) {
            Temporary<T> tmp = new Temporary<T>(name, t, numTemps++);
            params.add(name, tmp);
            return tmp;
        }

        public int getNumberOfTemps() {
            return numTemps;
        }

        public Iterable<Temporary<T>> getTemps() {
            return temps;
        }

        public Iterable<Temporary<T>> getParams() {
            return params;
        }

    }

    /**
     * The <code>Temporary</code> inner class represents a temporary variable
     * used within a method in a Virgil program. Temporaries each have an
     * method-unique identifier and a type associated with them. The
     * <code>Temporary</code> class also stores a String contents that is
     * the source-level variable that corresponds to this temporary.
     */
    public static class Temporary<T extends Type> implements Variable {

        protected final String name;
        public final int id;
        protected final T type;

        Temporary(String nm, T t, int i) {
            name = nm;
            type = t;
            id = i;
        }

        public T getType() {
            return type;
        }

        public String getName() {
            return name;
        }

        public String toString() {
            return name;
        }

        public boolean equals(Object o) {
            return this == o;
        }

        public int hashCode() {
            return name.hashCode();
        }
    }

    /**
     * The <code>Method.Family</code> class represents a method family, a set of
     * methods that are invocable through virtual dispatch.
     */
    public static class Family {
        public final Method rootMethod;
        public Field metaField;
        protected Hierarchy<Method> methods;

        public Family(Method root) {
            rootMethod = root;
            methods = new Hierarchy<Method>();
            methods.addRoot(root);
        }

        public void addMethod(Method p, Method m) {
            methods.add(p, m);
        }

    }
}
