/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Mar 24, 2006
 */

package vpc.core.concept;

import cck.util.Util;
import vpc.core.Value;
import vpc.tir.expr.Operator;
import vpc.tir.expr.Precedence;
import vpc.types.Capability;
import vpc.types.Type;

/**
 * The <code>PrimBool</code> class encapsulates the concept of a boolean
 * used throughout the compiler.
 *
 * @author Ben L. Titzer
 */
public class PrimBool {
    public static Type.Primitive TYPE = new IType();
    public static Val TRUE = toValue(true);
    public static Val FALSE = toValue(false);

    static {
        // add the binops to each of the primitive types
        TYPE.addBinOp(new Capability.BinOp("and", new AND()));
        TYPE.addBinOp(new Capability.BinOp("or", new OR()));
        TYPE.addUnaryOp(new Capability.UnaryOp("!", new NOT()));

        Precedence.register(AND.class, Precedence.PREC_LOGICAL_AND);
        Precedence.register(OR.class, Precedence.PREC_LOGICAL_OR);
        Precedence.register(NOT.class, Precedence.PREC_NOT);
        /*
        Precedence.register(EQU.class, Precedence.PREC_EQUAL);
        Precedence.register(NEQU.class, Precedence.PREC_NOT_EQUAL);
        */
    }

    public abstract static class BoolBinOp extends Operator.Op2 {
        protected BoolBinOp() {
            super(TYPE, TYPE, TYPE);
        }

        public Value apply2(Value v1, Value v2) {
            return toValue(apply(fromValue(v1), fromValue(v2)));
        }

        public abstract boolean apply(boolean v1, boolean v2);
    }

    public abstract static class BoolUnOp extends Operator.Op1 {
        protected BoolUnOp() {
            super(TYPE, TYPE);
        }

        public Value apply1(Value v1) {
            return toValue(apply(fromValue(v1)));
        }

        public abstract boolean apply(boolean v1);
    }

    /**
     * The <code>Logical.AND</code> class represents the logical and operation.
     * This operator does not implement shortcutted conditionals, because it operates on
     * values, not expressions.
     */
    public static class AND extends BoolBinOp {
        public boolean apply(boolean v1, boolean v2) {
            return v1 && v2;
        }
    }

    /**
     * The <code>Logical.OR</code> class represents the logical or operation on booleans.
     * This operator does not implement shortcutted conditionals, because it operates on
     * values, not expressions.
     */
    public static class OR extends BoolBinOp {
        public boolean apply(boolean v1, boolean v2) {
            return v1 || v2;
        }
    }

    /**
     * The <code>Logical.NOT</code> class represents a boolean negation.
     */
    public static class NOT extends BoolUnOp {
        public boolean apply(boolean v1) {
            return !v1;
        }
    }

    public static class IType extends Type.Primitive {

        IType() {
            super("boolean", EQUALITY);
        }

        public boolean canBeComparedTo(Type t) {
            return t == this;
        }
    }

    /**
     * The <code>BOOL</code> class represents a boolean value which can be either <code>true</code>
     * or <code>false</code>.
     */
    public static class Val extends Value {
        protected static Val TRUE = new Val(true);
        protected static Val FALSE = new Val(false);

        public final boolean value;

        private Val(boolean v) {
            super(TYPE);
            value = v;
        }

        public boolean equals(Object o) {
            return o instanceof Val && ((Val) o).value == value;
        }

        public String toString() {
            return "boolean:" + value;
        }
    }

    public static boolean fromValue(Value v) {
        if (v instanceof Val) return ((Val) v).value;
        throw Util.failure("Value of type " + v.dynType + " cannot be converted to Bool");
    }

    public static Val toValue(boolean val) {
        return val ? Val.TRUE : Val.FALSE;
    }
}
