/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Mar 24, 2006
 */

package vpc.core.concept;

import cck.util.Util;
import vpc.core.Value;
import vpc.tir.expr.Operator;
import vpc.tir.expr.Precedence;
import vpc.types.Capability;
import vpc.types.Type;

/**
 * The <code>PrimChar</code> class represents the character type and its values.
 *
 * @author Ben L. Titzer
 */
public class PrimChar {

    public static int CHAR_CACHE = 128;

    public static Type.Primitive TYPE = new IType();

    static {
        // add the binops to each of the primitive types
        TYPE.addBinOp(new Capability.BinOp("<", new LT()));
        TYPE.addBinOp(new Capability.BinOp(">", new GR()));
        TYPE.addBinOp(new Capability.BinOp("<=", new LTEQ()));
        TYPE.addBinOp(new Capability.BinOp(">=", new GREQ()));

        Precedence.register(LT.class, Precedence.PREC_LESS_THAN);
        Precedence.register(GR.class, Precedence.PREC_GREATER);
        Precedence.register(LTEQ.class, Precedence.PREC_LESS_EQUAL);
        Precedence.register(GREQ.class, Precedence.PREC_GREATER_EQUAL);
    }

    public abstract static class Compare extends Operator.Op2 {

        protected Compare() {
            super(TYPE, TYPE, PrimBool.TYPE);
        }

        public Value apply2(Value v1, Value v2) {
            return PrimBool.toValue(apply(fromValue(v1), fromValue(v2)));
        }

        public abstract boolean apply(char v1, char v2);

    }

    /**
     * The <code>Comparison.PrimChar.EQU</code> class represents an integer comparison for equality.
     */
    public static class EQU extends Compare {
        public boolean apply(char v1, char v2) {
            return v1 == v2;
        }
    }

    /**
     * The <code>Comparison.PrimChar.NEQU</code> class represents an integer comparison for inequality.
     */
    public static class NEQU extends Compare {
        public boolean apply(char v1, char v2) {
            return v1 != v2;
        }
    }

    /**
     * The <code>Comparison.PrimChar.GR</code> class represents an integer comparison for greater than.
     */
    public static class GR extends Compare {
        public boolean apply(char v1, char v2) {
            return v1 > v2;
        }
    }

    /**
     * The <code>Comparison.PrimChar.LT</code> class represents an integer comparison for less than.
     */
    public static class LT extends Compare {
        public boolean apply(char v1, char v2) {
            return v1 < v2;
        }
    }

    /**
     * The <code>Comparison.PrimChar.GREQ</code> class represents an integer comparison
     * for greater than or equal to.
     */
    public static class GREQ extends Compare {
        public boolean apply(char v1, char v2) {
            return v1 >= v2;
        }
    }

    /**
     * The <code>Comparison.PrimChar.LTEQ</code> class represents an integer comparison for less than
     * or equal to.
     */
    public static class LTEQ extends Compare {
        public boolean apply(char v1, char v2) {
            return v1 <= v2;
        }
    }

    private static class IType extends Type.Primitive {

        IType() {
            super("char", SWITCH, EQUALITY);
        }

        public boolean canBeComparedTo(Type t) {
            return t == this;
        }
    }

    /**
     * The <code>CHAR</code> class represents a character value.
     */
    public static class Val extends Value {
        public final char value;

        private static Val[] cache = new Val[CHAR_CACHE];

        private Val(char v) {
            super(TYPE);
            value = v;
        }

        public boolean equals(Object o) {
            if (o == this) return true;
            if (o instanceof Val) {
                return ((Val) o).value == value;
            } else if (o instanceof PrimInt32.Val) {
                return ((PrimInt32.Val) o).value == value;
            }
            return false;
        }

        public String toString() {
            return "char:" + (int) value;
        }
    }

    public static char fromValue(Value v) {
        if (v instanceof Val) return ((Val) v).value;
        throw Util.failure("Value of type " + v.dynType + " cannot be converted to PrimChar");
    }

    public static Val toValue(char c) {
        if (c < CHAR_CACHE) {
            Val ival = Val.cache[c];
            if (ival == null) return Val.cache[c] = new Val(c);
            else return ival;
        } else {
            return new Val(c);
        }
    }
}
