/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Mar 27, 2006
 */
package vpc.core.concept;

import vpc.core.Value;
import vpc.core.Heap;
import vpc.tir.expr.Operator;
import vpc.types.Type;
import cck.util.Util;

/**
 * @author Ben L. Titzer
 */
public class Reference {
    public static NullType NULL_TYPE = new NullType();
    public static final Val NULL_VALUE = new Val(NULL_TYPE, null);

    /**
     * The <code>PrimNull</code> type represents the type of the null value;
     * it is assignable to all object references. It has the appropriate
     * isXXX() methods overridden to reflect its semantics.
     */
    public static final class NullType extends Type {
        NullType() {
            super("null", NULL, REFERENCE, EQUALITY);
        }

        public boolean canBeComparedTo(Type t) {
            return t.isReference();
        }

    }

    public static class NullCheck extends Operator.Op1 {
        public NullCheck(Type t) {
            super(t, t);
        }
        public Value apply1(Value v1) throws NullCheckException {
            if ( isNull(v1) ) throw new NullCheckException();
            return v1;
        }
    }

    public static class NullCheckException extends Operator.Exception {
        public NullCheckException() {
            super("NullCheckException", "null check exception");
        }
    }

    /**
     * The <code>Reference.Val</code> class represents a value that contains a reference to a record. The
     * record has a <code>metadata</code> field that stores the type of the record, but this
     * class also stores the dynamic type of the reference. A reference can be <code>null</code>,
     * in which case the dynamic type is <code>Type.NULL</code> and the record reference is
     * <code>null</code>.
     */
    public static class Val extends Value {
        public final Heap.Record record;

        public Val(Type t, Heap.Record r) {
            super(t);
            record = r;
        }

        public Heap.Record asRecord() {
            return record;
        }

        public boolean equals(Object o) {
            if (o == this) return true; // degenerate aliasing case
            if (!(o instanceof Val)) return false; // incorrect type case
            return Heap.compareRecords(((Val) o).record, record);
        }

        public String toString() {
            return Heap.recordName(record);
        }

    }

    public static Heap.Record fromValue(Value v) {
        if (v instanceof Val) return ((Val) v).record;
        throw Util.failure("Value of type " + v.dynType + " cannot be converted to reference");
    }

    public static boolean isNull(Value v) {
        if (v == null) return true;
        if (v instanceof Val) return ((Val)v).record == null;
        return false;
    }

    public static Val toValue(Heap.Record rec) {
        if ( rec == null ) return NULL_VALUE;
        return new Val(rec.layout.type, rec);
    }

    public static Val toValue(Type t, Heap.Record rec) {
        return new Val(t, rec);
    }
}
