/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Jan 29, 2006
 */
package vpc.core.csr;

import cck.text.*;
import vpc.core.Variable;
import vpc.core.concept.Method;
import vpc.core.concept.Callable;
import vpc.tir.tir2c.*;
import vpc.types.Type;
import java.util.LinkedList;
import java.util.List;

/**
 * The <code>CSRFunction</code> class represents a function in the C source representation. It
 * may be the result of compiling a Virgil method in the TIR format, or it may be generated
 * as part of the linkage process. A <code>CSRFunction</code> instance contains a block of code
 * and a list of temporaries and includes the ability to print itself as (well-formatted) text
 * to be output to a .c file.
 *
 * @author Ben L. Titzer
 */
public class CSRFunction extends Method.BaseMethodRep<CIR.CStmt, CSRType> implements Printable {

    public final String name;
    public final CSRType retType;
    public final List<String> attributes;

    public final Method method;

    /**
     * This constructor for the <code>CSRFunction</code> class creates a new C function that corresponds
     * to the given source method. Using the source method's name and types, it will compute an
     * appropriate (mangled) name and (C) types.
     *
     * @param csr
     * @param m the method that this function corresponds to
     */
    public CSRFunction(CSRProgram csr, Method m) {
        method = m;
        name = CUtil.getCFunctionName(m);
        retType = csr.getCSRType(m.getReturnType().getType());
        attributes = new LinkedList<String>();
    }

    /**
     * This constructor for the <code>CSRFunction</code> class creates a new C function that
     * does not correspond to any source method, but has the specified C types and parameters.
     *
     * @param csr
     * @param n  the name of the function as a string
     * @param rt the return type of the function
     * @param p  the list of parameters to the function
     */
    public CSRFunction(CSRProgram csr, String n, CSRType rt, List<? extends Variable> p) {
        method = null;
        name = n;
        retType = rt;
        attributes = new LinkedList<String>();
        addParams(csr, p);
    }

    /**
     * This constructor for the <code>CSRFunction</code> class creates a new C function that
     * does not correspond to any source method, but has the specified C types and parameters.
     *
     * @param n  the name of the function as a string
     * @param rt the return type of the function
     */
    public CSRFunction(String n, CSRType rt) {
        method = null;
        name = n;
        retType = rt;
        attributes = new LinkedList<String>();
    }

    public void addParams(CSRProgram csr, Iterable<? extends Variable> l) {
        for ( Variable v : l )
            newParam(v.getName(), csr.getCSRType(v.getType()));
    }

    public <T extends Variable> void addParams(CSRProgram csr, T... l) {
        for ( Variable v : l )
            newParam(v.getName(), csr.getCSRType(v.getType()));
    }

    /**
     * The <code>print()</code> method prints this C function and its complete body to the specified
     * printer in a well-formatted (indented) way.
     *
     * @param pr the printer to which to output the code as text
     */
    public void print(Printer pr) {
        declareMethod(pr);
        pr.startblock("");
        for (Method.Temporary<CSRType> v : temps) {
            pr.println(v.getType().getLocalDecl(v.getName()) + ";");
        }
        body.print(pr);
        pr.endblock();
    }

    public void addAttribute(String a) {
        attributes.add(a);
    }

    public void printPrototype(Printer pr) {
        declareMethod(pr);
        pr.println(";");
    }

    private void declareMethod(Printer pr) {
        pr.print(retType.toString());
        if (!attributes.isEmpty()) {
            for (String str : attributes) {
                pr.print(" ");
                pr.print("__attribute((");
                pr.print(str);
                pr.print("))");
            }
        }

        pr.beginList(" " + name + "(");
        for (Method.Temporary<CSRType> v : params) {
            pr.print(v.getType().getLocalDecl(v.getName()));
        }
        pr.endList(")");
    }

    public static <T extends Variable> List<T> varList(T... vs) {
        LinkedList<T> list = new LinkedList<T>();
        for (T v : vs) list.add(v);
        return list;
    }

    /**
     * The <code>FuncPtr</code> class represents a C type corresponding to a function
     * pointer. The function pointer contains the type of the return value and the
     * arguments.
     */
    public static class IType extends CSRType implements Callable {
        List<CSRType> ptypes;
        CSRType rtype;

        IType(List<CSRType> pt, CSRType rt) {
            super(rt + " (*)(" + StringUtil.commalist(pt) + ")");
            ptypes = pt;
            rtype = rt;
        }

        public String getVarDecl(String varname) {
            return rtype + " (*" + varname + ")(" + StringUtil.commalist(ptypes) + ")";
        }

        public Type getReturnType() {
            return rtype;
        }

        public Type[] getArgumentTypes() {
            return ptypes.toArray(new Type[ptypes.size()]);
        }
    }
}
