/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Jan 30, 2007
 */

package vpc.core.csr;

import vpc.core.Program;
import vpc.core.concept.Function;
import vpc.core.virgil.*;
import vpc.types.Type;
import java.util.*;
import cck.util.Util;

/**
 * The <code>CSRProgram</code> class implements the functionality and data structures that
 * are global to the CSR (C source representation) of the program.
 *
 * @author Ben L. Titzer
 */
public class CSRProgram {

    public interface Impl {
        public CSRStruct.IType getDelegateStruct();
        public CSRStruct.IType getVirgilClassStruct(VirgilClass.IType c);
        public CSRStruct.IType getVirgilComponentStruct(VirgilComponent.IType c);
        public CSRStruct.IType getVirgilArrayStruct(VirgilArray.IType c);
        public CSRArray.IType getVirgilArrayType(VirgilArray.IType c);
        public CSRStruct.IType getVirgilMetaClassStruct(VirgilMetaClass.IType c);
        public CSRType getPrimitiveType(Type c);

        public CSRFunction.IType getCVirtualType(Type t);
    }

    public Impl impl;

    public final LinkedList<CSRStruct.IType> structs;
    public final List<CSRFunction> functions;

    protected final HashMap<Type, CSRStruct.IType> structMap;
    protected final HashMap<String, CSRType> types = new HashMap<String, CSRType>();

    protected CSRStruct.IType delegate;

    public CSRProgram(Program p) {
        structMap = new HashMap<Type, CSRStruct.IType>();
        structs = new LinkedList<CSRStruct.IType>();
        functions = new LinkedList<CSRFunction>();
    }

    public CSRStruct.IType getCSRStructType(Type t) {
        CSRStruct.IType cst = structMap.get(t);
        if ( cst != null ) return cst;
        if ( t instanceof VirgilArray.IType ) {
            // build a new struct for the array
            cst = impl.getVirgilArrayStruct((VirgilArray.IType)t);
        } else if ( t instanceof VirgilClass.IType ) {
            // build a new struct for the class
            cst = impl.getVirgilClassStruct((VirgilClass.IType)t);
        } else if ( t instanceof VirgilComponent.IType ) {
            // build a new struct for the component
            cst = impl.getVirgilComponentStruct((VirgilComponent.IType)t);
        } else if ( t instanceof VirgilMetaClass.IType ) {
            // build a new struct for the meta class
            cst = impl.getVirgilMetaClassStruct((VirgilMetaClass.IType)t);
        }
        return mapStruct(t, cst);
    }

    private CSRStruct.IType mapStruct(Type t, CSRStruct.IType cst) {
        structMap.put(t, cst);
        return cst;
    }

    public CSRArray.IType getCSRArrayType(Type t) {
        return impl.getVirgilArrayType((VirgilArray.IType)t);
    }

    public CSRStruct.IType getDelegateCSRStruct() {
        if ( delegate == null ) {
            delegate = impl.getDelegateStruct();
            structs.remove(delegate);
            // TODO: this hack is required because of ordering issues in declarations
            structs.addFirst(delegate);
        }
        return delegate;
    }

    public CSRType getCSRType(Type t) {
        if (t instanceof CSRType) return (CSRType)t;
        CSRType ct = impl.getPrimitiveType(t);
        if ( ct != null ) return ct;
        ct = getCSRStructType(t);
        if ( ct != null ) return CSRType.newPointer(this, ct);
        if ( t instanceof Function.IType ) return getDelegateCSRStruct();

        throw Util.failure("unknown type " + t);
    }

    public CSRStruct.IType newStruct(String name) {
        return getCachedType(new CSRStruct.IType(name));
    }

    public CSRStruct.IType newStruct(Type t, String name) {
        return mapStruct(t, getCachedType(new CSRStruct.IType(name)));
    }

    public <T extends CSRType> T getCachedType(T t) {
        CSRType ct = types.get(t.typename);
        if ( ct != null ) {
            return (T)ct;
        } else {
            if ( t instanceof CSRStruct.IType )
                structs.add((CSRStruct.IType)t);
            types.put(t.typename, t);
            return t;
        }
    }
}
