/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Jan 30, 2007
 */

package vpc.core.csr;

import vpc.tir.expr.Operator;
import vpc.core.Value;
import vpc.types.Type;
import java.util.List;
import java.util.LinkedList;
import cck.util.Util;

/**
 * The <code>CSRStruct</code> class implements a representation of a C struct type
 * and related member access operators for use in the TIR representation.
 *
 * @author Ben L. Titzer
 */
public class CSRStruct {

    /**
     * The <code>Struct</code> class represents a C struct type. A struct can contain
     * members (fields only), each with a name and a type.
     */
    public static class IType extends CSRType {
        public List<Field> fields;
        public final String shortName;

        public static class Field {
            public final CSRType type;
            public final String name;

            public Field(String n, CSRType t) {
                type = t;
                name = n;
            }
        }

        IType(String n) {
            super("struct " + n);
            shortName = n;
            fields = new LinkedList<Field>();
        }

        public void addField(String name, CSRType type) {
            fields.add(new Field(name, type));
        }

        public Field getField(String name) {
            for ( Field f : fields ) {
                if ( f.name.equals(name) ) return f;
            }
            return null;
        }

        public int hashCode() {
            return shortName.hashCode();
        }

        public boolean equals(Object o) {
            if ( o == this ) return true;
            if ( o instanceof IType ) {
                return ((IType)o).shortName.equals(shortName);
            }
            return false;
        }
    }

    public static class GetValueField extends Operator.Op1 implements Operator.Location {
        public final IType structType;
        public final IType.Field field;

        public GetValueField(IType t, IType.Field f) {
            super(t, f.type);
            structType = t;
            field = f;
        }

        public Value apply1(Value v1) throws Operator.Exception {
            throw Util.unimplemented();
        }

        public Operator getOperator() {
            return this;
        }

        public Operator setOperator() {
            return new SetValueField(structType, field);
        }
    }

    public static class SetValueField extends Operator.Op2 {
        public final IType structType;
        public final IType.Field field;

        public SetValueField(IType t, IType.Field f) {
            super(t, f.type, f.type);
            field = f;
            structType = t;
        }

        public Value apply2(Value v1, Value v2) throws Operator.Exception {
            throw Util.unimplemented();
        }
    }

    public static class GetRefField extends Operator.Op1 implements Operator.Location {
        public final CSRPointer.IType ptrType;
        public final IType structType;
        public final IType.Field field;

        public GetRefField(CSRPointer.IType t, IType.Field f) {
            super(t, f.type);
            ptrType = t;
            structType = (IType)t.ptype;
            field = f;
        }

        public Value apply1(Value v1) throws Operator.Exception {
            throw Util.unimplemented();
        }

        public Operator getOperator() {
            return this;
        }

        public Operator setOperator() {
            return new SetRefField(ptrType, field);
        }
    }

    public static class SetRefField extends Operator.Op2 {
        public final CSRPointer.IType ptrType;
        public final IType structType;
        public final IType.Field field;

        public SetRefField(CSRPointer.IType t, IType.Field f) {
            super(t, f.type, f.type);
            ptrType = t;
            structType = (IType)t.ptype;
            field = f;
        }

        public Value apply2(Value v1, Value v2) throws Operator.Exception {
            throw Util.unimplemented();
        }
    }

    public static class NewValue extends Operator {
        public final IType structType;

        public NewValue(IType t) {
            super(t);
            structType = t;
        }

        public Type[] getOperandTypes() {
            CSRType[] t = new CSRType[structType.fields.size()];
            int cntr = 0;
            for ( IType.Field f : structType.fields ) {
                t[cntr++] = f.type;
            }
            return t;
        }

        public int getArity() {
            return structType.fields.size();
        }

        public Value apply(Value... v) {
            throw Util.unimplemented();
        }
    }
}
