/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Mar 24, 2006
 */
package vpc.core.virgil;

import vpc.core.Heap;
import vpc.core.Value;
import vpc.core.concept.*;
import vpc.tir.TIRInterpreter;
import vpc.tir.expr.Operator;
import vpc.types.*;
import vpc.vst.parser.Token;
import java.util.Arrays;

/**
 * The <code>VirgilArray</code> class represents the array concept throughout the compiler.
 * An array concept includes operations such as reading / writing to array elements
 * (which may generate a
 * <code>NullCheckException</code> or
 * <code>BoundsCheckException</code>)
 * and getting the length
 * (which may generate a
 * <code>NullCheckException</code> exception).
 *
 * @author Ben L. Titzer
 */
public class VirgilArray {

    public static String buildArrayName(TypeToken et) {
        StringBuffer buf = new StringBuffer();
        if (isFunction(et)) {
            buf.append('(');
            buf.append(et.toString());
            buf.append(')');
        } else {
            buf.append(et.toString());
        }
        buf.append("[]");
        return buf.toString();
    }

    private static boolean isFunction(TypeToken et) {
        return et.toString().startsWith("function");
    }

    public static ITypeName getArrayTypeName(TypeCache typeCache, TypeName elemName) {
        String fullname = buildArrayName(elemName);
        if ( typeCache.hasTypeName(fullname) ) {
            return (ITypeName) typeCache.getTypeName(fullname);
        } else {
            return typeCache.addTypeName(new ITypeName(elemName));
        }
    }

    public static IType getArrayType(TypeCache typeCache, Type elemType) {
        TypeName elemName = typeCache.getTypeName(elemType);
        return (IType)getArrayTypeName(typeCache, elemName).getType();
    }

    public static class ITypeCon extends TypeCon {
        public ITypeCon() {
            super("function", UNBOUNDED);
        }
        public Type newType(TypeCache cache, Type... t) {
            assert t.length == 1;
            return new IType(t[0]);
        }
        public String render(TypeRef... p) {
            assert p.length == 1;
            return buildArrayName(p[0]);
        }
    }

    public static Heap.Layout getArrayLayout(Heap heap, IType at, int len) {
        String n = at.getElemType() + "[" + len + "]";
        Heap.Layout layout = heap.newLayout(n, at);
        Type et = at.getElemType();
        // fill up the layout with the elements
        TypeName etn = heap.program.typeCache.getTypeName(et);
        Field af = new Field(null, new Token(""), etn);
        Value val = TIRInterpreter.getDefaultValue(et);
        for (int cntr = 0; cntr < len; cntr++) layout.addCell(af, val);

        return layout;
    }

    public static class GetLength extends Operator.Op1 {
        public GetLength(IType t) {
            super(t, PrimInt32.TYPE);
        }

        public Value apply1(Value v1) throws Reference.NullCheckException {
            Heap.Record array = Reference.fromValue(v1);
            if (array == null) throw new Reference.NullCheckException();
            return PrimInt32.toValue(array.getSize());
        }
    }

    public static class GetElement extends Operator.Op2 implements Operator.Location {
        public final IType arrayType;

        public GetElement(IType t) {
            super(t, PrimInt32.TYPE, t.elemType);
            arrayType = t;
        }

        public Value apply2(Value v1, Value v2) throws Operator.Exception {
            Heap.Record array = Reference.fromValue(v1);
            int index = checkIndex(array, v2);
            return array.getValue(index);
        }

        public Operator getOperator() {
            return this;
        }

        public Operator setOperator() {
            return new SetElement(arrayType);
        }
    }

    public static class SetElement extends Operator.Op3 {
        public final IType arrayType;

        public SetElement(IType t) {
            super(t, PrimInt32.TYPE, t.elemType, t.elemType);
            arrayType = t;
        }

        public Value apply3(Value v1, Value v2, Value v3) throws Operator.Exception {
            Heap.Record array = Reference.fromValue(v1);
            int index = checkIndex(array, v2);
            array.setValue(index, v3);
            return v3;
        }
    }

    protected static int checkIndex(Heap.Record array, Value ind) throws Reference.NullCheckException, BoundsCheckException {
        if (array == null) throw new Reference.NullCheckException();
        int index = PrimInt32.fromValue(ind);
        if (index < 0 || index >= array.getSize()) throw new BoundsCheckException(array, index);
        return index;
    }

    public static class Alloc extends Operator {
        protected final IType arrayType;
        protected final Heap heap;
        protected final int dimensions;

        public Alloc(IType t, int d, Heap h) {
            super(t);
            arrayType = t;
            heap = h;
            dimensions = d;
        }

        public IType[] getOperandTypes() {
            IType[] t = new IType[dimensions];
            Arrays.fill(t, PrimInt32.TYPE);
            return t;
        }

        public Value apply(Value... v) {
            assert v.length == dimensions;
            return newArray(arrayType, v, 0);
        }

        public int getArity() {
            return dimensions;
        }

        protected Reference.Val newArray(IType at, Value[] dims, int wdim) {
            Type et = at.getElemType();
            int len = PrimInt32.fromValue(dims[wdim]);
            Heap.Record rec = allocArray(heap, at, len);
            if (wdim == dims.length - 1) {
                // inner most dimension = element type
                Value v = TIRInterpreter.getDefaultValue(et);
                for (int cntr = 0; cntr < len; cntr++)
                    rec.setValue(cntr, v);
            } else {
                // not the inner most dimension; recursively make inner arrays
                for (int cntr = 0; cntr < len; cntr++) {
                    Value v = newArray((VirgilArray.IType)et, dims, wdim + 1);
                    rec.setValue(cntr, v);
                }
            }
            return Reference.toValue(at, rec);
        }
    }

    public static class Init extends Operator {
        protected final IType arrayType;
        protected final int length;

        public Init(IType t, int vl) {
            super(t);
            arrayType = t;
            length = vl;
        }

        public Type[] getOperandTypes() {
            Type[] t = new IType[1 + length];
            Arrays.fill(t, arrayType.getElemType());
            t[0] = arrayType;
            return t;
        }

        public int getArity() {
            return length + 1;
        }

        public Value apply(Value... v) {
            assert v.length == length + 1;
            Heap.Record rec = Reference.fromValue(v[0]);
            for ( int cntr = 1; cntr < v.length; cntr++ ) {
                rec.setValue(cntr - 1, v[cntr]);
            }
            return v[0];
        }
    }

    public static Heap.Record allocArray(Heap h, IType at, int sz) {
        Heap.Record r = h.newRecordWithSize(sz);
        r.layout = getArrayLayout(h, at, sz);
        return r;
    }

    /**
     * The <code>VirgilArray.Type</code> class represents an array type. It contains an
     * element type and has the appropriate isXXX() methods overridden to
     * reflect its semantics.
     */
    public static class IType extends Type {
        protected final Type elemType;

        public IType(ITypeName tn) {
            super(tn.toString(), ARRAY, REFERENCE, EQUALITY, INDEXABLE);
            tn.resolveTo(this);
            elemType = tn.componentTypeName.getType();
        }

        public IType(Type et) {
            super(buildArrayName(et), ARRAY, REFERENCE, EQUALITY, INDEXABLE);
            elemType = et;
        }

        public Type getElemType() {
            return elemType;
        }

        public boolean canBeComparedTo(Type t) {
            if (t == this) return true;
            return t == Reference.NULL_TYPE;
        }
    }

    /**
     * The <code>VirgilArray.TypeName</code> class represents the name of an array
     * type, which includes the name of an element type.
     */
    public static class ITypeName extends TypeName {
        protected final TypeName componentTypeName;

        public ITypeName(TypeName cn) {
            super(buildArrayName(cn));
            componentTypeName = cn;
        }

        public boolean isResolved() {
            return componentTypeName.isResolved();
        }

        public Type getType() {
            if (type == null) {
                type = new IType(this);
            }
            return type;
        }
    }

    public static class BoundsCheckException extends Operator.Exception {
        public final Heap.Record array;
        public final int index;

        public BoundsCheckException(Heap.Record rec, int indx) {
            super("BoundsCheckException", "bounds check exception ()");
            array = rec;
            index = indx;
        }
    }
}
