/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Mar 24, 2006
 */
package vpc.core.virgil;

import cck.parser.AbstractToken;
import vpc.core.*;
import vpc.core.concept.*;
import vpc.tir.expr.Operator;
import vpc.types.*;
import vpc.util.HashList;

/**
 * The <code>VirgilClass</code> class encapsulates the concept of an object and object
 * type in the compiler. An instance of the <code>VirgilClass</code> class represents a
 * class within a program, including a reference to the super class and its declared
 * members.
 *
 * @author Ben L. Titzer
 */
public class VirgilClass extends CompoundDecl {
    protected final AbstractToken parent;
    public HashList<String, Method.Family> methodFamilies;
    protected VirgilClass.SourceRep sourceRep;
    public Field metaField;
    public int minTypeID;
    public int maxTypeID;

    public VirgilClass(AbstractToken n, AbstractToken pn, TypeName tn) {
        super(n, tn);
        methodFamilies = new HashList<String, Method.Family>();
        parent = pn;
    }

    public boolean hasParameterizedConstructor() {
        return constructor != null && constructor.getArgumentTypes().length > 0;
    }

    public SourceRep getSourceRep() {
        return sourceRep;
    }

    public void setSourceRep(SourceRep c) {
        sourceRep = c;
    }

    public Method resolveMethod(String name, Program p) {
        return p.virgil.resolveMethod(this, name);
    }

    public Field resolveField(String name, Program p) {
        return p.virgil.resolveField(this, name);
    }

    public AbstractToken getParent() {
        return parent;
    }

    public TypeName getParentName(TypeCache ts) {
        if ( parent == null ) return null;
        return ts.getTypeName(parent.image);
    }

    public Method newMethod(AbstractToken name, Function.ITypeName tnf, boolean priv) {
        Method m = super.newMethod(name, tnf, priv);
        if (!priv) m.family = new Method.Family(m);
        return m;
    }

    /**
     * The <code>VirgilClass.TypeQuery</code> class represents an operator on a
     * class reference that determines whether the receiver object is an instance of
     * the specified class--i.e. the "instanceof" operator. This operator returns
     * false if the specified object reference is null.
     */
    public static class TypeQuery extends Operator.Op1 {
        public final IType target;
        public TypeQuery(VirgilClass.IType stype, VirgilClass.IType target) {
            super(stype, PrimBool.TYPE);
            this.target = target;
        }

        public Value apply1(Value v1) {
            Heap.Record record = Reference.fromValue(v1);
            if (record == null) return PrimBool.toValue(false);
            IType other = (VirgilClass.IType)record.layout.type;
            return PrimBool.toValue(other.isSubtypeOf(target));
        }
    }

    /**
     * The <code>VirgilClass.TypeCast</code> class represents an operator on an
     * object reference that determines whether the receiver object is an instance of
     * the specified class and throws an exception if it is not--i.e. the
     * "dynamic typecast" operator. This operator returns a reference to the object
     * if the cast succeed and throws an exception if it fails.
     */
    public static class TypeCast extends Operator.Op1 {
        public final IType target;
        public TypeCast(VirgilClass.IType stype, VirgilClass.IType target) {
            super(stype, target);
            this.target = target;
        }

        public Value apply1(Value v1) throws TypeCheckException {
            Heap.Record record = Reference.fromValue(v1);
            if (record != null) {
                IType type = (VirgilClass.IType)record.layout.type;
                if (!type.isSubtypeOf(target))
                    throw new TypeCheckException(type, result);
            }
            return v1;
        }
    }

    public static class TypeCheckException extends Operator.Exception {
        public TypeCheckException(Type t, Type et) {
            super("TypeCheckException", "type check exception: "+t+", expected "+et);
        }
    }

    public interface OpVisitor<R, E> extends Operator.Visitor<R, E> {
        public R visit(GetField o, E... e);
        public R visit(SetField o, E... e);
        public R visit(GetMethod o, E... e);
    }

    static <R, E> OpVisitor<R, E> cast(Operator.Visitor<R, E> v) {
        return (OpVisitor<R, E>) v;
    }

    /**
     * The <code>VirgilClass.GetField</code> class represents an operator on a
     * class reference that retrieves the value of a named field. If the reference
     * to the object is null, this operator will throw a <code>NullCheckException</code>.
     */
    public static class GetField extends Operator.Op1 implements Operator.Location {
        public final IType type;
        public final Field field;

        public GetField(VirgilClass.IType et, Field f) {
            super(et, f.getType());
            field = f;
            type = et;
        }

        public Value apply1(Value v1) throws Reference.NullCheckException {
            Heap.Record record = Reference.fromValue(v1);
            if (record == null) throw new Reference.NullCheckException();
            return record.getValue(field.fieldIndex);
        }

        public <R, E> R accept(Operator.Visitor<R, E> v, E... e) {
            if ( v instanceof OpVisitor ) return cast(v).visit(this, e);
            else return super.accept(v, e);
        }

        public Operator getOperator() {
            return this;
        }

        public Operator setOperator() {
            return new SetField(type, field);
        }
    }

    /**
     * The <code>VirgilClass.SetField</code> class represents an operator that updates
     * the field of an object. If the reference to the object is null, this operator
     * will throw a <code>NullCheckException</code>.
     */
    public static class SetField extends Operator.Op2 {
        public final IType type;
        public final Field field;

        public SetField(VirgilClass.IType et, Field f) {
            super(et, f.getType(), f.getType());
            field = f;
            type = et;
        }

        public Value apply2(Value v1, Value v2) throws Reference.NullCheckException {
            Heap.Record record = Reference.fromValue(v1);
            if (record == null) throw new Reference.NullCheckException();
            record.setValue(field.fieldIndex, v2);
            return v2;
        }

        public <R, E> R accept(Operator.Visitor<R, E> v, E... e) {
            if ( v instanceof OpVisitor ) return cast(v).visit(this, e);
            else return super.accept(v, e);
        }
    }

    /**
     * The <code>VirgilClass.GetMethod</code> class represents an operator on a
     * class reference that creates a delegate for the named method. If the reference
     * to the object is null, this operator will throw a <code>NullCheckException</code>.
     * This operator implements virtual dispatch and static dispatch (like Java's invokespecial)
     * on the type of the receiver object.
     */
    public static class GetMethod extends Operator.Op1 {
        public final boolean virtual;
        public final IType type;
        public final Method method;

        public GetMethod(IType et, Method m, boolean v) {
            super(et, m.getType());
            type = et;
            method = m;
            virtual = v;
        }

        public Value apply1(Value v1) throws Reference.NullCheckException, MethodNotFoundException {
            Heap.Record record = Reference.fromValue(v1);
            if (record == null) throw new Reference.NullCheckException();
            if (virtual) {
                return new VirgilDelegate.Val(record, resolveMethod(record));
            } else {
                return new VirgilDelegate.Val(record, method);
            }
        }

        public <R, E> R accept(Operator.Visitor<R, E> v, E... e) {
            if ( v instanceof OpVisitor ) return cast(v).visit(this, e);
            else return super.accept(v, e);
        }

        private Method resolveMethod(Heap.Record record) throws MethodNotFoundException {
            Type t = record.layout.type;
            if ( !(t instanceof IType) )
                throw new MethodNotFoundException(record, method);
            Method m = ((IType)t).getDecl().resolveMethod(method.getName(), record.getProgram());
            if ( m == null )
                throw new MethodNotFoundException(record, method);
            return m;
        }

    }

    public static class Alloc extends Operator.Op0 {
        protected final VirgilClass decl;
        protected final Heap heap;

        public Alloc(VirgilClass.IType c, Heap h) {
            super(c);
            decl = c.getDecl();
            heap = h;
        }

        public Value apply0() {
            Heap.Layout l = heap.program.closure.getLayout(decl);
            return Reference.toValue(result, l.newInstance());
        }
    }


    /**
     * The <code>VirgilClass.Type</code> class represents a Virgil object type within
     * the type system. It consists of a list of fields and a list of methods
     * that make up the class as well as its parent class.
     */
    public static class IType extends Type {

        protected final IType parent;
        protected VirgilClass decl;

        public IType(String n, IType p) {
            super(n, OBJECT, REFERENCE, EQUALITY);
            parent = p;
        }

        public IType getParentType() {
            if (parent == null) return null;
            return parent;
        }

        public void setDecl(VirgilClass clz) {
            decl = clz;
        }

        public VirgilClass getDecl() {
            return decl;
        }

        public boolean isSubtypeOf(Type c) {
            if ( !(c instanceof IType) ) return false;
            for (IType w = this; w != null; w = w.getParentType()) {
                if (c == w) return true;
            }
            return false;
        }

        public boolean canBeComparedTo(Type t) {
            if (t == this) return true;
            if (t == Reference.NULL_TYPE) return true;
            if (!t.isObject()) return false;
            IType tc = (IType) t;
            return isSubtypeOf(tc) || tc.isSubtypeOf(this);
        }

    }

    public interface SourceRep {
    }

    public static class MethodNotFoundException extends Operator.Exception {
        public MethodNotFoundException(Heap.Record r, Method m) {
            super("MethodNotFoundException", "method not found: "+m.getFullName());
        }
    }

}
