/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Apr 6, 2006
 */
package vpc.core.virgil;

import cck.parser.AbstractToken;
import vpc.core.*;
import vpc.core.concept.Field;
import vpc.core.concept.Method;
import vpc.tir.expr.Operator;
import vpc.types.Type;
import vpc.types.TypeName;

/**
 * The <code>VirgilComponent</code> class represents a component declared
 * in a Virgil program.
 *
 * @author Ben L. Titzer
 */
public class VirgilComponent extends CompoundDecl {
    protected VirgilComponent.SourceRep sourceRep;
    protected Heap.Record record;

    public VirgilComponent(AbstractToken tn, TypeName tname) {
        super(tn, tname);
    }

    public SourceRep getComponentRep() {
        return sourceRep;
    }

    public void setComponentRep(SourceRep c) {
        sourceRep = c;
    }

    public void setRecord(Heap.Record r) {
        record = r;
    }

    public Heap.Record getRecord() {
        return record;
    }

    public interface SourceRep {
    }

    public interface OpVisitor<R, E> extends Operator.Visitor<R, E> {
        public R visit(GetField o, E... e);
        public R visit(SetField o, E... e);
        public R visit(GetMethod o, E... e);
    }

    static <R, E> OpVisitor<R, E> cast(Operator.Visitor<R, E> v) {
        return (OpVisitor<R, E>) v;
    }

    /**
     * The <code>VirgilComponent.GetField</code> class represents an operator that
     * retrieves the value of a named component field.
     */
    public static class GetField extends Operator.Op0 implements Operator.Location {
        public final VirgilComponent component;
        public final Field field;

        public GetField(VirgilComponent c, Field f) {
            super(f.getType());
            component = c;
            field = f;
        }

        public Value apply0() throws InitCheckException {
            Heap.Record record = getRecord(component);
            return record.getValue(field.fieldIndex);
        }

        public <R, E> R accept(Operator.Visitor<R, E> v, E... e) {
            if ( v instanceof OpVisitor ) return cast(v).visit(this, e);
            else return super.accept(v, e);
        }

        public Operator getOperator() {
            return this;
        }

        public Operator setOperator() {
            return new SetField(component, field);
        }
    }

    /**
     * The <code>VirgilComponent.SetField</code> class represents an operator that updates
     * the field of a component.
     */
    public static class SetField extends Operator.Op1 {
        public final VirgilComponent component;
        public final Field field;

        public SetField(VirgilComponent c, Field f) {
            super(f.getType(), f.getType());
            component = c;
            field = f;
        }

        public Value apply1(Value v1) throws InitCheckException {
            Heap.Record record = getRecord(component);
            record.setValue(field.fieldIndex, v1);
            return v1;
        }

        public <R, E> R accept(Operator.Visitor<R, E> v, E... e) {
            if ( v instanceof OpVisitor ) return cast(v).visit(this, e);
            else return super.accept(v, e);
        }
    }

    /**
     * The <code>VirgilComponent.GetMethod</code> class represents an operator on a
     * component that creates a delegate for the named component method.
     */
    public static class GetMethod extends Operator.Op0 {
        public final VirgilComponent component;
        public final Method method;

        public GetMethod(VirgilComponent c, Method m) {
            super(m.getType());
            component = c;
            method = m;
        }

        public Value apply0() throws InitCheckException {
            return new VirgilDelegate.Val(getRecord(component), method);
        }

        public <R, E> R accept(Operator.Visitor<R, E> v, E... e) {
            if ( v instanceof OpVisitor ) return cast(v).visit(this, e);
            else return super.accept(v, e);
        }
    }


    /**
     * The <code>Type</code> class represents the type of a component within
     * a Virgil program.
     */
    public static class IType extends Type {

        public IType(TypeName n) {
            super(n.toString(), COMPONENT);
        }

        public boolean canBeComparedTo(Type t) {
            return false;
        }
    }

    protected static Heap.Record getRecord(VirgilComponent component) throws InitCheckException {
        Heap.Record record = component.record;
        if (record == null) throw new InitCheckException(component);
        return record;
    }

    public static class InitCheckException extends Operator.Exception {
        public final VirgilComponent component;
        public InitCheckException(VirgilComponent c) {
            super("InitCheckException", "init check exception");
            component = c;
        }
    }
}
