/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Jun 18, 2006
 */
package vpc.core.virgil;

import cck.parser.AbstractToken;
import vpc.core.Member;
import vpc.core.Program;
import vpc.core.concept.Field;
import vpc.core.concept.Method;
import vpc.types.TypeName;
import vpc.util.HashList;
import vpc.util.Hierarchy;

/**
 * The <code>VirgilProgram</code> definition.
 *
 * @author Ben L. Titzer
 */
public class VirgilProgram {
    public final Program program;
    public final Hierarchy<VirgilClass> hierarchy;
    protected final HashList<TypeName, VirgilClass> classes;
    protected final HashList<TypeName, VirgilComponent> components;

    public int maxClassID;

    public VirgilProgram(Program p) {
        program = p;
        hierarchy = new Hierarchy<VirgilClass>();
        classes = new HashList<TypeName, VirgilClass>();
        components = new HashList<TypeName, VirgilComponent>();
    }

    public VirgilClass newClass(AbstractToken tn, AbstractToken pn) {
        VirgilClass cr = new VirgilClass(tn, pn, program.typeCache.getTypeName(tn.image));
        classes.add(cr.typename, cr);
        program.namespace.addDecl(cr.typename.toString(), cr);
        return cr;
    }

    public VirgilComponent newComponent(AbstractToken tn) {
        VirgilComponent cr = new VirgilComponent(tn, program.typeCache.getTypeName(tn.image));
        components.add(cr.typename, cr);
        program.namespace.addDecl(cr.typename.toString(), cr);
        return cr;
    }

    public Iterable<VirgilClass> getClasses() {
        return classes;
    }

    public Iterable<VirgilComponent> getComponents() {
        return components;
    }

    public VirgilClass getClassDecl(String name) {
        return classes.get(program.typeCache.getTypeName(name));
    }

    public VirgilComponent getComponentDecl(String name) {
        return components.get(program.typeCache.getTypeName(name));
    }

    public VirgilClass getParentClass(VirgilClass c) {
        VirgilClass parent = hierarchy.getParent(c);
        if ( parent != null ) return parent;
        TypeName pcn = c.getParentName(program.typeCache);
        if ( pcn == null ) return null;
        parent = classes.get(pcn);
        if ( parent != null ) hierarchy.add(parent, c);
        return parent;
    }

    public Method resolveMethod(VirgilClass c, String n) {
        while ( c != null ) {
            Method m = c.methods.get(n);
            if ( m != null ) return m;
            c = getParentClass(c);
        }
        return null;
    }

    public Field resolveField(VirgilClass c, String n) {
        while ( c != null ) {
            Field m = c.fields.get(n);
            if ( m != null ) return m;
            c = getParentClass(c);
        }
        return null;
    }

    public Method.Family resolveMethodFamily(VirgilClass c, String n) {
        for ( ; c != null; c = getParentClass(c)) {
            Method.Family mf = c.methodFamilies.get(n);
            if (mf != null) return mf;
        }
        return null;
    }

}
