/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Sep 23, 2006
 */
package vpc.core.virgil;

import cck.util.Arithmetic;
import vpc.core.concept.*;
import vpc.types.*;

/**
 * The <code>VirgilTypeSystem</code> class implements the type system for
 * the virgil language, including the special rules for raw types and classes
 * that cannot be fit into the general typesystem framework.
 *
 * @author Ben L. Titzer
 */
public class VirgilTypeSystem extends TypeSystem {

    public static final int RAW_INT_WIDTH = 32;
    public static final int RAW_CHAR_WIDTH = 8;

    protected static class RawTriple {
        PrimRaw.IType t1;
        PrimRaw.IType t2;
        PrimRaw.IType t3;
    }

    public VirgilTypeSystem() {

    }

    public Capability.BinOp resolveBinOp(String op, Type t1, Type t2) {
        if ( "&".equals(op) ) {
            RawTriple t = getMaxRawType(getRawTypes(t1, t2));
            if ( t.t3 == null ) return null;
            return new Capability.BinOp("&", new PrimRaw.AND(t.t3.width, Arithmetic.min(t.t1.width, t.t2.width)));
        } else if ( "|".equals(op) ) {
            RawTriple t = getMaxRawType(getRawTypes(t1, t2));
            if ( t.t3 == null ) return null;
            return new Capability.BinOp("|", new PrimRaw.OR(t.t3.width));
        } else if ( "^".equals(op) ) {
            RawTriple t = getMaxRawType(getRawTypes(t1, t2));
            if ( t.t3 == null ) return null;
            return new Capability.BinOp("^", new PrimRaw.XOR(t.t3.width));
        } else if ( "<<".equals(op) ) {
            PrimRaw.IType rt1 = getRawType(t1);
            if ( rt1 == null ) return null;
            return new Capability.BinOp("<<", new PrimRaw.SHL(rt1.width));
        } else if ( ">>".equals(op) ) {
            PrimRaw.IType rt1 = getRawType(t1);
            if ( rt1 == null ) return null;
            return new Capability.BinOp(">>", new PrimRaw.SHR(rt1.width));
        } else if ( "#".equals(op) ) {
            PrimRaw.IType rt1 = getRawType(t1);
            PrimRaw.IType rt2 = getRawType(t2);
            if ( rt1 == null || rt2 == null ) return null;
            return new Capability.BinOp("#", new PrimRaw.Concat(rt1, rt2));
        } else {
            return t1.resolveInfixBinOp(op);
        }
    }

    public Capability.UnaryOp resolveUnOp(String op, Type t1) {
        if ( "~".equals(op) ) {
            PrimRaw.IType rt1 = getRawType(t1);
            if ( rt1 == null ) return null;
            return new Capability.UnaryOp("~", new PrimRaw.Complement(rt1));
        } else return t1.resolvePrefixUnaryOp(op);
    }

    public boolean isAssignable(Type src, Type target) {
        // reflexivity: every type is assignable to itself
        if ( target == src ) return true;
        if ( isClass(target) ) return isAssignableClass(src, target);
        if ( isArray(target) ) return isAssignableArray(src, target);
        if ( isFunction(target) ) return isAssignableFunction(src, target);
        if ( isRaw(target) ) return isAssignableRaw(src, target);
        return false;
    }

    private boolean isAssignableClass(Type src, Type target) {
        if (isNull(src)) return true;
        if (!isClass(src)) return false;
        VirgilClass.IType cts = (VirgilClass.IType)src;
        return cts.isSubtypeOf(target);
    }

    private boolean isAssignableArray(Type src, Type target) {
        return isNull(src);
    }

    private boolean isAssignableFunction(Type src, Type target) {
        return isNull(src);
    }

    private boolean isAssignableRaw(Type src, Type target) {
        PrimRaw.IType tt = (PrimRaw.IType)target;
        if ( isRaw(src) ) {
            PrimRaw.IType st = (PrimRaw.IType)src;
            return st.width <= tt.width;
        }
        return false;
    }

    public boolean isComparable(Type t1, Type t2) {
        return t1.canBeComparedTo(t2);
    }

    public boolean isConvertible(Type src, Type target) {
        if ( src == target ) return true;
        if ( isRaw(src) ) return isConvertibleRaw(src, target);
        if ( isInt(src) ) return isConvertibleInt(target);
        if ( isChar(src) ) return isConvertibleChar(target);
        return false;
    }

    private boolean isConvertibleRaw(Type t1, Type t2) {
        if ( isRaw(t2) ) return true;

        PrimRaw.IType rt1 = getRawType(t1);
        if ( isInt(t2) ) return rt1.width <= RAW_INT_WIDTH;
        if ( isChar(t2) ) return rt1.width <= RAW_CHAR_WIDTH;

        return false;
    }

    private boolean isConvertibleInt(Type t2) {
        if ( isRaw(t2) ) return true;
        return isChar(t2);
    }

    private boolean isConvertibleChar(Type t2) {
        return isRaw(t2);
    }

    public boolean isPromotable(Type t1, Type t2) {
        // reflexivity: every type is promotable to itself
        if ( t1 == t2 ) return true;
        if ( isChar(t1) ) return isPromotableChar(t2);
        if ( isInt(t1) ) return isPromotableInt(t2);
        if ( isRaw(t1) ) return isPromotableRaw(t1, t2);
        return false;
    }

    private boolean isPromotableRaw(Type t1, Type t2) {
        // raw m => raw n, iff m <= n
        if ( isRaw(t2) ) {
            RawTriple t = getRawTypes(t1, t2);
            return t.t1.width <= t.t2.width;
        }
        return false;
    }

    private boolean isPromotableInt(Type t2) {
        if ( isRaw(t2) ) {
            PrimRaw.IType rt = getRawType(t2);
            return rt.width >= RAW_INT_WIDTH;
        }
        return false;
    }

    private boolean isPromotableChar(Type t2) {
        if ( isInt(t2) ) return true;
        if ( isRaw(t2) ) {
            PrimRaw.IType rt = getRawType(t2);
            return rt.width >= RAW_CHAR_WIDTH;
        }
        return false;
    }

    public Type unify(Type t1, Type t2) {
        if ( t1 == t2 ) return t1;
        if ( isNull(t1) ) return unifyNull(t2);
        if ( isNull(t2) ) return unifyNull(t1);
        if ( isClass(t1) ) return unifyClass(t1, t2);
        if ( isRaw(t1) || isRaw(t2) ) return unifyRaw(t1, t2);
        if ( isInt(t1) && isChar(t2) ) return t1;
        if ( isChar(t1) && isInt(t2) ) return t2;
        return null;
    }

    private Type unifyNull(Type t2) {
        if ( isClass(t2) ) return t2;
        if ( isArray(t2) ) return t2;
        if ( isFunction(t2) ) return t2;
        return null;
    }

    private Type unifyClass(Type t1, Type t2) {
        if ( !isClass(t2) ) return null;
        VirgilClass.IType ct1 = (VirgilClass.IType)t1;
        VirgilClass.IType ct2 = (VirgilClass.IType)t2;
        // TODO: this hierarchy search is inelegant (and might be slow)
        for ( ; ct1 != null; ct1 = ct1.getParentType() ) {
            if ( ct2.isSubtypeOf(ct1) ) return ct1;
        }
        return null;
    }

    private Type unifyRaw(Type t1, Type t2) {
        return getMaxRawType(getRawTypes(t1, t2)).t3;
    }

    protected RawTriple getRawTypes(Type t1, Type t2) {
        RawTriple triple = new RawTriple();
        triple.t1 = getRawType(t1);
        triple.t2 = getRawType(t2);
        return triple;
    }

    protected PrimRaw.IType getRawType(Type t1) {
        if ( isRaw(t1) ) return (PrimRaw.IType) t1;
        if ( isChar(t1) ) return PrimRaw.getType(RAW_CHAR_WIDTH);
        if ( isInt(t1) ) return PrimRaw.getType(RAW_INT_WIDTH);
        return null;
    }

    private RawTriple getMaxRawType(RawTriple t) {
        if ( t.t1 != null && t.t2 != null )
            t.t3 = PrimRaw.getMaxType(t.t1, t.t2);
        return t;
    }

    private RawTriple getMinRawType(RawTriple t) {
        if ( t.t1 != null && t.t2 != null )
            t.t3 = PrimRaw.getMinType(t.t1, t.t2);
        return t;
    }

    public static boolean isRaw(Type t) {
        return t instanceof PrimRaw.IType;
    }

    public static boolean isClass(Type t) {
        return t.isObject();
    }

    public static boolean isNull(Type t) {
        return t.isNull();
    }

    public static boolean isVoid(Type t) {
        return t.isVoid();
    }

    public static boolean isFunction(Type t) {
        return t.isFunction();
    }

    public static boolean isComponent(Type t) {
        return t.isComponent();
    }

    public static boolean isArray(Type t) {
        return t.isArray();
    }

    public static boolean isChar(Type t) {
        return t == PrimChar.TYPE;
    }

    public static boolean isInt(Type t) {
        return t == PrimInt32.TYPE;
    }

    public static boolean isBoolean(Type t) {
        return t == PrimBool.TYPE;
    }
}
