/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc.model;

import static cck.text.StringUtil.*;
import cck.text.TermUtil;
import cck.text.Terminal;
import cck.util.Options;
import vpc.Compiler;
import vpc.core.Program;
import java.io.File;
import java.io.IOException;
import static java.lang.System.currentTimeMillis;
import java.util.HashMap;
import java.util.LinkedList;

/**
 * The <code>Compilation</code> class encapsulates the functionality of
 * performing the various stages of the compilation process. A compilation
 * is built of <code>Compilation.Stage</code> instances that can be either
 * per module, per type, per class, per method, etc. A chain of these
 * stages is built and the Compilation class will schedule the
 * compilation appropriately (eventually in multiple threads).
 *
 * @author Ben L. Titzer
 */
public class Compilation {

    private LinkedList<Stage> stages = new LinkedList<Stage>();

    private Program program;
    private HashMap<Stage, StageStatistics> statistics;

    private long begin_millis;
    private long total_millis;

    public class StageStatistics {
        Stage stage;
        long milliseconds;
        boolean success;

        StageStatistics(Stage s) {
            stage = s;
        }

        public void report() {
            int color = success ? Terminal.COLOR_GREEN : Terminal.COLOR_RED;
            reportTime(color, Registry.getStageName(stage), milliseconds);
        }

    }

    public Compilation(Program p, Stage[] s, Options opts) {
        program = p;
        statistics = new HashMap<Stage, StageStatistics>();
        for (Stage value : s) {
            stages.addLast(value);
            if (opts != null) value.processOptions(opts);
        }
    }

    public void run() throws IOException {

        begin_millis = currentTimeMillis();
        try {
            for (Stage s : stages) {
                StageStatistics stat = new StageStatistics(s);

                long ms = currentTimeMillis();
                try {
                    s.visitProgram(program);
                    stat.success = true;
                } finally {
                    ms = currentTimeMillis() - ms;
                    stat.milliseconds = ms;
                    statistics.put(s, stat);
                }
            }
        } finally {
            total_millis = currentTimeMillis() - Compiler.startMillis;
        }
    }

    public void reportStatistics() {
        TermUtil.printSeparator(50, "Compilation Statistics");

        reportTime(-1, "startup", begin_millis - Compiler.startMillis);
        for (Stage s : stages) {
            StageStatistics stat = statistics.get(s);
            if (stat != null) stat.report();
        }

        TermUtil.printThinSeparator(50);
        Terminal.print(" Total              : " + rightJustify(milliToSecs(total_millis), 7) + " seconds");
        Terminal.println("\n");
    }

    private void reportTime(int color, String name, long ms) {
        float pct = (100.0f * ms) / total_millis;
        // \color:$1{\19{$2}}: \-7{$3} \-7{$4} %
        Terminal.print(" ");
        Terminal.print(color, leftJustify(name, 19));
        Terminal.print(": ");
        Terminal.print(rightJustify(milliToSecs(ms), 7));
        Terminal.print(rightJustify(toFixedFloat(pct, 2), 7));
        Terminal.print(" %");
        Terminal.nextln();
    }

    public void addFile(String s) {
        File f = new File(s);
        program.addFile(f);
    }

}
