/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Dec 8, 2006
 */

package vpc.model;

import java.util.HashSet;
import java.util.Set;

/**
 * The <code>IRState</code> class represents the internal representation state
 * for a compilation stage. Each IRState specifies an intermediate representation
 * and any attributes that exist or are guaranteed not to exist. This class
 * is used to enforce correct ordering of compilation within an intermediate
 * representation as well as to schedule the overall compilation process.
 *
 * @author Ben L. Titzer
 */
public class IRState {
    public final String name;
    public final Set<String> plusAttrs;
    public final Set<String> minusAttrs;
    private String sname;

    public IRState(String n) {
        name = n;
        plusAttrs = new HashSet<String>();
        minusAttrs = new HashSet<String>();
    }

    public boolean isMetBy(IRState other) {
        if ( !other.name.equals(name) ) return false;
        if ( !other.plusAttrs.containsAll(plusAttrs) ) return false;
        for ( String s : minusAttrs ) {
            if ( other.plusAttrs.contains(s) ) return false;
        }
        return true;
    }
    public String toString() {
        if ( sname == null ) {
            StringBuffer buf = new StringBuffer(name);
            appendAttrs(buf, plusAttrs, '+');
            appendAttrs(buf, minusAttrs, '-');
            sname = buf.toString();
        }
        return sname;
    }

    public void addPlusAttr(String s) {
        plusAttrs.add(s);
        // invalidate the name and hashing cache
        sname = null;
    }

    public void addMinusAttr(String s) {
        minusAttrs.add(s);
        // invalidate the name and hashing cache
        sname = null;
    }

    private void appendAttrs(StringBuffer buf, Set<String> pa, char h) {
        if (!pa.isEmpty() ) {
            buf.append(h);
            for ( String s : pa ) buf.append(s);
        }
    }

    public int hashCode() {
        return toString().hashCode();
    }

    public boolean equals(Object o) {
        if ( o == this ) return true;
        if ( !(o instanceof IRState) ) return false;
        IRState s = (IRState)o;
        if ( !name.equals(s.name) ) return false;
        if ( toString().equals(s.toString()) ) return true;
        if ( !plusAttrs.equals(s.plusAttrs) ) return false;
        return minusAttrs.equals(s.minusAttrs);
    }
}
