/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Jan 2, 2006
 */

package vpc.model;

import cck.util.ClassMap;
import cck.util.Options;
import vpc.core.Program;
import vpc.tir.opt.*;
import vpc.tir.opt.rma.TIR_RMA;
import vpc.tir.stages.*;
import vpc.tir.tir2c.TIREmitC;
import vpc.vst.VSTStages;
import java.io.IOException;
import java.util.*;

/**
 * The <code>Registry</code> class implements a registry of all the compiler
 * passes, including optimizations and translations between various program
 * representations. This entries in this registry are used to build the list
 * of compiler passes to apply to the program.
 *
 * @author Ben L. Titzer
 */
public class Registry {

    protected static Registry globalRegistry;

    public static Registry getDefaultRegistry() {
        if ( globalRegistry == null ) {
            globalRegistry = new Registry();
            globalRegistry.initDefault();
        }
        return globalRegistry;
    }

    protected final ClassMap stageMap;
    protected final List<String> stageList;
    protected final HashMap<String, String> signatures;
    protected final WeakHashMap<String, StageSignature> sigMap;

    public Registry() {
        stageMap = new ClassMap("Compilation Stage", Stage.class);
        stageList = new LinkedList<String>();
        signatures = new HashMap<String, String>();
        sigMap = new WeakHashMap<String, StageSignature>();
    }

    /**
     * The <code>Macro</code> class implements a compilation stage that
     * is the combination of a sequence of other compilation stages.
     */
    public static class Macro extends Stage {
        protected final Stage[] stages;

        public Macro(String str) {
            stages = Scheduler.getDefaultPath(str);
        }

        public void visitProgram(Program p) throws IOException {
            for (Stage s : stages) s.visitProgram(p);
        }

        public void processOptions(Options o) {
            for (Stage s : stages) s.processOptions(o);
        }
    }

    void initDefault() {
        // Parse: parse the source code to VST and verify structural properties
        addStage("virgil-parse", "vsc->vst+v",      VSTStages.Parse.class);
        // BuildTypes: build the type system from the declared types in the program
        addStage("virgil-gtb",   "vst+v->vst+t",    VSTStages.BuildTypes.class);
        // TypeCheck: type check the VST
        addStage("virgil-tc",    "vst+tv->vst+c",   VSTStages.TypeCheck.class);
        // BuildTIR: build TIR code from the VST
        addStage("virgil-tir",   "vst+tvc->tir",    VSTStages.BuildTIR.class);
        // PrintVST: debugging stage to output the VST tree
        addStage("virgil-emit",  "vst->vst",        VSTStages.PrintVST.class);
        // ObjectLayout: compute object layouts
        addStage("layout",     "tir->tir+l",        ObjectLayout.class);
        // MetaLayout: compute layouts for meta objects
        addStage("meta",       "tir+i->tir+m",      MetaLayout.class);
        // TIRInit: run the initializers
        addStage("tir-init",   "tir+l->tir+i",      TIRInit.class);
        // TIR_CCP: perform conditional constant propagation optimization
        addStage("tir-ccp",    "tir->tir-v",        TIR_CCP.class);
        // TIRSimplifier: perform TIR simplification
        addStage("tir-simp",   "tir->tir+s-v",      TIRSimplifier.class);
        // RawConversion: convert raw operations
        addStage("tir-raw",   "tir+s->tir+rs-v",    RawConversion.class);
        // ObjectConversion: convert object operations
        addStage("tir-obj",   "tir+rs->tir+or-v",   ObjectConversion.class);
        // TIRChecker: verify well-formedness of TIR code
        addStage("tir-check",  "tir->tir+v",        TIRChecker.class);
        // RMAnalyzer: perform RMA optimization
        addStage("tir-rma",    "tir+i-m->tir+lm-v", TIR_RMA.class);
        // TIREmit: emit TIR code as text
        addStage("tir-emit",   "tir->tir",          TIREmit.class);
        // TIREmitC: emit TIR code as C source code
        addStage("tir-emit-c", "tir+iorslm->csc",   TIREmitC.class);
        // EmitHeap: emit the heap as a dot graph
        addStage("emit-heap", "tir+i->tir",         EmitHeap.class);

        // macros: for shorter optimization sequences
        addMacro("virgil", "virgil-parse,virgil-gtb,virgil-tc,virgil-tir");
        addMacro("init", "layout,tir-simp,tir-init");

        addMacro("tir-opt0", "tir-raw,meta,tir-obj,tir-simp");
        addMacro("tir-opt1", "tir-raw,tir-rma,tir-obj,tir-simp");
        addMacro("vrma", "virgil-parse,virgil-gtb,virgil-tc,virgil-tir,layout,tir-simp,tir-init,tir-raw,tir-rma");
    }

    public void addMacro(String alias, String str) {
        stageMap.addInstance(alias, new Macro(str));
    }

    public void addStage(String alias, String sstr, Class<? extends Stage> clazz) {
        signatures.put(alias, sstr);
        stageMap.addClass(alias, clazz);
        stageList.add(alias);
    }

    public void addStage(String alias, String sstr, Stage stg) {
        signatures.put(alias, sstr);
        stageMap.addInstance(alias, stg);
        stageList.add(alias);
    }

    public StageSignature getSignature(String alias) {
        String sig = signatures.get(alias);
        if ( sig == null ) return null;
        StageSignature sobj = sigMap.get(sig);
        if ( sobj == null ) {
            sobj = StageSignature.parseSignature(sig);
            sigMap.put(sig, sobj);
        }
        return sobj;
    }

    public StageSignature getSignature(Stage stage) {
        return getSignature(stageMap.getAlias(stage));
    }

    public Stage getStage(String alias) {
        return (Stage) stageMap.getObjectOfClass(alias);
    }

    public static String getStageName(Stage s) {
        return globalRegistry.stageMap.getAlias(s);
    }
}
