/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Dec 11, 2006
 */

package vpc.model;

import java.util.*;

/**
 * The <code>Scheduler</code> class implements the algorithms that choose
 * the sequence of compilation passes to apply to a program. The scheduler
 * uses the model of the compiler in the <code>Registry</code> class to
 * determine the legal paths through the compiler. The scheduler then uses
 * tuning heuristics, options from the user, and the structure of the
 * program to choose and refine a set of compilation passes for the program.
 *
 * @author Ben L. Titzer
 */
public class Scheduler {

    private final Registry registry;
    private final HashSet<String> globalRequired;
    private final HashMap<String, StageSignature> specialSignatures;

    protected static class Path {
        final int length;
        final Path prev;
        final String node;
        final IRState state;

        Path(Path p, String n, IRState i) {
            length = p == null ? 1 : p.length + 1;
            prev = p;
            node = n;
            state = i;
        }

        public Stage[] buildReverse(Registry registry) {
            Stage[] result = new Stage[length];
            int cntr = length;
            for ( Path p = this; p != null; p = p.prev )
                result[--cntr] = registry.getStage(p.node);
            return result;
        }
    }

    public Scheduler() {
        this(null, null);
    }

    public Scheduler(Registry reg, Collection<String> req) {
        if ( reg == null ) reg = Registry.getDefaultRegistry();
        registry = reg;

        if ( req != null ) {
            globalRequired = new HashSet<String>();
            specialSignatures = new HashMap<String, StageSignature>();
            globalRequired.addAll(req);
            extendSignatures();
        } else {
            globalRequired = null;
            specialSignatures = null;
        }
    }

    private void extendSignatures() {
        for ( String s : globalRequired ) {
            // for every pass in the required set, alter its signature
            StageSignature sig = registry.getSignature(s);
            if ( sig != null ) {
                sig = sig.copy();
                sig.output.addPlusAttr("$"+s);
                specialSignatures.put(s, sig);
            }
        }
    }

    /**
     * The <code>getPath()</code> method builds a path of compiler stages that
     * correspond to the input string. The input string is expected to contain
     * a comma-seperated list of strings, with each string representing a distinct
     * compiler pass as it aliases in the <code>Registry</code> class.
     * @param strs a string representing a comma separated list of stages
     * @return an array of <code>Stage</code> instances that represent
     * the specified compilatiobn path
     */
    public Stage[] getPath(String strs) {
        return getPath(strs.split(","));
    }

    /**
     * The <code>getPath()</code> method builds a path of compiler stages that
     * correspond to the input string. The input is an array of strings,
     * with each string representing a distinct
     * compiler pass as it aliases in the <code>Registry</code> class.
     * @param stra an array of aliases for compilation stages
     * @return an array of <code>Stage</code> instances that represent
     * the specified compilatiobn path
     */
    public Stage[] getPath(String[] stra) {
        Stage[] stages = new Stage[stra.length];
        int cntr = 0;
        for (String stg : stra) {
            stages[cntr++] = registry.getStage(stg);
        }
        return stages;
    }
    /**
     * The <code>findPath()</code> method searches the internal model of the
     * compiler for a path of compilation stages that transforms the program
     * in the source format into the destination format. The source format
     * and the destination format are given as strings that are used to
     * build <code>IRState</code> instances.
     * @param a the source representation of the program
     * @param b the destination representation of the program
     * @return an array of <code>Stage</code> objects that represent a path
     * from the source format to the destination format if such a path
     * exists; null if no such path exists
     */
    public Stage[] findPath(String a, String b) {
        return findPath(StageSignature.parseIRState(a), StageSignature.parseIRState(b));
    }

    /**
     * The <code>findPath()</code> method searches the internal model of the
     * compiler for a path of compilation stages that transforms the program
     * in the source format into the destination format. The source format
     * and the destination format are given as <code>IRState</code> instances.
     * @param sa the source representation of the program
     * @param sb the destination representation of the program
     * @return an array of <code>Stage</code> objects that represent a path
     * from the source format to the destination format if such a path
     * exists; null if no such path exists
     */
    public Stage[] findPath(IRState sa, IRState sb) {
        // add the required stages to the input for the last state
        extendOutput(sb);

        // initialize the visited set and the queue
        Set<IRState> visited = new HashSet<IRState>();
        Queue<Path> queue = initQueue(sa, visited);

        for ( Path path = queue.poll(); path != null; path = queue.poll()) {
            StageSignature sig = getSignature(path.node);
            IRState ns = sig.process(path.state);
            if ( sb.isMetBy(ns) ) {
                // we found a candidate node: build a path
                return path.buildReverse(registry);
            } else {
                // add all the successors to this node to the queue
                addSuccessors(path, ns, visited, queue);
            }
        }
        return null;
    }

    private void extendOutput(IRState sb) {
        if ( globalRequired != null )
            for ( String s : globalRequired ) sb.addPlusAttr("$"+s);
    }

    private StageSignature getSignature(String stage) {
        StageSignature sig = null;
        if ( specialSignatures != null ) sig = specialSignatures.get(stage);
        if ( sig == null) sig = registry.getSignature(stage);
        return sig;
    }

    private Queue<Path> initQueue(IRState start, Set<IRState> known) {
        Queue<Path> queue = new LinkedList<Path>();
        addSuccessors(null, start, known, queue);
        return queue;
    }

    private void addSuccessors(Path path, IRState input, Set<IRState> known, Queue<Path> queue) {
        for ( String stage : registry.stageList ) {
            StageSignature sig = getSignature(stage);
            if ( sig != null && sig.canAccept(input) ) {
                // queue this stage if it produces a previously unknown state
                IRState ns = sig.process(input);
                if ( !known.contains(ns) ) {
                    // if this is a new stage, add the node to be explored
                    queue.offer(new Path(path, stage, input));
                    known.add(ns);
                }
            }
        }
    }

    public static Stage[] getDefaultPath(String str) {
        return new Scheduler().getPath(str);
    }
}
