/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Apr 15, 2006
 */
package vpc.test;

import cck.test.TestEngine;
import cck.text.Terminal;
import cck.text.StringUtil;
import cck.util.*;
import java.io.IOException;

/**
 * The <code>Main</code> class represents the main driver for the testing harness.
 * It parses the command line options and creates a <code>AutomatedTester</code>
 * instance that instantiates and runs each test case from the files specified
 * on the command line.
 *
 * @author Ben L. Titzer
 */
public class Main {
    static final Options options = new Options();

    static final Option.Bool COLORS = options.newOption("colors", true,
            "This option controls the printing of terminal colors.");
    static final Option.Bool DETAIL = options.newOption("detail", false,
            "When this option is specified, the automated testing framework will " +
            "report the details of any test case failures, including stacktraces " +
            "of unexpected exceptions, state mismatches, etc.");
    static final Option.Bool PROGRESS = options.newOption("progress", false,
            "When this option is specified, the automated testing framework will " +
            "report the result of each test as it is run.");
    static final Option.Bool STAT = options.newOption("statistics", false,
            "When this option is specified, the automated testing framework will " +
            "report statistics for each test case.");
    static final Option.Long THREADS = options.newOption("threads", 1,
            "This option specifies the number of worker threads to use in running " +
                    "the test cases.");

    protected static final ClassMap harnesses = new ClassMap("Test Harness", TestEngine.Harness.class);

    static {
        harnesses.addClass("virgil-seman", VSTHarness.Seman.class);
        harnesses.addClass("virgil-parse", VSTHarness.Parse.class);
        harnesses.addClass("tir-init", TIRInitHarness.class);
        harnesses.addClass("tir-to-c", TIRtoCHarness.class);
        harnesses.addClass("scheduler", SchedulerHarness.class);
    }

    public static void main(String[] args) {

        boolean success = false;

        try {
            // parse the command line options
            args = parseOptions(args);

            // no files specified on command line
            if (args.length == 0) Util.userError("no input files");

            // if all of the files exist, run the tests
            if (Util.verifyFilesExist(args)) {
                success = runTests(args);
            }

        } catch (Util.Error e) {
            e.report();
        } catch (Throwable t) {
            Terminal.print(Terminal.ERROR_COLOR, "Unexpected exception");
            Terminal.println(": " + t.getClass());
            t.printStackTrace();
        }
        if ( success ) System.exit(0);
        else System.exit(1);
    }

    public static String[] parseOptions(String[] args) {
        options.parseCommandLine(args);
        Terminal.useColors = COLORS.get();
        TestEngine.LONG_REPORT = DETAIL.get();
        TestEngine.PROGRESS_REPORT = PROGRESS.get();
        TestEngine.STATISTICS = STAT.get();
        TestEngine.THREADS = (int)THREADS.get();
        return options.getArguments();
    }

    private static boolean runTests(String[] files) throws IOException {
        return new TestEngine(harnesses).runTests(files);
    }
}
