/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Dec 14, 2006
 */

package vpc.test;

import cck.test.*;
import cck.text.StringUtil;
import vpc.core.Program;
import vpc.model.*;
import java.io.IOException;
import java.util.*;

/**
 * The <code>SchedulerHarness</code> class implements a test harness for
 * the scheduler that computes the sequences of optimizations and transformations
 * to apply to a program. This class implements a mock registry that allows
 * the core of the scheduler algorithm to be tested.
 *
 * @author Ben L. Titzer
 */
public class SchedulerHarness implements TestEngine.Harness {

    /**
     * The <code>newTestCase()</code> method creates a new test case of the right type given the file name and
     * the properties already extracted from the file by the testing framework.
     *
     * @param fname the name of the file
     * @param props a list of properties extracted from the file
     * @return an instance of the <code>TestCase</code> class
     * @throws Exception if there is a problem creating the testcase or reading it
     */
    public TestCase newTestCase(String fname, Properties props) throws Exception {
        return new SchedulerTestCase(fname, props);
    }

    static class MockStage extends Stage {

        public void visitProgram(Program p) throws IOException {
            // do nothing.
        }
    }

    public class SchedulerTestCase extends TestCase {
        Stage[] result;
        Stage[] expect;
        String expectString;
        String requiredString;
        protected final String input;
        protected final String output;

        public SchedulerTestCase(String fname, Properties props) {
            super(fname, props);
            input = expectProperty("Input");
            output = expectProperty("Output");
            expectString = expectProperty("Result");
            requiredString = props.getProperty("Required");
        }

        public void run() {
            Registry registry = extractRegistry();
            Collection<String> req = null;
            if ( requiredString != null ) {
                req = (List<String>)StringUtil.toList(requiredString);
            }
            Scheduler scheduler = new Scheduler(registry, req);
            if ( expectString.length() > 0 ) {
                expect = scheduler.getPath(expectString);
            }
            result = scheduler.findPath(input, output);
        }

        public TestResult match(Throwable t) {
            if (t == null) {
                // if no exception
                if (result == null) {
                    // if no path found
                    if (expect == null) return new TestResult.TestSuccess();
                    return new TestResult.TestFailure("no path found, expected " + expectString);
                }
                // no exception, path was found
                if (expect == null) return new TestResult.TestFailure("no path expected");

                // path lengths do not match
                if (result.length != expect.length) {
                    return new TestResult.TestFailure("path lengths do not match");
                }

                // check each component of the paths
                for (int cntr = 0; cntr < result.length; cntr++) {
                    if (!result[cntr].equals(expect[cntr]))
                        return new TestResult.TestFailure("found wrong path");
                }
            }
            // default: return the result of super matching
            return super.match(t);
        }

        private Registry extractRegistry() {
            Registry reg = new Registry();

            for ( Map.Entry e : properties.entrySet() ) {
                String key = (String)e.getKey();
                String sig = (String)e.getValue();

                if ( key.startsWith("$") ) {
                    String name = key.substring(1, key.length());
                    reg.addStage(name, sig, new MockStage());
                }
            }
            return reg;
        }
    }
}
