/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Jan 4, 2006
 */

package vpc.test;

import cck.test.*;
import cck.text.StringUtil;
import vpc.core.*;
import vpc.model.*;
import vpc.tir.stages.HeapParser;
import java.io.BufferedReader;
import java.io.FileReader;
import java.util.HashMap;
import java.util.Properties;

/**
 * The <code>TIRInitHarness</code> class implements a test harness for running the initializers
 * and building an initial object heap. The test case file contains a heap with records that
 * contains fields with values. This test harness compares records on a field-by-field basis
 * to check that the two heaps match each other exactly.
 *
 * @author Ben L. Titzer
 */
public class TIRInitHarness implements TestEngine.Harness {
    static String stages = "virgil,init";

    private static class HeapMismatch extends Exception {
        final TestResult result;

        HeapMismatch(TestResult r) {
            result = r;
        }
    }

    class InitTest extends TestCase.ExpectSourceError {

        Program program;
        Heap expected;

        InitTest(String fname, Properties props) {
            super(fname, props);
        }

        public void run() throws Exception {
            program = new Program(filename);
            Stage[] path = Scheduler.getDefaultPath(stages);
            Compilation c = new Compilation(program, path, null);
            c.addFile(filename);
            c.run();
            HeapParser parser = new HeapParser(program);
            expected = parser.parseHeap(new BufferedReader(new FileReader(filename)));
        }

        protected TestResult checkPass() {
            try {
                compareHeaps(program.heap, expected);
                return new TestResult.TestSuccess();
            } catch (HeapMismatch e) {
                return e.result;
            }
        }

        void compareHeaps(Heap h1, Heap eh) throws HeapMismatch {
            HashMap<Integer, Heap.Record> map1 = getRecordMap(h1);

            // compare the sizes of the heaps (number of records)
            int h1sz = map1.size();
            int h2sz = getRecordMap(eh).size();
            if (h1sz != h2sz)
                fail("heap sizes do not match" + " (" + h2sz + " != " + h1sz + ")");

            // check every record in the expected heap
            for (Heap.Record er : eh.getRecords()) {
                Heap.Record r = map1.get(er.uid);

                // is the record present in the program heap?
                if (r == null) fail("record " + er + " is not present");

                // compare layout names
                compareLayouts(r, r.layout, er.layout);

                // compare the sizes of the records
                int sz1 = r.getSize();
                int sz2 = er.getSize();
                if (sz1 != sz2) fail("record " + r + " size incorrect" + " (" + sz2 + " != " + sz1 + ")");

                // compare each field individually
                for (int cntr = 0; cntr < r.getSize(); cntr++) {
                    String name = r.layout.getCell(cntr).getName();
                    compareValues(r, name, r.getValue(cntr), er.getValue(cntr));
                }
            }
        }

        private void compareLayouts(Heap.Record r, Heap.Layout l, Heap.Layout el) throws HeapMismatch {
            if (l == null) fail("record " + r + " has null layout ");
            String ls = l.toString();
            String els = el.toString();
            if (!ls.equals(els))
                fail("record " + r + " has layout " + StringUtil.quote(l) + ", expected " + StringUtil.quote(els));
        }

        private void compareValues(Heap.Record r, String name, Value v1, Value v2) throws HeapMismatch {
            if (v1 == null) fail("record " + r + ", field \"" + name + "\" has null value ");
            if (!v1.equals(v2))
                fail("record " + r + ", field \"" + name + "\" has value " + qv(v1) + ", expected " + qv(v2));
        }

        private String qv(Value v1) {
            return StringUtil.quote(v1);
        }

        private void fail(String msg) throws HeapMismatch {
            throw new HeapMismatch(new TestResult.TestFailure(msg));
        }

        HashMap<Integer, Heap.Record> getRecordMap(Heap h) {
            HashMap<Integer, Heap.Record> map = new HashMap<Integer, Heap.Record>();
            for (Heap.Record r : h.getRecords())
                map.put(r.uid, r);
            return map;
        }
    }

    public TestCase newTestCase(String fname, Properties props) throws Exception {
        return new InitTest(fname, props);
    }
}
