/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Jan 30, 2006
 */

package vpc.test;

import cck.test.*;
import cck.text.StringUtil;
import cck.text.Terminal;
import cck.util.Options;
import cck.util.Util;
import vpc.core.Program;
import vpc.core.ProgramDecl;
import vpc.model.*;
import vpc.vst.parser.Token;
import java.io.IOException;
import java.text.CharacterIterator;
import java.text.StringCharacterIterator;
import java.util.*;

/**
 * The <code>TIRtoCHarness</code> class implements a test harness for the TIR to C translation.
 * The test harness checks that the generated C code compiles correctly (no errors) and
 * that when linked with a small main stub function and executed on the workstation, the
 * program actually computes the expected value.
 *
 * @author Ben L. Titzer
 */
public class TIRtoCHarness implements TestEngine.Harness {

    public static int ABORT_TYPE_CODE = 127;
    public static int ABORT_NULL_CODE = 126;
    public static int ABORT_BOUNDS_CODE = 125;
    public static int ABORT_DIV_CODE = 124;

    public static HashMap<String, Integer> exceptionCodes = new HashMap<String, Integer>();

    static {
        exceptionCodes.put("TypeCheckException", ABORT_TYPE_CODE);
        exceptionCodes.put("NullCheckException", ABORT_NULL_CODE);
        exceptionCodes.put("BoundsCheckException", ABORT_BOUNDS_CODE);
        exceptionCodes.put("DivideByZeroException", ABORT_DIV_CODE);
    }

    static class Run {
        int input;
        int expected;
        String name;
        int result;

        void parse(CharacterIterator i) {
           StringUtil.skipWhiteSpace(i);
           int val = StringUtil.readDecimalValue(i, 12);
           StringUtil.skipWhiteSpace(i);
           if (StringUtil.peekAndEat(i, '=')) {
               input = val;
               StringUtil.skipWhiteSpace(i);
               if (Character.isDigit(i.current())) {
                   expected = StringUtil.readDecimalValue(i, 12);
                   name = String.valueOf(expected);
               } else {
                   String except = StringUtil.readIdentifier(i);
                   Integer integer = exceptionCodes.get(except);
                   if (integer == null) throw Util.failure("unknown exception type: " + except);
                   expected = integer;
                   name = except;
               }
               StringUtil.skipWhiteSpace(i);
           } else {
               expected = val;
           }
        }
    }

    /**
     * The <code>TIRtoCTest</code> class implements a test case that checks the
     * semantic checker for the Virgil programming language. Each test case contains
     * one program and an expected that result that is either "PASS" or the name and
     * location of the semantic error that the virgil semantic checker should detect.
     */
    static class TIRtoCTest extends TestCase {
        private static String stages = "virgil,init,tir-opt0,tir-emit-c";

        private boolean compiled;

        private List<Run> runs;

        String outputFile;
        String binaryFile;
        protected String progpath;
        protected String progname;
        protected long vpcTime;
        protected long gccTime;
        protected long runTime;

        TIRtoCTest(String fname, Properties props) {
            super(fname, props);
            runs = parseRuns(expectProperty("Result"));

            progname = StringUtil.baseFileName(fname);
            progpath = "/tmp/" + progname;
            outputFile = progpath + ".c";
            binaryFile = progpath + ".bin";
        }

        private List<Run> parseRuns(String result) {
            List<Run> list = new LinkedList<Run>();
            CharacterIterator i = new StringCharacterIterator(result);
            while (true) {
                Run run = new Run();
                run.parse(i);
                list.add(run);
                if (!StringUtil.peekAndEat(i, ',')) break;
            }
            return list;
        }

        public void run() throws Exception {
            emitProgram();
            int val = runCC();
            if (val == 0) executeRuns();
        }

        private void executeRuns() throws IOException, InterruptedException {
            runTime = System.currentTimeMillis();
            try {
                compiled = true;
                for (Run r : runs) {
                    String[] command = {binaryFile, Integer.toString(r.input)};
                    Process test = Runtime.getRuntime().exec(command);
                    r.result = test.waitFor();
                }
            } finally {
                runTime = System.currentTimeMillis() - runTime;
            }
        }

        private int runCC() throws IOException, InterruptedException {
            gccTime = System.currentTimeMillis();
            try {
                String[] command = {"cc", "-o", binaryFile, outputFile};
                Process compiler = Runtime.getRuntime().exec(command);
                return compiler.waitFor();
            } finally {
                gccTime = System.currentTimeMillis() - gccTime;
            }
        }

        private void emitProgram() throws IOException {
            vpcTime = System.currentTimeMillis();
            try {
                Program program = new Program(progpath);
                addMainEntryPoint(program);
                Stage[] path = Scheduler.getDefaultPath(stages);
                Compilation c = new Compilation(program, path, getOptions());
                c.addFile(filename);
                c.run();
            } finally {
                vpcTime = System.currentTimeMillis() - vpcTime;
            }
        }

        private void addMainEntryPoint(Program program) {
            program.programDecl = new ProgramDecl(new Token(progname));
            Token mn = new Token("main");
            Token c = new Token(progname);
            Token m = new Token("main");
            ProgramDecl.EntryPoint n = new ProgramDecl.EntryPoint(mn, c, m);
            program.programDecl.addEntryPoint(n);
        }

        private Options getOptions() {
            Properties p = new Properties();
            p.setProperty("linkage", "c:test");
            Options o = new Options();
            o.process(p);
            return o;
        }

        /**
         * The <code>match()</code> method of a test case is called after the test case completes. If
         * the <code>run()</code> method throws an exception (Throwable), this method will be passed
         * the exception that was generated. If no exception was generated, then the exception passed
         * will be null. This method should return a new instance of the <code>TestResult</code> class
         * that represents the results of the test case.
         *
         * @param t the exception (if any) that was thrown while running the test case
         * @return a new <code>TestResult</code> indicating success or error
         */
        public TestResult match(Throwable t) {
            // default behavior: no exception = pass
            if (t == null) {
                if (!compiled) return new TestResult.TestFailure("program did not compile correctly");
                for (Run r : runs) {
                    if (r.result != r.expected)
                        return new TestResult.TestFailure("program(" + r.input + ") = " + r.result + ", expected " + r.name);
                }
                return new TestResult.TestSuccess();
            }

            // internal error encountered.
            if (t instanceof Util.InternalError) return new TestResult.InternalError((Util.InternalError) t);

            // default: unexpected exception
            return new TestResult.UnexpectedException(t);
        }

        public void reportStatistics() {
            Terminal.print(result.getColor(), getFileName());
            Terminal.print(": ");
            Terminal.print(StringUtil.rightJustify(StringUtil.milliToSecs(vpcTime), 6)+" ");
            Terminal.print(StringUtil.rightJustify(StringUtil.milliToSecs(gccTime), 6)+" ");
            Terminal.print(StringUtil.rightJustify(StringUtil.milliToSecs(runTime), 6)+"  (vpc, gcc, run)");
            Terminal.nextln();
        }

    }

    /**
     * The <code>newTestCase()</code> method creates a new test case of the right type given the file name and
     * the properties already extracted from the file by the testing framework.
     *
     * @param fname the name of the file
     * @param props a list of properties extracted from the file
     * @return an instance of the <code>TestCase</code> class
     * @throws Exception if there is a problem creating the testcase or reading it
     */
    public TestCase newTestCase(String fname, Properties props) throws Exception {
        return new TIRtoCTest(fname, props);
    }

}
