/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Sep 11, 2005
 */
package vpc.test;

import cck.test.TestCase;
import cck.test.TestEngine;
import vpc.core.Program;
import vpc.model.*;
import vpc.types.TypeCache;
import vpc.vst.parser.VirgilParser;
import java.io.FileInputStream;
import java.util.Properties;

/**
 * The <code>VSTHarness</code> class implements a test harness
 * for virgil semantic tests. This class implements a harness that loads,
 * parses, and semantically checks each test case as a separate program.
 * Each program contains a description of what error(s) it expects the
 * semantic analyzer to generate, and this test harness compares the
 * expected results against the actual results to determine whether the
 * test case passed.
 *
 * @author Ben L. Titzer
 */
public class VSTHarness {

    /**
     * The <code>SemanTest</code> class implements a test case that checks the
     * semantic checker for the Virgil programming language. Each test case contains
     * one program and an expected that result that is either "PASS" or the name and
     * location of the semantic error that the virgil semantic checker should detect.
     */
    static class SemanTest extends TestCase.ExpectSourceError {
        private static String stages = "virgil-parse,virgil-gtb,virgil-tc";

        SemanTest(String fname, Properties props) {
            super(fname, props);
        }

        public void run() throws Exception {
            Program program = new Program("test");
            Stage[] path = Scheduler.getDefaultPath(stages);
            Compilation c = new Compilation(program, path, null);
            c.addFile(filename);
            c.run();
        }
    }

    /**
     * The <code>ParseTest</code> class implements a test case that checks the
     * parser for the Virgil programming language. Each test case contains
     * one program and an expected that result that is either "PASS" or the
     * location of the parse error.
     */
    static class ParseTest extends TestCase.ExpectSourceError {

        ParseTest(String fname, Properties props) {
            super(fname, props);
        }

        public void run() throws Exception {
            VirgilParser.parseModule(new FileInputStream(filename), filename, new TypeCache());
        }
    }

    public static class Seman implements TestEngine.Harness {
        public TestCase newTestCase(String fname, Properties props) throws Exception {
            return new SemanTest(fname, props);
        }
    }

    public static class Parse implements TestEngine.Harness {
        public TestCase newTestCase(String fname, Properties props) throws Exception {
            return new ParseTest(fname, props);
        }
    }
}
