/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc.tir;

import vpc.core.concept.PrimVoid;

import java.util.LinkedList;
import java.util.List;

/**
 * The <code>TIRBlock</code> class represents an ordered list of statements that comprise
 * a block of code in the program.
 *
 * @author Ben L. Titzer
 */
public class TIRBlock extends TIRExpr {

    public final String label;
    public final List<TIRExpr> list;

    public TIRBlock(String l) {
        label = l;
        list = new LinkedList<TIRExpr>();
        setType(PrimVoid.TYPE); // empty block has type void
    }

    /**
     * The <code>addExpr</code> method is used to add an expression to the end of
     * this block. It simply allocates a node and puts it on the end of the list.
     *
     * @param e the expression to add to the block
     */
    public void addExpr(TIRExpr e) {
        list.add(e);
        setType(e.getType());
    }

    public <R, E> R accept(TIRExprVisitor<R, E> v, E env) {
        return v.visit(this, env);
    }

    public String toString() {
        StringBuffer buf = new StringBuffer("block \"" + label + "\" {");
        for (TIRExpr e : list) {
            buf.append(e);
            buf.append(";\n");
        }
        buf.append("}\n");
        return buf.toString();
    }

    public boolean isEmpty() {
        return list.isEmpty();
    }

    public boolean isSingleton() {
        return list.size() == 1;
    }

    public TIRExpr getFirst() {
        if (list.isEmpty() ) return null;
        return list.get(0);
    }

    /**
     * The <code>Continue</code> class represents a continue statement that returns to the
     * beginning of a TIRBlock.
     */
    public static class Continue extends TIRExpr {
        public String toString() {
            return "continue";
        }

        public <R, E> R accept(TIRExprVisitor<R, E> v, E env) {
            return v.visit(this, env);
        }
    }

    /**
     * The <code>Break</code> class represents a break statement that exits to the end
     * of a TIRBlock.
     */
    public static class Break extends TIRExpr {
        public String toString() {
            return "break";
        }

        public <R, E> R accept(TIRExprVisitor<R, E> v, E env) {
            return v.visit(this, env);
        }
    }
}
