/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Jan 20, 2006
 */

package vpc.tir;

import cck.parser.SourcePoint;
import cck.parser.StackTrace;
import cck.parser.SourceException;
import cck.text.StringUtil;
import cck.text.Terminal;
import cck.util.Util;
import vpc.core.*;
import vpc.core.concept.*;
import vpc.core.virgil.VirgilComponent;
import vpc.core.virgil.VirgilDelegate;
import vpc.core.virgil.VirgilArray;
import vpc.types.Type;
import vpc.tir.expr.Operator;

/**
 * The <code>TIRInterpreter</code> class implements an interpreter for evaluating
 * expressions in the TIR representation (i.e. <code>TIRExpr</code> instances). This
 * interpreter implements the semantics of all the basic TIR constructs and relies
 * on the behavior of <code>Operator.apply()</code> for the functionality of
 * unknown operators.
 *
 * @author Ben L. Titzer
 * @see Operator
 */
public class TIRInterpreter extends TIRExprVisitor<Value, TIRInterpreter.Frame> {

    public static final TIRExpr[] NOARGS = new TIRExpr[0];
    public static final Value[] NONE = new Value[0];
    public static final PrimBool.Val TRUE = PrimBool.toValue(true);
    public static final PrimBool.Val FALSE = PrimBool.toValue(false);
    public static final PrimInt32.Val ZERO_INT = PrimInt32.toValue(0);
    public static final PrimChar.Val ZERO_CHAR = PrimChar.toValue((char) 0);

    public static class Frame extends StackTrace {
        protected final Value[] locals;

        public Frame(Frame p, String name, int size) {
            super(p, name);
            locals = new Value[size];
        }
    }

    private static class LoopExit extends RuntimeException {
    }

    private static class LoopReentry extends RuntimeException {
    }

    private static class MethodReturn extends RuntimeException {
        Value value;
    }

    protected final LoopExit LOOPEXIT = new LoopExit();
    protected final LoopReentry LOOPREENTRY = new LoopReentry();
    protected final MethodReturn RETURN = new MethodReturn();

    protected final Program program;
    protected final boolean trace;

    protected Frame frame;
    protected int indent;

    public TIRInterpreter(Program p, boolean tflag) {
        program = p;
        trace = tflag;
    }

    public Value eval(TIRExpr e) {
        return e.accept(this, frame);
    }

    public Value[] eval(TIRExpr[] e) {
        if (e.length == 0) return NONE;
        Value[] vs = new Value[e.length];
        for (int cntr = 0; cntr < e.length; cntr++)
            vs[cntr] = eval(e[cntr]);
        return vs;
    }

    private Reference.Val eval(Heap.Record rec) {
        return Reference.toValue(rec);
    }

    public Value invokeComponentMethod(Method m, Value[] args) {
        SourcePoint pt = program.getDefaultSourcePoint();
        VirgilComponent c = (VirgilComponent)m.getCompoundDecl();
        Heap.Record rec = c.getRecord();
        return invoke(pt, m, rec, args);
    }

    private static Type typeOf(Heap.Record rec) {
        return rec.layout.type;
    }

    public static Value getDefaultValue(Type t) {
        if (t.isReference()) return Reference.NULL_VALUE;
        if (t.isVoid()) return PrimVoid.VALUE;
        if (t == PrimInt32.TYPE) return ZERO_INT;
        if (t == PrimBool.TYPE) return FALSE;
        if (t == PrimChar.TYPE) return ZERO_CHAR;
        if (t instanceof PrimRaw.IType ) {
            PrimRaw.IType rt = (PrimRaw.IType)t;
            return PrimRaw.toValue(rt.width, 0);
        }
        else throw Util.failure("no default value for type " + StringUtil.quote(t));
    }

    public Value visit(TIRBlock st, Frame frame) {
        Value last = PrimVoid.VALUE;
        for (TIRExpr e : st.list)
            last = eval(e);
        return last;
    }

    public Value visit(TIRNop n, Frame frame) {
        return PrimVoid.VALUE;
    }

    public Value visit(TIROperator o, Frame frame) {
        try {
            Value[] vals = eval(o.operands);
            try {
                return o.operator.apply(vals);
            } catch (VirgilComponent.InitCheckException e) {
                // retry operator (once only) if failure due to init dependency
                initComponent(e.component);
                return o.operator.apply(vals);
            }
        } catch (Operator.Exception e) {
            throw new SourceException(e.type, makeStackTrace(o.getSourcePoint(), frame), e.msg, StringUtil.EMPTY_STRING_ARRAY);
        }
    }

    public Value visit(TIRBlock.Break b, Frame frame) {
        throw LOOPEXIT;
    }

    public Value visit(TIRBlock.Continue c, Frame frame) {
        throw LOOPREENTRY;
    }

    public Value visit(TIRIfExpr s, Frame frame) {
        Value val = eval(s.condition);
        if (PrimBool.fromValue(val)) return eval(s.true_target);
        else return eval(s.false_target);
    }

    public Value visit(TIRLoop.For l, Frame frame) {
        eval(l.initial);
        while (PrimBool.fromValue(eval(l.condition))) {
            try {
                eval(l.body);
            } catch (LoopExit e) {
                break;
            } catch (LoopReentry e) {
                // do nothing.
            }
            eval(l.update);
        }
        return PrimVoid.VALUE;
    }

    public Value visit(TIRLoop.DoWhile l, Frame frame) {
        do {
            try {
                eval(l.body);
            } catch (LoopExit e) {
                break;
            } catch (LoopReentry e) {
                // do nothing.
            }
        } while (PrimBool.fromValue(eval(l.condition)));
        return PrimVoid.VALUE;
    }

    public Value visit(TIRLoop.While l, Frame frame) {
        while (PrimBool.fromValue(eval(l.condition))) {
            try {
                eval(l.body);
            } catch (LoopExit e) {
                break;
            } catch (LoopReentry e) {
                // do nothing.
            }
        }
        return PrimVoid.VALUE;
    }

    public Value visit(TIRReturn r, Frame frame) {
        RETURN.value = eval(r.value); // evaluate return value
        throw RETURN;
    }

    public Value visit(TIRThrow r, Frame frame) {
        throw new SourceException(r.exception, makeStackTrace(r.point, frame), "", StringUtil.EMPTY_STRING_ARRAY);
    }

    public Value visit(TIRSwitch s, Frame frame) {
        boolean eval = false;
        Object val = eval(s.expr);
        // linear search for now; could be hashtable or array
        for (TIRSwitch.Case c : s.cases) {
            for (TIRConst v : c.value) {
                if (val.equals(eval(v))) {
                    eval(c.body);
                    eval = true;
                }
            }
        }
        // no match; execute default case
        if (!eval && s.defcase != null) eval(s.defcase.body);
        return PrimVoid.VALUE;
    }

    public Value visit(TIRCall d, Frame frame) {
        Value val = eval(d.func);
        SourcePoint callpt = d.getSourcePoint();
        if (Reference.isNull(val) || VirgilDelegate.isNull(val)) {
            throw new SourceException("NullCheckException",
                    makeStackTrace(d.getSourcePoint(), frame),
                    "null check exception",
                    StringUtil.EMPTY_STRING_ARRAY);
        }
        VirgilDelegate.Val m = VirgilDelegate.fromValue(val);

        return invoke(callpt, m.method, m.record, d.arguments);
    }

    public Value visit(TIRLocal.Set l, Frame frame) {
        // a = value;
        Value val = eval(l.value);
        frame.locals[l.temp.id] = val;
        return val;
    }

    public Value visit(TIRLocal.Get l, Frame frame) {
        return frame.locals[l.temp.id];
    }

    public Value visit(TIRConst.Value c, Frame frame) {
        return c.value;
    }

    public Value visit(TIRConst.String c, Frame frame) {
        char[] chs = StringUtil.getStringChars(c.value());
        VirgilArray.IType at = (VirgilArray.IType)c.getType();
        Heap.Layout layout = VirgilArray.getArrayLayout(program.heap, at, chs.length);
        Heap.Record rec = layout.newInstance();
        for (int cntr = 0; cntr < chs.length; cntr++) {
            rec.setValue(cntr, PrimChar.toValue(chs[cntr]));
        }
        return Reference.toValue(rec);
    }

    public Value visit(TIRCompare.Equal c, Frame frame) {
        Value l = eval(c.left);
        Value r = eval(c.right);
        return l.equals(r) ? TRUE : FALSE;
    }

    public Value visit(TIRCompare.NotEqual c, Frame frame) {
        Value l = eval(c.left);
        Value r = eval(c.right);
        return l.equals(r) ? FALSE : TRUE;
    }

    public Value visit(TIRExpr e, Frame frame) {
        throw fail("eval(" + e.getClass() + ") unimplemented", e);
    }

    public Heap.Record initComponent(VirgilComponent comp) {
        Heap.Record compRecord = comp.getRecord();
        if (compRecord == null) {
            compRecord = buildComponentRecord(comp);
            comp.setRecord(compRecord);
            Constructor constructor = comp.getConstructor();
            if (constructor != null) {
                invoke(null, constructor, compRecord, NOARGS);
            }
        }
        return compRecord;
    }

    private Heap.Record buildComponentRecord(VirgilComponent comp) {
        Heap.Layout layout = program.closure.getLayout(comp);
        if (layout == null)
            throw fail("No layout for component " + comp.getName());
        Heap.Record compRecord = layout.newInstance();
        compRecord.setRoot(true);
        return compRecord;
    }

    private Value invoke(SourcePoint pt, Method m, Heap.Record rec, Value[] args) {
        TIRRep body = TIRUtil.getRep(m);
        Frame nf = newFrame(pt, m.getFullName(), body);
        initFrame(nf, rec, args);
        return trace_call(body, nf);
    }

    private Value invoke(SourcePoint callpt, Method m, Heap.Record rec, TIRExpr[] args) {
        TIRRep body = TIRUtil.getRep(m);
        Frame nf = newFrame(callpt, m.getFullName(), body);
        initFrame(nf, rec, args);
        return trace_call(body, nf);
    }

    private Value trace_call(TIRRep meth, Frame nf) {
        if ( trace ) {
            Terminal.print(StringUtil.space(indent*4));
            Terminal.println("-> "+nf.method);
            indent++;
        }
        Value value = call(meth, nf);
        if ( trace ) {
            indent--;
            Terminal.print(StringUtil.space(indent*4));
            if ( value == PrimVoid.VALUE ) Terminal.println("<- "+nf.method);
            else Terminal.println("<- "+nf.method+" = "+value);
        }
        return value;
    }

    private Value call(TIRRep meth, Frame nf) {
        try {
            frame = nf;
            eval(meth.getBody());
            return PrimVoid.VALUE;
        } catch (MethodReturn r) {
            return r.value;
        } finally {
            frame = (Frame) frame.prev;
        }
    }

    private void initFrame(Frame nf, Heap.Record this_, Value[] args) {
        nf.locals[0] = eval(this_);
        System.arraycopy(args, 0, nf.locals, 1, args.length);
    }

    private void initFrame(Frame nf, Heap.Record this_, TIRExpr[] args) {
        nf.locals[0] = eval(this_);
        for (int cntr = 0; cntr < args.length; cntr++) {
            nf.locals[cntr + 1] = eval(args[cntr]);
        }
    }

    private Frame newFrame(SourcePoint callpt, String name, TIRRep meth) {
        if (frame != null) frame.setSourcePoint(callpt);
        return new Frame(frame, name, meth.getNumberOfTemps());
    }

    private StackTrace makeStackTrace(SourcePoint p, Frame frame) {
        frame.setSourcePoint(p);
        return frame;
    }

    private SourceException fail(String msg, TIRExpr e) {
        return fail(msg, e.getSourcePoint());
    }

    private SourceException fail(String msg) {
        return fail(msg, (SourcePoint)null);
    }

    private SourceException fail(String msg, SourcePoint pt) {
        return new SourceException("Failure", makeStackTrace(pt, frame), msg, StringUtil.EMPTY_STRING_ARRAY);
    }
}
