/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc.tir;

import cck.text.StringUtil;
import vpc.core.concept.Method;

/**
 * The <code>TIRLocal</code> class and its inner classes represent reads
 * and writes to local variables in TIR code.
 *
 * @author Ben L. Titzer
 */
public abstract class TIRLocal extends TIRExpr {

    public final Method.Temporary temp;

    protected TIRLocal(Method.Temporary t) {
        temp = t;
    }

    /**
     * The <code>Local</code> inner class represents a read from a local
     * in the original program. It contains a <code>Method.Temporary</code>
     * that indicates which temporary to read.
     */
    public static class Get extends TIRLocal {
        public Get(Method.Temporary t) {
            super(t);
            setType(t.getType());
        }

        public <R, E> R accept(TIRExprVisitor<R, E> v, E env) {
            return v.visit(this, env);
        }

        public String toString() {
            return StringUtil.embed("read_local", temp);
        }
    }

    /**
     * The <code>Local</code> inner class represents a write to a local
     * variable in the program. It contains a <code>Method.Temporary</code>
     * that represents the temporary to write.
     */
    public static class Set extends TIRLocal {
        public final TIRExpr value;

        public Set(Method.Temporary t, TIRExpr val) {
            super(t);
            value = val;
            setType(t.getType());
        }

        public String toString() {
            return StringUtil.embed("write_local", temp, value);
        }

        public <R, E> R accept(TIRExprVisitor<R, E> v, E env) {
            return v.visit(this, env);
        }
    }
}
