/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc.tir;


/**
 * The <code>Switch</code> class represents a multi-way branch in
 * the program. Multi-way branches are restricted so that each case
 * must be of a constant, compile-time computable value.
 *
 * @author Ben L. Titzer
 */
public class TIRSwitch extends TIRExpr {
    public final TIRExpr expr;
    public final Case[] cases;
    public final Case defcase;

    /**
     * The <code>Case</code> inner class represents a case within
     * the a switch statement. It contains a <code>Const</code> that
     * represents that value of the case statement as well as a
     * <code>Block</code> that represents the body of the case.
     */
    public static class Case {
        public final TIRConst.Value[] value;
        public final TIRExpr body;

        public Case(TIRConst.Value[] v, TIRExpr b) {
            value = v;
            body = b;
        }
    }

    public TIRSwitch(TIRExpr e, Case[] c, Case def) {
        expr = e;
        cases = c;
        defcase = def;
    }

    public <R, E> R accept(TIRExprVisitor<R, E> v, E env) {
        return v.visit(this, env);
    }

    public String toString() {
        StringBuffer buf = new StringBuffer("switch(");
        buf.append(expr);
        buf.append(") {");
        for (Case c : cases) {
            buf.append(TIRUtil.exprsToString(c.value));
            buf.append(' ');
            buf.append(c.body);
            buf.append("; ");
        }
        if (defcase != null) {
            buf.append("[] ");
            buf.append(defcase.body);
        }
        buf.append("}");
        return buf.toString();
    }
}