/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Jan 29, 2006
 */
package vpc.tir;

import cck.parser.SourcePoint;
import cck.text.StringUtil;
import cck.util.Util;
import vpc.core.Value;
import vpc.core.Heap;
import vpc.core.concept.*;
import vpc.tir.expr.Operator;
import vpc.types.Type;
import java.util.List;

/**
 * The <code>TIRUtil</code> class contains a collection of utility methods for
 * working with code in the TIR format.
 *
 * @author Ben L. Titzer
 */
public class TIRUtil {

    public static final String THIS_PARAM = "__this";

    /**
     * The <code>getRep()</code> method gets the TIR method representation for a given
     * method declaration by looking it up in the representation map.
     *
     * @param method the method for which to retrieve the TIR code
     * @return a reference to the <code>TIRRep</code> instances that represents
     *         the code for the specified method
     */
    public static TIRRep getRep(Method method) {
        return (TIRRep) method.getMethodRep(TIRRep.REP_NAME);
    }

    public static String exprsToString(TIRExpr[] args) {
        StringBuffer b = new StringBuffer("[");
        for (int cntr = 0; cntr < args.length; cntr++) {
            if (cntr != 0) b.append(", ");
            b.append(args[cntr].toString());
        }
        b.append("]");
        return b.toString();
    }

    public static TIRLocal.Get lift(TIRRep method, Method.Temporary temp, TIRExpr e, TIRBlock block) {
        if ( e instanceof TIRLocal.Set ) {
            // this statement is already an assignment; lift and return the temp
            block.addExpr(e);
            TIRLocal.Get rd = new TIRLocal.Get(((TIRLocal.Set)e).temp);
            copy(e, rd);
            return rd;
        } else {
            // this expression requires a new temporary (if one is not supplied)
            if ( temp == null )
                temp = method.newTemporary(e.getType());
            TIRLocal.Set wr = new TIRLocal.Set(temp, e);
            copy(e, wr);
            block.addExpr(wr);
            TIRLocal.Get rd = new TIRLocal.Get(temp);
            copy(e, rd);
            return rd;
        }
    }

    public static TIROperator nullCheck(TIRExpr e) {
        Reference.NullCheck check = new Reference.NullCheck(e.getType());
        TIROperator op = $OP(check, e);
        op.setType(e.getType());
        op.setSourcePoint(e.getSourcePoint());
        return op;
    }

    public static Util.Error fail(String msg, TIRExpr e) {
        String nstr = StringUtil.quote(e.getClass());
        SourcePoint sourcePoint = e.getSourcePoint();
        return Util.failure("TIRBuilder failure: " + msg + " on node " + nstr + " @ " + sourcePoint);
    }

    public static <R extends TIRExpr> R copy(TIRExpr oe, R ne) {
        ne.setType(oe.getType());
        ne.setSourcePoint(oe.getSourcePoint());
        return ne;
    }

    public static TIRLocal.Set dup(TIRLocal.Set e, TIRExpr v) {
        return copy(e, new TIRLocal.Set(e.temp, v));
    }

    public static TIRLocal.Get dup(TIRLocal.Get e) {
        return copy(e, new TIRLocal.Get(e.temp));
    }

    public static TIRNop dup(TIRNop e) {
        return copy(e, new TIRNop());
    }

    public static TIROperator dup(TIROperator e, TIRExpr... o) {
        return copy(e, new TIROperator(e.operator, o));
    }

    public static TIRCompare.Equal dup(TIRCompare.Equal e, TIRExpr left, TIRExpr right) {
        return copy(e, new TIRCompare.Equal(left, right));
    }

    public static TIRCompare.NotEqual dup(TIRCompare.NotEqual e, TIRExpr left, TIRExpr right) {
        return copy(e, new TIRCompare.NotEqual(left, right));
    }

    public static TIRConst.Value dup(TIRConst.Value e) {
        return copy(e, new TIRConst.Value(e.value));
    }

    public static TIRConst.String dup(TIRConst.String e) {
        return copy(e, new TIRConst.String(e.orig, e.value));
    }

    public static TIRCall dup(TIRCall e, TIRExpr f, TIRExpr... a) {
        return copy(e, new TIRCall(f, a));
    }

    public static TIRReturn dup(TIRReturn e, TIRExpr v) {
        return copy(e, new TIRReturn(v));
    }

    public static TIRThrow dup(TIRThrow e) {
        return copy(e, new TIRThrow(e.exception, e.point));
    }

    public static TIRIfExpr dup(TIRIfExpr e, TIRExpr c, TIRExpr t, TIRExpr f) {
        return copy(e, new TIRIfExpr(c, t, f));
    }

    public static TIRSwitch dup(TIRSwitch e, TIRExpr v, TIRSwitch.Case[] c, TIRSwitch.Case d) {
        return copy(e, new TIRSwitch(v, c, d));
    }

    public static TIRBlock dup(TIRBlock e, List<TIRExpr> i) {
        TIRBlock b = copy(e, new TIRBlock(e.label));
        for ( TIRExpr ie : i ) b.addExpr(ie);
        return b;
    }

    public static TIRLoop.For dup(TIRLoop.For e, TIRExpr i, TIRExpr c, TIRExpr u, TIRExpr b) {
        return copy(e, new TIRLoop.For(i, c, u, b));
    }

    public static TIRLoop.While dup(TIRLoop.While e, TIRExpr c, TIRExpr b) {
        return copy(e, new TIRLoop.While(c, b));
    }

    public static TIRLoop.DoWhile dup(TIRLoop.DoWhile e, TIRExpr c, TIRExpr b) {
        return copy(e, new TIRLoop.DoWhile(c, b));
    }

    public static TIRBlock.Break dup(TIRBlock.Break e) {
        return copy(e, new TIRBlock.Break());
    }

    public static TIRBlock.Continue dup(TIRBlock.Continue e) {
        return copy(e, new TIRBlock.Continue());
    }

    public static boolean isNoOp(TIRExpr nstmt) {
        return nstmt instanceof TIRNop;
    }

    public static boolean isGet(TIRExpr nstmt) {
        return nstmt instanceof TIRLocal.Get;
    }

    public static boolean isValue(TIRExpr e) {
        return e instanceof TIRConst.Value;
    }

    public static TIROperator $OP(Operator op, TIRExpr... list) {
        return new TIROperator(op, list);
    }

    public static TIRCompare.NotEqual $ISNOTNULL(TIRExpr e) {
        return $NEQ(e, $NULL());
    }

    public static TIRCompare.Equal $ISNULL(TIRExpr e) {
        return $EQ(e, $NULL());
    }

    public static TIRCompare.NotEqual $NEQ(TIRExpr a, TIRExpr b) {
        return new TIRCompare.NotEqual(a, b);
    }

    public static TIRCompare.Equal $EQ(TIRExpr a, TIRExpr b) {
        return new TIRCompare.Equal(a, b);
    }

    public static TIRExpr $AND(TIRExpr a, TIRExpr b) {
        TIRIfExpr e = new TIRIfExpr(a, b, $VAL(PrimBool.FALSE));
        e.setType(PrimBool.TYPE);
        return e;
    }

    public static TIRExpr $OR(TIRExpr a, TIRExpr b) {
        TIRIfExpr e = new TIRIfExpr(a, $VAL(PrimBool.TRUE), b);
        e.setType(PrimBool.TYPE);
        return e;
    }

    public static TIRExpr $IF(TIRExpr c, TIRExpr a, TIRExpr b) {
        TIRIfExpr e = new TIRIfExpr(c, a, b);
        e.setType(a.getType());
        return e;
    }

    public static TIRLocal.Get $GET(Method.Temporary t) {
        return new TIRLocal.Get(t);
    }

    public static TIRCall $CALL(TIRExpr e, TIRExpr... args) {
        return new TIRCall(e, args);
    }

    public static TIRLocal.Set $SET(Method.Temporary t, TIRExpr e) {
        return new TIRLocal.Set(t, e);
    }

    public static TIRConst.Value $REF(Type t, Heap.Record r) {
        return new TIRConst.Value(Reference.toValue(t, r));
    }

    public static TIRConst.Value $REF(Heap.Record r) {
        return new TIRConst.Value(Reference.toValue(r));
    }

    public static TIRConst.Value $NULL() {
        return new TIRConst.Value(Reference.NULL_VALUE);
    }

    public static TIRConst.Value $VAL(Value v) {
        return new TIRConst.Value(v);
    }

    public static TIRConst.Value $VAL(int v) {
        return new TIRConst.Value(PrimInt32.toValue(v));
    }

    public static TIRConst.Value $VAL(boolean b) {
        return new TIRConst.Value(PrimBool.toValue(b));
    }

    public static TIROperator $GR(TIRExpr a, TIRExpr b) {
        return $OP(new PrimInt32.GR(), a, b);
    }

    public static TIROperator $GREQ(TIRExpr a, TIRExpr b) {
        return $OP(new PrimInt32.GREQ(), a, b);
    }

    public static TIROperator $LTEQ(TIRExpr a, TIRExpr b) {
        return $OP(new PrimInt32.LTEQ(), a, b);
    }

    public static TIROperator $LT(TIRExpr a, TIRExpr b) {
        return $OP(new PrimInt32.LT(), a, b);
    }

    public static TIRConst.Value $FUNCREF(Method method) {
        return $VAL(new Function.Val(method));
    }

    public static Value[] reduce(TIRExpr... list) {
        if ( list.length == 0 ) return TIRInterpreter.NONE;
        if ( !isValue(list[0]) ) return null;
        Value[] vals = new Value[list.length];
        for ( int cntr = 0; cntr < list.length; cntr++ ) {
            TIRExpr e = list[cntr];
            if ( !isValue(e) ) return null;
            vals[cntr] = ((TIRConst.Value)e).getValue();
        }
        return vals;
    }

    public static Value toValue(TIRExpr e) {
        if ( !isValue(e) ) return null;
        return ((TIRConst.Value)e).getValue();
    }
}
