/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Mar 1, 2006
 */

package vpc.tir.opt;

import vpc.tir.*;
import vpc.tir.expr.Operator;

/**
 * The <code>AccessAnalyzer</code> visits the code of each method to extract the writes
 * to fields and use of methods.
 */
public class DepthFirstOpVisitor<R, E> extends TIRExprVisitor<R, E> {

    private Operator.Visitor<R, TIRExpr> opVisitor;

    public DepthFirstOpVisitor(Operator.Visitor<R, TIRExpr> a) {
        opVisitor = a;
    }

    public R visit(TIRExpr e, E env) {
        return null;
    }

    public R visit(TIROperator o, E env) {
        for (TIRExpr e : o.operands) e.accept(this, env);
        // call the appropriate visit method on the operator visitor
        return o.operator.accept(opVisitor, o.operands);
    }

    public R visit(TIRCompare c, E env) {
        c.left.accept(this, env);
        c.right.accept(this, env);
        return null;
    }

    public R visit(TIRCall c, E env) {
        c.func.accept(this, env);
        for (TIRExpr e : c.arguments) e.accept(this, env);
        return null;
    }

    public R visit(TIRLocal.Set l, E env) {
        l.value.accept(this, env);
        return null;
    }

    public R visit(TIRReturn r, E env) {
        r.value.accept(this, env);
        return null;
    }

    public R visit(TIRIfExpr e, E env) {
        e.condition.accept(this, env);
        e.true_target.accept(this, env);
        e.false_target.accept(this, env);
        return null;
    }

    public R visit(TIRSwitch e, E env) {
        e.expr.accept(this, env);
        for (TIRSwitch.Case c : e.cases)
            c.body.accept(this, env);
        if (e.defcase != null) e.defcase.body.accept(this, env);
        return null;
    }

    public R visit(TIRBlock e, E env) {
        for (TIRExpr i : e.list) i.accept(this, env);
        return null;
    }

    public R visit(TIRLoop.For e, E env) {
        e.initial.accept(this, env);
        e.condition.accept(this, env);
        e.body.accept(this, env);
        e.update.accept(this, env);
        return null;
    }

    public R visit(TIRLoop.While e, E env) {
        e.condition.accept(this, env);
        e.body.accept(this, env);
        return null;
    }

    public R visit(TIRLoop.DoWhile e, E env) {
        e.body.accept(this, env);
        e.condition.accept(this, env);
        return null;
    }
}
