/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Jan 31, 2007
 */

package vpc.tir.opt;

import vpc.tir.*;
import vpc.tir.expr.Operator;
import static vpc.tir.TIRUtil.*;
import java.util.List;
import java.util.LinkedList;

/**
 * The <code>AccessAnalyzer</code> visits the code of each method to extract the writes
 * to fields and use of methods.
 */
public class DepthFirstTransformer<E> extends TIRExprVisitor<TIRExpr, E> {

    public TIRExpr transform(TIRExpr e, E env) {
        return e.accept(this, env);
    }

    public TIRExpr[] transform(TIRExpr[] oargs, E env) {
        TIRExpr[] nargs = new TIRExpr[oargs.length];
        for (int cntr = 0; cntr < nargs.length; cntr++)
            nargs[cntr] = transform(oargs[cntr], env);
        return nargs;
    }

    public TIRSwitch.Case transform(TIRSwitch.Case v, E env) {
        return new TIRSwitch.Case(v.value, transform(v.body, env));
    }

    public List<TIRExpr> transform(List<TIRExpr> ol, E env) {
        List<TIRExpr> nl = new LinkedList<TIRExpr>();
        for (TIRExpr i : ol)
            nl.add(transform(i, env));
        return nl;
    }

    public TIRExpr visit(TIRExpr e, E env) {
        return e;
    }

    public TIRExpr visit(TIROperator o, E env) {
        return dup(o, transform(o.operands, env));
    }

    public TIRExpr visit(TIRCompare.Equal c, E env) {
        return dup(c, transform(c.left, env), transform(c.right, env));
    }

    public TIRExpr visit(TIRCompare.NotEqual c, E env) {
        return dup(c, transform(c.left, env), transform(c.right, env));
    }

    public TIRExpr visit(TIRCall c, E env) {
        return dup(c, transform(c.func, env), transform(c.arguments, env));
    }

    public TIRExpr visit(TIRLocal.Set l, E env) {
        return dup(l, transform(l.value, env));
    }

    public TIRExpr visit(TIRReturn r, E env) {
        return dup(r, transform(r.value, env));
    }

    public TIRExpr visit(TIRIfExpr e, E env) {
        return dup(e, transform(e.condition, env), transform(e.true_target, env), transform(e.false_target, env));
    }

    public TIRExpr visit(TIRSwitch e, E env) {
        TIRExpr expr = transform(e.expr, env);
        TIRSwitch.Case defcase = null;
        TIRSwitch.Case[] nc = new TIRSwitch.Case[e.cases.length];
        for (int i = 0; i < e.cases.length; i++) {
            nc[i] = transform(e.cases[i], env);
        }
        if (e.defcase != null) {
            defcase = transform(e.defcase, env);
        }
        return dup(e, expr, nc, defcase);
    }

    public TIRExpr visit(TIRBlock e, E env) {
        return dup(e, transform(e.list, env));
    }

    public TIRExpr visit(TIRLoop.For e, E env) {
        TIRExpr init = transform(e.initial, env);
        TIRExpr cond = transform(e.condition, env);
        TIRExpr body = transform(e.body, env);
        TIRExpr update = transform(e.update, env);
        return dup(e, init, cond, update, body);
    }

    public TIRExpr visit(TIRLoop.While e, E env) {
        TIRExpr cond = transform(e.condition, env);
        TIRExpr body = transform(e.body, env);
        return dup(e, cond, body);
    }

    public TIRExpr visit(TIRLoop.DoWhile e, E env) {
        TIRExpr cond = transform(e.condition, env);
        TIRExpr body = transform(e.body, env);
        return dup(e, cond, body);
    }
}
