/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Nov 16, 2006
 */

package vpc.tir.opt;

import cck.parser.SourcePoint;
import cck.text.StringUtil;
import cck.text.Terminal;
import cck.util.Util;
import vpc.core.Program;
import vpc.core.concept.Method;
import vpc.core.concept.PrimVoid;
import vpc.model.Stage;
import vpc.tir.*;
import java.util.Iterator;
import java.io.IOException;

/**
 * The <code>TIRChecker</code> class performs a number of consistency checks on
 * TIR code that helps to maintain invariants across compilation passes.
 *
 * @author Ben L. Titzer
 */
public class TIRChecker extends Stage {

    private Method currentMethod;

    public void visitProgram(Program p) throws IOException {
        for ( Method m : p.closure.methods ) {
            currentMethod = m;
            TIRRep r = TIRUtil.getRep(m);
            Visitor visitor = new Visitor(r);
            visitor.visitBody(r.getBody());
        }
    }

    protected class Visitor extends TIRExprVisitor<TIRExpr, Visitor.Context> {

        public class Context {
            final boolean expression;
            final TIRBlock block;
            public Context(Context p, boolean b, TIRBlock bl) {
                expression = b;
                block = bl;
            }
        }

        Visitor(TIRRep r) {
        }

        public TIRExpr visit(TIRExpr e, Context c) {
            throw TIRUtil.fail("unknown TIR expression kind in checker", e);
        }

        public TIRExpr visit(TIRNop e, Context c) {
            return assertStmt(e, c);
        }

        public TIRExpr visit(TIROperator e, Context c) {
            basicChecks(e, c);
            int arity = e.operands.length;
            if ( arity != e.operator.getArity() ) {
                // check the number of operands match the arity of the operator.
                fail("arity mismatch", e, c);
            }
            visitExprs(e.operands, c);
            return e;
        }

        public TIRExpr visit(TIRCompare.Equal e, Context c) {
            assertExpr(e, c);
            visitExpr(e.left, c);
            visitExpr(e.right, c);
            return e;
        }

        public TIRExpr visit(TIRCompare.NotEqual e, Context c) {
            assertExpr(e, c);
            visitExpr(e.left, c);
            visitExpr(e.right, c);
            return e;
        }

        public TIRExpr visit(TIRConst e, Context c) {
            basicChecks(e, c);
            return e;
        }

        public TIRExpr visit(TIRLocal.Get e, Context c) {
            basicChecks(e, c);
            return nonVoid(e, c);
        }

        public TIRExpr visit(TIRCall e, Context c) {
            basicChecks(e, c);
            visitExpr(e.func, c);
            visitExprs(e.arguments, c);
            return e;
        }

        public TIRExpr visit(TIRLocal.Set e, Context c) {
            basicChecks(e, c);
            visitExpr(e.value, c);
            return e;
        }

        public TIRExpr visit(TIRReturn e, Context c) {
            basicChecks(e, c);
            assertStmt(e, c);
            visitExpr(e.value, c);
            return e;
        }

        public TIRExpr visit(TIRIfExpr e, Context c) {
            basicChecks(e, c);
            visitExpr(e.condition, c);
            if ( c.expression ) {
                visitExpr(e.true_target, c);
                visitExpr(e.false_target, c);
            } else {
                visitStmt(e.true_target, c);
                visitStmt(e.false_target, c);
            }
            return e;
        }

        public TIRExpr visit(TIRSwitch e, Context c) {
            basicChecks(e, c);
            visitExpr(e.expr, c);
            for (TIRSwitch.Case oc : e.cases) {
                visitStmt(oc.body, c);
            }
            visitStmt(e.defcase.body, c);
            return e;
        }

        public TIRExpr visit(TIRBlock e, Context c) {
            basicChecks(e, c);
            Context nc = new Context(c, c.expression, e);
            Iterator<TIRExpr> i = e.list.iterator();
            int cntr = 0, last = e.list.size() - 1;
            while ( i.hasNext() ) {
                TIRExpr se = i.next();
                if ( c.expression && cntr == last ) visitExpr(se, nc);
                else visitStmt(se, nc);
                cntr++;
            }
            return e;
        }

        public TIRExpr visit(TIRLoop.For e, Context c) {
            assertStmt(e, c);
            visitStmt(e.initial, c);
            visitExpr(e.condition, c);
            visitStmt(e.update, c);
            visitStmt(e.body, c);
            return e;
        }

        public TIRExpr visit(TIRLoop.While e, Context c) {
            assertStmt(e, c);
            visitExpr(e.condition, c);
            visitStmt(e.body, c);
            return e;
        }

        public TIRExpr visit(TIRLoop.DoWhile e, Context c) {
            assertStmt(e, c);
            visitStmt(e.body, c);
            visitExpr(e.condition, c);
            return e;
        }

        public TIRExpr visit(TIRBlock.Break e, Context c) {
            return assertStmt(e, c);
        }

        public TIRExpr visit(TIRBlock.Continue e, Context c) {
            return assertStmt(e, c);
        }

        protected TIRExpr[] visitExprs(TIRExpr[] e, Context c) {
            for ( TIRExpr ie : e ) visitExpr(ie, c);
            return e;
        }

        protected TIRExpr visitExpr(TIRExpr e, Context c) {
            return e.accept(this, new Context(c, true, c.block));
        }

        protected TIRExpr visitStmt(TIRExpr e, Context c) {
            return e.accept(this, new Context(c, false, c.block));
        }

        protected TIRExpr visitBody(TIRExpr e) {
            Context nc = new Context(null, false, null);
            return e.accept(this, nc);
        }
    }

    private TIRExpr assertExpr(TIRExpr e, Visitor.Context c) {
        basicChecks(e, c);
        nonVoid(e, c);
        if ( !c.expression ) fail("should be an expression", e, c);
        return e;
    }

    private TIRExpr nonVoid(TIRExpr e, Visitor.Context c) {
        if ( e.getType() == PrimVoid.TYPE ) fail("should have non-void type", e, c);
        return e;
    }

    private TIRExpr assertStmt(TIRExpr e, Visitor.Context c) {
        basicChecks(e, c);
        if ( c.expression ) fail("should be a statement", e, c);
        return e;
    }

    private void basicChecks(TIRExpr e, TIRChecker.Visitor.Context c) {
        if ( e.getType() == null ) fail("no type", e, c);
        if ( e.getSourcePoint() == null ) warn("no source information", e, c);
    }

    private void fail(String s, TIRExpr e, Visitor.Context c) {
        String nstr = StringUtil.quote(e.getClass());
        SourcePoint sp = e.getSourcePoint();
        String mn = currentMethod.getFullName();
        throw Util.failure("TIRChecker failure", s + " on node " + nstr + " @ " + sp +" in "+mn);
    }

    private void warn(String s, TIRExpr e, Visitor.Context c) {
        String nstr = StringUtil.quote(e.getClass());
        String mn = currentMethod.getFullName();
        Terminal.print(Terminal.ERROR_COLOR, "TIRChecker warning");
        Terminal.println(": " + s + " on node " + nstr + " in "+mn);
    }
}
