/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Oct 5, 2006
 */
package vpc.tir.opt;

import vpc.core.*;
import vpc.core.csr.*;
import vpc.core.concept.*;
import vpc.core.virgil.VirgilClass;
import vpc.core.virgil.VirgilComponent;
import vpc.model.Stage;
import vpc.tir.*;
import static vpc.tir.TIRUtil.*;
import vpc.tir.expr.Operator;
import java.util.HashSet;
import java.io.IOException;

import cck.util.Util;

/**
 * The <code>TIRSimplifier</code> class simplifies complicated blocks of code,
 * flattening unnecessary nesting and lifting out subexpressions that can have
 * side-effects into a canonical order. This avoids ambiguous evaluation
 * order problems in generating C code, which was previously handled with
 * complex lifting built into the code generator.
 *
 * @author Ben L. Titzer
 */
public class TIRSimplifier extends Stage {

    protected static final HashSet<Class> simple;

    public static final boolean LIFT_THROUGH = true;
    public static final boolean CONST_FOLDING = true;

    static {
        simple = new HashSet<Class>();

        // operators for PrimInt32 have no side effects
        simple.add(PrimInt32.ADD.class);
        simple.add(PrimInt32.SUB.class);
        simple.add(PrimInt32.MUL.class);
        simple.add(PrimInt32.NEG.class);
        simple.add(PrimInt32.EQU.class);
        simple.add(PrimInt32.NEQU.class);
        simple.add(PrimInt32.GR.class);
        simple.add(PrimInt32.LT.class);
        simple.add(PrimInt32.GREQ.class);
        simple.add(PrimInt32.LTEQ.class);

        // operators for PrimChar have no side effects
        simple.add(PrimChar.EQU.class);
        simple.add(PrimChar.NEQU.class);
        simple.add(PrimChar.GR.class);
        simple.add(PrimChar.LT.class);
        simple.add(PrimChar.GREQ.class);
        simple.add(PrimChar.LTEQ.class);

        simple.add(PrimBool.NOT.class);

        // raw operators do not have side effects (except setbit)
        simple.add(PrimRaw.AND.class);
        simple.add(PrimRaw.OR.class);
        simple.add(PrimRaw.XOR.class);
        simple.add(PrimRaw.SHL.class);
        simple.add(PrimRaw.SHR.class);
        simple.add(PrimRaw.GetBit.class);
        simple.add(PrimRaw.Complement.class);

        // primitive conversions do not have side effects
        simple.add(PrimConversion.AdjustRaw.class);
        simple.add(PrimConversion.Char_Int32.class);
        simple.add(PrimConversion.Char_Raw.class);
        simple.add(PrimConversion.Int32_Char.class);
        simple.add(PrimConversion.Int32_Raw.class);
        simple.add(PrimConversion.Raw_Char.class);
        simple.add(PrimConversion.Raw_Int32.class);

        // the getmethod operators don't directly cause side effects
        simple.add(VirgilClass.GetMethod.class);
        simple.add(VirgilComponent.GetMethod.class);

        // the pointer, array, and struct read operators are simple
        simple.add(CSRArray.GetElement.class);
        simple.add(CSRPointer.GetPtr.class);
        simple.add(CSRStruct.GetRefField.class);
        simple.add(CSRStruct.GetValueField.class);
    }

    public void visitProgram(Program p) throws IOException {
        for ( Method m : p.closure.methods ) {
            visitMethod(m);
        }
    }

    public void visitMethod(Method m) {
        TIRRep r = getRep(m);
        Transformer transformer = new Transformer(r);
        r.setBody(transformer.simplifyBlock(false, r.getBody()));
        m.addMethodRep(TIRRep.REP_NAME, r);
    }

    protected class Transformer extends TIRExprVisitor<TIRExpr, Transformer.Context> {

        public class Context {
            final boolean expression;
            final TIRBlock block;
            final Method.Temporary temp;
            public Context(boolean b, TIRBlock bl) {
                expression = b;
                block = bl;
                temp = null;
            }
            public Context(boolean b, Method.Temporary t, TIRBlock bl) {
                expression = b;
                block = bl;
                temp = t;
            }
            public TIRExpr lift(TIRExpr e) {
                if ( expression )
                    return TIRUtil.lift(method, temp, e, block);
                else {
                    block.addExpr(e);
                    return null;
                }
            }
            public TIRExpr liftThrough(TIRExpr e, TIRExpr r) {
                if ( expression ) {
                    if ( temp == null && TIRUtil.isGet(r)) {
                        TIRUtil.lift(method, ((TIRLocal.Get)r).temp, e, block);
                    } else {
                        TIRUtil.lift(method, temp, e, block);
                    }
                    return r;
                }
                // liftThrough should never be called on a statement
                throw Util.unreachable();
            }
            public TIRExpr simple(TIRExpr e) {
                if ( expression ) return e;
                else return null;
            }
        }

        protected final TIRRep method;

        Transformer(TIRRep r) {
            method = r;
        }

        public TIRExpr visit(TIRExpr e, Context c) {
            throw TIRUtil.fail("unknown TIR expression kind in simplifier", e);
        }

        public TIRExpr visit(TIRNop e, Context c) {
            assert !c.expression;
            return e; // NOP : simple, terminal
        }

        public TIRExpr visit(TIROperator e, Context c) {
            TIRExpr ne = copy(e, $_OP(e.operator, simplifyExprs(e.operands, c)));
            if ( isSimple(ne) ) {
                // no side effects, return this expression in its simple form
                return c.simple(ne);
            }
            if ( LIFT_THROUGH && isCheckOnly(ne) ) {
                assert e.operands.length == 1;
                return c.liftThrough(ne, ((TIROperator)ne).operands[0]);
            }
            // lift the operator to the top level
            return c.lift(ne);
        }

        public TIRExpr visit(TIRCompare.Equal e, Context c) {
            // simple, recurse
            TIRExpr ne = copy(e, $_EQ(simplifyExpr(e.left, c), simplifyExpr(e.right, c)));
            return c.simple(ne);
        }

        public TIRExpr visit(TIRCompare.NotEqual e, Context c) {
            // simple, recurse
            TIRExpr ne = copy(e, $_NEQ(simplifyExpr(e.left, c), simplifyExpr(e.right, c)));
            return c.simple(ne);
        }

        public TIRExpr visit(TIRConst e, Context c) {
            // simple, terminal
            return c.simple(e); // Value : simple, terminal
        }

        public TIRExpr visit(TIRConst.String e, Context c) {
            // lift a string constant
            return c.lift(dup(e));
        }

        public TIRExpr visit(TIRLocal.Get e, Context c) {
            // simple, terminal
            return c.simple(e); // Local : simple, terminal
        }

        public TIRExpr visit(TIRCall e, Context c) {
            TIRExpr nf = simplifyExpr(e.func, c);
            TIRExpr[] na = simplifyExprs(e.arguments, c);
            return c.lift(dup(e, nf, na));
        }

        public TIRExpr visit(TIRLocal.Set e, Context c) {
            return c.lift(dup(e, simplifyVar(e.temp, e.value, c)));
        }

        public TIRExpr visit(TIRReturn e, Context c) {
            assert !c.expression;
            return dup(e, simplifyExpr(e.value, c));
        }

        public TIRExpr visit(TIRThrow e, Context c) {
            return dup(e);
        }

        public TIRExpr visit(TIRIfExpr e, Context c) {
            TIRExpr nc = simplifyExpr(e.condition, c);
            if ( isValue(nc) ) {
                // perform constant folding
                Value cv = toValue(nc);
                if ( cv.equals(PrimBool.TRUE) ) {
                    return simplifyBranch(e.true_target, c);
                }
                if ( cv.equals(PrimBool.FALSE) ) {
                    return simplifyBranch(e.false_target, c);
                }
            }
            if ( c.expression ) {
                // if this if is an expression, lift to the top level and introduce a temp
                Method.Temporary temp = method.newTemporary(e.getType());
                TIRExpr nt = simplifyBlock(false, new TIRLocal.Set(temp, e.true_target));
                TIRExpr nf = simplifyBlock(false, new TIRLocal.Set(temp, e.false_target));
                c.block.addExpr(dup(e, nc, nt, nf));
                // the overall result is the value of the temporary
                return new TIRLocal.Get(temp);
            } else {
                // otherwise, simplify this if expression as a statement
                TIRExpr nt = simplifyBlock(c.expression, e.true_target);
                TIRExpr nf = simplifyBlock(c.expression, e.false_target);
                return c.lift(dup(e, nc, nt, nf));
            }
        }

        public TIRExpr visit(TIRSwitch e, Context c) {
            TIRExpr nv = simplifyExpr(e.expr, c);
            TIRSwitch.Case[] nc = new TIRSwitch.Case[e.cases.length];
            for ( int cntr = 0; cntr < nc.length; cntr++ ) {
                TIRSwitch.Case oc = e.cases[cntr];
                nc[cntr] = new TIRSwitch.Case(oc.value, simplifyBlock(false, oc.body));
            }
            TIRSwitch.Case ndef = new TIRSwitch.Case(e.defcase.value, simplifyBlock(false, e.defcase.body));
            return c.lift(dup(e, nv, nc, ndef));
        }

        public TIRExpr visit(TIRBlock e, Context c) {
            int cntr = 0;
            int size = e.list.size();
            for ( TIRExpr se : e.list ) {
                if ( c.expression && cntr == size - 1 ) {
                    // if the outer block is used as an expression, the last is an expression
                    return simplifyExpr(se, c);
                }
                else {
                    TIRExpr n = se.accept(this, new Context(false, c.block));
                    if ( n != null ) c.block.addExpr(copy(se, n));
                }
                cntr++;
            }
            assert !c.expression;
            return null;
        }

        public TIRExpr visit(TIRLoop.For e, Context c) {
            assert !c.expression;
            TIRExpr ni = simplifyStmt(e.initial, c);
            TIRExpr nc = simplifyBlock(true, e.condition);
            TIRExpr nu = simplifyBlock(false, e.update);
            TIRExpr nb = simplifyBlock(false, e.body);
            return c.lift(dup(e, ni, nc, nu, nb));
        }

        public TIRExpr visit(TIRLoop.While e, Context c) {
            assert !c.expression;
            TIRExpr nc = simplifyBlock(true, e.condition);
            TIRExpr nt = simplifyBlock(false, e.body);
            return c.lift(dup(e, nc, nt));
        }

        public TIRExpr visit(TIRLoop.DoWhile e, Context c) {
            assert !c.expression;
            TIRExpr nt = simplifyBlock(false, e.body);
            TIRExpr nc = simplifyBlock(true, e.condition);
            return c.lift(dup(e, nc, nt));
        }

        public TIRExpr visit(TIRBlock.Break e, Context c) {
            assert !c.expression;
            return e;
        }

        public TIRExpr visit(TIRBlock.Continue e, Context c) {
            assert !c.expression;
            return e;
        }

        protected TIRExpr[] simplifyExprs(TIRExpr[] e, Context c) {
            TIRExpr[] ne = new TIRExpr[e.length];
            for ( int cntr = 0; cntr < e.length; cntr++ ) {
                ne[cntr] = simplifyExpr(e[cntr], c);
            }
            return ne;
        }

        private TIRExpr simplifyVar(Method.Temporary var, TIRExpr e, Context c) {
            TIRExpr n = e.accept(this, new Context(true, var, c.block));
            return copy(e, n);
        }

        protected TIRExpr simplifyExpr(TIRExpr e, Context c) {
            TIRExpr n = e.accept(this, new Context(true, c.block));
            return copy(e, n);
        }

        protected TIRExpr simplifyStmt(TIRExpr e, Context c) {
            TIRExpr n = e.accept(this, new Context(false, c.block));
            if ( n == null ) n = new TIRNop();
            return copy(e, n);
        }

        protected TIRExpr simplifyBranch(TIRExpr e, Context c) {
            TIRExpr n = e.accept(this, new Context(c.expression, c.block));
            if ( n == null ) return n;
            return copy(e, n);
        }

        protected TIRExpr simplifyBlock(boolean expr, TIRExpr e) {
            TIRBlock b = new TIRBlock(method.newLabel("K"));
            Context c = new Context(expr, b);
            TIRExpr ne = e.accept(this, c);
            if ( ne != null ) b.addExpr(ne);
            if ( b.isEmpty() ) return copy(e, new TIRNop());
            if ( b.isSingleton() ) return b.getFirst();
            return copy(e, b);
        }

    }

    protected static boolean isSimple(Operator op) {
        return simple.contains(op.getClass());
    }

    protected static boolean isSimple(TIRExpr e) {
        if ( !(e instanceof TIROperator) ) return false;
        return isSimple(((TIROperator)e).operator);
    }

    protected static boolean isCheckOnly(TIRExpr e) {
        if ( !(e instanceof TIROperator) ) return false;
        return ((TIROperator) e).operator.getClass() == Reference.NullCheck.class;
    }

    public static TIRExpr $_OP(Operator op, TIRExpr... list) {
        if ( CONST_FOLDING && isSimple(op) ) {
            try {
                Value[] vals = TIRUtil.reduce(list);
                if ( vals != null )
                    return $VAL(op.apply(vals));
            } catch ( Throwable t) {
                // do nothing and return basic operator in case of exception
            }
        }
        return $OP(op, list);
    }

    public static TIRExpr $_NEQ(TIRExpr a, TIRExpr b) {
        if ( CONST_FOLDING && isValue(a) && isValue(b) ) {
            // constant folding
            Value av = toValue(a);
            Value bv = toValue(b);
            return $VAL(!av.equals(bv));
        }
        return $NEQ(a, b);
    }

    public static TIRExpr $_EQ(TIRExpr a, TIRExpr b) {
        if ( CONST_FOLDING && isValue(a) && isValue(b) ) {
            // constant folding
            Value av = toValue(a);
            Value bv = toValue(b);
            return $VAL(av.equals(bv));
        }
        return $EQ(a, b);
    }

}
