/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Feb 28, 2006
 */

package vpc.tir.opt.rma;

import cck.text.Printer;
import vpc.core.*;
import vpc.core.concept.*;
import vpc.core.virgil.*;
import static vpc.core.virgil.VirgilTypeSystem.*;
import vpc.tir.TIRExpr;
import vpc.tir.TIRUtil;
import vpc.tir.expr.Operator;
import vpc.tir.opt.DepthFirstOpVisitor;
import vpc.types.Type;
import java.util.*;

/**
 * The <code>RMAnalyzer</code> class implements Reachable Members Analysis (RMA),
 * which is similar to Rapid Type Analysis, but instead of reachable types,
 * this analysis additionally computes which members of which components and
 * classes are accessible.
 *
 * @author Ben L. Titzer
 */
public class RMAnalyzer {

    protected boolean debug;
    protected OpVisitor visitor;
    protected DepthFirstOpVisitor<Void, Object> analyzer;
    protected Printer printer;
    protected Program program;
    protected LinkedList<Unit> worklist;
    protected HashMap<Type, TypeInfo> typeInfo;
    protected HashSet<Method> visitedMethods;

    /**
     * The <code>TypeInfo</code> class represents the information tracked for a
     * particular type, which includes its instantiated subtypes, live members,
     * live instances, and references to the new layout and meta layouts.
     */
    protected class TypeInfo {
        protected final TypeInfo parent;          // super type (if any)
        protected final Type type;                // this type
        protected final HashSet<Type> subtypes;   // instantiated subtypes
        protected final HashSet<Field> fields;   // used members
        protected final HashSet<Method> methods;  // used methods
        protected HashSet<Heap.Record> instances; // live instances

        protected TypeInfo(Type t, TypeInfo p) {
            parent = p;
            type = t;
            subtypes = new HashSet<Type>();
            fields = new HashSet<Field>();
            methods = new HashSet<Method>();
        }
    }

    /**
     * The <code>Unit</code> class represents a unit of work for the analysis.
     */
    protected abstract static class Unit {
        protected abstract boolean isDone();
        protected abstract void analyze();
    }

    /**
     * The <code>InstUnit</code> class represents a unit of work for a newly discovered
     * (reachable) record instance.
     * PENDING/DONE: if this record exists in the live record list of its type
     * ACTION: analyze each used field in the record and post values as necessary
     */
    protected class InstUnit extends Unit {
        private final Heap.Record record;

        protected InstUnit(Heap.Record r) {
            record = r;
        }

        protected boolean isDone() {
            if (record == null) return true;
            TypeInfo info = getTypeInfo(record.layout.type);
            post(new TypeUnit(info));
            return !info.instances.add(record);
        }

        protected void analyze() {
            Type type = record.layout.type;
            TypeInfo info = getTypeInfo(type);
            if (type.isArray()) {
                // if this an array, (conservatively) assume all members are reachable
                for (int cntr = 0; cntr < record.getSize(); cntr++)
                    postValue(record.getValue(cntr));
            } else {
                // record the values of fields of this record that are used
                for (Member m : info.fields) {
                    if (m.isField()) {
                        Value value = getFieldValue(record, m);
                        postValue(value);
                    }
                }
            }
        }

        public String toString() {
            return "Record(" + record + ")";
        }
    }

    /**
     * The <code>MethodImpl</code> class represents a unit of work for a newly discovered
     * (reachable) method.
     * PENDING/DONE: if this method exists in the global <code>visitedMethods</code> list
     * ACTION: analyze the code of the method to discover member uses
     */
    protected class MethodImpl extends Unit {
        private final Method method;

        protected MethodImpl(Method m) {
            method = m;
        }

        protected boolean isDone() {
            return !visitedMethods.add(method);
        }

        protected void analyze() {
            TIRUtil.getRep(method).getBody().accept(analyzer, null);
        }

        public String toString() {
            return "Method(" + method.getFullName() + ")";
        }
    }

    /**
     * The <code>TypeUnit</code> class represents a unit of work for a newly discovered
     * (reachable) type.
     * PENDING/DONE: if the type info has a non-null list of record instances
     * ACTION: add this type to all supertypes' subtype lists; post parent's used members
     * on this type
     */
    protected class TypeUnit extends Unit {
        final TypeInfo info;

        TypeUnit(TypeInfo ti) {
            info = ti;
        }

        protected boolean isDone() {
            if (info.instances != null) return true;
            info.instances = new HashSet<Heap.Record>();
            return false;
        }

        protected void analyze() {
            // add this to the live subtypes of each of the parent types
            for (TypeInfo i = info; i != null; i = i.parent) {
                i.subtypes.add(info.type);
                // for each member in each parent, post that member for this type
                for (Field f : i.fields)
                    post(new FieldUse(getTypeInfo(info.type), f));
                for (Method m : i.methods)
                    post(new MethodUse(getTypeInfo(info.type), m));
            }
        }

        public String toString() {
            return "Type(" + info.type + ")";
        }
    }

    /**
     * The <code>FieldUse</code> class represents a unit of work for a newly discovered
     * used member on a particular type.
     * PENDING/DONE: if the type info already contains the member in its members list
     * ACTION: for all records of this type, inspect this field; post this member for all
     * live subtypes of this type
     */
    protected class FieldUse extends Unit {
        final TypeInfo info;
        final Field field;

        FieldUse(TypeInfo ti, Field m) {
            info = ti;
            field = m;
        }

        protected boolean isDone() {
            return !info.fields.add(field);
        }

        protected void analyze() {
            if (info.instances != null) {
                for (Heap.Record r : info.instances)
                    postValue(getFieldValue(r, field));
            }
            // repost the field use to all live subtypes
            for (Type t : info.subtypes) {
                post(new FieldUse(getTypeInfo(t), field));
            }
        }

        public String toString() {
            return "Member(" + info.type + "," + field.getName() + ")";
        }
    }

    /**
     * The <code>MethodUse</code> class represents a unit of work for a newly discovered
     * used member on a particular type.
     * PENDING/DONE: if the type info already contains the member in its members list
     * ACTION: for all records of this type, inspect this field; post this member for all
     * live subtypes of this type
     */
    protected class MethodUse extends Unit {
        final TypeInfo info;
        final Method method;

        MethodUse(TypeInfo ti, Method m) {
            info = ti;
            method = m;
        }

        protected boolean isDone() {
            // TODO: is there an equality problem with contains and virtual methods?
            return !info.methods.add(method);
        }

        protected void analyze() {
            post(new MethodImpl(program.resolveMethod(info.type, method)));
            for (Type t : info.subtypes) {
                post(new MethodUse(getTypeInfo(t), method));
            }
        }

        public String toString() {
            return "MethodUse(" + info.type + "," + method.getName() + ")";
        }
    }


    /**
     * The <code>analyzeProgram()</code> method begins the analysis by starting at the entrypoints
     * of the program and iteratively finding the reachable code and data.
     *
     * @param p the program to analyze
     */
    public void analyzeProgram(Program p) {
        visitor = new OpVisitor();
        analyzer = new DepthFirstOpVisitor<Void, Object>(visitor);
        ProgramDecl pd = p.programDecl;
        program = p;
        if (pd != null) { // only run the analysis if the entrypoints are known
            // build a cache of live (reachable) methods
            visitedMethods = new HashSet<Method>();
            typeInfo = new HashMap<Type, TypeInfo>();
            worklist = new LinkedList<Unit>();
            // get the entry points
            getEntryPoints(pd);
            // analyze all the units of work
            for (Unit u = worklist.poll(); u != null; u = worklist.poll()) {
                if (debug) printer.startblock(u.toString());
                u.analyze();
                if (debug) printer.endblock();
            }
        }
    }

    private void getEntryPoints(ProgramDecl pd) {
        for (ProgramDecl.EntryPoint ept : pd.entryPoints)
            post(new MethodImpl(ept.resolve(program)));
    }

    private void postValue(Value v) {
        if (v instanceof Reference.Val) {
            post(new InstUnit(Reference.fromValue(v)));
        } else if (v instanceof VirgilDelegate.Val) {
            VirgilDelegate.Val d = VirgilDelegate.fromValue(v);
            post(new InstUnit(d.record));
            post(new MethodImpl(d.method));
        }
    }

    private void post(Unit u) {
        if (!u.isDone()) {
            worklist.add(u);
            if (debug) printer.println("+ " + u);
        } else {
            if (debug) printer.println("- " + u);
        }
    }

    protected TypeInfo getTypeInfo(Type t) {
        TypeInfo u = typeInfo.get(t);
        if (u != null) return u;

        TypeInfo parent = null;
        if (isClass(t)) {
            // if it is an object type, get its parent info
            VirgilClass.IType pn = ((VirgilClass.IType) t).getParentType();
            if (pn != null) parent = getTypeInfo(pn);
        }
        if (isClass(t) || isArray(t) || isComponent(t)) {
            // if it is a member type
            u = new TypeInfo(t, parent);
            typeInfo.put(t, u);
            return u;
        }
        return null;
    }

    private void postComponentField(VirgilComponent comp, Field field) {
        Type type = comp.getTypeName().getType();
        // get the member info
        post(new InstUnit(comp.getRecord()));

        // add an access for this member
        post(new FieldUse(getTypeInfo(type), field));
    }

    /**
     * The <code>RMAnalyzer.Operator</code> class implements a visitor that performs specific
     * actions on the operations in the program that correspond to accessing fields
     * and invoking methods.
     */
    protected class OpVisitor implements VirgilClass.OpVisitor<Void, TIRExpr>,
                                            VirgilComponent.OpVisitor<Void, TIRExpr> {

        public Void visit(VirgilClass.GetField f, TIRExpr... e) {
            post(new FieldUse(getTypeInfo(e[0].getType()), f.field));
            return null;
        }

        public Void visit(VirgilClass.SetField f, TIRExpr... e) {
            post(new FieldUse(getTypeInfo(e[0].getType()), f.field));
            return null;
        }

        public Void visit(VirgilClass.GetMethod f, TIRExpr... e) {
            if ( f.virtual )
                post(new MethodUse(getTypeInfo(e[0].getType()), f.method));
            else
                post(new MethodImpl(f.method));
            return null;
        }

        public Void visit(VirgilComponent.GetField f, TIRExpr... e) {
            postComponentField(f.component, f.field);
            return null;
        }

        public Void visit(VirgilComponent.SetField f, TIRExpr... e) {
            postComponentField(f.component, f.field);
            return null;
        }

        public Void visit(VirgilComponent.GetMethod f, TIRExpr... e) {
            post(new MethodImpl(f.method));
            return null;
        }

        public Void visit(Operator o, TIRExpr... e) {
            // default behavior is to simply do nothing
            return null;
        }
    }

    private Value getFieldValue(Heap.Record record, Member m) {
        return record.getValue(((Field) m).fieldIndex);
    }
}
