/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Dec 26, 2006
 */
package vpc.tir.opt.rma;

import vpc.core.*;
import vpc.core.concept.*;
import vpc.core.virgil.*;
import vpc.tir.TIRInterpreter;
import vpc.tir.opt.rma.RMAnalyzer.TypeInfo;
import vpc.tir.stages.MetaLayout;
import vpc.types.Type;

/**
 * The <code>RMTransformer</code> definition.
 *
 * @author Ben L. Titzer
 */
public class RMTransformer {

    protected final RMAnalyzer rma;
    protected boolean orphans;

    public RMTransformer(RMAnalyzer r) {
        rma = r;
    }

    protected void transform() {
        Program p = rma.program;
        Closure nc = p.closure.copy();
        nc.resetLayouts();
        // layouts: remove fields (unread), (unwritten+const)
        for (VirgilComponent cd : p.virgil.getComponents()) rebuildLayout(cd, nc);
        for (VirgilClass cd : p.virgil.getClasses()) rebuildLayout(cd, nc);
        // heap: rebuild with new layout
        for (Heap.Record r : p.heap.getRecords()) rebuildRecord(r, nc);
        // closure: reset the live methods
        nc.methods = rma.visitedMethods;
        p.closure = nc;
    }

    protected void rebuildLayout(VirgilComponent cd, Closure nc) {
        Type type = cd.getTypeName().getType();
        RMAnalyzer.TypeInfo info = rma.getTypeInfo(type);
        if (!nc.componentLayouts.containsKey(cd)) {
            Heap.Layout nl = rma.program.heap.newLayout(cd.getName(), type);
            addFields(info, nl);
            nc.componentLayouts.put(cd, nl);
        }
    }

    protected Heap.Layout rebuildMetaLayout(VirgilClass cd, Closure nc) {
        TypeInfo info = rma.getTypeInfo(cd.getTypeName().getType());
        Heap.Layout nl = nc.getMetaLayout(cd);
        if ( nl == null ) {
            nl = MetaLayout.newMetaLayout(rma.program.virgil, cd);
            addMethods(info, cd, nl);
            nc.metaLayouts.put(cd, nl);
        }
        return nl;
    }

    protected void rebuildLayout(VirgilClass cd, Closure nc) {
        Type type = cd.getTypeName().getType();
        TypeInfo info = rma.getTypeInfo(type);
        if (!nc.classLayouts.containsKey(cd)) {
            Heap.Layout nl = rma.program.heap.newLayout(cd.getName(), type);
            if ( !MetaLayout.isOrphan(orphans, nc, cd)) {
                // add the meta field is this is not an orphan
                Heap.Layout ml = rebuildMetaLayout(cd, nc);
                Reference.Val val = Reference.toValue(ml.newInstance());
                MetaLayout.addMetaField(nl, cd, val, nc);
            }
            addFields(info, nl);
            nc.classLayouts.put(cd, nl);
        }
    }

    protected void addFields(TypeInfo info, Heap.Layout layout) {
        if (info == null) return;
        // add info for parent fields first
        addFields(info.parent, layout);

        for (Field f : info.fields) {
            if (layout.findCell(f) == null) {
                // if the cell is not already in the layout, add it
                layout.addCell(f, TIRInterpreter.getDefaultValue(f.getType()));
            }
        }
    }

    protected void addMethods(TypeInfo info, VirgilClass rcvr, Heap.Layout meta) {
        // add the parent members first
        if (info.parent != null) addMethods(info.parent, rcvr, meta);
        else MetaLayout.addMetaTID(rma.program.virgil, meta, rcvr);

        for (Method m : info.methods) {
            // get the method family
            Method.Family fam = m.family;
            // is a meta field required?
            if ( MetaLayout.requiresMetaField(fam, meta)) {
                // add a meta field to the layout
                MetaLayout.addMetaField(rma.program.virgil, rcvr, meta, fam);
            }
        }
    }

    protected void rebuildRecord(Heap.Record r, Closure nc) {
        Type type = r.layout.type;
        if (VirgilTypeSystem.isClass(type)) {
            VirgilClass cd = ((VirgilClass.IType) type).getDecl();
            Heap.Layout nl = nc.getLayout(cd);
            nl.migrate(r);
        } else if (VirgilTypeSystem.isComponent(type)) {
            VirgilComponent cd = rma.program.virgil.getComponentDecl(type.toString());
            Heap.Layout nl = nc.getLayout(cd);
            nl.migrate(r);
        }
    }
}
