/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Feb 24, 2006
 */
package vpc.tir.stages;

import cck.text.Printer;
import cck.text.StringUtil;
import vpc.core.*;
import vpc.core.concept.Reference;
import vpc.core.virgil.VirgilDelegate;
import vpc.model.Stage;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.IOException;

/**
 * @author Ben L. Titzer
 */
public class EmitHeap extends Stage {

    public void visitProgram(Program program) throws IOException {
        FileOutputStream fos = new FileOutputStream(program.name + ".dot");
        Printer p = new Printer(new PrintStream(fos));
        p.startblock("digraph " + program.name);
        p.println("rankdir=LR;");
        emitRecordDecls(program, p);
        emitEdges(program, p);
        p.endblock();
        p.close();
    }

    private void emitRecordDecls(Program program, Printer p) {
        for (Heap.Record r : program.heap.getRecords()) {
            String rn = getRecordLabel(r);
            p.print(rn + "[shape=record,label=\"");
            int cntr = 0;
            for (Heap.Cell f : r.layout.getCells()) {
                if (cntr > 0) p.print(" | ");
                String fname = getFieldName(f, cntr);
                p.print("<" + fname + "> " + fname);
                emitValue(r, f, p);
                cntr++;
            }
            p.println("\"];");
        }
    }

    private void emitValue(Heap.Record r, Heap.Cell f, Printer p) {
        Value fval = r.getValue(f.getIndex());
        if (fval instanceof Reference.Val) {
            p.print(": " + f.getType());
        } else if (fval instanceof VirgilDelegate.Val) {
            p.print(": " + f.getType());
        } else {
            p.print(" = " + fval);
        }
    }

    private void emitEdges(Program program, Printer p) {
        for (Heap.Record r : program.heap.getRecords()) {
            String rn = getRecordLabel(r);
            int cntr = 0;
            for (Heap.Cell f : r.layout.getCells()) {
                emitEdge(p, rn, r.getValue(f.getIndex()), getFieldName(f, cntr));
                cntr++;
            }
        }
    }

    private String getRecordLabel(Heap.Record r) {
        return StringUtil.quote(Heap.recordName(r));
    }

    private String getFieldName(Heap.Cell f, int cntr) {
        String n = f.getName();
        return "".equals(n) ? String.valueOf(cntr) : n;
    }

    private void emitEdge(Printer p, String rn, Value val, String name) {
        if (val instanceof Reference.Val && Reference.fromValue(val) != null) {
            p.println(rn + ":" + name + " -> " + StringUtil.quote(val) + "[label=" + name + "];");
        } else if (val instanceof VirgilDelegate.Val) {
            VirgilDelegate.Val d = VirgilDelegate.fromValue(val);
            if (d.record != null) {
                p.println(rn + ":" + name + " -> " + getRecordLabel(d.record) + "[label=" + name + "];");
            }
            if (d.method != null) {
                String mname = StringUtil.quote(d.method.getFullName());
                p.println(mname + "[shape=diamond];");
                p.println(rn + ":" + name + " -> " + mname + "[label=" + name + "];");
            }
        }
    }
}