/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Jan 11, 2006
 */

package vpc.tir.stages;

import cck.parser.AbstractToken;
import cck.text.StringUtil;
import cck.util.Util;
import vpc.core.*;
import vpc.core.concept.*;
import vpc.core.virgil.*;
import vpc.types.Type;
import vpc.types.TypeName;
import vpc.util.Maybe;
import java.io.BufferedReader;
import java.io.IOException;
import java.text.CharacterIterator;
import java.text.StringCharacterIterator;
import java.util.*;

/**
 * The <code>HeapParser</code> class can parse and load a heap from a textual file.
 * This functionality is useful for optimizing a pre-existing program, or comparing
 * two heaps in the process of testing the heap initialization code (TIRInterpreter).
 *
 * @author Ben L. Titzer
 */
public class HeapParser {

    protected final Program program;
    protected final VirgilProgram vp;
    protected final Closure closure;

    HashMap<Integer, Heap.Record> records;
    HashMap<TypeName, Heap.Layout> layouts;
    Heap heap;

    //============ H E A P   P A R S E R ====================================

    // HeapDecl: "heap \{\n" -> new Heap()
    // RecordDecl: "record #{rnum:INT} : {rsize:INT} : {rtype:Type} \{\n" -> getRecord(rnum, rsize, rtype)
    // Field: "field {fname:IDENT} : {ftype:Type} = {fval:Value} ;\n" -> addField(fname, ftype, fval)
    // End: "}" -> null

    // Type:      "( {type:Type} )"                             -> type;
    // Type-int:  "int"                                         -> Type.INT
    // Type-bool: "boolean"                                     -> Type.BOOLEAN
    // Type-char: "char"                                        -> Type.CHAR
    // Type-proc: "function ( {ptl:Type list} )"                -> Type.newFunc(ptl)
    // Type-fun:  "function ( {ptl:Type list} ) : {rtype:Type}" -> Type.newFunc(ptl, rtype)
    // Type-arr:  "{etype: Type} [ ]"                           -> Type.newArray(etype);

    // Value-int:  "int : {ival:INT}"                    -> Value.INT.valueOf(ival)
    // Value-bool: "boolean : {bval:BOOL}"               -> Value.BOOL.valueOf(bval)
    // Value-char: "char : {cval:INT}"                   -> Value.CHAR.valueOf(bval)
    // Value-null: "#null"                               -> NULL
    // Value-ref:  "#{rnum:INT} : {rtype:Type}"          -> new Value.REF(getRecord(rnum))
    // Value-del:  "( {ref:Value-ref} , {meth:Method} )" -> new Value.Val(ref, meth)
    // Method:     "{tn:IDENT} : {mn:IDENT} ( )"         -> resolveMethod(tn, mn)

    // Heap: "{h:HeapDecl} {Record*} {End}"
    // Record: "{rdecl:RecordDecl} {Field*} {End}"

    public HeapParser(Program p) {
        program = p;
        vp = program.virgil;
        closure = vp.program.closure;
    }

    public Heap parseHeap(BufferedReader input) throws Exception {
        heap = new Heap(program);
        records = new HashMap<Integer, Heap.Record>();
        layouts = new HashMap<TypeName, Heap.Layout>();
        skipToHeap(input);
        return heap;
    }

    void skipToHeap(BufferedReader input) throws Exception {
        String line;
        while ((line = input.readLine()) != null) {
            line = line.trim();
            if (line.startsWith("heap")) {
                readHeap(input);
                break;
            }
        }
    }

    private void readHeap(BufferedReader input) throws Exception {
        CharacterIterator line;
        while ((line = readLine(input)) != null) {
            if (StringUtil.peekAndEat(line, '}')) break;
            expectKeyword(line, "record");
            StringUtil.skipWhiteSpace(line);
            // read the record number
            StringUtil.expectChar(line, '#');
            int rnum = StringUtil.readDecimalValue(line, 12);
            StringUtil.expectChar(line, ':');
            int sz = StringUtil.readDecimalValue(line, 12);
            StringUtil.expectChar(line, ':');
            // read the layout
            Heap.Layout layout = parseLayout(line);
            Heap.Record r = getRecord(layout, rnum);
            // parse the fields in the record
            parseRecord(r, input);
        }
    }

    private Heap.Layout parseLayout(CharacterIterator line) throws Exception {
        TypeName tn = parseType(line);
        if (StringUtil.peekAndEat(line, '[')) {
            int len = StringUtil.readIntegerValue(line);
            StringUtil.expectChar(line, '[');
            return newArrayLayout(len, tn);
        }
        return getLayout(tn);
    }

    private Heap.Layout newArrayLayout(int len, TypeName etn) {
        VirgilArray.IType atype = VirgilArray.getArrayType(program.typeCache, etn.getType());
        return VirgilArray.getArrayLayout(heap, atype, len);
    }

    private Heap.Layout getLayout(TypeName t) {
        Heap.Layout layout = layouts.get(t);
        if (layout != null) return layout;
        VirgilComponent compDecl = vp.getComponentDecl(t.toString());
        if (compDecl != null) {
            layout = closure.getLayout(compDecl);
        } else {
            VirgilClass classDecl = vp.getClassDecl(t.toString());
            if (classDecl != null) {
                layout = closure.getLayout(classDecl);
            }
        }

        if (layout == null) throw Util.failure("unknown layout for type: " + t);
        Heap.Layout nl = heap.newLayout(layout.toString(), layout.type);
        layout.copy(nl);
        layouts.put(t, nl);
        return nl;
    }

    void parseRecord(Heap.Record r, BufferedReader input) throws Exception {
        CharacterIterator line;
        int cntr = 0;
        while ((line = readLine(input)) != null) {
            if (StringUtil.peekAndEat(line, '}')) break;
            parseField(r, line, cntr++);
        }
    }

    void parseField(Heap.Record r, CharacterIterator i, int num) throws Exception {
        expectKeyword(i, "field");
        StringUtil.skipWhiteSpace(i);
        String name = StringUtil.readIdentifier(i);
        StringUtil.expectChar(i, ':');
        Type staticType = parseType(i).getType();
        StringUtil.skipWhiteSpace(i);
        StringUtil.expectChar(i, '=');
        Value val = parseValue(i);

        r.setValue(num, val);
    }

    TypeName parseType(CharacterIterator i) throws Exception {
        StringUtil.skipWhiteSpace(i);
        TypeName tn;
        if (StringUtil.peekAndEat(i, '(')) {
            tn = parseType(i);
            StringUtil.expectChar(i, ')');
        } else {
            String ident = StringUtil.readIdentifier(i);
            if ("function".equals(ident)) {
                tn = parseFuncType(i);
            } else if ("raw".equals(ident)) {
                tn = parseRawType(i);
            } else {
                tn = program.typeCache.getTypeName(ident);
            }
        }
        tn = parseTypeSuffix(i, tn);
        return tn;
    }

    private TypeName parseRawType(CharacterIterator i) throws Exception {
        TypeName tn;
        StringUtil.expectChar(i, '.');
        int w = StringUtil.readDecimalValue(i, 12);
        tn = program.typeCache.getTypeName(PrimRaw.getType(w));
        return tn;
    }

    private TypeName parseTypeSuffix(CharacterIterator i, TypeName t) {
        while (StringUtil.peekAndEat(i, '[')) {
            if (StringUtil.peekAndEat(i, ']')) {
                t = VirgilArray.getArrayTypeName(program.typeCache, t);
            } else {
                i.previous();
                return t;
            }
        }
        return t;
    }

    private TypeName parseFuncType(CharacterIterator i) throws Exception {
        List<TypeName> pt = new LinkedList<TypeName>();
        StringUtil.expectChar(i, '(');
        parseParamTypes(i, pt);
        TypeName rt = StringUtil.peekAndEat(i, ':') ? parseType(i) : program.typeCache.getTypeName("void");
        return Function.getFuncTypeName(program.typeCache, rt, pt.toArray(Function.NOARGS));
    }

    private void parseParamTypes(CharacterIterator i, List<TypeName> pt) throws Exception {
        if (StringUtil.peekAndEat(i, ')')) return;
        while (true) {
            pt.add(parseType(i));
            if (StringUtil.peekAndEat(i, ')')) return;
            StringUtil.expectChar(i, ',');
        }
    }

    Value parseValue(CharacterIterator i) throws Exception {
        StringUtil.skipWhiteSpace(i);
        if (i.current() == '#') return parseRef(i);

        if (i.current() == '[') return parseDelegate(i);

        Type dynType = parseType(i).getType();
        StringUtil.expectChar(i, ':');

        if (dynType == PrimBool.TYPE) {
            String v = StringUtil.readIdentifier(i);
            return PrimBool.toValue(Boolean.valueOf(v));
        } else if (dynType == PrimInt32.TYPE) {
            int v = StringUtil.readIntegerValue(i);
            return PrimInt32.toValue(v);
        } else if (dynType == PrimChar.TYPE) {
            int v = StringUtil.readIntegerValue(i);
            return PrimChar.toValue((char) v);
        } else if ( dynType instanceof PrimRaw.IType ) {
            Maybe<AbstractToken> mt = PrimRaw.Lexer.consume(i);
            Maybe<PrimRaw.Val> mv = PrimRaw.Converter.convert(mt.getValue());
            return mv.getValue();
        } else if (dynType.isFunction()) {
            return parseDelegate(i);
        }

        throw Util.failure("cannot parse value for type " + dynType);
    }

    private Value parseDelegate(CharacterIterator i) throws Exception {

        StringUtil.expectChar(i, '[');
        Reference.Val ref = parseRef(i);
        StringUtil.expectChar(i, ',');
        String ident = StringUtil.readIdentifier(i);
        StringUtil.expectChar(i, ':');
        String meth = StringUtil.readIdentifier(i);
        StringUtil.expectChars(i, "()]");
        CompoundDecl decl = (CompoundDecl) program.getTypeDecl(ident);
        if (decl == null) throw Util.failure("unresolved type: " + ident);
        Method m = decl.getLocalMethod(meth);
        if (m == null) throw Util.failure("unknown method: " + meth);
        return new VirgilDelegate.Val(ref.record, m);
    }

    private Reference.Val parseRef(CharacterIterator i) throws Exception {
        // going to be a reference
        StringUtil.expectChar(i, '#');
        char ch = i.current();
        if (Character.isDigit(ch)) {
            // reference to a record
            int rnum = StringUtil.readDecimalValue(i, 12);
            StringUtil.expectChar(i, ':');
            Heap.Layout layout = parseLayout(i);
            return Reference.toValue(getRecord(layout, rnum));
        } else {
            expectKeyword(i, "null");
            return Reference.NULL_VALUE;
        }
    }

    private void expectKeyword(CharacterIterator i, String kw) {
        String str = StringUtil.readIdentifier(i);
        if (!str.equals(kw))
            Util.failure("parse error at " + i.getIndex() + ", expected keyword " + StringUtil.quote(kw));
    }

    private Heap.Record getRecord(Heap.Layout l, int rnum) {
        Heap.Record r = records.get(rnum);
        if (r == null) {
            r = l.newInstanceWithUID(rnum);
            r.setRoot(true);
            records.put(rnum, r);
        }
        return r;
    }

    private CharacterIterator readLine(BufferedReader r) throws IOException {
        String line = r.readLine();
        StringCharacterIterator i = new StringCharacterIterator(line);
        StringUtil.skipWhiteSpace(i);
        return i;
    }
}
