/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Mar 1, 2006
 */

package vpc.tir.stages;

import cck.util.Option;
import vpc.core.*;
import vpc.core.concept.*;
import vpc.core.virgil.*;
import vpc.model.Stage;
import vpc.types.Type;
import vpc.types.TypeName;
import vpc.vst.parser.Token;
import java.util.*;
import java.io.IOException;

/**
 * The <code>MetaLayout</code> class computes the layout for each object and component
 * in the program.
 *
 * @author Ben L. Titzer
 */
public class MetaLayout extends Stage {

    public static final String TID_FIELD = "__id";
    public static final String META_POSTFIX = "__meta";
    public static final String META_FIELD = "__meta";

    protected final Option.Bool ORPHANS = options.newOption("orphans", true,
            "This option controls whether the compiler will support efficient " +
            "implementation of orphan classes by removing the object header " +
            "and devirtualizing all calls to orphan methods.");

    public void visitProgram(Program p) throws IOException {
        Closure nc = p.closure.copy();
        // reset the class layouts
        nc.classLayouts = new HashMap<VirgilClass, Heap.Layout>();

        // compute the class hierarchy
        numberTIBs(nc);

        // compute the (augmented) layout for each class
        for (VirgilClass cd : p.virgil.getClasses()) {
            extendObjectLayout(cd, p.closure, nc);
        }

        // migrate each object record by adding the meta field
        for (Heap.Record r : p.heap.getRecords()) {
            migrate(r, nc);
        }

        p.closure = nc;
    }

    private void numberTIBs(Closure nc) {
        nc.maxClassID = 0;
        assignIds(nc, 0, computeRoots(nc));
    }

    private HashSet<VirgilClass> computeRoots(Closure nc) {
        HashSet<VirgilClass> roots = new HashSet<VirgilClass>();
        for ( VirgilClass c : nc.hierarchy.getNodes() ) {
            if ( nc.hierarchy.getParent(c) == null ) roots.add(c);
        }
        return roots;
    }

    private int assignIds(Closure nc, int id, Iterable<VirgilClass> decls) {
        int max = id;
        if (decls != null) {
            for (VirgilClass d : decls) {
                int thisId = nc.maxClassID++;
                d.minTypeID = thisId;
                d.maxTypeID = max = assignIds(nc, thisId, nc.hierarchy.getChildren(d));
            }
        }
        return max;
    }

    public void extendObjectLayout(VirgilClass cd, Closure oc, Closure nc) {
        if ( !nc.classLayouts.containsKey(cd) ) {
            Heap.Layout ol = oc.getLayout(cd);
            if (isOrphan(ORPHANS.get(), nc, cd)) {
                // if the class is an orphan, no need to extend the layout
                nc.classLayouts.put(cd, ol);
            } else {
                // extend the layout with the meta field
                Reference.Val meta = computeMetaLayout(nc.program.virgil, cd, nc);
                Heap.Layout nl = newLayout(nc.program.virgil, cd);
                addMetaField(nl, cd, meta, nc);
                nc.classLayouts.put(cd, ol.copy(nl));
            }
        }
    }

    public static boolean isOrphan(boolean supported, Closure nc, VirgilClass cd) {
        if ( !supported ) return false;
        return nc.hierarchy.isOrphan(cd);
    }

    public static Reference.Val computeMetaLayout(VirgilProgram vp, VirgilClass cd, Closure nc) {
        Type mt = getMetaType(vp, cd);
        Heap.Layout meta = vp.program.heap.newLayout(mt.toString(), mt);
        addMetaMembers(vp, cd, meta);
        nc.metaLayouts.put(cd, meta);
        Heap.Record r = meta.newInstance();
        return Reference.toValue(mt, r);
    }

    public static Type getMetaType(VirgilProgram vp, VirgilClass cd) {
        TypeName tn = vp.program.typeCache.getTypeName(cd.getName() + META_POSTFIX);
        if (tn.isResolved()) return tn.getType();
        TypeName ptn = null;
        VirgilClass pcd = vp.getParentClass(cd);
        if (pcd != null) ptn = vp.program.typeCache.getTypeName(getMetaType(vp, pcd));
        return new VirgilMetaClass.IType(cd, tn, ptn);
    }

    public static void addMetaMembers(VirgilProgram vp, VirgilClass cd, Heap.Layout meta) {
        List<VirgilClass> chain = vp.hierarchy.getChain(cd);
        addMetaTID(vp, meta, cd);
        for (VirgilClass pcd : chain) {
            for (Method meth : pcd.getMethods()) {
                Method.Family fam = meth.family;
                if (requiresMetaField(fam, meta)) {
                    addMetaField(vp, cd, meta, fam);
                }
            }
        }
    }

    public static boolean requiresMetaField(Method.Family fam, Heap.Layout meta) {
        if ( fam == null ) return false;
        return fam.metaField == null || meta.findCell(fam.metaField) == null;
    }

    public static void addMetaField(VirgilProgram vp, VirgilClass cd, Heap.Layout meta, Method.Family fam) {
        Method root = fam.rootMethod;
        String name = root.getName();
        if ( fam.metaField == null ) {
            fam.metaField = new Field(root.getCompoundDecl(), new Token(name), root.getTypeName());
        }
        Value val = new Function.Val(vp.resolveMethod(cd, name));
        meta.addCell(fam.metaField, val);
    }

    public static void addMetaTID(VirgilProgram vp, Heap.Layout meta, VirgilClass cd) {
        TypeName intName = vp.program.typeCache.getTypeName(PrimInt32.TYPE);
        meta.addCell(new Field(cd, new Token(TID_FIELD), intName), getTypeID(cd));
    }

    public static PrimInt32.Val getTypeID(VirgilClass cd) {
        return PrimInt32.toValue(cd.minTypeID);
    }

    public static Heap.Layout newLayout(VirgilProgram vp, CompoundDecl cd) {
        return vp.program.heap.newLayout(cd.getName(), cd.getTypeName().getType());
    }

    public static Heap.Layout newMetaLayout(VirgilProgram vp, VirgilClass cd) {
        Type mt = getMetaType(vp, cd);
        return vp.program.heap.newLayout(mt.toString(), mt);
    }

    public static void addMetaField(Heap.Layout l, VirgilClass cd, Reference.Val meta, Closure nc) {
        Token mft = new Token(META_FIELD);
        Type mt = nc.getMetaLayout(cd).type;
        TypeName tn = nc.program.typeCache.getTypeName(mt);
        cd.metaField = new Field(cd, mft, tn);
        l.addCell(cd.metaField, meta);
    }

    private void migrate(Heap.Record r, Closure nc) {
        // migrate an object records to use its new layout
        Type type = r.layout.type;
        if (type instanceof VirgilClass.IType) {
            VirgilClass.IType ct = (VirgilClass.IType)type;
            VirgilClass cd = ct.getDecl();
            Heap.Layout nl = nc.getLayout(cd);
            nl.migrate(r);
        }
    }
}
