/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Feb 4, 2007
 */
package vpc.tir.stages;

import vpc.core.Program;
import vpc.core.csr.*;
import vpc.core.virgil.*;
import vpc.core.concept.*;
import vpc.tir.*;
import vpc.tir.opt.DepthFirstTransformer;
import static vpc.tir.TIRUtil.*;
import vpc.model.Stage;
import vpc.types.Type;
import vpc.util.ArrayUtil;
import cck.parser.SourcePoint;

/**
 * The <code>ObjectConversion</code> class defines a compiler pass that
 * translates object-level and array-level operations to the lower-level
 * (CSR) representation in terms of structs and pointers.
 *
 * @author Ben L. Titzer
 */
public class ObjectConversion extends Stage {

    public void visitProgram(Program p) {
        for ( Method m : p.closure.methods ) {
            visitMethod(p, m);
        }
    }

    public void visitMethod(Program p, Method m) {
        TIRRep mrep = getRep(m);
        Transformer transformer = new Transformer(p);
        transformer.rep = mrep;
        TIRExpr nbody = transformer.transform(mrep.getBody(), null);
        mrep.setBody(nbody);
        m.addMethodRep(TIRRep.REP_NAME, mrep);
    }

    protected static class Transformer extends DepthFirstTransformer<Object> {

        protected CSRProgram csr;
        protected Program program;
        protected TIRRep rep;

        Transformer(Program p) {
            program = p;
            csr = p.csr;
        }

        public TIRExpr visit(TIROperator o, Object env) {

            switch ( getOpcode(o) ) {
                case VirgilOp.NULL_CHECK:
                    TIRExpr nobj = transform(o.operands, env)[0];
                    return $IF($ISNULL(nobj), throwExpr(o.getType(), "NullCheckException", o.getSourcePoint()), nobj);
                case VirgilOp.CLASS_GETFIELD: {
                    // a.f  =>  a-> f
                    VirgilClass.GetField op = (VirgilClass.GetField) o.operator;
                    CSRStruct.IType cst = csr.getCSRStructType(op.type);
                    return GETREF(cst, op.field, transform(o.operands, env));
                }
                case VirgilOp.CLASS_SETFIELD: {
                    // a.f = b  =>  a->f = b
                    VirgilClass.SetField op = (VirgilClass.SetField) o.operator;
                    CSRStruct.IType cst = csr.getCSRStructType(op.type);
                    return SETREF(cst, op.field, transform(o.operands, env));
                }
                case VirgilOp.CLASS_GETMETHOD: {
                    VirgilClass.GetMethod gm = (VirgilClass.GetMethod) o.operator;
                    if ( metaDispatch(gm) ) {
                        // e.m  =>  {e, e->meta->m}
                        CSRStruct.IType cst = csr.getCSRStructType(gm.type);
                        TIRExpr[] no = transform(o.operands, env);
                        Method.Family methodFam = gm.method.family;
                        CSRStruct.IType metacsr = getMetaCSRStructType(gm.type);
                        TIRExpr mref = GETREF(metacsr, methodFam.metaField, GETREF(cst, MetaLayout.META_FIELD, no));
                        return DELEG(no[0], mref);
                    } else {
                        // e.m  =>  {e, #E_m}
                        TIRExpr[] no = transform(o.operands, env);
                        return DELEG(no[0], $FUNCREF(gm.method));
                    }
                }
                case VirgilOp.COMPONENT_GETFIELD: {
                    // C.f  =>  #C->f
                    VirgilComponent.GetField gf = (VirgilComponent.GetField)o.operator;
                    VirgilComponent.IType type = (VirgilComponent.IType)gf.component.getTypeName().getType();
                    CSRStruct.IType cst = csr.getCSRStructType(type);
                    TIRConst.Value ref = $REF(type, gf.component.getRecord());
                    return GETREF(cst, gf.field, ref);
                }
                case VirgilOp.COMPONENT_SETFIELD: {
                    // C.f = a  =>  #C->f = a
                    VirgilComponent.SetField gf = (VirgilComponent.SetField)o.operator;
                    VirgilComponent.IType type = (VirgilComponent.IType)gf.component.getTypeName().getType();
                    CSRStruct.IType cst = csr.getCSRStructType(type);
                    TIRConst.Value ref = $REF(type, gf.component.getRecord());
                    TIRExpr[] no = new TIRExpr[] { ref, transform(o.operands, env)[0] };
                    return SETREF(cst, gf.field, no);
                }
                case VirgilOp.COMPONENT_GETMETHOD: {
                    // C.m  =>  {#C, #C_m}
                    VirgilComponent.GetMethod gm = (VirgilComponent.GetMethod)o.operator;
                    VirgilComponent.IType type = (VirgilComponent.IType)gm.component.getTypeName().getType();
                    TIRConst.Value ref = $REF(type, gm.component.getRecord());
                    return DELEG(ref, TIRUtil.$FUNCREF(gm.method));
                }
                case VirgilOp.ARRAY_GETLENGTH: {
                    CSRStruct.IType cst = csr.getCSRStructType(o.operands[0].getType());
                    return GETREF(cst, "length", transform(o.operands, env));
                }
                case VirgilOp.ARRAY_INIT: // TODO: implement initialization for arrays
                    return $REF(o.getType(), null);
                case VirgilOp.NEW_ARRAY: // TODO: implement allocation for arrays
                    return $REF(o.getType(), null);
                case VirgilOp.NEW_OBJECT: // TODO: implement allocation for objects
                    return $REF(o.getType(), null);
                case VirgilOp.TYPE_QUERY:
                    return typeQuery((VirgilClass.TypeQuery)o.operator, transform(o.operands, env)[0]);
                case VirgilOp.TYPE_CAST:
                    return typeCast(o, (VirgilClass.TypeCast)o.operator, transform(o.operands, env)[0]);
                case VirgilOp.ARRAY_GETELEMENT: {
                    CSRStruct.IType cst = csr.getCSRStructType(o.operands[0].getType());
                    CSRArray.IType car = csr.getCSRArrayType(o.operands[0].getType());
                    TIRExpr arr = transform(o.operands[0], env);
                    TIRExpr indx = checkIndex(o, env, cst, arr);
                    return $OP(new CSRArray.GetElement(car), GETARRAY(cst, arr, env), indx);
                }
                case VirgilOp.ARRAY_SETELEMENT: {
                    Type atype = o.operands[0].getType();
                    CSRStruct.IType cst = csr.getCSRStructType(atype);
                    CSRArray.IType car = csr.getCSRArrayType(atype);
                    TIRExpr arr = transform(o.operands[0], env);
                    TIRExpr indx = checkIndex(o, env, cst, arr);
                    TIRExpr val = transform(o.operands[2], env);
                    return $OP(new CSRArray.SetElement(car), GETARRAY(cst, arr, env), indx, val);
                }
                default:
                    return super.visit(o, env);
            }
        }

        private TIRExpr checkIndex(TIROperator o, Object env, CSRStruct.IType csr, TIRExpr arr) {
            TIRExpr indx = transform(o.operands[1], env);
            TIRExpr check = $AND($GREQ(indx, $VAL(0)), $LT(indx, GETREF(csr, "length", arr)));
            return $IF(check, indx, throwExpr(indx.getType(), "BoundsCheckException", o.getSourcePoint()));
        }

        private TIROperator GETARRAY(CSRStruct.IType csr, TIRExpr a, Object env) {
            return GETREF(csr, "values", a);
        }

        private TIRExpr typeQuery(VirgilClass.TypeQuery tq, TIRExpr e) {
            VirgilClass ttype = tq.target.getDecl();
            VirgilClass.IType ftype = (VirgilClass.IType)e.getType();
            if (ttype.minTypeID == ttype.maxTypeID) {
                if ( ttype.getParent() == null ) {
                    // orphan:  e <: T  =>  e != null
                    // TODO: correct implementation of type query for induced orphans
                    return $ISNOTNULL(e);
                }
                // leaf:  e <: T  =>  e != null && e->meta->tid == leaf;
                CSRStruct.IType cst = csr.getCSRStructType(ftype);
                CSRStruct.IType metacst = getMetaCSRStructType(ftype);
                TIRExpr mid = GETREF(metacst, MetaLayout.TID_FIELD, GETREF(cst, MetaLayout.META_FIELD, e));
                return $AND($ISNOTNULL(e), $EQ(mid, $VAL(ttype.minTypeID)));
            } else {
                // nonleaf:  e <: T  =>  e != null && e->meta->tid >= min && e->meta->tid <= max
                CSRStruct.IType cst = csr.getCSRStructType(ftype);
                CSRStruct.IType metacst = getMetaCSRStructType(ftype);
                TIRExpr mid = GETREF(metacst, MetaLayout.TID_FIELD, GETREF(cst, MetaLayout.META_FIELD, e));
                return $AND(
                        $AND(
                                $ISNOTNULL(e),
                                TIRUtil.$GREQ(mid, $VAL(ttype.minTypeID))),
                        TIRUtil.$LTEQ(mid, $VAL(ttype.maxTypeID)));
            }
        }

        private TIRExpr typeCast(TIROperator op, VirgilClass.TypeCast tq, TIRExpr e) {
            VirgilClass ttype = tq.target.getDecl();
            VirgilClass.IType ftype = (VirgilClass.IType)e.getType();
            if (ttype.minTypeID == ttype.maxTypeID) {
                if ( ttype.getParent() == null ) {
                    // orphan:  e <: T  =>  e != null
                    return e;
                }
                // leaf:  e <: T  =>  if ( e == null || e->meta->tid == leaf ) e; else $except;
                CSRStruct.IType cst = csr.getCSRStructType(ftype);
                CSRStruct.IType metacst = getMetaCSRStructType(ftype);
                TIRExpr mid = GETREF(metacst, MetaLayout.TID_FIELD, GETREF(cst, MetaLayout.META_FIELD, e));
                TIRExpr ok = $OR($ISNULL(e), $EQ(mid, $VAL(ttype.minTypeID)));
                TIRExpr check = $IF(ok, e, throwExpr(e.getType(), "TypeCheckException", op.getSourcePoint()));
                check.setType(ttype.getTypeName().getType());
                return check;
            } else {
                // nonleaf:  e <: T  =>  if ( e == null || e->meta->tid >= min && e->meta->tid <= max ) e; else $except;
                CSRStruct.IType cst = csr.getCSRStructType(ftype);
                CSRStruct.IType metacst = getMetaCSRStructType(ftype);
                TIRExpr mid = GETREF(metacst, MetaLayout.TID_FIELD, GETREF(cst, MetaLayout.META_FIELD, e));
                TIRExpr ok = $OR(
                        $ISNULL(e),
                        $AND(
                            TIRUtil.$GREQ(mid, $VAL(ttype.minTypeID)),
                            TIRUtil.$LTEQ(mid, $VAL(ttype.maxTypeID))));
                TIRExpr check = $IF(ok, e, throwExpr(e.getType(), "TypeCheckException", op.getSourcePoint()));
                check.setType(ttype.getTypeName().getType());
                return check;
            }
        }

        private TIRExpr throwExpr(Type t, String except, SourcePoint pt) {
            TIRThrow threxpr = new TIRThrow(except, pt);
            threxpr.setType(t);
            return threxpr;
        }

        private CSRStruct.IType getMetaCSRStructType(VirgilClass.IType ftype) {
            Type metaType = MetaLayout.getMetaType(program.virgil, ftype.getDecl());
            return csr.getCSRStructType(metaType);
        }

        private TIROperator DELEG(TIRExpr oref, TIRExpr mref) {
            return $OP(new CSRStruct.NewValue(csr.getDelegateCSRStruct()), oref, mref);
        }

        private TIROperator SETREF(CSRStruct.IType cst, Field fn, TIRExpr... no) {
            CSRPointer.IType ptr = CSRType.newPointer(csr, cst);
            CSRStruct.IType.Field csrfield = getCSRField(cst, fn);
            return $OP(new CSRStruct.SetRefField(ptr, csrfield), no);
        }

        private TIROperator GETREF(CSRStruct.IType cst, Field field, TIRExpr... no) {
            CSRPointer.IType ptr = CSRType.newPointer(csr, cst);
            CSRStruct.IType.Field csrfield = getCSRField(cst, field);
            return $OP(new CSRStruct.GetRefField(ptr, csrfield), no);
        }

        private TIROperator GETREF(CSRStruct.IType cst, String field, TIRExpr... no) {
            CSRPointer.IType ptr = CSRType.newPointer(csr, cst);
            CSRStruct.IType.Field csrfield = cst.getField(field);
            return $OP(new CSRStruct.GetRefField(ptr, csrfield), no);
        }

        private TIROperator GETVAL(CSRStruct.IType cst, String field, TIRExpr... no) {
            CSRStruct.IType.Field csrfield = cst.getField(field);
            return $OP(new CSRStruct.GetValueField(cst, csrfield), no);
        }

        private CSRStruct.IType.Field getCSRField(CSRStruct.IType csr, Field field) {
            return csr.getField(field.getUniqueName());
        }

        public TIRExpr visit(TIRCall c, Object env) {
            CSRFunction.IType ftype = csr.impl.getCVirtualType(c.func.getType());
            TIRExpr func;
            TIRExpr ptr;
            switch ( CSRGen.getOpcode(c.func) ) {
                case VirgilOp.CLASS_GETMETHOD: {
                    TIROperator op = (TIROperator)c.func;
                    VirgilClass.GetMethod gm = (VirgilClass.GetMethod) op.operator;
                    if ( metaDispatch(gm) ) {
                        // call (e.m, args)  =>  call(e->meta->m, e, args)
                        CSRStruct.IType cst = csr.getCSRStructType(gm.type);
                        TIRExpr[] no = transform(op.operands, env);
                        Method.Family methodFam = gm.method.family;
                        CSRStruct.IType metacsr = getMetaCSRStructType(gm.type);
                        TIRExpr mref = GETREF(metacsr, methodFam.metaField, GETREF(cst, MetaLayout.META_FIELD, no));
                        ptr = no[0];
                        func = mref;
                    } else {
                        // call (e.m, args)  =>  call(E_m, e, args)
                        ptr = transform(op.operands, env)[0];
                        func = $FUNCREF(gm.method);
                    }
                    break;
                }
                case VirgilOp.COMPONENT_GETMETHOD: {
                    // call(C.m, args)  =>  call(C_m, C, args)
                    TIROperator op = (TIROperator)c.func;
                    VirgilComponent.GetMethod gm = (VirgilComponent.GetMethod)op.operator;
                    ptr = $REF(gm.component.getRecord());
                    func = $FUNCREF(gm.method);
                    break;
                }
                default: {
                    // call(e, args)  =>  call(if ( (v = e.meth) == null ) throw NCE; else v, e.func, args)
                    TIRExpr deleg = transform(c.func, env);
                    Method.Temporary<Type> temp = rep.newTemporary(ftype);
                    CSRStruct.IType cst = csr.getDelegateCSRStruct();

                    ptr = GETVAL(cst, "ref", deleg);
                    func = $IF($ISNULL($SET(temp, GETVAL(cst, "method", deleg))),
                            throwExpr(temp.getType(), "NullCheckException", c.func.getSourcePoint()),
                            $GET(temp));
                }
            }
            func.setType(ftype);
            return $CALL(func, ArrayUtil.prepend(ptr, transform(c.arguments, env)));
        }

        private boolean metaDispatch(VirgilClass.GetMethod gm) {
            if ( !gm.virtual ) return false;
            Method.Family f = gm.method.family;
            return f != null && f.metaField != null;
        }

        private int getOpcode(TIROperator to) {
            Integer val = CSRGen.opMap.get(to.operator.getClass());
            if ( val == null ) return -1;
            return val;
        }
    }
}
