/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Mar 1, 2006
 */

package vpc.tir.stages;

import vpc.core.*;
import vpc.core.concept.*;
import vpc.core.virgil.*;
import vpc.model.Stage;
import vpc.tir.TIRInterpreter;
import vpc.types.Type;

import java.io.IOException;

/**
 * The <code>ObjectLayout</code> class computes the layout for each object and component
 * in the program.
 *
 * @author Ben L. Titzer
 */
public class ObjectLayout extends Stage {

    public void visitProgram(Program p) throws IOException {
        // compute layout for each component
        VirgilProgram vp = p.virgil;
        for (VirgilComponent cd : vp.getComponents()) computeLayout(vp, cd);

        // compute layout for each class
        for (VirgilClass cd : vp.getClasses()) computeLayout(vp, cd);

        // use the hierarchy of the declared program
        p.closure.hierarchy = vp.hierarchy;

    }

    private void computeLayout(VirgilProgram vp, VirgilComponent cd) {
        if (vp.program.closure.getLayout(cd) == null) {
            Heap.Layout nl = newLayout(vp, cd);
            vp.program.closure.componentLayouts.put(cd, nl);
            addFields(cd, nl);
            addMethods(cd, vp);
        }
    }

    private Heap.Layout computeLayout(VirgilProgram vp, VirgilClass cd) {
        Heap.Layout nl = vp.program.closure.getLayout(cd);
        if (nl == null) {
            nl = newLayout(vp, cd);
            vp.program.closure.classLayouts.put(cd, nl);
            VirgilClass pcd = vp.getParentClass(cd);
            if (pcd != null) {
                // if there is a parent class, copy its layout into this one
                Heap.Layout pl = computeLayout(vp, pcd);
                pl.copy(nl);
            }
            addFields(cd, nl);
            addMethods(cd, vp);
        }
        return nl;
    }

    private void addFields(CompoundDecl cd, Heap.Layout nl) {
        for (Field f : cd.getFields()) {
            Type type = f.getType();
            Value value = TIRInterpreter.getDefaultValue(type);
            Heap.Cell field = nl.addCell(f, value);
            f.fieldIndex = field.getIndex();
        }
    }

    private void addMethods(CompoundDecl cd, VirgilProgram vp) {
        Constructor cons = cd.getConstructor();
        if ( cons != null ) vp.program.closure.methods.add(cons);
        for (Method m : cd.getMethods()) {
            vp.program.closure.methods.add(m);
        }
    }

    private Heap.Layout newLayout(VirgilProgram vp, CompoundDecl cd) {
        return vp.program.heap.newLayout(cd.getName(), cd.getTypeName().getType());
    }

}
