/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Jan 2, 2006
 */

package vpc.tir.stages;

import vpc.core.*;
import vpc.core.concept.Constructor;
import vpc.core.concept.Method;
import vpc.core.virgil.VirgilClass;
import vpc.core.virgil.VirgilComponent;
import vpc.model.Stage;
import vpc.tir.*;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.IOException;
import java.util.Iterator;

/**
 * The <code>TIREmit</code> class implements a compilation phase that emits the
 * TIR representation of the program as text.
 *
 * @author Ben L. Titzer
 */
public class TIREmit extends Stage {

    public void visitProgram(Program p) throws IOException {
        FileOutputStream fos = new FileOutputStream(p.name + ".tir");
        IREmitVisitor visitor = new IREmitVisitor(new PrintStream(fos));
        for (VirgilClass c : p.virgil.getClasses()) visitor.visit(c);
        for (VirgilComponent c : p.virgil.getComponents()) visitor.visit(c);
        visitor.emitHeap(p.heap);
    }

    public static class IREmitVisitor extends ProgramPrinter {
        private TIRPrinter irp;

        public IREmitVisitor(PrintStream o) {
            super(TIRRep.REP_NAME, o);
            irp = new TIRPrinter(printer);
        }

        public void printMethodRep(Method m, Method.MethodRep rep) {
            printRep((TIRRep) rep);
        }

        private void printRep(TIRRep tcr) {
            printer.startblock();

            printTemps("params", tcr.getParams());

            TIRExpr b = tcr.getBody();
            if (b != null) {
                printTemps("temps", tcr.getTemps());
                printer.print("body ");
                tcr.getBody().accept(irp, null);
            }
            printer.endblock();
        }

        public void printConstructorRep(Constructor c, Constructor.MethodRep rep) {
            printRep((TIRRep) rep);
        }

        private void printTemps(String name, Iterable<? extends Variable> it) {
            Iterator<? extends Variable> i = it.iterator();
            if (!i.hasNext()) return;
            printer.startblock(name);
            while (i.hasNext()) {
                Variable t = i.next();
                printer.println(t.getName() + ": " + t.getType() + ";");
            }
            printer.endblock();
        }
    }
}
