/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Mar 15, 2006
 */
package vpc.tir.tir2c;

import cck.text.*;
import cck.util.Arithmetic;
import cck.util.Util;
import vpc.core.Heap;
import vpc.core.Program;
import vpc.core.concept.*;
import vpc.core.virgil.*;
import vpc.types.Type;

import java.util.HashMap;
import java.util.Map;

/**
 * The <code>CAccountant</code> class accounts for the RAM consumption in a Virgil program.
 * It can be used to generate a report about what types of objects are consuming the most
 * memory under various representation scenarios.
 *
 * @author Ben L. Titzer
 */
public abstract class CAccountant {

    public static final int REF_SIZE = 2;
    public static final int INT_SIZE = 2;
    public static final int BYTE_SIZE = 8;
    public static final int ALIGN = 8; // align on 8-bit boundaries

    protected class Usage {
        protected Heap.Layout layout;
        protected int count;

        protected int RAM_prim;
        protected int RAM_data;
        protected int RAM_meta;
        protected int RAM_align;
        protected int ROM_data;
        protected int ROM_meta;
        protected int ROM_ctbl;

        protected void addTo(Usage u) {
            u.RAM_prim += RAM_prim;
            u.RAM_data += RAM_data;
            u.RAM_align += RAM_align;
            u.RAM_meta += RAM_meta;
            u.ROM_data += ROM_data;
            u.ROM_meta += ROM_meta;
            u.ROM_ctbl += ROM_ctbl;
        }
    }

    protected static class BitSize {
        protected final boolean primitive;
        protected final int size;

        protected BitSize(boolean p, int s) {
            primitive = p;
            size = s;
        }
    }

    protected HashMap<Heap.Layout, Usage> layoutUsage = new HashMap<Heap.Layout, Usage>();
    protected Usage total = new Usage();

    public void account(Program p, CImpl impl) {

        for (Heap.Record r : p.heap.getRecords()) {
            computeUsage(r, impl).addTo(total);
        }

        reportRAMUsage();
        reportROMUsage();
    }

    protected static final boolean L = false;
    protected static final boolean R = true;
    protected static final int[] WIDTH = {24, 6, 6, 6, 6, 6, 6, 1, 9};
    protected static final boolean[] JUST = {L, R, R, R, R, R, R, L, L};

    private void reportRAMUsage() {
        Columnifier ramCol = new Columnifier(null, WIDTH, JUST);
        ramCol.setLeftMargin(1);

        ramCol.printTitle(Terminal.COLOR_GREEN, new String[] {"RAM Object", "Fields", "Prim", "Ref", "Meta", "Align", "Total", "", "Count" });
        for (Usage u : layoutUsage.values()) {
            reportRAMUsage(u, ramCol);
        }
        TermUtil.printThinSeparator();
        reportRAMUsage(total, ramCol);
    }

    private void reportRAMUsage(Usage u, Columnifier col) {
        int ram = u.RAM_prim + u.RAM_data + u.RAM_meta + u.RAM_align;
        if ( ram == 0 ) return;
        String col1 = u.layout == null ? "Total" : String.valueOf(u.layout.type);
        String col2 = u.layout == null ? "" : String.valueOf(u.layout.size());
        String col3 = String.valueOf(u.RAM_prim);
        String col4 = String.valueOf(u.RAM_data);
        String col5 = String.valueOf(u.RAM_meta);
        String col6 = String.valueOf(u.RAM_align);
        String col7 = String.valueOf(ram);
        String col8 = "";
        String col9 = u.layout == null ? "" : "x " + u.count;
        col.println(new String[] { col1, col2, col3, col4, col5, col6, col7, col8, col9 });
    }

    private void reportROMUsage() {

    }

    protected Usage computeUsage(Heap.Record r, CImpl impl) {
        Heap.Layout layout = r.layout;
        Usage u = layoutUsage.get(layout);
        if (u == null) {
            u = new Usage();
            u.layout = layout;
            computeSizes(impl, layout, u);
            layoutUsage.put(layout, u);
        }
        u.count++;
        return u;
    }

    protected void reset() {
        total = new Usage();
    }

    protected void computeSizes(CImpl impl, Heap.Layout layout, Usage u) {
        Type type = layout.type;
        if (type.isArray()) {
            computeArrayLayoutSize(type, layout, u);
        } else if (type instanceof VirgilMetaClass.IType) {
            computeMetaLayoutSize(layout, u);
        } else {
            computeObjectLayoutSize(layout, u);
        }
    }

    private void computeArrayLayoutSize(Type type, Heap.Layout layout, Usage u) {
        VirgilArray.IType atype = (VirgilArray.IType) type;
        int esize = 0;
        BitSize bs = getBitSize(atype.getElemType());
        for (int cntr = 0; cntr < layout.size(); cntr++) {
            addRAM(bs, u);
            esize += bs.size;
        }
        u.RAM_meta += INT_SIZE * BYTE_SIZE;
        u.RAM_align += alignment(esize);
    }

    private void addRAM(BitSize bs, Usage u) {
        if (bs.primitive) u.RAM_prim += bs.size;
        else u.RAM_data += bs.size;
    }

    private void computeObjectLayoutSize(Heap.Layout layout, Usage u) {
        int objsize = 0;
        for (Heap.Cell f : layout.getCells()) {
            BitSize bs = getBitSize(f.getType());
            objsize += bs.size;
            if (f.getType() instanceof VirgilMetaClass.IType) u.RAM_meta += bs.size;
            else addRAM(bs, u);

        }
        // objects must be byte aligned in memory
        u.RAM_align += alignment(objsize);
    }

    private void computeMetaLayoutSize(Heap.Layout layout, Usage u) {
        int metasize = 0;
        for (Heap.Cell f : layout.getCells()) {
            metasize += getMetaSize(f);
        }
        u.ROM_meta += metasize;
        // meta objects must be byte aligned
        u.ROM_meta += alignment(metasize);
    }

    protected int alignment(int size) {
        int mod = size % ALIGN;
        if (mod > 0) return ALIGN - mod;
        else return 0;
    }

    protected BitSize getBitSize(Type t) {
        if (t == PrimInt32.TYPE) return byteSize(true, INT_SIZE);
        if (t == PrimBool.TYPE) return byteSize(true, 1);
        if (t == PrimChar.TYPE) return byteSize(true, 1);
        if (t.isArray()) return byteSize(false, REF_SIZE);
        if (t.isObject()) return byteSize(false, REF_SIZE);
        if (t.isFunction()) return byteSize(false, 2 * REF_SIZE);
        if (t instanceof PrimRaw.IType) return bitSize(true, ((PrimRaw.IType)t).width);
        throw Util.failure("unknown size of type " + t);
    }

    protected BitSize byteSize(boolean primitive, int bytesize) {
        return new BitSize(primitive, bytesize * BYTE_SIZE);
    }

    protected BitSize bitSize(boolean primitive, int bs) {
        return new BitSize(primitive, bs);
    }

    protected int getMetaSize(Heap.Cell f) {
        if (f.getType().isFunction()) return REF_SIZE * BYTE_SIZE;
        else return getBitSize(f.getType()).size;
    }

    public static class Standard extends CAccountant {

    }

    public static class Compressed extends CAccountant {

        protected HashMap<Type, Integer> refSize;
        public HashMap<Type, Integer> typeCount;

        public void account(Program p, CImpl impl) {
            reset();
            compress(p, impl);
            super.account(p, impl);
        }

        protected void compress(Program p, CImpl impl) {
            countTypes(p);
            compressRefs();
        }

        private void compressRefs() {
            refSize = new HashMap<Type, Integer>();
            int ctable = 0;
            for (Map.Entry<Type, Integer> m : typeCount.entrySet()) {
                int count = m.getValue();
                Type type = m.getKey();
                if (!(type instanceof VirgilMetaClass.IType)) count += 1; // to account for null
                refSize.put(type, Arithmetic.highestBit(count) + 1);
                ctable += REF_SIZE * count * BYTE_SIZE;
            }
            total.ROM_ctbl += ctable;
        }

        private void countTypes(Program p) {
            typeCount = new HashMap<Type, Integer>();
            for (Heap.Record r : p.heap.getRecords()) {
                Type t = r.layout.type;
                if (t instanceof VirgilClass.IType) increment(getRootType((VirgilClass.IType) t));
                if (t instanceof VirgilMetaClass.IType) increment(getRootMetaType((VirgilMetaClass.IType) t));
                if (t instanceof VirgilArray.IType) increment(t);
            }
        }

        private VirgilMetaClass.IType getRootMetaType(VirgilMetaClass.IType mt) {
            for (VirgilMetaClass.IType pt = mt; pt != null; pt = pt.getParentMetaClass()) mt = pt;
            return mt;
        }

        private VirgilClass.IType getRootType(VirgilClass.IType mt) {
            for (VirgilClass.IType pt = mt; pt != null; pt = pt.getParentType()) mt = pt;
            return mt;
        }

        private Type getRootType(Type t) {
            if (t instanceof VirgilClass.IType) {
                return getRootType((VirgilClass.IType) t);
            } else if (t instanceof VirgilMetaClass.IType) {
                return getRootMetaType((VirgilMetaClass.IType) t);
            }
            return t;
        }

        private void increment(Type rtype) {
            Integer i = typeCount.get(rtype);
            if (i != null) i = i + 1; // another instance seen
            else i = 1; // first one
            typeCount.put(rtype, i);
        }

        protected BitSize getBitSize(Type t) {
            // in the compressed format, we use one bit for booleans
            if (t == PrimBool.TYPE) return bitSize(true, 1);
            // otherwise, look for the reference size
            Integer i = refSize.get(getRootType(t));
            if (i != null) return bitSize(false, i);
            // use 1 bit for references with no possible referencable objects
            if (t instanceof VirgilClass.IType) return bitSize(false, 1);
            // use 1 bit for references with no possible referencable objects
            if (t instanceof VirgilArray.IType) return bitSize(false, 1);
            // if no compressed version found, return the default
            return super.getBitSize(t);
        }

    }
}
