/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Feb 20, 2006
 */
package vpc.tir.tir2c;

import cck.text.Printable;
import cck.text.Printer;
import java.util.LinkedList;
import java.util.List;

/**
 * The <code>CIR</code> class represents a container class for classes that implement
 * an internal representation of C code. The C code internal representation is
 * used to compile from TIR to C.
 *
 * @author Ben L. Titzer
 */
public class CIR {

    /**
     * The <code>CExpr</code> class represents a C expression. It contains a
     * precedence number that can be used when generating a textual representation
     * when it is necessary to nest the expression in parentheses due to the precedence
     * rules of the C language.
     */
    public static class CExpr {
        protected final int precedence;
        protected final List<String> exprs;

        public CExpr(int prec, String... e) {
            exprs = new LinkedList<String>();
            for (String s : e) exprs.add(s);
            precedence = prec;
        }

        public String nest(int prec) {
            if (exprs.size() > 1) {
                return asCompoundExpr();
            } else if (precedence < prec) {
                return "(" + getFirst() + ")";
            } else {
                return getFirst();
            }
        }

        public String toString() {
            return nest(0);
        }

        protected void addExpr(String s) {
            exprs.add(s);
        }

        private String asCompoundExpr() {
            StringBuffer buf = new StringBuffer("({");
            for (String s : exprs) {
                buf.append(s);
                buf.append("; ");
            }
            buf.append(" })");
            return buf.toString();
        }

        private String getFirst() {
            return exprs.iterator().next();
        }
    }

    /**
     * The <code>CStmt</code> class represents a C statement.
     */
    public static abstract class CStmt implements Printable {
        protected List<CStmt> unNest() {
            return null;
        }

        protected abstract String asString();
    }

    public static class CSingle extends CStmt {
        protected final String rep;

        public CSingle(String r) {
            rep = r;
        }

        public void print(Printer p) {
            p.print(rep);
            p.println(";");
        }

        public String asString() {
            return rep + ";";
        }
    }

    /**
     * The <code>CIfStmt</code> class represents a C if statement that contains a condition
     * (rendered as a string), and two statements, one for the true branch, and one for the
     * false branch.
     */
    public static class CIfStmt extends CStmt {
        protected final String before;
        protected final CStmt trueBody;
        protected final CStmt falseBody;

        public CIfStmt(String b, CStmt ts, CStmt fs) {
            before = b;
            trueBody = ts;
            falseBody = fs;
        }

        public void print(Printer p) {
            p.print(before);
            trueBody.print(p);
            p.nextln();
            if (falseBody != null) {
                p.print("else ");
                falseBody.print(p);
            }
        }

        public String asString() {
            String ec = falseBody == null ? "" : " else " + falseBody.asString();
            return before + " " + trueBody.asString() + ec;
        }
    }

    /**
     * The <code>CBlock</code> class represents a block of C statements. It may contain
     * code before and after the block. For example, a "for" loop is represented as a block
     * with the code "for (...) " at the beginning.
     */
    public static class CBlock extends CStmt {
        protected String before;
        protected CStmt body;
        protected String after;

        public CBlock(String bf, CStmt b, String a) {
            before = bf;
            body = b;
            after = a;
        }

        public void print(Printer p) {
            p.startblock(before);
            body.print(p);
            p.endblock(after);
        }

        public String asString() {
            return before + " " + body.asString() + " " + after;
        }
    }

    /**
     * The <code>CNested</code> class represents a nested list of statements that can be
     * un-nested when generating code.
     */
    public static class CNested extends CStmt {
        protected final List<CStmt> body;

        public CNested(List<CStmt> b) {
            body = b;
        }

        public void print(Printer p) {
            p.startblock();
            for (CStmt c : body) c.print(p);
            p.endblock();
        }

        protected List<CStmt> unNest() {
            return body;
        }

        public String asString() {
            StringBuffer buf = new StringBuffer("{");
            for (CStmt st : body) buf.append(st.asString());
            buf.append("}");
            return buf.toString();
        }
    }
}
