/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Jan 29, 2006
 */
package vpc.tir.tir2c;

import cck.text.StringUtil;
import cck.util.Util;
import vpc.core.*;
import vpc.core.csr.*;
import vpc.core.concept.*;
import vpc.core.virgil.*;
import vpc.test.TIRtoCHarness;
import vpc.types.Type;
import java.util.*;

/**
 * The <code>CImpl</code> class represents a collection of all the information
 * about the program that is needed to generate C code. For example, the implementation
 * of subtype tests, the representation of types, and the C data structures and
 * functions needed to implement the program are represented here.
 *
 * @author Ben L. Titzer
 */
public class CImpl extends CSRGen implements CSRProgram.Impl {

    public static final String NULL = "(void*)0";
    public static final String NULL_DELEGATE = "((struct delegate){ (void*)0, (void*)0 })";

    public static final CSRType.Unit RAW64 = new CSRType.Unit("unsigned long long");
    public static final CSRType.Unit RAW32 = new CSRType.Unit("unsigned int");
    public static final CSRType.Unit RAW16 = new CSRType.Unit("unsigned short");
    public static final CSRType.Unit RAW8 = new CSRType.Unit("unsigned char");

    public static final CSRType.Unit INT32 = new CSRType.Unit("int");
    public static final CSRType.Unit CHAR = new CSRType.Unit("char");
    public static final CSRType.Void VOID = new CSRType.Void();
    public static final CSRPointer.VoidPtr VOIDPTR = new CSRPointer.VoidPtr();
    public static final CSRType.Unit BOOL = CHAR;

    public final Program program;
    public final CSRProgram csr;

    public CImpl(Program p) {
        program = p;
        csr = p.csr;
    }

    protected void makeSafetyChecks(Program p) {
        makeZeroCheck();
    }

    private void makeZeroCheck() {
        CSRFunction nc = new CSRFunction("zero_check", INT32);
        nc.addParams(csr, INTVAR("val"), INTVAR("loc"));
        nc.setBody(IF(EXPR("val == 0"),
                STMT("throw("+ TIRtoCHarness.ABORT_DIV_CODE+", loc)"),
                RETURN(EXPR("val"))));
        csr.functions.add(nc);
    }

    protected void layoutComponents(Program p) {
        for (VirgilComponent d : p.virgil.getComponents()) {
            CSRStruct.IType cst = csr.getCSRStructType(d.getTypeName().getType());
            assert cst != null;
        }
    }

    protected void layoutClasses(Program p) {
        for (VirgilClass d : p.virgil.getClasses()) {
            Heap.Layout ml = p.closure.getMetaLayout(d);
            if ( ml != null) {
                CSRStruct.IType mst = csr.getCSRStructType(ml.type);
                assert mst != null;
            }
            CSRStruct.IType cst = csr.getCSRStructType(d.getTypeName().getType());
            assert cst != null;
        }
    }

    protected void addFields(Heap.Layout layout, CSRStruct.IType st) {
        for (Heap.Cell f : layout.getCells())
            st.addField(CUtil.getCellName(f), csr.getCSRType(f.getType()));
    }

    protected void layoutDecl(CompoundDecl d, CSRStruct.IType st) {
        Heap.Layout layout = program.closure.getLayout(d);
        if (!st.fields.isEmpty() ) return;
        addFields(layout, st);
        if (st.fields.isEmpty() ) {
            st.addField("empty", INT32);
        }
    }

    public CSRFunction.IType getCVirtualType(Type t) {
        Function.IType ft = (Function.IType) t;
        CSRType rtype = csr.getCSRType(ft.getReturnType());
        LinkedList<CSRType> ptypes = new LinkedList<CSRType>();
        ptypes.add(VOIDPTR);
        for (Type tn : ft.getArgumentTypes())
            ptypes.add(csr.getCSRType(tn));
        return CSRType.newFuncPtr(csr, ptypes, rtype);
    }

    public static String renderRef(Heap.Record r) {
        if (r == null) return NULL;
        else return "(void *)&obj_" + r.uid;
    }

    protected String renderValue(Value value, Type type) {
        if (value instanceof PrimInt32.Val) return Integer.toString(PrimInt32.fromValue(value));
        if (value instanceof PrimBool.Val) return PrimBool.fromValue(value) ? "1" : "0";
        if (value instanceof PrimChar.Val) return Integer.toString(PrimChar.fromValue(value));
        if (value instanceof Reference.Val) {
            return renderRef(Reference.fromValue(value), type);
        }
        if (value instanceof VirgilDelegate.Val) {
            VirgilDelegate.Val deleg = VirgilDelegate.fromValue(value);
            return "{ " + renderRef(deleg.record) + ", " + renderMethRef(deleg.method) + " }";
        }
        if (value instanceof Function.Val) {
            return renderMethRef(((Function.Val) value).method);
        }
        if ( value instanceof PrimRaw.Val ) {
            PrimRaw.Val v = (PrimRaw.Val)value;
            return StringUtil.to0xHex(v.value, (v.width + 3) / 4);
        }
        throw Util.failure("cannot render value: " + value);
    }

    protected String renderRef(Heap.Record r, Type t) {
        if (r == null) {
            if (t.isFunction()) return NULL_DELEGATE;
            else return NULL;
        } else {
            CSRType vt = csr.getCSRType(r.layout.type);
            CSRType ct = csr.getCSRType(t);
            if ( vt == ct ) return "(&obj_" + r.uid+")";
            else return "(("+ct+")&obj_" + r.uid+")";
        }
    }

    protected static String renderMethRef(Method m) {
        if (m == null) return NULL;
        else return CUtil.getCFunctionName(m);
    }

    public CSRStruct.IType getDelegateStruct() {
        CSRStruct.IType str = csr.newStruct("delegate");
        str.addField("ref", VOIDPTR);
        str.addField("method", VOIDPTR);
        return str;
    }

    public CSRStruct.IType getVirgilClassStruct(VirgilClass.IType c) {
        VirgilClass d = c.getDecl();
        CSRStruct.IType classStruct = csr.newStruct(c, CUtil.mangleTypeName(d));
        layoutDecl(d, classStruct);
        return classStruct;
    }

    public CSRStruct.IType getVirgilComponentStruct(VirgilComponent.IType c) {
        VirgilComponent d = program.virgil.getComponentDecl(c.toString());
        CSRStruct.IType compStruct = csr.newStruct(c, CUtil.mangleTypeName(d));
        layoutDecl(d, compStruct);
        return compStruct;
    }

    public CSRStruct.IType getVirgilArrayStruct(VirgilArray.IType c) {
        CSRArray.IType at = getVirgilArrayType(c);
        CSRType et = at.elemType;
        CSRStruct.IType arrstruct = csr.newStruct(c, getElemName(et) + "_array");
        if (arrstruct.fields.isEmpty() ) {
            // only add these fields if the struct has not been filled out yet
            arrstruct.addField("length", INT32);
            arrstruct.addField("values", at);
        }
        return arrstruct;
    }

    public CSRArray.IType getVirgilArrayType(VirgilArray.IType at) {
        Type t = at.getElemType();
        if (t == PrimInt32.TYPE || t == INT32) return CSRType.newArray(csr, INT32);
        if (t == PrimChar.TYPE || t == CHAR) return CSRType.newArray(csr, CHAR);
        if (t == PrimBool.TYPE || t == BOOL ) return CSRType.newArray(csr, BOOL);
        if (t.isFunction()) return CSRType.newArray(csr, csr.getDelegateCSRStruct());
        if ( t instanceof PrimRaw.IType ) {
            PrimRaw.IType rt = (PrimRaw.IType)t;
            if ( rt.width <= 8 ) return CSRType.newArray(csr, RAW8);
            if ( rt.width <= 16 ) return CSRType.newArray(csr, RAW16);
            if ( rt.width <= 32 ) return CSRType.newArray(csr, RAW32);
            if ( rt.width <= 64 ) return CSRType.newArray(csr, RAW64);
        }
        return CSRType.newArray(csr, VOIDPTR);
    }

    public CSRStruct.IType getVirgilMetaClassStruct(VirgilMetaClass.IType c) {

        Heap.Layout layout = program.closure.getMetaLayout(c.classDecl);
        if ( layout != null ) {
            String name = CUtil.mangleTypeName((VirgilMetaClass.IType)layout.type);
            CSRStruct.IType meta = csr.newStruct(layout.type, name);
            for (Heap.Cell f : layout.getCells()) {
                Type type = f.getType();
                CSRType ctype = type.isFunction() ? getCVirtualType(type) : csr.getCSRType(type);
                meta.addField(f.getName(), ctype);
            }
            return meta;
        }
        return null;
    }

    public CSRType getPrimitiveType(Type t) {
        if (t == Reference.NULL_TYPE) return VOIDPTR;
        if (t == PrimVoid.TYPE) return VOID;
        if (t == PrimInt32.TYPE) return INT32;
        if (t == PrimChar.TYPE) return CHAR;
        if (t == PrimBool.TYPE) return CHAR;
        if ( t instanceof PrimRaw.IType ) {
            PrimRaw.IType rt = (PrimRaw.IType)t;
            if ( rt.width <= 8 ) return RAW8;
            if ( rt.width <= 16 ) return RAW16;
            if ( rt.width <= 32 ) return RAW32;
            if ( rt.width <= 64 ) return RAW64;
        }
        return null;
    }

    private String getElemName(CSRType et) {
        String nm;
        if (et == INT32) nm = "int";
        else if (et == CHAR) nm = "char";
        else if (et == BOOL ) nm = "char";
        else if (et == csr.getDelegateCSRStruct()) nm = "func";
        else if ( et == VOIDPTR ) nm = "ptr";
        else if ( et == RAW8 ) nm = "raw8";
        else if ( et == RAW16 ) nm = "raw16";
        else if ( et == RAW32 ) nm = "raw32";
        else if ( et == RAW64 ) nm = "raw64";
        else {
            throw Util.failure("unknown array element type: "+et);
        }
        return nm;
    }

}
