/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Feb 17, 2006
 */

package vpc.tir.tir2c;

import vpc.core.*;
import vpc.core.csr.*;
import vpc.core.concept.Method;
import vpc.core.concept.PrimChar;
import vpc.core.virgil.VirgilError;
import vpc.core.virgil.VirgilArray;
import vpc.hil.Interrupt;
import vpc.types.Type;

/**
 * The <code>CLinkage</code> class represents a linkage model for generating C code
 * as output of the Virgil compiler. The linkage model defines how entrypoints to
 * the Virgil program are connected with C glue code. For example, what version of the
 * \"main\" C function should be generated, how interrupts should be attached to
 * entrypoints, etc.
 *
 * @author Ben L. Titzer
 */
public abstract class CLinkage extends CSRGen {

    public void generateLinkage(Program p, TIREmitC c) {
        addFunction(c, newMain(p));
        addFunction(c, newThrowFunc(p.csr));
    }

    protected void addFunction(TIREmitC c, CSRFunction cf) {
        c.impl.program.csr.functions.add(cf);
    }

    protected CSRFunction newThrowFunc(CSRProgram csr) {
        CSRFunction func = new CSRFunction("throw", CImpl.INT32);
        func.addParams(csr, INTVAR("code"), INTVAR("loc"));
        addThrowFuncImpl(func);
        return func;
    }

    protected void addStringParams(CSRProgram csr, CSRFunction cf) {
        cf.addParams(csr, INTVAR("argc"), VAR("argv", ptr(csr, ptr(csr, CImpl.CHAR))));
    }

    protected abstract CSRFunction newMain(Program p);

    protected abstract void addThrowFuncImpl(CSRFunction func);

    /**
     * The <code>CTestLinkage</code> class represents a linkage model for the built-in
     * testing framework. This linkage model generates a C "main" function that converts
     * its first argument to an integer and then passes it to the main method declared
     * in the test case. The return value of the program's entrypoint point is then
     * returned by the main, making it the exit code of the process.
     */
    public static class CTestLinkage extends CLinkage {

        protected CSRFunction newMain(Program p) {
            CSRFunction cf = new CSRFunction("main", CImpl.INT32);
            addStringParams(p.csr, cf);
            String entry = CUtil.getCFunctionName(ProgramDecl.resolveEntryPoint("main", p));
            cf.setBody(new CIR.CSingle("return " + entry + "(" + CImpl.NULL + ", atoi(argv[1]))"));
            return cf;
        }

        protected void addThrowFuncImpl(CSRFunction func) {
            func.setBody(STMT("exit(code)"));
        }

    }

    /**
     * The <code>CAVRLinkage</code> class represents a linkage model for the AVR
     * microcontroller. This linkage model defines main with no arguments and generates
     * the appropriate vector declarations that connect the declared interrupt entrypoints
     * in the program to the actual interrupt handlers.
     */
    public static class CAVRLinkage extends CLinkage {
        public void generateLinkage(Program p, TIREmitC c) {
            super.generateLinkage(p, c);
            if (p.targetDevice != null) generateHandlers(p, c);
        }

        protected CSRFunction newMain(Program p) {
            CSRFunction cf = new CSRFunction(p.csr, "main", CImpl.INT32, CSRFunction.varList());
            String entry = CUtil.getCFunctionName(ProgramDecl.resolveEntryPoint("main", p));
            cf.setBody(new CIR.CSingle(entry + "(" + CImpl.NULL + "); return 0"));
            return cf;
        }

        private void generateHandlers(Program p, TIREmitC c) {
            for (ProgramDecl.EntryPoint pt : p.programDecl.entryPoints) {
                String ename = pt.getName();
                if ("main".equals(ename) ) continue;
                Interrupt i = p.targetDevice.interrupts.get(ename);
                if (i == null)
                    throw VirgilError.UnresolvedEntrypoint.gen(pt.getSourcePoint(), ename);
                Method md = pt.resolve(p);
                CSRFunction cf = new CSRFunction(p.csr, "__vector_" + i.number.image, CImpl.VOID, CSRFunction.varList());
                cf.addAttribute("signal");
                String entry = CUtil.getCFunctionName(md);
                cf.setBody(new CIR.CSingle(entry + "(" + CImpl.NULL + ")"));
                c.impl.program.csr.functions.add(cf);
            }
        }

        protected void addThrowFuncImpl(CSRFunction func) {
            func.setBody(STMT("asm(\"break\")"));
        }
    }

    /**
     * The <code>CUserLinkage</code> class represents a linkage model for generating
     * user-space C programs that run as a native process. It generates a main method
     * that converts the arguments into the character array representation used by
     * the Virgil language implementation.
     */
    public static class CUserLinkage extends CLinkage {
        protected CSRFunction newMain(Program p) {
            CSRFunction cf = new CSRFunction("main", CImpl.INT32);
            addStringParams(p.csr, cf);
            String entry = CUtil.getCFunctionName(ProgramDecl.resolveEntryPoint("main", p));
            CSRPointer.IType charArrayType = getArrayType(p, PrimChar.TYPE);
            CSRPointer.IType ptrArrayType = getArrayType(p, getArrayType(p, PrimChar.TYPE));
            cf.setBody(NEST(
                    DECLARE(cf, "cntr", CImpl.INT32, EXPR("1")),
                    DECLARE(cf, "aa", ptrArrayType,
                            EXPR("("+ptrArrayType+")malloc(sizeof("+ptrArrayType.ptype+")+argc*sizeof(void*))")),
                    ASSIGN("aa->length", "argc-1"),
                    FOR(null,
                            EXPR("cntr < argc"),
                            EXPR("cntr++"),
                            NEST(DECLARE(cf, "len", CImpl.INT32,
                                    EXPR("strlen(argv[cntr])")),
                                    DECLARE(cf, "buf", charArrayType,
                                            EXPR("("+charArrayType+")malloc(sizeof("+charArrayType.ptype+")+len+1)")),
                                    ASSIGN("buf->length", "len"),
                                    STMT("strcpy(buf->values, argv[cntr])"),
                                    ASSIGN("aa->values[cntr-1]", "buf"))),
                    RETURN(CALL(entry,
                            EXPR(CImpl.NULL),
                            EXPR("aa")))));
            return cf;
        }

        private CSRPointer.IType getArrayType(Program p, Type et) {
            return (CSRPointer.IType)p.csr.getCSRType(VirgilArray.getArrayType(p.typeCache, et));
        }

        protected void addThrowFuncImpl(CSRFunction func) {
            func.setBody(STMT("exit(code)"));
        }

    }

    private static CSRPointer.IType ptr(CSRProgram csr, CSRType type) {
        return CSRType.newPointer(csr, type);
    }

}
