/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Jan 27, 2006
 */
package vpc.tir.tir2c;

import cck.text.StringUtil;
import cck.util.Util;
import cck.parser.SourcePoint;
import vpc.core.Value;
import vpc.core.csr.*;
import vpc.core.concept.*;
import vpc.core.virgil.*;
import vpc.hil.DeviceDecl;
import vpc.hil.Register;
import vpc.test.TIRtoCHarness;
import vpc.tir.*;
import vpc.tir.expr.Operator;
import static vpc.tir.expr.Precedence.*;
import vpc.tir.stages.MetaLayout;
import vpc.types.Type;
import java.util.*;

/**
 * The <code>TIRCompiler</code> class implements a translation between TIR code
 * and C code. It can compile a TIR representation of a method into a
 * <code>CSRFunction</code>.
 *
 * @author Ben L. Titzer
 */
public class TIRCompiler {

    protected final ExprCompiler EXP;
    protected final StmtCompiler STMT;
    protected final CImpl impl;
    protected final CSRProgram csr;
    protected final CSRStruct.IType delegate;

    protected CSRFunction func;

    protected static final boolean ELIDE_ELSE = true;

    public TIRCompiler(CImpl ci) {
        impl = ci;
        csr = ci.program.csr;
        EXP = new ExprCompiler();
        STMT = new StmtCompiler();
        delegate = csr.getDelegateCSRStruct();
    }

    protected static String $eval(String pat, Object... objs) {
        return StringUtil.stringReplace(pat, new Properties(), objs);
    }

    protected static CIR.CExpr $expr(int prec, String pat, Object... objs) {
        return new CIR.CExpr(prec, StringUtil.stringReplace(pat, new Properties(), objs));
    }

    protected static CIR.CStmt $stmt(String pat, Object... objs) {
        return new CIR.CSingle(StringUtil.stringReplace(pat, new Properties(), objs));
    }

    /**
     * The <code>compile()</code> method compiles a given C function. It translates
     * the TIR representation of the method to a <code>CIR</code> implementation
     * that can be used to output C code.
     *
     * @param cf the C function to compile.
     */
    public void compile(CSRFunction cf) {
        func = cf;
        TIRRep tr = TIRUtil.getRep(cf.method);
        for (Method.Temporary t : tr.getTemps()) {
            cf.newVariable(t.getName(), csr.getCSRType(t.getType()));
        }
        cf.setBody(tr.getBody().accept(STMT, CImpl.VOID));
    }

    protected class ExprCompiler extends TIRExprVisitor<CIR.CExpr, CSRType> {

        public CIR.CExpr visit(TIRExpr e, CSRType ct) {
            throw Util.failure("compile(" + e.getClass() + ") unimplemented in TIR->C @ " + e.getSourcePoint());
        }

        public CIR.CExpr visit(TIRBlock s, CSRType ctype) {
            if (s.isEmpty()) {
                throw TIRUtil.fail("cannot convert empty block to expression", s);
            } else if (s.isSingleton()) {
                return s.getFirst().accept(EXP, ctype);
            }
            CIR.CExpr ce = new CIR.CExpr(PREC_TERM);
            int cntr = 1, sz = s.list.size();
            for (TIRExpr e : s.list) {
                if (cntr++ == sz) ce.addExpr(CEX(e, ctype));
                else ce.addExpr(e.accept(STMT, csr.getCSRType(e.getType())).asString());
            }
            return ce;
        }

        public CIR.CExpr visit(TIRIfExpr i, CSRType ctype) {
            return $expr(PREC_IF_EXPR, "($1) ? ($2) : ($3)",
                    CEX(i.condition, CImpl.BOOL),
                    CEX(i.true_target, ctype),
                    CEX(i.false_target, ctype));
        }

        public CIR.CExpr visit(TIROperator to, CSRType ctype) {
            Operator op = to.operator;

            switch ( CSRGen.getOpcode(to) ) {
                // check against the known operators for integers
                case CSRGen.INT32_ADD: return binop(to, "+", PREC_ADD);
                case CSRGen.INT32_SUB: return binop(to, "-", PREC_SUBTRACT);
                case CSRGen.INT32_MUL: return binop(to, "*", PREC_MULTIPLY);
                case CSRGen.INT32_DIV: return divide(to, "/", PREC_DIVIDE);
                case CSRGen.INT32_MOD: return divide(to, "%", PREC_MODULUS);
                case CSRGen.INT32_LT: return binop(to, "<", PREC_LESS_THAN);
                case CSRGen.INT32_GR: return binop(to, ">", PREC_GREATER);
                case CSRGen.INT32_LTEQ: return binop(to, "<=", PREC_LESS_EQUAL);
                case CSRGen.INT32_GREQ: return binop(to, ">=", PREC_GREATER_EQUAL);
                case CSRGen.INT32_NEG: return unop(to, "-", PREC_MINUS);

                // check against the raw operators
                case CSRGen.RAW_OR: return binop(to, "|", PREC_BITWISE_OR);
                case CSRGen.RAW_AND: return binop(to, "&", PREC_BITWISE_AND);
                case CSRGen.RAW_XOR: return binop(to, "^", PREC_BITWISE_XOR);
                case CSRGen.RAW_SHR: return binop(to, ">>", PREC_SHIFT_RIGHT);
                case CSRGen.RAW_COMP: return unop(to, "~", PREC_COMP);
                case CSRGen.RAW_SHL: return binop(to, "<<", PREC_SHIFT_LEFT);

                // check against the known operators for characters
                case CSRGen.CHAR_LT: return binop(to, "<", PREC_LESS_THAN);
                case CSRGen.CHAR_GR: return binop(to, ">", PREC_GREATER);
                case CSRGen.CHAR_LTEQ: return binop(to, "<=", PREC_LESS_EQUAL);
                case CSRGen.CHAR_GREQ: return binop(to, ">=", PREC_GREATER_EQUAL);

                // check against operators for booleans
                case CSRGen.BOOL_AND: return binop(to, "&&", PREC_LOGICAL_AND);
                case CSRGen.BOOL_OR: return binop(to, "||", PREC_LOGICAL_OR);
                case CSRGen.BOOL_NOT: return unop(to, "!", PREC_NOT);

                // check against struct operators
                case CSRGen.CSRSTRUCT_GETREFFIELD:
                    return getMember(to, ((CSRStruct.GetRefField)op).field, "->");
                case CSRGen.CSRSTRUCT_GETVALFIELD:
                    return getMember(to, ((CSRStruct.GetValueField)op).field, ".");
                case CSRGen.CSRSTRUCT_SETREFFIELD:
                    return setMember(to, ((CSRStruct.SetRefField)op).field, "->");
                case CSRGen.CSRSTRUCT_SETVALFIELD:
                    return setMember(to, ((CSRStruct.SetValueField)op).field, ".");
                case CSRGen.CSRSTRUCT_NEWVAL:
                    CSRStruct.NewValue nv = (CSRStruct.NewValue)op;
                    StringBuffer buf = new StringBuffer("((");
                    buf.append(nv.structType);
                    buf.append("){");
                    int cntr = 0;
                    for (TIRExpr e : to.operands) {
                        if (cntr > 0) buf.append(", ");
                        buf.append(CEX(e));
                        cntr++;
                    }
                    buf.append("})");
                    return new CIR.CExpr(0, buf.toString());

                    // check against C array operators
                case CSRGen.CSRARRAY_GETELEM:
                    return $expr(PREC_INDEX, "$1[$2]",
                            CEX(PREC_INDEX, to.operands[0]),
                            CEX(PREC_INDEX, to.operands[1]));
                case CSRGen.CSRARRAY_SETELEM:
                    return $expr(PREC_ASSIGN, "$1[$2] = $3",
                            CEX(PREC_INDEX, to.operands[0]),
                            CEX(to.operands[1]),
                            CEX(PREC_ASSIGN, to.operands[2]));

                    // check against C pointer operators
                case CSRGen.CSRPTR_GET:
                    return $expr(PREC_DEREF, "*$1",
                            CEX(PREC_DEREF, to.operands[0]));
                case CSRGen.CSRPTR_SET:
                    return $expr(PREC_ASSIGN, "*$1 = $2",
                            CEX(PREC_UNKNOWN, to.operands[0]),
                            CEX(PREC_ASSIGN, to.operands[1]));

                    // check against operators for devices
                case VirgilOp.DEVICE_GETREGISTER:
                    return new CIR.CExpr(PREC_TERM, REG(((DeviceDecl.GetRegister)op).register));
                case VirgilOp.DEVICE_SETREGISTER:
                    return $expr(PREC_ASSIGN, "$1 = $2",
                            REG(((DeviceDecl.SetRegister)op).register),
                            CEX(PREC_ASSIGN, to.operands[0]));

                    // check against operators for casts
                case CSRGen.CAST_INT_RAW:
                    return cast("unsigned int", to.operands[0]);
                case CSRGen.CAST_INT_CHAR:
                    return cast("char", to.operands[0]);
                case CSRGen.CAST_CHAR_RAW:
                    return cast("unsigned char", to.operands[0]);
                case CSRGen.CAST_CHAR_INT:
                    return cast("int", to.operands[0]);
                case CSRGen.CAST_RAW_INT:
                    return cast("int", to.operands[0]);
                case CSRGen.CAST_RAW_CHAR:
                    return cast("char", to.operands[0]);
            }

            throw Util.failure("compile(op:" + op.getClass() + ") unsupported @ " + to.getSourcePoint());
        }

        private CIR.CExpr getMember(TIROperator to, CSRStruct.IType.Field field, String op) {
            return $expr(PREC_ARROW, "$1"+ op +"$2",
                    CEX(PREC_ARROW, to.operands[0]),
                    field.name);
        }

        private CIR.CExpr setMember(TIROperator to, CSRStruct.IType.Field field, String op) {
            return $expr(PREC_ASSIGN, "$1"+op +"$2 = $3",
                    CEX(PREC_ARROW, to.operands[0]),
                    field.name,
                    CEX(PREC_ASSIGN, to.operands[1], field.type));
        }

        private CIR.CExpr cast(String t, TIRExpr e) {
            return $expr(PREC_CAST, "($1)$2", t, CEX(PREC_CAST, e));
        }

        private CIR.CExpr binop(TIROperator b, String op, int prec) {
            return binop(b.operands[0], op, prec, b.operands[1]);
        }

        private CIR.CExpr unop(TIROperator b, String op, int prec) {
            return new CIR.CExpr(prec, op + CEX(prec, b.operands[0]));
        }

        private CIR.CExpr divide(TIROperator b, String op, int prec) {
            return $expr(prec, "$1 $2 zero_check($3, $4)",
                    CEX(prec, b.operands[0]),
                    op,
                    CEX(prec, b.operands[1]),
                    getFLID(b));
        }

        public CIR.CExpr visit(TIRCompare.Equal c, CSRType ctype) {
            Type type = getCompareType(c);
            if (type.isFunction()) return compareDelegates(c, "==", "&&");
            else return binop(c.left, "==", PREC_EQUAL, c.right);
        }

        public CIR.CExpr visit(TIRCompare.NotEqual c, CSRType ctype) {
            Type type = getCompareType(c);
            if (type.isFunction()) return compareDelegates(c, "!=", "||");
            else return binop(c.left, "!=", PREC_EQUAL, c.right);
        }

        public CIR.CExpr visit(TIRConst.Value v, CSRType ctype) {
            if ( ctype == delegate && isNull(v) )
                return new CIR.CExpr(PREC_TERM, CImpl.NULL_DELEGATE);
            else return new CIR.CExpr(PREC_TERM, impl.renderValue(v.getValue(), ctype));
        }

        public CIR.CExpr visit(TIRThrow e, CSRType ctype) {
            Integer code = TIRtoCHarness.exceptionCodes.get(e.exception);
            return $expr(PREC_TERM, "($1)throw($2, $3)", ctype, code, getFLID(e.point));
        }

        public CIR.CExpr visit(TIRConst.String c, CSRType ctype) {
            // TODO: implement string constants
            return new CIR.CExpr(PREC_CAST, "(" + ctype + ")0");
        }

        public CIR.CExpr visit(TIRLocal.Get r, CSRType ctype) {
            return new CIR.CExpr(PREC_TERM, r.temp.getName());
        }

        public CIR.CExpr visit(TIRLocal.Set l, CSRType ctype) {
            return new CIR.CExpr(PREC_ASSIGN, compile(l));
        }

        public CIR.CExpr visit(TIRCall c, CSRType ctype) {

            StringBuffer buf = new StringBuffer();
            Type[] tn = ((Callable)c.func.getType()).getArgumentTypes();
            for (int cntr = 0; cntr < tn.length; cntr++) {
                if (cntr > 0) buf.append(", ");
                buf.append(CEX(c.arguments[cntr], csr.getCSRType(tn[cntr])));
            }
            return $expr(PREC_APPLY, "$1($2)", CEX(c.func), buf.toString());
        }

        private CIR.CExpr binop(TIRExpr left, String op, int prec, TIRExpr right) {
            return $expr(prec, "$1 $2 $3", CEX(prec, left), op, CEX(prec, right));
        }
    }

    protected class StmtCompiler extends TIRExprVisitor<CIR.CStmt, CSRType> {

        public CIR.CStmt visit(TIRExpr e, CSRType ct) {
            // default behavior is to assign to a new temporary
            CSRType ctype = csr.getCSRType(e.getType());
            return $stmt("$1 = $2", func.newTemporary(ctype), CEX(e));
        }

        public CIR.CStmt toStmt(TIRExpr e, CSRType ct) {
            // default behavior is to assign to a new temporary
            return new CIR.CSingle(CEX(e));
        }

        public CIR.CStmt visit(TIRNop n, CSRType ctype) {
            return new CIR.CSingle("");
        }

        public CIR.CStmt visit(TIROperator to, CSRType ctype) {
            switch ( CSRGen.getOpcode(to) ) {
                case CSRGen.CSRSTRUCT_SETREFFIELD: // fall through
                case CSRGen.CSRSTRUCT_SETVALFIELD: // fall through
                case CSRGen.CSRARRAY_SETELEM: // fall through
                case VirgilOp.DEVICE_SETREGISTER: // fall through
                case CSRGen.CSRPTR_SET: // these expressions are fine as statements
                    return toStmt(to, ctype);
            }

            // default: bail to the expression compiler and lift value to temp
            return visit((TIRExpr)to, ctype);
        }

        public CIR.CStmt visit(TIRThrow e, CSRType ctype) {
            Integer code = TIRtoCHarness.exceptionCodes.get(e.exception);
            return $stmt("throw($1, $2)", code, getFLID(e.point));
        }

        public CIR.CStmt visit(TIRBlock s, CSRType ctype) {
            List<CIR.CStmt> l = new LinkedList<CIR.CStmt>();
            for (TIRExpr e : s.list) {
                CIR.CStmt st = e.accept(STMT, CImpl.VOID);
                if (st == null) continue;
                List<CIR.CStmt> il = st.unNest();
                if (il != null) l.addAll(il);
                else l.add(st);
            }
            return new CIR.CNested(l);
        }

        public CIR.CStmt visit(TIRLocal.Set l, CSRType ctype) {
            return new CIR.CSingle(compile(l));
        }

        public CIR.CStmt visit(TIRBlock.Break e, CSRType ctype) {
            return new CIR.CSingle("break");
        }

        public CIR.CStmt visit(TIRLoop.For f, CSRType ctype) {
            String irep = compileNop(f.initial);
            CIR.CExpr cond = f.condition.accept(EXP, CImpl.VOID);
            String urep = compileNop(f.update);
            CIR.CStmt body = f.body.accept(STMT, CImpl.VOID);
            String fstr = "for ( " + irep + "; " + nest(cond) + "; " + urep + " )";
            return new CIR.CBlock(fstr, body, "");
        }

        private String compileNop(TIRExpr expr) {
            if (expr instanceof TIRNop)
                return "";
            else if ( expr instanceof TIRBlock ) {
                if ( ((TIRBlock)expr).isEmpty() ) return "";
            }
            return CEX(expr);
        }

        public CIR.CStmt visit(TIRLoop.While w, CSRType ctype) {
            CIR.CStmt body = w.body.accept(STMT, CImpl.VOID);
            String fstr = "while ( " + CEX(w.condition) + " )";
            return new CIR.CBlock(fstr, body, "");
        }

        public CIR.CStmt visit(TIRLoop.DoWhile w, CSRType ctype) {
            CIR.CStmt body = w.body.accept(STMT, CImpl.VOID);
            return new CIR.CBlock("do", body, "while ( " + CEX(w.condition) + " );");
        }

        public CIR.CStmt visit(TIRIfExpr i, CSRType ctype) {
            CIR.CStmt tt = i.true_target.accept(STMT, CImpl.VOID);
            CIR.CStmt ft = compileElseClause(i);
            return new CIR.CIfStmt("if ( " + CEX(i.condition) + " ) ", tt, ft);
        }

        private CIR.CStmt compileElseClause(TIRIfExpr i) {
            if ( ELIDE_ELSE && TIRUtil.isNoOp(i.false_target) ) return null;
            return i.false_target.accept(STMT, CImpl.VOID);
        }

        public CIR.CStmt visit(TIRSwitch s, CSRType ctype) {
            String before = "switch ( " + CEX(s.expr) + " )";
            List<CIR.CStmt> stmts = new LinkedList<CIR.CStmt>();
            for (TIRSwitch.Case c : s.cases) {
                for (TIRConst.Value cv : c.value) {
                    Value value = cv.getValue();
                    stmts.add(new CIR.CSingle("case " + impl.renderValue(value, value.dynType) + ":"));
                }
                stmts.add(c.body.accept(STMT, CImpl.VOID));
                stmts.add(new CIR.CSingle("break"));
            }
            if (s.defcase != null) {
                stmts.add(new CIR.CSingle("default:"));
                stmts.add(s.defcase.body.accept(STMT, CImpl.VOID));
            }
            CIR.CStmt block = new CIR.CNested(stmts);
            return new CIR.CBlock(before, block, "");
        }


        public CIR.CStmt visit(TIRBlock.Continue e, CSRType ctype) {
            return new CIR.CSingle("continue");
        }

        public CIR.CStmt visit(TIRReturn e, CSRType ctype) {
            if (func.retType == CImpl.VOID ) return new CIR.CSingle("return");
            else return $stmt("return $1", CEX(e.value, func.retType));
        }

        public CIR.CStmt visit(TIRCall c, CSRType ctype) {
            return new CIR.CSingle(c.accept(EXP, ctype).nest(0));
        }

    }

    protected String compile(TIRLocal.Set l) {
        return $eval("$1 = $2", l.temp.getName(), CEX(PREC_ASSIGN, l.value, csr.getCSRType(l.temp.getType())));
    }

    private String CEX(TIRExpr e) {
        return nest(e.accept(EXP, csr.getCSRType(e.getType())));
    }

    private String CEX(TIRExpr e, CSRType ct) {
        return nest(e.accept(EXP, ct));
    }

    private String CEX(int prec, TIRExpr e) {
        return e.accept(EXP, csr.getCSRType(e.getType())).nest(prec);
    }

    private String CEX(int prec, TIRExpr e, CSRType ct) {
        return e.accept(EXP, ct).nest(prec);
    }

    private String REG(Register r) {
        String addr = StringUtil.addrToString(r.getAbsoluteAddress());
        return "*((volatile "+csr.getCSRType(r.getType())+" *)" + addr + ")";
    }

    private static String nest(CIR.CExpr e) {
        return e.nest(0);
    }

    private Type getCompareType(TIRCompare c) {
        Type type = c.left.getType();
        if (type.isNull()) return c.right.getType();
        else return type;
    }

    private CIR.CExpr compareDelegates(TIRCompare c, String comp, String join) {
        TIRExpr left = c.left;
        TIRExpr right = c.right;
        if (isNull(left)) {
            // special case of comparing a delegate against the null constant
            return $expr(PREC_EQUAL, "$1 $2 $3.method", CImpl.NULL, comp, CEX(PREC_EQUAL, right));
        }
        if (isNull(right)) {
            // special case of comparing a delegate against the null constant
            return $expr(PREC_EQUAL, "$1 $2 $3.method", CImpl.NULL, comp, CEX(PREC_EQUAL, left));
        }

        Type funcType = c.left.getType();
        String t1 = newTemp(funcType);
        String t2 = newTemp(funcType);
        return $expr(PREC_LOGICAL_OR, "($1 = $2).ref $3 ($4 = $5).ref $6 $1.method $3 $4.method",
                t1,                           // 1 left tmp
                CEX(PREC_LOGICAL_OR, c.left), // 2 left
                comp,                         // 3 comparison operator
                t2,                           // 4 right tmp
                CEX(PREC_LOGICAL_OR,c.right), // 5 right
                join);                        // 6 join operator
    }

    private String newTemp(Type type) {
        return func.newTemporary(csr.getCSRType(type)).getName();
    }

    private boolean isNull(TIRExpr e) {
        if (e instanceof TIRConst.Value) {
            Value v = ((TIRConst.Value)e).getValue();
            return Reference.isNull(v) || VirgilDelegate.isNull(v);
        }
        return false;
    }

    private String getFLID(TIRExpr e) {
        return "0";
    }

    private String getFLID(SourcePoint pt) {
        return "0";
    }

}
