/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Jan 26, 2006
 */

package vpc.tir.tir2c;

import cck.text.Printer;
import cck.text.StringUtil;
import cck.util.*;
import vpc.Version;
import vpc.core.*;
import vpc.core.virgil.VirgilDelegate;
import vpc.core.csr.*;
import vpc.core.concept.*;
import vpc.model.Stage;
import vpc.tir.TIRUtil;
import vpc.types.Type;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.IOException;
import java.util.*;

/**
 * The <code>TIREmitC</code> class implements a compiler stage that generates C code
 * from TIR.
 *
 * @author Ben L. Titzer
 */
public class TIREmitC extends Stage {

    protected CImpl impl;

    protected static final ClassMap accountMap = new ClassMap("Space Accounting", CAccountant.class);
    protected static final ClassMap linkageMap = new ClassMap("Linkage Model", CLinkage.class);

    static {
        linkageMap.addClass("c:test", CLinkage.CTestLinkage.class);
        linkageMap.addClass("c:avr", CLinkage.CAVRLinkage.class);
        linkageMap.addClass("c:user", CLinkage.CUserLinkage.class);
        accountMap.addClass("standard", CAccountant.Standard.class);
        accountMap.addClass("compressed", CAccountant.Compressed.class);
    }

    protected Option.Str LINKAGE = options.newOption("linkage", "c:user",
            "This option selects the linkage model for emitting C code that can be compiled " +
            "by the appropriate platform C compiler. For example, the linkage model selects " +
            "what version of the \"main\" method will be generated and how other entrypoints " +
            "to the program will be connected with generate C glue code.");
    protected Option.Str ACCOUNT = options.newOption("accounting", "",
            "This option selectes an accounting scheme that reports the amount of heap space " +
            "taken by data structures, including objects, meta objects, compression tables, etc.");

    public void visitProgram(Program p) throws IOException {
        Printer pr = new Printer(new PrintStream(new FileOutputStream(p.name + ".c")));

        pr.println("//==========================================================");
        pr.println("// C code generated by: VPC [" + Version.getVersion() + "]");
        pr.println("// "+ new Date());
        pr.println("//==========================================================");
        pr.println("");

        impl = (CImpl)p.csr.impl;
        // layout the component and class structs
        impl.layoutComponents(impl.program);
        impl.layoutClasses(impl.program);
        // make some safety check functions
        impl.makeSafetyChecks(impl.program);
        // generate main method and interrupt handlers
        CLinkage link = getLinkageModel();
        link.generateLinkage(p, this);
        // compile Virgil methods
        compileMethods(p, pr);

        // declare and define all structs
        for (CSRStruct.IType st : impl.csr.structs) declareStructure(pr, st);
        for (CSRStruct.IType st : impl.csr.structs) defineStructure(pr, st);

        // declare prototypes for all functions
        declareFunctions(p, pr);
        // emit the heap object definitions
        defineHeap(p, pr);
        // emit the code for each function
        defineFunctions(p, pr);

        if (!ACCOUNT.isBlank()) {
            CAccountant a = (CAccountant) accountMap.getObjectOfClass(ACCOUNT.get());
            a.account(p, impl);
        }
    }

    private void printSeparator(Printer pr, String title) {
        pr.print("//=========={ ");
        pr.print(title);
        pr.print(" }");
        pr.println(StringUtil.dup('=', 65 - title.length()));
    }

    void compileMethods(Program p, Printer pr) {
        printSeparator(pr, "Function prototypes");
        // declare all the component methods
        for (Method m : p.closure.methods) {
            CSRFunction cf = new CSRFunction(p.csr, m);
            for (Variable v : TIRUtil.getRep(m).getParams()) {
                cf.newParam(v.getName(), p.csr.getCSRType(v.getType()));
            }
            new TIRCompiler(impl).compile(cf);
            p.csr.functions.add(cf);
        }
        pr.println("");
    }

    void declareFunctions(Program p, Printer pr) {
        for ( CSRFunction cf : p.csr.functions )
            cf.printPrototype(pr);
    }

    void defineHeap(Program p, Printer pr) {
        printSeparator(pr, "Heap object declarations");
        // declare forward references
        for (Heap.Record r : p.heap.getRecords()) {
            Type t = r.layout.type;
            CSRType ct = p.csr.getCSRStructType(t);
            pr.println(ct.getVarDecl("obj_" + r.uid) + ";");
        }
        pr.println("");
        printSeparator(pr, "Heap object definitions");
        // declare actual heap records
        for (Heap.Record r : p.heap.getRecords()) {
            emitRecord(pr, r.layout.type, r, p);
        }
    }

    private void emitRecord(Printer pr, Type t, Heap.Record r, Program p) {
        CSRType ct = p.csr.getCSRStructType(t);
        pr.startblock(ct.getVarDecl("obj_" + r.uid) + " =");
        List<String> fields = extractFields(p, t, r);
        Iterator<String> it = fields.iterator();
        while (it.hasNext()) {
            pr.print(it.next());
            if (it.hasNext()) pr.print(",");
            pr.nextln();
        }

        pr.endblock(";");
    }

    private List<String> extractFields(Program p, Type t, Heap.Record r) {
        List<String> fields = new LinkedList<String>();
        if (t.isArray()) {
            fields.add(nameComment("length") + Integer.toString(r.getSize()));
            StringBuffer buf = new StringBuffer(nameComment("values")+"{ ");
            int cntr = 0;
            for (Heap.Cell f : r.layout.getCells()) {
                if (cntr++ > 0) buf.append(", ");
                buf.append(renderCellValue(p, r, f));
            }
            buf.append(" }");
            fields.add(buf.toString());
        } else if (t.isObject() || t.isComponent()) {
            for (Heap.Cell f : r.layout.getCells()) {
                String val = renderCellValue(p, r, f);
                fields.add(nameComment(f.getName()) + val);
            }
        } else {
            throw Util.failure("unknown record type: " + t);
        }
        return fields;
    }

    private String renderCellValue(Program p, Heap.Record r, Heap.Cell f) {
        Value value = f.getValue(r);
        String val = impl.renderValue(value, f.getType());
        Class vc = value.getClass();
        if ( requiresInitializationCast(value, vc) ) {
            val = "(void*)"+val;
        }
        return val;
    }

    private boolean requiresInitializationCast(Value value, Class vc) {
        if (Reference.isNull(value) || VirgilDelegate.isNull(value)) {
            return false;
        } else {
            return vc == Reference.Val.class || vc == Function.Val.class;
        }
    }

    private String nameComment(String fname) {
        return "/* " + StringUtil.leftJustify(fname, 16) + " */ ";
    }

    void defineFunctions(Program p, Printer pr) {
        printSeparator(pr, "C functions implementing methods");
        for (CSRFunction cf : impl.program.csr.functions) {
            // emit C code
            cf.print(pr);
        }
        pr.println("");
    }

    public void declareStructure(Printer pr, CSRStruct.IType st) {
        pr.println(st.toString() + ";");
    }

    public void defineStructure(Printer pr, CSRStruct.IType st) {
        pr.startblock(st.toString());
        for (CSRStruct.IType.Field f : st.fields)
            pr.println(f.type.getVarDecl(f.name) + ";");
        pr.endblock(";");
    }

    protected CLinkage getLinkageModel() {
        return (CLinkage) linkageMap.getObjectOfClass(LINKAGE.get());
    }
}
