/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc.types;

import vpc.util.Flags;
import java.util.HashMap;

/**
 * The <code>Type</code> class represents a type present in the system,
 * whether it is a primitive type, class, component, or structure. Its
 * inner classes provide the implementations that represent the types
 * available in the Virgil programming language.
 *
 * @author Ben L. Titzer
 */
public abstract class Type implements TypeToken {

    public static final Flags typeFlags = new Flags();

    public static final int ARRAY = typeFlags.addFlag("array");
    public static final int VOID = typeFlags.addFlag("void");
    public static final int OBJECT = typeFlags.addFlag("object");
    public static final int COMPONENT = typeFlags.addFlag("component");
    public static final int REFERENCE = typeFlags.addFlag("reference");
    public static final int NULL = typeFlags.addFlag("null");
    public static final int FUNCTION = typeFlags.addFlag("function");
    public static final int SWITCH = typeFlags.addFlag("switch");
    public static final int EQUALITY = typeFlags.addFlag("equality");
    public static final int INDEXABLE = typeFlags.addFlag("indexable");

    public static Type[] NOARGS = new Type[0];

    protected int flags;
    protected final String name;

    protected Type(String n, int... fl) {
        name = n;
        flags = Flags.setFlags(fl);
    }

    public Capability.BinOp resolveInfixBinOp(String op) {
        return null;
    }

    public Capability.UnaryOp resolvePrefixUnaryOp(String op) {
        return null;
    }

    public Capability.AutoOp resolveAutoOp(String op) {
        return null;
    }

    public boolean isArray() {
        return Flags.getFlag(flags, ARRAY);
    }

    public boolean isVoid() {
        return Flags.getFlag(flags, VOID);
    }

    public boolean isObject() {
        return Flags.getFlag(flags, OBJECT);
    }

    public boolean isComponent() {
        return Flags.getFlag(flags, COMPONENT);
    }

    public boolean isReference() {
        return Flags.getFlag(flags, REFERENCE);
    }

    public boolean isNull() {
        return Flags.getFlag(flags, NULL);
    }

    public boolean isFunction() {
        return Flags.getFlag(flags, FUNCTION);
    }

    public boolean isSwitchable() {
        return Flags.getFlag(flags, SWITCH);
    }

    public boolean isEqualityType() {
        return Flags.getFlag(flags, EQUALITY);
    }

    public boolean isIndexable() {
        return Flags.getFlag(flags, INDEXABLE);
    }

    public String toString() {
        return name;
    }

    public abstract boolean canBeComparedTo(Type t);

    /**
     * The <code>Primitive</code> class represents a primitive type such as
     * a boolean, an integer, or a character. It has the appropriate isXXX()
     * methods overridden to return the correct results.
     */
    public abstract static class Primitive extends Type {

        protected final HashMap<String, Capability.AutoOp> auto;
        protected final HashMap<String, Capability.BinOp> binops;
        protected final HashMap<String, Capability.UnaryOp> unaryops;

        protected Primitive(String n, int... fl) {
            super(n, fl);
            binops = new HashMap<String, Capability.BinOp>();
            unaryops = new HashMap<String, Capability.UnaryOp>();
            auto = new HashMap<String, Capability.AutoOp>();
        }

        public Capability.BinOp resolveInfixBinOp(String op) {
            return binops.get(op);
        }

        public Capability.UnaryOp resolvePrefixUnaryOp(String op) {
            return unaryops.get(op);
        }

        public Capability.AutoOp resolveAutoOp(String op) {
            return auto.get(op);
        }

        public Capability.BinOp addBinOp(Capability.BinOp b) {
            binops.put(b.opname, b);
            return b;
        }

        public Capability.AutoOp addAutoOp(Capability.AutoOp b) {
            auto.put(b.opname, b);
            return b;
        }

        public Capability.UnaryOp addUnaryOp(Capability.UnaryOp o) {
            unaryops.put(o.opname, o);
            return o;
        }
    }
}
