/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Jan 23, 2007
 */

package vpc.types;

/**
 * The <code>TypeCon</code> class models a type constructor, which is a function
 * that can produce new types from parameter types. For example, type constructors
 * model functions, arrays, and parameterized classes.
 *
 * @author Ben L. Titzer
 */
public abstract class TypeCon {

    public final String name;
    public final int arity;

    public static final int UNBOUNDED = -1;

    protected TypeCon(String name, int a) {
        this.name = name;
        arity = a;
    }

    /**
     * The <code>newType()</code> method constructs a new type from the given types.
     * @param cache the type cache to use in order to assure reference equality
     * @param p the list of types that are parameters to this type constructor
     * @return a new type that represents the type that results from applying this
     * type constructor to the given list of types
     */
    public abstract Type newType(TypeCache cache, Type... p);

    /**
     * The <code>getArity()</code> method gets the arity, or number of expected type
     * parameters, of this constructor. Some type constructors can accept an unbounded
     * number of type parameters. In this case, this method will return <code>UNBOUNDED</code>.
     * @return the number of type parameters expected by this type constructor
     */
    public int getArity() {
        return arity;
    }

    /**
     * The <code>render()</code> method creates a string representation of the type
     * that would result from applying this type constructor to the list of types
     * given.
     * @param p the list of type references to the type constructor
     * @return a canonical string representation of the type that would result
     * from applying this type constructor to the specifed parameter types
     */
    public String render(TypeRef... p) {
        assert arity == UNBOUNDED || p.length == arity;
        if ( p.length > 0 ) {
            StringBuffer buf = new StringBuffer(name);
            buf.append("<");
            for ( int cntr = 0; cntr < p.length; cntr++ ) {
                if ( cntr > 0 ) buf.append(", ");
                buf.append(p[cntr]);
            }
            buf.append(">");
            return buf.toString();
        } else {
            return name;
        }
    }

    /**
     * The <code>TypeCon.Singleton</code> class represents an implementation of the
     * <code>TypeCon</code> class that can be used to create a nullary type constructor
     * from any given type.
     */
    public static class Singleton extends TypeCon {
        public final Type type;

        public Singleton(String n, Type t) {
            super(n, 0);
            type = t;
        }

        public Type newType(TypeCache cache, Type... p) {
            assert p.length == 0;
            return type;
        }

        public String render(TypeRef... p) {
            assert p.length == 0;
            return name;
        }
    }
}
