/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Feb 9, 2006
 */

package vpc.types;

import cck.parser.AbstractToken;
import cck.parser.ErrorReporter;
import cck.util.Util;
import java.util.HashMap;

/**
 * The <code>TypeEnv</code> class represents a type environment for a program,
 * module, file, type declaration, or method. A type environment contains local
 * type declarations and aliases as well as <code>TypeRef</code> instances that
 * represent the free (i.e. unbound or unknown) types within the environment.
 * <p/>
 * To complete a type environment, all of its unbound type references must be
 * resolved. The type environment is then suitable for use in typechecking.
 *
 * @author Ben L. Titzer
 */
public class TypeEnv {

    protected final HashMap<String, Type> types;
    protected final TypeEnv parent;

    /**
     * The constructor for the <code>TypeEnv</code> class creates a new type environment.
     * It accepts as a parameter the parent, or enclosing, type environment.
     *
     * @param p the parent type environment for the newly created type environment
     */
    public TypeEnv(TypeEnv p) {
        types = new HashMap<String, Type>();
        parent = p;
    }

    /**
     * The <code>bindLocal()</code> method binds the specified type name to the
     * specified type only within this environment.
     *
     * @param str the name of the type
     * @param t   the type to bind in this environment
     */
    public void bindLocal(String str, Type t) {
        types.put(str, t);
    }

    /**
     * The <code>bindLocal()</code> method binds the specified type name to the
     * specified type globally; that is in the topmost parent type environment.
     *
     * @param str the name of the type
     * @param t   the type to bind in the global environment
     */
    public void bindGlobal(String str, Type t) {
        if (parent != null) parent.bindGlobal(str, t);
        else bindLocal(str, t);
    }

    /**
     * The <code>resolve()</code> method attempts to resolve the specified type reference
     * in this type environment and recursively in parent environments.
     *
     * @param tr the type reference to resolve
     * @return the type that this type reference refers to, if it exists; null otherwise
     */
    public Type resolve(TypeRef tr) {
        Type t = types.get(tr.toString());
        return t == null && parent != null ? parent.resolve(tr) : t;
    }

    public TypeRef newTypeRef(AbstractToken tname) {
        return new TypeRef(tname);
    }

    public void resolveTypeRefs(ErrorReporter er) {
        throw Util.unimplemented();
    }
}
