/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Feb 9, 2006
 */

package vpc.types;

import cck.parser.AbstractToken;
import cck.parser.SourcePoint;
import cck.util.Util;

/**
 * The <code>TypeRef</code> class represents a reference to a type that may
 * or may not have been resolved yet.
 *
 * @author Ben L. Titzer
 */
public class TypeRef implements TypeToken {

    protected final String name;
    protected final SourcePoint point;
    protected Type type;

    /**
     * The constructor for the <code>TypeRef</code> class creates a new type reference
     * with the specified token, which indicates the source point and the name of the type
     * reference.
     *
     * @param rn the token representing the type reference's name and source position
     */
    public TypeRef(AbstractToken rn) {
        name = rn.image;
        point = rn.getSourcePoint();
    }

    /**
     * The constructor for the <code>TypeRef</code> class creates a new type reference
     * with the specified token, which indicates the source point and the name of the type
     * reference.
     *
     * @param rn a string representing the type reference's name
     * @param sp the point in the source code where this type reference occurs
     */
    public TypeRef(String rn, SourcePoint sp) {
        name = rn;
        point = sp;
    }

    public TypeRef(TypeCon tc) {
        name = tc.name;
        point = new SourcePoint("", 0, 0, 0);
    }

    /**
     * The <code>isResolved()</code> method checks whether this type reference has been
     * resolved yet.
     *
     * @return true if this type reference has been resolved to some type; false otherwise
     */
    public boolean isResolved() {
        return type != null;
    }

    /**
     * The <code>getType()</code> method returns the type that this type reference is bound to,
     * if this reference is bound. If the reference is not bound (i.e. not resolved), this method
     * will generate an error.
     *
     * @return the type that this reference is bound to, if any
     * @throws InternalError if the reference is not bound
     */
    public Type getType() {
        if (type == null) throw Util.failure("unresolved type at " + point);
        return type;
    }

    /**
     * The <code>toString()</code> method converts this type reference to a canonical string
     * representation that is used to look up the type in type caches.
     *
     * @return a string representation of this type reference
     */
    public String toString() {
        return name;
    }

    /**
     * The <code>getSourcePoint()</code> method returns the source point at which this type
     * reference was declared. This is used to report an error when the type cannot be resolved.
     *
     * @return the source point corresponding to this type reference
     */
    public SourcePoint getSourcePoint() {
        return point;
    }

}
