/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Sep 17, 2006
 */
package vpc.types;

/**
 * The <code>TypeSystem</code> class defines type rules for a particular language.
 * It includes such functionality as the rules regarding automatic type conversion,
 * assignability, subtyping, source capabilities such as binary operators, etc.
 *
 * @author Ben L. Titzer
 */
public abstract class TypeSystem {

    public Capability.AutoOp resolveAutOp(String op, Type t1) {
        return t1.resolveAutoOp(op);
    }

    public Capability.BinOp resolveBinOp(String op, Type t1, Type t2) {
        return t1.resolveInfixBinOp(op);
    }

    public Capability.UnaryOp resolveUnOp(String op, Type t1) {
        return t1.resolvePrefixUnaryOp(op);
    }

    /**
     * The <code>isAssignable()</code> method implements the assignability relation
     * for this type system. This method should determine whether a statement which
     * assigns a value of type "src" to the destination expression of type "dst"
     * is correct.
     *
     * @param src the type of the expression being assigned to the location
     * @param target
     * @return true if the expression can be assigned to the location; false otherwise
     */
    public abstract boolean isAssignable(Type src, Type target);

    /**
     * The <code>isComparable()</code> method implements the comparability relation
     * on types. Two types are comparable if they can be compared with the "==" and
     * "!=" operators.
     * @param t1 the type of the left hand expression
     * @param t2 the type of the right hand expression
     * @return true if the expressions can be compared; false otherwise
     */
    public abstract boolean isComparable(Type t1, Type t2);

    /**
     * The <code>isConvertible()</code> method implements the convertibility relationship
     * on types. A type is convertible to another type if an explicit type conversion
     * operation on an expression of the source type is runtime convertible to a value
     * of the target type.
     *
     * @param src the type of the source expression
     * @param target the target type of the conversion expression
     * @return true if the expression can be converted to the destination type; false otherwise
     */
    public abstract boolean isConvertible(Type src, Type target);

    /**
     * The <code>isPromotable()</code> method implements the promotable relationship
     * on types. A type is promotable to another type if an implicit type conversion
     * operation can be inserted by the compiler to translate a value of the source type
     * to a value of the destination type.
     *
     * @param src the type of the source expression
     * @param target the target type of the conversion expression
     * @return true if the expression can be promoted automatically to the destination type; false otherwise
     */
    public abstract boolean isPromotable(Type src, Type target);

    /**
     * The <code>unify()</code> method implements the unification operation on types
     * that is used in type inferrence.
     * @param t1 the first type
     * @param t2 the second type
     * @return a type to which the two types can be unified if it exists; null otherwise
     */
    public abstract Type unify(Type t1, Type t2);
}
