/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Sep 14, 2006
 */

package vpc.util;

import java.util.LinkedList;

/**
 * The <code>Delegator</code> class implements a flexible delegator pattern
 * that delegates a method call across a number of possible objects
 * according to a search strategy.
 *
 * </p>
 * The <code>Delegator</code> works by maintaining a list of objects
 * that are searched in order. Upon invocation of the delegator, it
 * begins by calling the <code>invoke()</code> method of the first object; if that
 * call returns non-null, then that value is returned. If that call
 * returns null, then the next object's <code>invoke()</code> method is called,
 * and so on.
 *
 * </p>
 * The <code>Delegator</code> class also implements the <code>Delegate</code>
 * interface, allowing it to be one of the objects in the delegation chain.
 * This allows recursive delegation, making possible depth-first
 * delegation over a tree structure.
 *
 * @author Ben L. Titzer
 */
public class Delegator<R, P> implements Delegate<R, P> {

    protected final LinkedList<Delegate<R,P>> delegates;

    /**
     * The default constructor for a delegator takes no arguments.
     */
    public Delegator() {
        delegates = new LinkedList<Delegate<R,P>>();
    }

    /**
     * The <code>add()</code> method adds a new object to this delegation
     * chain at the end.
     * @param d the new delegate to add to the delegation chain
     */
    public void add(Delegate<R, P> d) {
        delegates.add(d);
    }

    /**
     * The <code>addFirst()</code> method adds a new object to this delegation
     * chain at the beginning.
     * @param d the new delegate to add to the delegation chain
     */
    public void addFirst(Delegate<R, P> d) {
        delegates.addFirst(d);
    }

    /**
     * The <code>remove()</code> method remove the specified delegate
     * from this delegation chain, if it exists.
     * @param d the delegate to remove from this chain
     */
    public void remove(Delegate<R, P> d) {
        delegates.remove(d);
    }

    /**
     * The <code>invoke()</code> method starts the delegation process. It will
     * invoke the <code>invoke()</code> methods of the delegated objects according
     * to the search strategy until the first returns non-null.
     * @param p the parameter to the invoke method
     * @return the result of first method to return non-null; null if no delegates
     * exist, or none of them returned non-null
     */
    public R invoke(P p) {
        for ( Delegate<R,P> i : delegates ) {
            R r = i.invoke(p);
            if ( r != null ) return r;
        }
        return null;
    }
}
