/**
 * Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created May 27, 2006
 */
package vpc.util;

import cck.util.Util;

/**
 * The <code>Maybe</code> class is a utility object that represents a value
 * that may or not be valid, such as the result of parsing or converting a
 * integer from a string. This is useful in situations where the handling of
 * exceptions needs to be deferred without the clutter of try/catch blocks.
 * <p/>
 * </p>
 * The <code>Maybe</code> class either contains a valid value of the specified
 * generic parameter type, or a <code>Throwable</code> that indicates why the
 * value is not valid. Attempts to access the value through the
 * <code>getValue()</code> method when the value is not valid will cause
 * failure; thus clients should always consult the <code>exists()</code> method
 * before calling <code>getValue()</code>
 *
 * @author Ben L. Titzer
 */
public class Maybe<T> {

    protected final T value;
    protected final Throwable error;

    /**
     * The default constructor for the <code>Maybe</code> class creates an object
     * representing an invalid value.
     */
    public Maybe() {
        value = null;
        error = null;
    }

    /**
     * This constructor will create a new <code>Maybe</code> object containing a
     * valid value.
     *
     * @param v the value to store in this <code>Maybe</code> object.
     */
    public Maybe(T v) {
        value = v;
        error = null;
    }

    /**
     * This constructor will create a new <code>Maybe</code> object that contains
     * an invalid value, with the specified exception.
     *
     * @param e the exception that represents the reason the value is invalid
     */
    public Maybe(Throwable e) {
        value = null;
        error = e;
    }

    /**
     * The <code>exists()</code> method checks whether the value exists.
     *
     * @return true if a valid value exists in this object; false otherwise
     */
    public boolean exists() {
        return value != null;
    }

    /**
     * The <code>getValue()</code> method retrieves the value stored in this object,
     * if it exists. If the value does not exist, this method will fail.
     *
     * @return the value stored in this object, if it is valid
     * @throws Util.InternalError if the value is not valid
     */
    public T getValue() {
        if (!exists()) throw Util.failure("Value does not exist: " + error);
        else return value;
    }

    /**
     * The <code>getError()</code> method returns the error corresponding to the
     * reason that this value object is not valid.
     *
     * @return the error for an invalid value
     */
    public Throwable getError() {
        return error;
    }
}
