/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Jan 26, 2006
 */

package vpc.vst;

import cck.parser.SourceError;
import cck.parser.SourcePoint;
import cck.text.StringUtil;
import cck.util.Util;
import vpc.core.Program;
import vpc.core.virgil.VirgilClass;
import vpc.core.virgil.VirgilComponent;
import vpc.model.Stage;
import vpc.vst.parser.*;
import vpc.vst.stages.*;
import vpc.vst.tir.TIRBuilder;
import vpc.vst.tree.VSTModule;
import vpc.vst.tree.VSTPrinter;
import java.io.*;

/**
 * The <code>VSTStages</code> class implements a namespace class (i.e. contains inner classes)
 * for each of the compilation stages associated with the Virgil syntax tree representation,
 * including parsing, verifying, building types, typechecking, and translating VST to TIR.
 *
 * @author Ben L. Titzer
 */
public class VSTStages {
    /**
     * The <code>Parse</code> class implements the phase of compilation
     * relating to parsing the source file and building the core program
     * representation for later phases of the compilation.
     *
     * @author Ben L. Titzer
     */
    public static class Parse extends Stage {

        public void visitProgram(Program program) {
            for (File f : program.getFiles()) {
                try {
                    VSTModule m = VirgilParser.parseModule(new FileInputStream(f), f.getName(), program.typeCache);
                    Verifier.buildAndVerifyModule(program, m);
                } catch (ParseException e) {
                    Token tok = e.currentToken.next;
                    SourcePoint pp = new SourcePoint(f.getName(), tok.beginLine, tok.beginColumn, tok.endColumn);
                    String report = "parse error at token \"" + tok + "\"\n" + e.getMessage();
                    throw new SourceError("ParseError", pp, report, StringUtil.EMPTY_STRING_ARRAY);
                } catch (IOException e) {
                    Util.userError("IOException: " + e.getMessage());
                }
            }
        }
    }

    /**
     * The <code>BuildTypes</code> stage of compilation walks the Virgil syntax tree (without
     * inspecting method bodies) and builds representations of each of the declared types,
     * including classes and components. It does a number of checks on the well-formedness of
     * the types declared. When this stage is completed, all of the types needed in the
     * typechecking phase are found, resolved, and built, so the type system is complete and
     * consistent.
     */
    public static class BuildTypes extends Stage {
        public void visitProgram(Program p) {
            VSTTypeBuilder gtb = new VSTTypeBuilder(p, new VSTErrorReporter());
            gtb.bindComponentTypes(p.virgil.getComponents());
            gtb.bindClassTypes(p.virgil.getClasses());
            for (VirgilClass d : p.virgil.getClasses()) gtb.visit(d);
            for (VirgilComponent d : p.virgil.getComponents()) gtb.visit(d);
        }
    }

    /**
     * The <code>TypeCheck</code> stage of compilation walks the Virgil syntax tree (including
     * the method bodies and expressions initializing fields) and type checks each expression
     * against the type system that was built in the <code>BuildTypes</code> stage.
     */
    public static class TypeCheck extends Stage {
        public void visitProgram(Program p) {
            TypeChecker t = new TypeChecker(p, new VSTErrorReporter());
            for (VirgilClass d : p.virgil.getClasses()) t.typeCheckClass(d);
            for (VirgilComponent d : p.virgil.getComponents()) t.typeCheckComponent(d);
        }
    }

    /**
     * The <code>PrintVST</code> stage is an ending stage that emits the Virgil syntax tree
     * in a textual format. This is used to output the syntax tree in a format close to
     * the original text.
     */
    public static class PrintVST extends Stage {
        private VSTPrinter printer;

        public PrintVST(VSTPrinter p) {
            printer = p;
        }

        public void visitProgram(Program program) {
            for (File f : program.getFiles()) {

                try {
                    VSTModule m = VirgilParser.parseModule(new FileInputStream(f), f.getName(), program.typeCache);
                    m.accept(printer);
                } catch (ParseException e) {
                    Token tok = e.currentToken.next;
                    SourcePoint pp = new SourcePoint(f.getName(), tok.beginLine, tok.beginColumn, tok.endColumn);
                    String report = "parse error at token \"" + tok + "\"\n" + e.getMessage();
                    throw new SourceError("ParseError", pp, report, StringUtil.EMPTY_STRING_ARRAY);
                } catch (IOException e) {
                    Util.userError("IOException: " + e.getMessage());
                }
            }
        }
    }

    /**
     * The <code>BuildTIR</code> stage of compilation translates a Virgil syntax tree program
     * into a program in the <code>TIR</code> format.
     */
    public static class BuildTIR extends Stage {

        public void visitProgram(Program p) {
            TIRBuilder.buildIRForProgram(p);
        }
    }
}
