/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created Jan 22, 2006
 */
package vpc.vst;

import cck.util.Util;
import vpc.core.*;
import vpc.core.concept.*;
import vpc.core.virgil.VirgilClass;
import vpc.core.virgil.VirgilComponent;
import vpc.types.Type;
import vpc.vst.tree.*;

/**
 * The <code>VSTUtil</code> class encapsulates a number of utility methods that are used
 * through the implementation of various VST analysis and transformation passes.
 *
 * @author Ben L. Titzer
 */
public class VSTUtil {
    public static final String REPRESENTATION_KEY = "vst";

    public static VSTClassDecl vstRepOf(VirgilClass pcd) {
        if (pcd == null) return null;
        VSTClassDecl d = (VSTClassDecl) pcd.getSourceRep();
        if (d == null) throw Util.failure("no VST for ClassDecl " + pcd.getTypeName());
        return d;
    }

    public static VSTComponentDecl vstRepOf(VirgilComponent pcd) {
        if (pcd == null) return null;
        VSTComponentDecl d = (VSTComponentDecl) pcd.getComponentRep();
        if (d == null) throw Util.failure("no VST for ClassDecl " + pcd.getTypeName());
        return d;
    }

    public static VSTMethodDecl vstRepOf(Method m) {
        if (m == null) return null;
        VSTMethodDecl d = (VSTMethodDecl) m.getMethodRep(REPRESENTATION_KEY);
        if (d == null) throw Util.failure("no VST for CompoundDecl.Method " + m.getFullName());
        return d;
    }

    public static VSTFieldDecl vstRepOf(Field f) {
        if (f == null) return null;
        VSTFieldDecl d = (VSTFieldDecl) f.getFieldRep();
        if (d == null) throw Util.failure("no VST for CompoundDecl.Field " + f.getFullName());
        return d;
    }

    public static VSTMemberDecl vstRepOf(Member m) {
        if (m.isField()) return vstRepOf((Field) m);
        else return vstRepOf((Method) m);
    }

    public static VSTConstructorDecl vstRepOf(Constructor c) {
        if (c == null) return null;
        VSTConstructorDecl cd = (VSTConstructorDecl) c.getMethodRep(REPRESENTATION_KEY);
        if (cd == null) throw Util.failure("no VST for Program.Constructor " + c.getCompoundDecl());
        return cd;
    }

    public static Type typeOf(VSTTypeRef r) {
        return r.getTypeName().getType();
    }

    public static Type typeOf(VSTLocalVarDecl d) {
        return d.getType();
    }

    public static Type typeOf(VSTFieldDecl f) {
        return f.getTypeName().getType();
    }

    public static Type typeOf(VSTExpr e) {
        return e.getType();
    }

    public static Type typeOf(Member m) {
        return m.getTypeName().getType();
    }

    public static boolean isPrivate(Member m) {
        VSTMemberDecl md = vstRepOf(m);
        return md.getMode().isPrivate();
    }

    public static Member getVisibleMember(Program p, String name, CompoundDecl c, CompoundDecl where) {
        if (c == null) return null;
        Member pm = c.resolveMember(name, p);
        if (pm == null) return null;

        VSTMemberDecl md = vstRepOf(pm);
        CompoundDecl cd = pm.getCompoundDecl();
        if (cd == where) return pm;
        else return md.getMode().isPrivate() ? null : pm;
    }

    public static boolean isDeviceAccess(VSTMemberExpr.Binding b) {
        return b instanceof VSTMemberExpr.DeviceRegister;
    }

    public static boolean isArrayLength(VSTMemberExpr.Binding b) {
        return b instanceof VSTMemberExpr.ArrayLength;
    }

    public static boolean isClassMember(VSTMemberExpr.Binding b) {
        return b instanceof VSTMemberExpr.ClassMember;
    }

    public static boolean isComponentMember(VSTMemberExpr.Binding b) {
        return b instanceof VSTMemberExpr.ComponentMember;
    }
}
