/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc.vst.stages;

import cck.text.StringUtil;
import static cck.text.StringUtil.quote;
import cck.util.Util;
import vpc.core.Decl;
import vpc.core.Program;
import vpc.core.concept.*;
import vpc.core.virgil.*;
import static vpc.core.virgil.VirgilTypeSystem.*;
import vpc.tir.expr.Operator;
import vpc.types.*;
import vpc.util.Maybe;
import vpc.vst.VSTErrorReporter;
import vpc.vst.VSTUtil;
import vpc.vst.tree.*;
import java.util.HashMap;

/**
 * The <code>TypeChecker</code> class implements a typechecker that visits Virgil
 * syntax tree nodes, infers types, and checks that the types of all expressions
 * are correct according to the type rules of the Virgil language.
 *
 * @author Ben L. Titzer
 */
public class TypeChecker extends TypeCheckerBase {

    private VirgilClass.IType thisClass;
    private Type thisReturnType;

    public TypeChecker(Program p, VSTErrorReporter error) {
        typeSystem = p.typeSystem;
        program = p;
        ERROR = error;
    }

    public void typeCheckClass(VirgilClass pcd) {
        VSTClassDecl d = VSTUtil.vstRepOf(pcd);
        setContainer(pcd);
        visit(d);
    }

    public void visit(VSTClassDecl d) {
        if ( d.parent != null ) resolveClassType(d.parent);
        thisClass = (VirgilClass.IType) program.typeCache.getTypeName(d.token.image).getType();
        super.visit(d);
        thisClass = null;
    }

    public void typeCheckComponent(VirgilComponent pcd) {
        VSTComponentDecl d = VSTUtil.vstRepOf(pcd);
        setContainer(pcd);
        visit(d);
    }

    public void visit(VSTFieldDecl d) {
        checkInitializer(d.getName(), d.type, d.init);
    }

    public void visit(VSTMethodDecl d) {
        visitParamDecls(d.params);
        thisReturnType = d.returnType.getTypeName().getType();
        if ( isComponent(thisReturnType) )
            ERROR.ExpectedReturnType(d, d.returnType);
        super.visit(d);
    }

    public void visit(VSTParamDecl d) {
        checkVarType(d.type);
    }

    public void visit(VSTLocalVarDecl d) {
        d.type = checkInitializer(d.getName(), d.type, d.init);
    }

    private VSTTypeRef checkInitializer(String name, VSTTypeRef decl, VSTExpr init) {
        if (decl == null) {
            // if the type is not declared, infer and check
            TypeName typeName = program.typeCache.getTypeName(inferType(init));
            VSTTypeRef tref = new VSTTypeRef(init.getToken(), typeName);
            checkVarType(tref);
            return tref;
        } else {
            checkVarType(decl);
            if (init != null) {
                Type dst = decl.getTypeName().getType();
                if (isArray(dst) && init instanceof VSTArrayInitializer)
                    typeCheckArrayInitializer(name, (VirgilArray.IType) dst, (VSTArrayInitializer) init);
                else typeCheckExpr("initialization of", name, init, dst);
            }
            return decl;
        }
    }

    public void visit(VSTArrayInitializer a) {
        // this method is only called in very specific circumstances:
        // when the type of an array initializer must be inferred
        // in order to determine the type of a local variable that
        // has been declared without a type signature.

        if (a.list.size() < 1) ERROR.CannotInferEmptyArrayType(a);

        Type comp = null;
        for (VSTExpr e : a.list) {
            Type t = inferType(e);
            if (comp == null) comp = t;
            else {
                Type u = typeSystem.unify(comp, t);
                if (u == null) ERROR.CannotUnifyElemTypes(e, comp, t);
                comp = u;
            }
        }
        a.setType(getArrayType(program.typeCache.getTypeName(comp)));
    }

    private void typeCheckArrayInitializer(String name, VirgilArray.IType outertype, VSTArrayInitializer outerexpr) {
        Type innertype = outertype.getElemType();
        for (VSTExpr innerexpr : outerexpr.list) {
            if (isArray(innertype) && innerexpr instanceof VSTArrayInitializer)
                typeCheckArrayInitializer(name, (VirgilArray.IType) innertype, (VSTArrayInitializer) innerexpr);
            else typeCheckExpr("initialization of array", name, innerexpr, innertype);
        }
        outerexpr.setType(outertype);
    }

    private void typeCheckExpr(String op, Object o, VSTExpr e, Type dst) {
        // conditionals in for loops can be null
        if (e == null) return;
        // visit the node to compute its type
        e.accept(this);
        // if the type cannot be assigned or promoted, this is a type error
        if (promoteExpr(e, dst) == null) {
            // give more detailed messages if there is a parameter
            String msg = o == null ? op : op + " " + quote(o);
            ERROR.TypeMismatch(e, msg, dst, e.getType());
        }
    }

    private Type promoteExpr(VSTExpr e, Type dst) {
        Type etype = e.getType();
        if (isVoid(etype)) {
            // no expression is allowed to be of type void
            ERROR.ExprCannotBeVoid(e);
        } else if ( typeSystem.isAssignable(etype, dst) ) {
            // if assignable, no promotion is necessary
            return dst;
        } else if ( typeSystem.isPromotable(etype, dst) ) {
            // if promotable, record the need for promotion code
            if ( etype != dst ) e.setPromotionType(dst);
            return dst;
        }
        return null;
    }

    public void visit(VSTAssign a) {
        typeCheckAssign(a, checkLvalue(a.target), a.value);
    }

    protected VSTExpr checkLvalue(VSTExpr e) {
        // TODO: calling inferType may cause exponential worst-case time
        if (e instanceof VSTIndexExpr ) {
            VSTIndexExpr vi = (VSTIndexExpr)e;
            if ( VirgilTypeSystem.isRaw(inferType(vi.array)) ) {
                checkLvalue(vi.array);
                return e;
            }
        }
        if (!e.isLvalue()) ERROR.NotAnLvalue(e);
        return e;
    }

    public void visit(VSTCompoundAssign a) {
        VSTExpr target = checkLvalue(a.target);
        Type dtype = inferType(target);
        String op = binopFromCompoundAssign(a.getToken().image);

        Capability.BinOp binop = typeSystem.resolveBinOp(op, dtype, inferType(a.value));

        if (binop == null) ERROR.UnresolvedAssignOp(a, dtype);

        checkOperand(target, binop, binop.binop.type1);
        checkOperand(a.value, binop, binop.binop.type2);

        a.setBinOp(binop);
        a.setType(dtype);

        if ( !typeSystem.isAssignable(binop.binop.getResultType(), dtype) ) {
            ERROR.TypeMismatch(a, "compound assignment", dtype, binop.binop.getResultType());
        }
    }

    private String binopFromCompoundAssign(String image) {
        return image.substring(0, image.length() - 1);
    }

    private void typeCheckAssign(VSTExpr a, VSTExpr target, VSTExpr value) {
        Type targettype = inferType(target);
        if (target instanceof VSTMemberExpr) {
            checkMemberAssignment(target, a);
        } else if (target instanceof VSTVarUse) {
            VSTVarUse u = (VSTVarUse) target;
            if (u.isMethod()) ERROR.CannotAssignToMember(a);
        }
        typeCheckExpr("assignment", null, value, targettype);
        a.setType(targettype);
    }

    private void checkMemberAssignment(VSTExpr target, VSTExpr a) {
        VSTMemberExpr e = (VSTMemberExpr) target;
        if (e.binding instanceof VSTMemberExpr.ComponentMember)
            if (((VSTMemberExpr.ComponentMember) e.binding).member instanceof Method) ERROR.CannotAssignToMember(a);
        if (e.binding instanceof VSTMemberExpr.ClassMember)
            if (((VSTMemberExpr.ClassMember) e.binding).member instanceof Method) ERROR.CannotAssignToMember(a);
        if (e.binding instanceof VSTMemberExpr.ArrayLength) ERROR.CannotAssignToMember(a);
    }

    public void visit(VSTVarUse u) {
        Decl d = resolveVarUse(u);
        if (d instanceof Typeable) {
            u.setType(((Typeable) d).getType());
        } else {
            ERROR.ExpectedExprType(u);
        }
    }

    public void visit(VSTTypeQueryExpr e) {
        Type etype = inferType(e.expr);

        if (isClass(etype) || isNull(etype)) {
            // if the expression type is an object (or null), this an "instanceof" test
            VirgilClass.IType qtc = resolveClassType(e.type);
            if ( !classesAreRelated(etype, qtc) )
                ERROR.InvalidTypeQuery(e, quote(qtc)+" is not a subclass of "+quote(etype));
        } else {
            // Virgil does not support type queries on non-object types
            ERROR.InvalidTypeQuery(e, "object expression expected");
        }
    }

    private boolean classesAreRelated(Type et, VirgilClass.IType qtc) {
        if ( isNull(et) ) return true;
        if ( !isClass(et) ) return false;
        VirgilClass.IType etc = (VirgilClass.IType)et;
        return qtc.isSubtypeOf(etc) || etc.isSubtypeOf(qtc);
    }

    public void visit(VSTTypeCastExpr e) {
        Type etype = inferType(e.expr);
        Type rtype;

        if (isClass(etype) || isNull(etype)) {
            // if the expression type is an object, this a dynamic object cast
            rtype = visitObjectTypeCast(e, etype);
        } else {
            // this is a type conversion, not a type cast
            rtype = visitTypeConversion(e, etype);
        }
        // the result type is the type specified in the cast
        e.setType(rtype);
    }

    private Type visitTypeConversion(VSTTypeCastExpr e, Type etype) {
        Type dtype = resolveType(e.type);
        if ( !typeSystem.isPromotable(etype, dtype) && !typeSystem.isConvertible(etype, dtype) )
            ERROR.InvalidTypeCast(e, "cannot convert "+quote(etype)+" to "+quote(dtype));
        return dtype;
    }

    private Type visitObjectTypeCast(VSTTypeCastExpr e, Type etype) {
        VirgilClass.IType qtc = resolveClassType(e.type);
        if ( !classesAreRelated(etype, qtc) )
            ERROR.InvalidTypeCast(e, quote(qtc)+" is not a subclass of "+quote(etype));
        return qtc;
    }

    public void visit(VSTIntLiteral i) {
        Maybe<PrimInt32.Val> mv = i.value;
        if (!mv.exists()) ERROR.InvalidLiteral(i);
    }

    public void visit(VSTRawLiteral i) {
        Maybe<PrimRaw.Val> mv = i.value;
        if (!mv.exists()) ERROR.InvalidLiteral(i);
    }

    public void visit(VSTIndexExpr e) {
        Type type = inferType(e.array);
        if ( !type.isIndexable() )
            ERROR.ExpectedIndexable(e.array, type);
        typeCheckExpr("index expression", null, e.index, PrimInt32.TYPE);
        if ( type instanceof VirgilArray.IType ) {
            VirgilArray.IType arraytype = (VirgilArray.IType) type;
            e.setType(arraytype.getElemType());
        } else if ( type instanceof PrimRaw.IType) {
            e.setType(PrimRaw.getType(1));
        } else {
            throw Util.failure("unknown indexable type: "+type);
        }
    }

    public void visit(VSTAppExpr a) {
        Type t = inferType(a.func);
        if (!isFunction(t)) ERROR.ExpectedFunction(a.func, t);

        Function.IType f = (Function.IType) t;
        Type[] atypes = f.getArgumentTypes();

        if (a.params.size() != atypes.length) ERROR.ArgumentCountMismatch(a, atypes.length, a.params.size());

        int cntr = 0;
        for (VSTExpr arg : a.params) {
            Type dst = atypes[cntr];
            typeCheckExpr("invocation", null, arg, dst);
            cntr++;
        }

        a.setType(f.getReturnType());
    }

    public void visit(VSTSuperClause sc) {
        if (sc.target == null) throw Util.failure("unbound super call");

        TypeName[] atypes = sc.target.getArgumentTypes();

        if (sc.params.size() != atypes.length) ERROR.ArgumentCountMismatch(sc, atypes.length, sc.params.size());

        int cntr = 0;
        for (VSTExpr arg : sc.params) {
            Type dst = atypes[cntr].getType();
            typeCheckExpr("super constructor invocation", null, arg, dst);
            cntr++;
        }

    }

    public void visit(VSTNewArrayExpr nae) {
        TypeName rname = resolveTypeName(nae.type);
        for (VSTExpr e : nae.list) {
            typeCheckExpr("array dimension", null, e, PrimInt32.TYPE);
            rname = getArrayTypeName(rname);
        }
        nae.setType(rname.getType());
    }

    public void visit(VSTNewObjectExpr ne) {
        VirgilClass.IType tc = resolveClassType(ne.type);
        ne.decl = prRepOf(tc);
        Constructor cs = ne.decl.getConstructor();
        TypeName[] atypes = Function.NOARGS;
        if (cs != null) atypes = cs.getArgumentTypes();

        if (ne.params.size() != atypes.length) ERROR.ArgumentCountMismatchInNew(ne, atypes.length, ne.params.size());

        int cntr = 0;
        for (VSTExpr arg : ne.params) {
            Type dst = atypes[cntr].getType();
            typeCheckExpr("constructor invocation", null, arg, dst);
            cntr++;
        }

        ne.setType(tc);
    }

    public void visit(VSTMemberExpr fe) {
        if (fe.expr instanceof VSTVarUse) {
            // the expression matches var.member => could mean component.member
            resolveVarMember(fe);
        } else {
            // the expression is (e).member => infer type and check
            resolveExprMember(fe);
        }
    }


    public void visit(VSTReturnStmt rs) {
        if (rs.expr != null) typeCheckExpr("return", null, rs.expr, thisReturnType);
    }

    public void visit(VSTStringLiteral l) {
        TypeName cname = program.typeCache.getTypeName("char");
        l.setType(getArrayType(cname));
    }

    private Type getArrayType(TypeName cname) {
        return getArrayTypeName(cname).getType();
    }

    private VirgilArray.ITypeName getArrayTypeName(TypeName elemName) {
        return VirgilArray.getArrayTypeName(program.typeCache, elemName);
    }

    public void visit(VSTIfStmt s) {
        typeCheckExpr("if statement", null, s.cond, PrimBool.TYPE);
        visitIfNonNull(s.trueStmt);
        visitIfNonNull(s.falseStmt);
    }

    public void visit(VSTSwitchStmt s) {
        Type valuetype = inferType(s.value);
        if (!valuetype.isSwitchable()) ERROR.NonSwitchableType(s.value, valuetype);

        HashMap<String, VSTExpr> cases = new HashMap<String, VSTExpr>();

        for (VSTSwitchCase c : s.cases) {
            for (VSTExpr val : c.list) {
                String cval = evaluateCaseValue(val, valuetype);
                VSTExpr prev = cases.get(cval);
                if (prev != null) {
                    ERROR.DuplicateCase(val, prev);
                }
                cases.put(cval, val);
            }

            visitIfNonNull(c.stmt);
        }

        // visit the default case (if there is one)
        if (s.defcase != null) visitIfNonNull(s.defcase.stmt);
    }

    private String evaluateCaseValue(VSTExpr e, Type valuetype) {
        typeCheckExpr("switch case", null, e, valuetype);
        if (!e.isComputable()) ERROR.NotComputable(e);
        try {
            return e.computeConstValue().toString();
        } catch (Operator.Exception ex) {
            ERROR.UnexpectedException(e, ex);
            return "";
        }
    }

    public void visit(VSTWhileStmt s) {
        typeCheckExpr("while statement", null, s.cond, PrimBool.TYPE);
        visitIfNonNull(s.body);
    }

    public void visit(VSTForStmt s) {
        if (s.init != null) visitExprs(s.init);
        typeCheckExpr("for statement", null, s.cond, PrimBool.TYPE);
        if (s.update != null) visitExprs(s.update);
        s.body.accept(this);
    }

    public void visit(VSTTernaryExpr e) {
        typeCheckExpr("ternary expression", null, e.cond, PrimBool.TYPE);
        Type tt = inferType(e.trueExpr);
        Type ft = inferType(e.falseExpr);
        Type united = typeSystem.unify(tt, ft);

        if (united == null) ERROR.CannotUnifyBranchTypes(e, tt, ft);

        e.setType(united);
    }

    public void visit(VSTBinOp b) {

        Capability.BinOp binop = typeSystem.resolveBinOp(b.getOp(), inferType(b.left), inferType(b.right));

        if (binop == null) ERROR.UnresolvedBinOp(b);

        checkOperand(b.left, binop, binop.binop.type1);
        checkOperand(b.right, binop, binop.binop.type2);

        b.setBinOp(binop);
        b.setType(binop.binop.getResultType());
    }

    private void checkOperand(VSTExpr e, Capability.BinOp op, Type xlt) {
        if ( promoteExpr(e, xlt) == null ) {
            String opstr = "operator " + op.opname +"("+op.binop.type1+", "+op.binop.type2+")";
            ERROR.TypeMismatch(e, opstr, xlt, e.getType());
        }
    }

    public void visit(VSTPreOp p) {
        VSTExpr expr = checkLvalue(p.expr);
        Capability.AutoOp at = resolveAutoOp(p, p.getToken().toString(), expr);
        p.setType(at.binop.getResultType());
        p.setAutoOp(at);
    }

    public void visit(VSTPostOp p) {
        VSTExpr expr = checkLvalue(p.expr);
        Capability.AutoOp at = resolveAutoOp(p, p.getToken().toString(), expr);
        p.setType(at.binop.getResultType());
        p.setAutoOp(at);
    }

    private Capability.AutoOp resolveAutoOp(VSTExpr p, String op, VSTExpr expr) {
        Capability.AutoOp at = typeSystem.resolveAutOp(op, inferType(expr));
        if ( at == null ) ERROR.UnresolvedOperator(p, op+" on type "+ StringUtil.quote(expr.getType()));
        return at;
    }

    public void visit(VSTComparison c) {
        Type lt = inferType(c.left);
        Type rt = inferType(c.right);
        if (!lt.isEqualityType()) ERROR.NotAnEqualityType(c.left, lt);
        if (!rt.isEqualityType()) ERROR.NotAnEqualityType(c.right, rt);
        if ( !typeSystem.isComparable(lt, rt) ) {
            if ( typeSystem.isPromotable(lt, rt) ) c.left.setPromotionType(rt);
            else if ( typeSystem.isPromotable(rt, lt) ) c.right.setPromotionType(lt);
            else ERROR.CannotCompareValues(c, lt, rt);
        }
        //if (!typeSystem.isComparable(lt, rt)) ERROR.CannotCompareValues(c, lt, rt);
    }

    public void visit(VSTUnaryOp u) {
        Type t = inferType(u.expr);
        Capability.UnaryOp uo = typeSystem.resolveUnOp(u.getToken().toString(), t);

        if (uo == null) ERROR.UnresolvedUnaryOp(u, t);

        if ( promoteExpr(u.expr, uo.unop.type1) == null ) {
            String opstr = "operator " + uo.opname +"("+uo.unop.type1+")";
            ERROR.TypeMismatch(u, opstr, uo.unop.type1, t);
        }

        u.setUnaryOp(uo);
        u.setType(uo.unop.getResultType());
    }

    public void visit(VSTThisLiteral t) {
        if (thisClass == null) ERROR.ThisMustBeInClass(t);
        t.setType(thisClass);
    }

}
