/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Feb 9, 2006
 */

package vpc.vst.stages;

import vpc.core.*;
import vpc.core.concept.PrimInt32;
import vpc.core.virgil.*;
import vpc.hil.DeviceDecl;
import vpc.hil.Register;
import vpc.types.*;
import vpc.vst.VSTErrorReporter;
import vpc.vst.VSTUtil;
import vpc.vst.tree.*;
import vpc.vst.visitor.VSTDepthFirstVisitor;

/**
 * @author Ben L. Titzer
 */
public class TypeCheckerBase extends VSTDepthFirstVisitor {

    protected TypeSystem typeSystem;
    protected CompoundDecl container;
    protected Program program;
    protected VSTErrorReporter ERROR;
    protected NameSpace namespace;

    protected Type resolveType(VSTTypeRef tref) {
        if (tref == null) return null;
        if (!tref.getTypeName().isResolved()) ERROR.UnresolvedType(tref);
        return tref.getTypeName().getType();
    }

    protected TypeName resolveTypeName(VSTTypeRef tref) {
        if (tref == null) return null;
        if (!tref.getTypeName().isResolved()) ERROR.UnresolvedType(tref);
        return tref.getTypeName();
    }

    protected VirgilClass.IType resolveClassType(VSTTypeRef tref) {
        Type type = resolveType(tref);
        if (!VirgilTypeSystem.isClass(type))
            ERROR.ExpectedObjectType(tref);
        return (VirgilClass.IType)type;
    }

    protected Type inferType(VSTExpr e) {
        e.accept(this);
        return e.getType();
    }

    protected Type checkVarType(VSTTypeRef tref) {
        TypeName tn = tref.getTypeName();
        if (!tn.isResolved()) ERROR.UnresolvedType(tref);
        Type t = tn.getType();
        if (VirgilTypeSystem.isVoid(t)
                || VirgilTypeSystem.isComponent(t)
                || VirgilTypeSystem.isNull(t)) ERROR.ExpectedVarType(tref);
        return t;
    }

    protected VirgilClass prRepOf(VirgilClass.IType c) {
        return program.virgil.getClassDecl(c.toString());
    }

    protected Decl resolveVarUse(VSTVarUse u) {
        if (u.decl != null) return u.decl;
        else {
            Decl decl = namespace.resolveDecl(u.getName());
            if (decl == null) ERROR.UnresolvedIdentifier(u);
            return u.decl = decl;
        }
    }

    protected void resolveExprMember(VSTMemberExpr fe) {
        Type type = inferType(fe.expr);
        if (VirgilTypeSystem.isClass(type)) {
            resolveObjectMember(type, fe);
        } else if (type.isArray()) {
            resolveArrayMember(fe, type);
        } else ERROR.UnresolvedMember(fe, type);
    }

    private void resolveArrayMember(VSTMemberExpr fe, Type type) {
        if ("length".equals(fe.getMemberName())) {
            fe.bindToArrayLength();
            fe.setType(PrimInt32.TYPE);
        } else ERROR.UnresolvedMember(fe, type);
    }

    private void resolveObjectMember(Type type, VSTMemberExpr fe) {
        VirgilClass pcd = prRepOf((VirgilClass.IType) type);
        Member m = VSTUtil.getVisibleMember(program, fe.getMemberName(), pcd, container);
        if (m == null) ERROR.UnresolvedMember(fe, pcd);
        fe.bindToClass(m);
        fe.setType(m.getTypeName().getType());
    }

    protected void resolveVarMember(VSTMemberExpr fe) {
        VSTVarUse u = (VSTVarUse) fe.expr;
        Decl d = resolveVarUse(u);
        if (d instanceof Typeable) {
            // if the declaration is typeable, resolve the type member
            resolveExprMember(fe);
        } else {
            if (d instanceof VirgilComponent) {
                // if the resolved declaration is a component, this is a field access
                VirgilComponent cd = (VirgilComponent) d;
                Member m = VSTUtil.getVisibleMember(program, fe.getMemberName(), cd, container);
                if (m == null) ERROR.UnresolvedMember(fe, cd);
                fe.bindToComponent(cd, m);
                fe.setType(m.getTypeName().getType());
            } else if (d instanceof DeviceDecl) {
                // if the resolved declaration is a device, this is a register access
                DeviceDecl dev = (DeviceDecl) d;
                Register r = dev.registers.get(fe.getMemberName());
                if (r == null) ERROR.UnresolvedMember(fe, dev);
                fe.bindToRegister(dev, r);
                fe.setType(r.getType());
            } else ERROR.UnresolvedIdentifier(u);
        }
    }

    protected void setContainer(CompoundDecl pcd) {
        container = pcd;
        namespace = container.internalNameSpace;
    }
}
