/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc.vst.stages;

import cck.util.Util;
import vpc.core.*;
import vpc.core.concept.*;
import vpc.core.virgil.*;
import vpc.types.*;
import vpc.vst.VSTErrorReporter;
import vpc.vst.VSTUtil;
import vpc.vst.tree.*;
import java.util.HashSet;
import java.util.List;

/**
 * The inner class <code>VSTTypeBuilder</code> contains the portion
 * of the verifier that will build the global type system, including
 * the inheritance tree, and build several important properties of the
 * whole program. The whole program must be present for this analysis.
 *
 * @author Ben L. Titzer
 */
public class VSTTypeBuilder {
    private final HashSet<VirgilClass> verified;
    private final HashSet<TypeName> stack;
    private final HashSet<VirgilClass> built;
    private VSTErrorReporter ERROR;
    private Program program;

    public VSTTypeBuilder(Program p, VSTErrorReporter error) {
        verified = new HashSet<VirgilClass>();
        stack = new HashSet<TypeName>();
        built = new HashSet<VirgilClass>();
        ERROR = error;
        program = p;
    }

    public void bindClassTypes(Iterable<VirgilClass> d) {
        for (VirgilClass cd : d) buildClassType(cd);
    }

    public void bindComponentTypes(Iterable<VirgilComponent> d) {
        for (VirgilComponent cd : d) buildComponentType(cd);
    }

    public void visit(VirgilClass pcd) {
        buildClass(pcd);
    }

    public static Function.ITypeName buildConstructorTypeName(TypeCache typeCache, TypeName[] args) {
        return new Function.ITypeName(typeCache.getTypeName("void"), args);
    }

    public static Function.ITypeName buildFuncTypeName(TypeCache typeCache, VSTMethodDecl m) {
        return Function.getFuncTypeName(typeCache, m.getReturnType(), buildFuncArgTypes(m.params));
    }

    public static TypeName[] buildFuncArgTypes(List<VSTParamDecl> d) {
        TypeName[] tn = new TypeName[d.size()];
        int cntr = 0;
        for (VSTParamDecl p : d) tn[cntr++] = p.getTypeName();
        return tn;
    }

    private class ClassBuilder {

        private VirgilClass classDecl;
        private VirgilClass parentDecl;

        ClassBuilder(VirgilClass c) {
            classDecl = c;
            parentDecl = program.virgil.getParentClass(c);
        }

        public void visit(Field f) {
            VSTFieldDecl fd = VSTUtil.vstRepOf(f);
            String name = fd.getName();
            if (VSTUtil.getVisibleMember(program, name, parentDecl, classDecl) != null) ERROR.MemberDefinedInSuper(fd);
            resolvedCheck(fd);
        }


        public void visit(Method meth) {
            String name = meth.getName();
            Member pm = VSTUtil.getVisibleMember(program, name, parentDecl, classDecl);
            if (pm != null) {
                if (pm instanceof Method) {
                    Method pmeth = (Method) pm;
                    overrideCheck(meth, pmeth);
                    Method.Family mf = pmeth.family;
                    mf.addMethod(pmeth, meth);
                    meth.family = mf;
                } else {
                    ERROR.MemberDefinedInSuper(VSTUtil.vstRepOf(meth));
                }
            }
            resolvedCheck(VSTUtil.vstRepOf(meth));
        }

        public void checkConstructor(Constructor c) {
            VSTClassDecl vstcd = VSTUtil.vstRepOf(classDecl);
            if (c == null) {
                if (mustCallSuper()) ERROR.NoDefaultConstructor(vstcd);
                else classDecl.newConstructor(buildConstructorTypeName(program.typeCache, Function.NOARGS));
            } else {
                VSTConstructorDecl cd = VSTUtil.vstRepOf(c);
                resolvedCheck(cd);

                if (cd.supclause != null) {
                    if (parentDecl == null) ERROR.NoSuperClass(classDecl, cd.supclause);
                    cd.supclause.target = parentDecl.getConstructor();
                } else if (mustCallSuper()) {
                    ERROR.NoDefaultConstructor(cd);
                }
            }
        }

        private boolean mustCallSuper() {
            return parentDecl != null && parentDecl.hasParameterizedConstructor();
        }

    }

    private void overrideCheck(Method m, Method pm) {

        VSTMethodDecl md = VSTUtil.vstRepOf(m);

        if (m.getReturnType() != pm.getReturnType()) ERROR.CannotOverrideReturnType(md);

        TypeName[] pa = pm.getArgumentTypes();

        if (pa.length != md.params.size()) ERROR.CannotOverrideArity(md);

        int cntr = 0;
        for (VSTParamDecl pd : md.params) {
            if (pa[cntr] != pd.getTypeRef().getTypeName()) ERROR.CannotOverrideParamType(md, pd);
            cntr++;
        }
    }

    private void buildClass(VirgilClass pcd) {
        if (built.contains(pcd)) return;

        VirgilClass ppcd = program.virgil.getParentClass(pcd);
        if (ppcd != null) buildClass(ppcd);

        ClassBuilder cb = new ClassBuilder(pcd);
        for (Field f : pcd.getFields()) cb.visit(f);
        for (Method m : pcd.getMethods()) cb.visit(m);
        cb.checkConstructor(pcd.getConstructor());
        if (ppcd != null) pcd.internalNameSpace.add(ppcd.externalNameSpace);
        pcd.internalNameSpace.add(program.namespace);
        built.add(pcd);
    }

    public void visit(VirgilComponent pcd) {
        // record which module we are in for error reporting purposes.
        ComponentBuilder cb = new ComponentBuilder(pcd);
        for (Field f : pcd.getFields()) cb.visit(f);
        for (Method m : pcd.getMethods()) cb.visit(m);
        cb.checkConstructor(pcd.getConstructor());
        pcd.internalNameSpace.add(program.namespace);
    }

    private class ComponentBuilder {
        private VirgilComponent componentDecl;

        ComponentBuilder(VirgilComponent pcd) {
            componentDecl = pcd;
        }

        public void visit(Field f) {
            resolvedCheck(VSTUtil.vstRepOf(f));
        }

        public void visit(Method m) {
            resolvedCheck(VSTUtil.vstRepOf(m));
        }

        public void checkConstructor(Constructor c) {
            VSTConstructorDecl cd;
            if (c != null) {
                cd = VSTUtil.vstRepOf(c);
                resolvedCheck(cd);
                if (cd != null && cd.supclause != null) ERROR.SuperClauseMustBeInClass(cd.supclause);
            } else {
                componentDecl.newConstructor(buildConstructorTypeName(program.typeCache, Function.NOARGS));
            }
        }

    }

    private Type buildType(VSTTypeRef typeRef) {
        TypeName typeName = typeRef.getTypeName();
        if ( typeName.isResolved() ) return typeName.getType();
        TypeDecl td = program.getTypeDecl(typeName);
        if ( td == null )
            ERROR.UnresolvedType(typeRef);
        return buildType(td);
    }

    private Type buildType(TypeDecl d) {
        if ( d.getTypeName().isResolved() ) return d.getTypeName().getType();
        if ( d instanceof VirgilClass ) return buildClassType((VirgilClass)d);
        if ( d instanceof VirgilComponent ) return buildComponentType((VirgilComponent)d);
        throw Util.failure("unknown type declaration kind "+d.getClass());
    }

    private VirgilComponent.IType buildComponentType(VirgilComponent c) {
        TypeName compName = c.getTypeName();
        VirgilComponent.IType type = new VirgilComponent.IType(compName);
        compName.resolveTo(type);
        return type;
    }

    private VirgilClass.IType buildClassType(VirgilClass classDecl) {
        if ( verified.contains(classDecl) ) {
            return (VirgilClass.IType)classDecl.getTypeName().getType();
        }
        TypeName className = classDecl.getTypeName();
        TypeName parentName = classDecl.getParentName(program.typeCache);
        VirgilClass.IType parentType = null;

        if (parentName != null) {
            // is there a declared parent?
            stack.add(className);
            if (stack.contains(parentName)) {
                // check for cyclic inheritance first
                ERROR.CyclicInheritance(VSTUtil.vstRepOf(classDecl).parent);
            }
            Type t = buildType(VSTUtil.vstRepOf(classDecl).parent);
            if (!VirgilTypeSystem.isClass(t)) {
                // check that the parent type is a class
                ERROR.ExpectedObjectType(VSTUtil.vstRepOf(classDecl).parent);
            }
            parentType = (VirgilClass.IType)t;

            VirgilClass parentDecl = parentType.getDecl();
            stack.remove(className);
            program.virgil.hierarchy.add(parentDecl, classDecl);
        }

        VirgilClass.IType type = new VirgilClass.IType(className.toString(), parentType);
        className.resolveTo(type);
        type.setDecl(classDecl);
        verified.add(classDecl);
        return type;
    }

    private void resolvedCheck(VSTFieldDecl fd) {
        if (!fd.getTypeName().isResolved()) ERROR.UnresolvedType(fd.getType());
    }

    private void resolvedCheck(VSTMethodDecl md) {
        if (!md.getReturnType().isResolved()) ERROR.UnresolvedType(md.returnType);
        for (VSTParamDecl pd : md.params) {
            if (!pd.getTypeName().isResolved()) ERROR.UnresolvedType(pd.getTypeRef());
        }
    }

    private void resolvedCheck(VSTConstructorDecl md) {
        for (VSTParamDecl pd : md.params) {
            if (!pd.getTypeName().isResolved()) ERROR.UnresolvedType(pd.getTypeRef());
        }
    }

} // End of VSTTypeBuilder
