/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Jan 26, 2006
 */

package vpc.vst.tir;

import vpc.core.CompoundDecl;
import vpc.core.concept.Field;
import vpc.core.concept.Method;
import vpc.tir.*;
import vpc.types.Type;
import vpc.vst.VSTUtil;
import vpc.vst.tree.*;
import java.util.*;

/**
 * @author Ben L. Titzer
 */
public abstract class BuilderEnv {
    protected TIRRep method;
    protected HashMap<VSTVarDecl, Method.Temporary> varMap;
    protected HashSet<String> collisions;
    protected TIRBuilder tirBuilder;
    protected CompoundDecl compound;
    protected Method.Temporary thisParam;

    protected BuilderEnv(TIRBuilder tirBuilder, CompoundDecl comp) {
        this.tirBuilder = tirBuilder;
        compound = comp;
    }

    abstract TIROperator getField(Field field);

    abstract TIROperator getMethod(Method method);

    abstract VSTSuperClause getSuperClause(VSTConstructorDecl cd);

    abstract TIRExpr newThisRef();

    TIRRep buildMethod(VSTMethodDecl md) {
        return buildCodeRep(md.params, md.block);
    }

    TIRRep buildConstructor(VSTConstructorDecl cd, Iterable<Field> fields) {
        method = new TIRRep();
        varMap = new HashMap<VSTVarDecl, Method.Temporary>();
        // add parameters to the constructor
        expandParams(cd == null ? TIRBuilder.NOPARAMS : cd.params, compound.getTypeName().getType());

        VSTCompiler bb = new VSTCompiler(this);
        TIRBlock block = newBlock("L");
        addSuperClause(cd, block, bb);
        addFieldInits(fields, block, bb);
        addConstructorBody(cd, block, bb);
        method.setBody(block);
        return method;
    }

    public TIRBlock newBlock(String lbl) {
        return new TIRBlock(method.newLabel(lbl));
    }

    private void addSuperClause(VSTConstructorDecl cd, TIRBlock block, VSTCompiler bb) {
        VSTSuperClause sup = getSuperClause(cd);
        if (sup != null) block.addExpr(bb.visit(sup));
    }

    private void addConstructorBody(VSTConstructorDecl cd, TIRBlock block, VSTCompiler bb) {
        if (cd != null && cd.block != null) {
            for (VSTStmt s : cd.block.stmts) block.addExpr(s.accept(bb));
        }
    }

    private void addFieldInits(Iterable<Field> fields, TIRBlock block, VSTCompiler bb) {
        for (Field f : fields) {
            VSTFieldDecl fd = VSTUtil.vstRepOf(f);
            if (fd.init != null) {
                TIROperator r = getField(f);
                // TODO: this source point location should be the assignment token
                block.addExpr(bb.convertGet(fd.getToken(), r, bb.buildExpr(fd.init)));
            }
        }
    }

    private TIRRep buildCodeRep(List<VSTParamDecl> params, VSTBlock block) {
        method = new TIRRep();
        varMap = new HashMap<VSTVarDecl, Method.Temporary>();
        // add parameters to the method.
        expandParams(params, compound.getTypeName().getType());
        buildBody(block);
        return method;
    }

    private void expandParams(List<VSTParamDecl> params, Type thisType) {
        thisParam = method.newParam(TIRUtil.THIS_PARAM, thisType);
        for (VSTParamDecl p : params) {
            Type type = p.getType();
            Method.Temporary t = method.newParam(p.getName(), type);
            varMap.put(p, t);
        }
    }

    private void buildBody(VSTBlock body) {
        // build the main block of the method.
        VSTCompiler bb = new VSTCompiler(this);
        if ( body != null ) {
            TIRExpr nbody = bb.buildStmtBlock(body.stmts);
            method.setBody(nbody);
            nbody.setSourcePoint(body.getToken().getSourcePoint());
        } else {
            method.setBody(TIRBuilder.NOOP);
        }
    }

    protected Method.Temporary newVariable(VSTVarDecl v) {
        Method.Temporary<Type> t = method.newVariable(v.getName(), v.getType());
        varMap.put(v, t);
        return t;
    }

} // end of BuilderEnv
