/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc.vst.tir;

import cck.parser.SourcePoint;
import cck.text.StringUtil;
import cck.util.Util;
import vpc.core.CompoundDecl;
import vpc.core.Program;
import vpc.core.concept.Constructor;
import vpc.core.concept.Method;
import vpc.core.virgil.VirgilClass;
import vpc.core.virgil.VirgilComponent;
import vpc.tir.*;
import vpc.vst.VSTUtil;
import vpc.vst.tree.*;
import java.util.LinkedList;
import java.util.List;

/**
 * The <code>TIRBuilder</code> class builds the IR representation from
 * the syntax tree. The principle differences between the VST and the
 * IR tree is simplification, resolution, and specialization. VSTs
 * represent the code in a form that is closer to its lexical structure
 * than its semantic level (e.g. assignments to locals and assignments
 * to fields are represented in the same way in the VST; in the IR
 * tree, a distinction is made).
 * <p/>
 * The <code>TIRBuilder</code> class makes use of some information
 * supplied by the verification phase, such as the type of expressions
 * and resolution of variable references.
 *
 * @author Ben L. Titzer
 */
public class TIRBuilder {

    protected static final TIRExpr NOOP = TIRNop.NOOP;
    protected static final List<VSTParamDecl> NOPARAMS = new LinkedList<VSTParamDecl>();

    protected final Program program;

    protected TIRBuilder(Program p) {
        program = p;
    }

    public static void buildIRForProgram(Program p) {
        TIRBuilder irb = new TIRBuilder(p);
        irb.buildIR(p);
    }

    private void buildIR(Program p) {
        for (VirgilClass pcd : p.virgil.getClasses())
            buildCompound(pcd, new ClassBuilderEnv(this, pcd));
        for (VirgilComponent pcd : p.virgil.getComponents())
            buildCompound(pcd, new ComponentBuilderEnv(this, pcd));
    }

    public void buildCompound(CompoundDecl pcd, BuilderEnv mbuilder) {
        for (Method m : pcd.getMethods())
            buildMethod(m, mbuilder);
        buildConstructor(pcd, mbuilder);
    }

    public void buildMethod(Method pm, BuilderEnv mbuilder) {
        if (pm == null) return;
        VSTMethodDecl md = VSTUtil.vstRepOf(pm);
        TIRRep rep = mbuilder.buildMethod(md);
        pm.addMethodRep(TIRRep.REP_NAME, rep);
    }

    public void buildConstructor(CompoundDecl pcd, BuilderEnv mbuilder) {
        Constructor pc = pcd.getConstructor();
        VSTConstructorDecl md = (VSTConstructorDecl) pc.getMethodRep(VSTUtil.REPRESENTATION_KEY);
        TIRRep rep = mbuilder.buildConstructor(md, pcd.getFields());
        pc.addMethodRep(TIRRep.REP_NAME, rep);
        pc.addMethodRep(TIRRep.REP_NAME, rep);
    }

    public static Util.Error fail(String msg, VSTNode n) {
        String nstr = StringUtil.quote(n.getClass());
        SourcePoint sourcePoint = n.getToken().getSourcePoint();
        return Util.failure("TIRBuilder failure: " + msg + " on node " + nstr + " @ " + sourcePoint);
    }

}
