/**
 * Copyright (c) 2005, Regents of the University of California
x * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Creation date: Jan 26, 2006
 */

package vpc.vst.tir;

import cck.text.StringUtil;
import cck.util.Util;
import vpc.core.*;
import vpc.core.concept.*;
import vpc.core.virgil.*;
import static vpc.core.virgil.VirgilTypeSystem.*;
import vpc.hil.DeviceDecl;
import vpc.tir.*;
import static vpc.tir.TIRUtil.*;
import vpc.tir.expr.Operator;
import vpc.types.Type;
import static vpc.vst.VSTUtil.*;
import vpc.vst.parser.Token;
import vpc.vst.tree.*;
import vpc.vst.visitor.VSTAccumulator;
import vpc.util.ArrayUtil;
import java.util.List;

/**
 * The <code>VSTCompiler</code> class implements a visitor that is able
 * to build blocks of IR code from blocks of abstract syntax trees.
 */
public class VSTCompiler implements VSTAccumulator<TIRExpr> {

    private BuilderEnv builderEnv;

    public VSTCompiler(BuilderEnv builderEnv) {
        this.builderEnv = builderEnv;
    }

    public TIRExpr buildExpr(VSTExpr e) {
        TIRExpr tirExpr = setTypeAndSourcePoint(e, typeOf(e), e.accept(this));
        return promote(e, tirExpr);
    }

    private TIRExpr promote(VSTExpr e, TIRExpr ne) {
        Type pt = e.getPromotionType();
        if ( pt != null ) {
            Type ot = typeOf(e);
            Operator op = null;
            if ( isInt(ot) ) {
                if ( isRaw(pt) )  op = new PrimConversion.Int32_Raw((PrimRaw.IType)pt);
            } else if ( isChar(ot) ) {
                if ( isInt(pt) ) op = new PrimConversion.Char_Int32();
                if ( isRaw(pt) ) op = new PrimConversion.Char_Raw((PrimRaw.IType)pt);
            } else if ( isRaw(ot) && isRaw(pt) ) {
                op = new PrimConversion.AdjustRaw((PrimRaw.IType)ot, (PrimRaw.IType)pt);
            }

            if ( op == null)
                throw fail("Invalid promotion from "+ot+" to "+pt, ne);

            return setTypeAndSourcePoint(e, pt, $OP(op, ne));
        }
        return ne;
    }

    private TIRExpr buildStmt(VSTStmt s) {
        if (s == null) {
            return newNoOp(null);
        } else {
            return setTypeAndSourcePoint(s, PrimVoid.TYPE, s.accept(this));
        }
    }

    TIRExpr buildStmtBlock(List<VSTStmt> l) {
        if (l == null) return newNoOp(null);
        TIRBlock block = builderEnv.newBlock("L_");
        for (VSTStmt stmt : l) {
            TIRExpr nstmt = buildStmt(stmt);
            if (!isNoOp(nstmt)) block.addExpr(nstmt);
        }
        return block;
    }

    TIRExpr buildExprBlock(List<VSTExpr> l) {
        if (l == null) return newNoOp(null);
        TIRBlock block = builderEnv.newBlock("E_");
        for (VSTExpr expr : l) {
            TIRExpr nexpr = buildExpr(expr);
            if (!isNoOp(nexpr)) block.addExpr(nexpr);
        }
        return block;
    }

    private TIRExpr[] buildExprList(List<VSTExpr> l) {
        TIRExpr[] res = new TIRExpr[l.size()];
        int cntr = 0;
        for (VSTExpr e : l) {
            res[cntr++] = buildExpr(e);
        }
        return res;
    }

    private Method.Temporary getTemporary(VSTVarDecl var) {
        return builderEnv.varMap.get(var);
    }

    private TIRLocal.Get liftTemp(TIRExpr e, TIRBlock block) {
        return lift(builderEnv.method, null, e, block);
    }

    private TIRExpr liftExpr(TIRExpr e, TIRBlock block) {
        if (e instanceof TIRLocal.Get || e instanceof TIRConst ) return e;
        else return lift(builderEnv.method, null, e, block);
    }

    private TIRExpr liftLvalue(TIRExpr e, TIRBlock block) {
        if (e instanceof TIRLocal.Get ) return e;
        if (e instanceof TIROperator ) {
            TIROperator to = (TIROperator)e;
            TIRExpr[] nl = new TIRExpr[to.operands.length];
            if ( to.operator instanceof PrimRaw.GetBit ) {
                // in the case of a get bit, we need to recursively lift the lvalue
                nl[0] = liftLvalue(to.operands[0], block);
                nl[1] = liftExpr(to.operands[1], block);
            } else {
                // otherwise, we can just lift all expressions to temporaries
                for ( int cntr = 0; cntr < to.operands.length; cntr++ ) {
                    nl[cntr] = liftExpr(to.operands[cntr], block);
                }
            }
            TIROperator tirOp = $OP(to.operator, nl);
            tirOp.setSourcePoint(e.getSourcePoint());
            return tirOp;
        }
        throw TIRUtil.fail("cannot lift location", e);
    }

    public TIRExpr buildBinOp(VSTNode w, Operator oper, TIRExpr le, TIRExpr re) {
        Token tok = w.getToken();
        if (oper != null) {
            TIRExpr e;
            Class clz = oper.getClass();
            if (clz == PrimBool.OR.class) e = $OR(le, re);
            else if (clz == PrimBool.AND.class) e = $AND(le, re);
            else e = $OP(oper, le, re);

            return setSourcePoint(tok, e);
        }

        throw TIRBuilder.fail("unknown binary op " + StringUtil.quote(tok.image), w);
    }

    public TIRExpr visit(VSTReturnStmt r) {
        if (r.expr == null) {
            // translate a void return into returning the special VOID value
            return new TIRReturn($VAL(PrimVoid.VALUE));
        } else {
            // translate a return with a value
            return new TIRReturn(buildExpr(r.expr));
        }
    }

    public TIRExpr visit(VSTEmptyStmt s) {
        return newNoOp(s.getToken());
    }

    public TIRExpr visit(VSTBlock b) {
        return buildStmtBlock(b.stmts); // flatten into this block
    }

    public TIRExpr visit(VSTSuperClause s) {
        TIRExpr[] args = buildExprList(s.params);
        TIRExpr tref = builderEnv.newThisRef();
        Operator op = opDirectDispatch((VirgilClass.IType)tref.getType(), s.target);
        return new TIRCall(setSourcePoint(s, $OP(op, tref)), args);
    }

    public TIRExpr visit(VSTIfStmt s) {
        return $IF(buildExpr(s.cond), buildStmt(s.trueStmt), buildStmt(s.falseStmt));
    }

    public TIRExpr visit(VSTWhileStmt s) {
        // while ( c ) e => block l { if ( c ) { e; continue l; } }
        // break in e => break l
        // continue in e => continue l
        return new TIRLoop.While(buildExpr(s.cond), buildStmt(s.body));
    }

    public TIRExpr visit(VSTDoWhileStmt s) {
        // do e while ( c ) => block l { e; if ( c ) { continue l; } }
        // break in e => break l
        // continue in e => continue l
        return new TIRLoop.DoWhile(buildExpr(s.cond), buildStmt(s.body));
    }

    public TIRExpr visit(VSTContinueStmt s) {
        return new TIRBlock.Continue();
    }

    public TIRExpr visit(VSTBreakStmt s) {
        return new TIRBlock.Break();
    }

    public TIRExpr visit(VSTExprStmt s) {
        return buildExpr(s.expr);
    }

    public TIRExpr visit(VSTForStmt s) {
        // for ( i; c; u ) e =>
        //     block b { i; block o { block l { if ( !c ) break o; e; } u; continue o; } } }
        // break in e => break o
        // continue in e => break l
        TIRExpr i = buildExprBlock(s.init);
        TIRExpr c = s.cond == null ? $VAL(true) : buildExpr(s.cond);
        TIRExpr b = buildStmt(s.body);
        TIRExpr u = buildExprBlock(s.update);
        return new TIRLoop.For(i, c, u, b);
    }

    public TIRExpr visit(VSTSwitchStmt s) {
        TIRExpr e = buildExpr(s.value);
        TIRSwitch.Case[] cs = new TIRSwitch.Case[s.cases.size()];
        int cntr = 0;
        for (VSTSwitchCase c : s.cases) {
            TIRConst.Value[] vals = null;
            if (c.list != null) vals = buildConsts(c.list);
            cs[cntr] = new TIRSwitch.Case(vals, buildStmt(c.stmt));
            cntr++;
        }
        TIRSwitch.Case dc = null;
        if (s.defcase != null) // build the default case
            dc = new TIRSwitch.Case(null, buildStmt(s.defcase.stmt));
        return new TIRSwitch(e, cs, dc);
    }

    private TIRConst.Value[] buildConsts(List<VSTExpr> l) {
        TIRConst.Value[] ret = new TIRConst.Value[l.size()];
        int cntr = 0;
        for (VSTExpr expr : l) {
            TIRExpr e = buildExpr(expr);
            if (e instanceof TIRConst.Value) ret[cntr] = (TIRConst.Value) e;
            else throw TIRUtil.fail("value required", e);
            cntr++;
        }
        return ret;
    }

    public TIRExpr visit(VSTLocalVarDecl d) {
        Method.Temporary t = builderEnv.newVariable(d);
        if (d.init == null) return newNoOp(d.getToken());
        else return new TIRLocal.Set(t, buildExpr(d.init));
    }

    public TIRExpr visit(VSTSwitchCase c) {
        throw TIRBuilder.fail("should not visit cases directly", c);
    }

    public TIRExpr visit(VSTArrayInitializer e) {
        // build an expression representing the element initializers
        TIRExpr[] elems = buildExprList(e.list);
        VirgilArray.IType at = (VirgilArray.IType) typeOf(e);
        VirgilArray.Alloc allocop = opArrayAlloc(at, 1);
        TIRConst.Value len = $VAL(elems.length);
        setSourcePoint(e, len);
        TIRExpr op = setSourcePoint(e, $OP(allocop, len));
        return $OP(new VirgilArray.Init(at, elems.length), ArrayUtil.prepend(op, elems));
    }

    public TIRExpr visit(VSTAssign e) {
        TIRExpr l = buildExpr(e.target);
        TIRExpr v = buildExpr(e.value);
        return convertAssign(e.getToken(), l, v);
    }

    public TIRExpr convertAssign(Token assign, TIRExpr l, TIRExpr v) {
        if ( l instanceof TIRLocal.Get ) {
            return setSourcePoint(assign, new TIRLocal.Set(((TIRLocal.Get) l).temp, v));
        } else if ( l instanceof TIROperator ) {
            return convertGet(assign, (TIROperator)l, v);
        }
        throw TIRUtil.fail("not an lvalue", l);
    }

    public TIRExpr visit(VSTBinOp e) {
        TIRExpr le = buildExpr(e.left);
        TIRExpr re = buildExpr(e.right);

        return buildBinOp(e, e.binop.binop, le, re);
    }

    public TIRExpr visit(VSTComparison c) {
        if (c.isEqOp()) return new TIRCompare.Equal(buildExpr(c.left), buildExpr(c.right));
        else return new TIRCompare.NotEqual(buildExpr(c.left), buildExpr(c.right));
    }

    public TIRExpr visit(VSTIndexExpr e) {
        Type it = typeOf(e.array);
        TIRExpr ne = buildExpr(e.array);
        TIRExpr i = buildExpr(e.index);
        if ( isArray(it) ) {
            Operator op = new VirgilArray.GetElement((VirgilArray.IType)it);
            return $OP(op, (TIRExpr)nullCheck(ne), i);
        } else if ( isRaw(it) ) {
            Operator op = new PrimRaw.GetBit((PrimRaw.IType)it);
            return $OP(op, ne, i);
        }
        throw TIRUtil.fail("invalid index expression of type "+it, ne);
    }

    public TIRExpr visit(VSTTypeQueryExpr e) {
        TIRExpr expr = buildExpr(e.expr);
        VirgilClass.IType stype = (VirgilClass.IType)expr.getType();
        VirgilClass.IType target = (VirgilClass.IType)typeOf(e.type);
        return $OP(new VirgilClass.TypeQuery(stype, target), expr);
    }

    public TIRExpr visit(VSTTypeCastExpr e) {
        TIRExpr ne = buildExpr(e.expr);
        Type ot = ne.getType();
        Type ct = typeOf(e.type);

        Operator op = null;
        if ( isClass(ot) ) {
            op = new VirgilClass.TypeCast((VirgilClass.IType)ot, (VirgilClass.IType)ct);
        } else if ( isInt(ot) ) {
            if ( isRaw(ct) )  op = new PrimConversion.Int32_Raw((PrimRaw.IType)ct);
            if ( isChar(ct) )  op = new PrimConversion.Int32_Char();
        } else if ( isChar(ot) ) {
            if ( isInt(ct) ) op = new PrimConversion.Char_Int32();
            if ( isRaw(ct) ) op = new PrimConversion.Char_Raw((PrimRaw.IType)ct);
        } else if ( isRaw(ot) ) {
            if ( isRaw(ct) )
                op = new PrimConversion.AdjustRaw((PrimRaw.IType)ot, (PrimRaw.IType)ct);
            if ( isChar(ct) )
                op = new PrimConversion.Raw_Char((PrimRaw.IType)ot);
            if ( isInt(ct) )
                op = new PrimConversion.Raw_Int32((PrimRaw.IType)ot);
        }

        if ( op == null)
            throw TIRUtil.fail("Invalid promotion from "+ot+" to "+ct, ne);

        return setTypeAndSourcePoint(e, ct, $OP(op, ne));
    }

    public TIRExpr visit(VSTBooleanLiteral e) {
        return $VAL(e.getValue());
    }

    public TIRExpr visit(VSTCharLiteral e) {
        return $VAL(e.getValue());
    }

    public TIRExpr visit(VSTIntLiteral e) {
        return $VAL(e.getValue());
    }

    public TIRExpr visit(VSTRawLiteral e) {
        return $VAL(e.getValue());
    }

    public TIRExpr visit(VSTStringLiteral e) {
        return new TIRConst.String(e.original(), e.value());
    }

    public TIRExpr visit(VSTNullLiteral e) {
        return $NULL();
    }

    public TIRExpr visit(VSTThisLiteral e) {
        return builderEnv.newThisRef();
    }

    public TIRExpr visit(VSTAppExpr e) {
        TIRExpr ref = buildExpr(e.func);
        return new TIRCall(ref, buildExprList(e.params));
    }

    public TIRExpr visit(VSTNewArrayExpr e) {
        TIRExpr[] dims = buildExprList(e.list);
        VirgilArray.IType at = (VirgilArray.IType)typeOf(e);
        return $OP(opArrayAlloc(at, dims.length), dims);
    }

    public TIRExpr visit(VSTNewObjectExpr e) {
        TIRExpr[] args = buildExprList(e.params);
        TIRBlock b = builderEnv.newBlock("NEW_");
        setSourcePoint(e.getToken(), b);
        VirgilClass.IType ct = (VirgilClass.IType)typeOf(e.type);
        TIRExpr ao = setSourcePoint(e.type, $OP(new VirgilClass.Alloc(ct, builderEnv.tirBuilder.program.heap)));
        TIRExpr obj = lift(builderEnv.method, null, ao, b);
        TIRExpr rc = setSourcePoint(e.type, $OP(opDirectDispatch(ct, ct.getDecl().getConstructor()), obj));
        b.addExpr(setSourcePoint(e.type, new TIRCall(rc, args)));
        b.addExpr(obj);
        return b;
    }

    public TIRExpr visit(VSTCompoundAssign e) {
        return readModifyWrite(e.getToken(), false, buildExpr(e.target), e.binop.binop, buildExpr(e.value));
    }

    public TIRExpr visit(VSTTernaryExpr e) {
        return $IF(buildExpr(e.cond), buildExpr(e.trueExpr), buildExpr(e.falseExpr));
    }

    public TIRExpr visit(VSTUnaryOp e) {
        return $OP(e.unaryop.unop, buildExpr(e.expr));
    }

    public TIRExpr visit(VSTPreOp p) {
        TIRExpr val = buildConstFor(p, p.autoop.value);
        return readModifyWrite(p.getToken(), false, buildExpr(p.expr), p.autoop.binop, val);
    }

    public TIRExpr visit(VSTPostOp p) {
        TIRExpr val = buildConstFor(p, p.autoop.value);
        return readModifyWrite(p.getToken(), true, buildExpr(p.expr), p.autoop.binop, val);
    }

    public TIRExpr visit(VSTMemberExpr e) {
        VSTMemberExpr.Binding b = e.binding;
        if (isComponentMember(b)) {
            return $OP(opComponentMember(b));
        } else if (isClassMember(b)) {
            return $OP(opClassMember(b, e), nullCheck(buildExpr(e.expr)));
        } else if (isArrayLength(b)) {
            return $OP(opArrayLength(e), nullCheck(buildExpr(e.expr)));
        } else if (isDeviceAccess(b)) {
            return $OP(opGetRegister(b));
        }
        throw TIRBuilder.fail("unbound member", e);
    }

    public TIRExpr visit(VSTVarUse u) {
        if (u.decl == null) throw TIRBuilder.fail("unbound VarUse", u);

        if (u.isLocal() || u.isParam()) {
            return new TIRLocal.Get(getTemporary((VSTVarDecl)u.decl));
        } else if (u.isField()) {
            Field field = (Field) u.decl;
            return builderEnv.getField(field);
        } else if (u.isMethod()) {
            Method method = (Method) u.decl;
            return builderEnv.getMethod(method);
        }

        throw TIRBuilder.fail("invalid VarUse type " + u.decl.getClass(), u);
    }

    private TIRExpr readModifyWrite(Token loc, boolean before, TIRExpr expr, Operator mod, TIRExpr val) {
        TIRBlock block = builderEnv.newBlock("RMW_");
        setSourcePoint(loc, block);
        expr = liftLvalue(expr, block);
        if ( before ) {
            TIRLocal.Get temp = liftTemp(expr, block);
            TIRExpr op = setSourcePoint(loc, $OP(mod, temp, val));
            block.addExpr(convertAssign(loc, expr, op));
            block.addExpr(temp);
        } else {
            TIRExpr op = setSourcePoint(loc, $OP(mod, expr, val));
            block.addExpr(convertAssign(loc, expr, op));
        }
        return block;
    }

    protected TIRExpr convertGet(Token assign, TIROperator read, TIRExpr val) {
        if ( read.operator.getClass() == PrimRaw.GetBit.class )
            return convertGetBit(assign, read, val);
        else if ( read.operator instanceof Operator.Location )
            return convertGetLocation(assign, read, val);
        throw Util.failure("unknown operator "+read.operator.getClass());
    }

    private TIRExpr convertGetLocation(Token assign, TIROperator read, TIRExpr val) {
        Operator setOp = ((Operator.Location)read.operator).setOperator();
        TIRExpr[] writeOps = ArrayUtil.append(read.operands, val);
        return setSourcePoint(assign, $OP(setOp, writeOps));
    }

    private TIRExpr convertGetBit(Token assign, TIROperator read, TIRExpr val) {
        // TODO: this read-modify-write operation is kind of ugly
        TIRExpr[] ops = read.operands;
        PrimRaw.GetBit f = (PrimRaw.GetBit)read.operator;
        PrimRaw.SetBit mod = new PrimRaw.SetBit((PrimRaw.IType)f.type1);
        TIRBlock block = builderEnv.newBlock("R");
        setSourcePoint(assign, block);
        TIRExpr expr = liftLvalue(ops[0], block);
        TIRExpr bit = liftExpr(ops[1], block);
        TIRExpr nval = liftExpr(val, block);
        TIROperator op = $OP(mod, expr, bit, nval);
        setSourcePoint(assign, op);
        block.addExpr(convertAssign(assign, expr, op));
        block.addExpr(nval);
        return block;
    }

    private TIRExpr setSourcePoint(VSTNode vn, TIRExpr expr) {
        expr.setSourcePoint(vn.getToken().getSourcePoint());
        return expr;
    }

    private TIRExpr setTypeAndSourcePoint(VSTNode vn, Type t, TIRExpr expr) {
        expr.setSourcePoint(vn.getToken().getSourcePoint());
        expr.setType(t);
        return expr;
    }

    private TIRExpr setSourcePoint(Token t, TIRExpr expr) {
        expr.setSourcePoint(t.getSourcePoint());
        return expr;
    }

    private Operator opComponentMember(VSTMemberExpr.Binding b) {
        VSTMemberExpr.ComponentMember mb = (VSTMemberExpr.ComponentMember) b;
        Member m = mb.member;
        if (m.isField()) return new VirgilComponent.GetField(mb.component, (Field) m);
        else return new VirgilComponent.GetMethod(mb.component, (Method) m);
    }

    private Operator opClassMember(VSTMemberExpr.Binding b, VSTMemberExpr e) {
        Member m = ((VSTMemberExpr.ClassMember) b).member;
        VirgilClass.IType type = (VirgilClass.IType)typeOf(e.expr);
        if (m.isField())
            return new VirgilClass.GetField(type, (Field) m);
        else {
            if ( isPrivate(m) ) return opDirectDispatch(type, (Method)m);
            else return opVirtualDispatch(type, (Method)m);
        }
    }

    private Operator opGetRegister(VSTMemberExpr.Binding b) {
        VSTMemberExpr.DeviceRegister binding = (VSTMemberExpr.DeviceRegister) b;
        return new DeviceDecl.GetRegister(binding.register);
    }

    private Operator opArrayLength(VSTMemberExpr e) {
        return new VirgilArray.GetLength((VirgilArray.IType)e.expr.getType());
    }

    private Operator opDirectDispatch(VirgilClass.IType ct, Method meth) {
        return new VirgilClass.GetMethod(ct, meth, false);
    }

    private Operator opVirtualDispatch(VirgilClass.IType ct, Method meth) {
        return new VirgilClass.GetMethod(ct, meth, true);
    }

    private VirgilArray.Alloc opArrayAlloc(VirgilArray.IType at, int dims) {
        Heap heap = builderEnv.tirBuilder.program.heap;
        return new VirgilArray.Alloc(at, dims, heap);
    }

    private TIRExpr newNoOp(Token t) {
        TIRExpr n = new TIRNop();
        if ( t != null ) {
            setSourcePoint(t, new TIRNop());
        }
        return n;
    }

    private static TIRExpr buildConstFor(VSTNode e, Value v) {
        TIRExpr value = $VAL(v);
        value.setSourcePoint(e.getToken().getSourcePoint());
        return value;
    }
}
