/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc.vst.tree;

import vpc.core.Value;
import vpc.tir.expr.Operator;
import vpc.tir.expr.Precedence;
import vpc.types.Capability;
import vpc.vst.parser.Token;
import vpc.vst.visitor.VSTAccumulator;
import vpc.vst.visitor.VSTVisitor;

/**
 * "expr1 <op> expr2"
 * <p/>
 * The <code>VSTBinOp</code> class represents a binary infix operation that
 * appears in the source code of a Virgil program. For example, addition,
 * subtraction, multiplication, logical operators, etc, are BinOps. The
 * identity of each operation is stored internally as a character--there
 * aren't any fancy inner classes.
 *
 * @author Ben L. Titzer
 */
public class VSTBinOp extends VSTExpr {

    public VSTExpr left, right;
    public Capability.BinOp binop;

    public VSTBinOp(Token tok, VSTExpr l, VSTExpr r) {
        super(tok);
        left = l;
        right = r;
    }

    public boolean matches(String s) {
        return token.image.equals(s);
    }

    public boolean isComputable() {
        return left.isComputable() && right.isComputable();
    }

    public Value computeConstValue() throws Operator.Exception {
        return binop.binop.apply2(left.computeConstValue(), right.computeConstValue());
    }

    public String getOp() {
        return token.image;
    }

    public int getPrecedence() {
        return Precedence.get(binop.binop.getClass());
    }

    public void accept(VSTVisitor v) {
        v.visit(this);
    }

    public <V> V accept(VSTAccumulator<V> a) {
        return a.visit(this);
    }

    public void setBinOp(Capability.BinOp t) {
        binop = t;
    }
}
