/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc.vst.tree;

import cck.util.Util;
import vpc.core.Value;
import vpc.tir.expr.Operator;
import vpc.types.Type;
import vpc.vst.parser.Token;
import vpc.vst.visitor.VSTAccumulator;

/**
 * The <code>VSTExpr</code> interface represents an expression in the syntax
 * tree representation of the program. It contains methods that accept visitors,
 * test various properties, and allow types to be applied to expressions.
 */
public abstract class VSTExpr implements VSTNode {

    public static final int PREC_ASSIGN = 1;
    public static final int PREC_TERNARY = 2;
    public static final int PREC_COND_OR = 3;
    public static final int PREC_COND_AND = 4;
    public static final int PREC_TYPECAST = 5;
    public static final int PREC_OR = 5;
    public static final int PREC_XOR = 5;
    public static final int PREC_AND = 6;
    public static final int PREC_EQ = 7;
    public static final int PREC_TYPEQUERY = 8;
    public static final int PREC_REL = 9;
    public static final int PREC_SHIFT = 10;
    public static final int PREC_ADD = 11;
    public static final int PREC_MUL = 12;
    public static final int PREC_POST = 13;
    public static final int PREC_UNARY = 14;
    public static final int PREC_PRE = 15;
    public static final int PREC_FIELD = 16;
    public static final int PREC_ELEM = 16;
    public static final int PREC_METH = 16;

    public static final int PREC_NEW = 17;
    public static final int PREC_VAR = 17;
    public static final int PREC_LIT = 17;
    public static final int PREC_INIT = 17;

    public Token token;
    public Type type;
    public Type promotionType;

    public abstract int getPrecedence();

    public abstract <V> V accept(VSTAccumulator<V> a);

    protected VSTExpr(Token tok) {
        token = tok;
    }

    public void setType(Type t) {
        type = t;
    }

    public Type getType() {
        if (type == null)
            throw Util.failure("node without type: " + getClass() + " @ " + token.getSourcePoint());
        return type;
    }

    public void setPromotionType(Type t) {
        promotionType = t;
    }

    public Type getPromotionType() {
        return promotionType;
    }

    public boolean isLvalue() {
        return false;
    }

    public boolean isStmt() {
        return false;
    }

    public boolean isComputable() {
        return false;
    }

    public Value computeConstValue() throws Operator.Exception {
        throw Util.failure("not computable");
    }

    public Token getToken() {
        return token;
    }
}
