/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc.vst.tree;

import cck.text.Printer;
import vpc.core.concept.Function;
import vpc.vst.visitor.VSTDepthFirstVisitor;
import java.util.List;

/**
 * The <class>VSTPrinter</code> is a utility that allows an abstract syntax tree
 * to be printed out. It also serves as a demonstration of how to implement a
 * visitor for the VST of a Virgil program.
 *
 * @author Ben L. Titzer
 */
public class VSTPrinter extends VSTDepthFirstVisitor {

    protected final Printer p;

    public VSTPrinter(Printer p) {
        this.p = p;
    }

    public void visit(VSTClassDecl d) {
        p.print("class " + d.token + " ");
        if (d.parent != null) {
            p.print("extends ");
            visit(d.parent);
            p.print(" ");
        }
        p.startblock();
        super.visit(d);
        p.endblock();
    }

    public void visit(VSTArrayInitializer a) {
        p.print("{ ");
        visitExprs(a.list);
        p.print(" }");
    }

    public void visit(VSTComponentDecl d) {
        p.println("component " + d.token + " ");
        p.startblock();
        super.visit(d);
        p.endblock();
    }

    public void visit(VSTBlock b) {
        p.println("{");
        p.indent();
        visitStmts(b.stmts);
        p.unindent();
        p.nextln();
        p.println("}");
    }

    public void visit(VSTWhileStmt s) {
        p.print("while ( ");
        s.cond.accept(this);
        p.print(" ) ");
        s.body.accept(this);
        p.nextln();
    }

    public void visit(VSTSwitchStmt s) {
        p.print("switch ( ");
        s.value.accept(this);
        p.print(" ) ");
        p.startblock();
        visitSwitchCases(s.cases);
        visitIfNonNull(s.defcase);
        p.endblock();
    }

    public void visit(VSTSwitchCase c) {
        if (c.list != null) {
            p.print("case ( ");
            visitExprs(c.list);
            p.print(" ) ");
        } else {
            p.print("default ");
        }
        c.stmt.accept(this);
        p.nextln();
    }

    public void visit(VSTDoWhileStmt s) {
        p.print("do ");
        s.body.accept(this);
        p.print("while ( ");
        s.cond.accept(this);
        p.println(" );");
    }

    public void visitExprs(List<VSTExpr> l) {
        boolean prev = false;
        for (VSTExpr e : l) {
            if (prev) p.print(", ");
            e.accept(this);
            prev = true;
        }
    }

    public void visitParamDecl(List<VSTParamDecl> d) {
        boolean prev = false;
        for (VSTParamDecl pd : d) {
            if (prev) p.print(", ");
            pd.accept(this);
            prev = true;
        }
    }


    public void visit(VSTForStmt s) {
        p.print("for ( ");
        visitExprs(s.init);
        p.print("; ");
        if (s.cond != null) s.cond.accept(this);
        p.print("; ");
        visitExprs(s.update);
        if (!s.update.isEmpty()) p.print(" "); // slight space adjustment
        p.print(") ");
        s.body.accept(this);
        p.nextln();
    }

    public void visit(VSTIfStmt s) {
        p.print("if ( ");
        s.cond.accept(this);
        p.print(" ) ");
        s.trueStmt.accept(this);
        if (s.falseStmt != null && !(s.falseStmt instanceof VSTEmptyStmt)) {
            p.print("else ");
            s.falseStmt.accept(this);
        }
        p.nextln();
    }

    public void visit(VSTBreakStmt s) {
        p.println("break;");
    }

    public void visit(VSTContinueStmt s) {
        p.println("continue;");
    }

    public void visit(VSTReturnStmt s) {
        p.print("return");
        if (s.expr != null) {
            p.print(" ");
            s.expr.accept(this);
        }
        p.println(";");
    }

    public void visit(VSTEmptyStmt s) {
        p.println(";");
    }

    public void visit(VSTExprStmt s) {
        s.expr.accept(this);
        p.println(";");
    }

    public void visit(VSTAssign a) {
        embed(a, a.target);
        p.print(" = ");
        embed(a, a.value);
    }

    public void visit(VSTCompoundAssign a) {
        embed(a, a.target);
        p.print(" " + a.token + " ");
        embed(a, a.value);
    }

    // helper for parentheses
    protected void embed(VSTExpr out, VSTExpr in) {
        if (in.getPrecedence() < out.getPrecedence()) p.print("(");
        in.accept(this);
        if (in.getPrecedence() < out.getPrecedence()) p.print(")");
    }

    public void visit(VSTTernaryExpr e) {
        embed(e, e.cond);
        p.print(" ? ");
        embed(e, e.trueExpr);
        p.print(" : ");
        embed(e, e.falseExpr);
    }

    public void visit(VSTBinOp b) {
        embed(b, b.left);
        p.print(" " + b.token + " ");
        embed(b, b.right);
    }

    public void visit(VSTComparison c) {
        embed(c, c.left);
        p.print(" " + c.token + " ");
        embed(c, c.right);
    }

    public void visit(VSTUnaryOp u) {
        p.print(u.token.toString());
        embed(u, u.expr);
    }

    public void visit(VSTMemberExpr e) {
        embed(e, e.expr);
        p.print("." + e.token);
    }

    public void visit(VSTIndexExpr e) {
        embed(e, e.array);
        p.print("[");
        e.index.accept(this);
        p.print("]");
    }

    public void visit(VSTAppExpr c) {
        embed(c, c.func);
        p.print("(");
        if (c.params != null) visitExprs(c.params);
        p.print(")");
    }

    public void visit(VSTNewObjectExpr n) {
        p.print("new ");
        visit(n.type);
        p.print("(");
        if (n.params != null) visitExprs(n.params);
        p.print(")");
    }

    public void visit(VSTNewArrayExpr n) {
        p.print("new ");
        visit(n.type);
        for (VSTExpr e : n.list) {
            p.print("[");
            e.accept(this);
            p.print("]");
        }
    }

    public void visit(VSTVarUse u) {
        p.print(u.token.toString());
    }

    public void visit(VSTIntLiteral l) {
        p.print(l.token.toString());
    }

    public void visit(VSTBooleanLiteral l) {
        p.print(l.token.toString());
    }

    public void visit(VSTThisLiteral l) {
        p.print(l.token.toString());
    }

    public void visit(VSTNullLiteral l) {
        p.print(l.token.toString());
    }

    public void visit(VSTCharLiteral l) {
        p.print(l.token.toString());
    }

    public void visit(VSTStringLiteral l) {
        p.print(l.token.toString());
    }

    public void visit(VSTTypeRef r) {
        p.print(r.typeName.toString());
    }

    public void visit(VSTMethodDecl m) {
        p.println("");
        visitModifiers(m.modifiers);
        p.print("method " + m.getToken() + "(");
        visitParamDecls(m.params);
        p.print(")" + Function.asReturnType(m.returnType.getTypeName()));
        if (m.block != null) {
            p.print(" ");
            m.block.accept(this);
        } else p.println(";");

    }

    public void visit(VSTConstructorDecl c) {
        p.print("constructor(");
        visitParamDecls(c.params);
        p.print(")");
        if (c.supclause != null) {
            p.print(": super(");
            visitExprs(c.supclause.params);
            p.print(")");
        }
        if (c.block != null) {
            p.print(" ");
            c.block.accept(this);
        } else p.println(";");
    }

    public void visit(VSTFieldDecl f) {
        visitModifiers(f.modifiers);
        p.print("field " + f.getToken() + ": ");
        visit(f.type);
        if (f.init != null) {
            p.print(" = ");
            f.init.accept(this);
        }
        p.println(";");
    }

    public void visit(VSTLocalVarDecl d) {
        p.print("local " + d.token);

        if (d.type != null) {
            // locals with explicit initialization have their types inferred.
            p.print(": ");
            visit(d.type);
        }

        if (d.init != null) {
            p.print(" = ");
            d.init.accept(this);
        }
        p.println(";");
    }

    public void visit(VSTParamDecl d) {
        p.print(d.getName() + ": ");
        visit(d.getTypeRef());
    }

    private void visitModifiers(List<VSTModifier> l) {
        for (VSTModifier m : l) {
            p.print(m.token + " ");
        }
    }


}
