/**
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the University of California, Los Angeles nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package vpc.vst.visitor;

import vpc.vst.tree.*;
import java.util.List;


/**
 * The <code>VSTDepthFirstVisitor</code> class is a basic implementation of the
 * <code>VSTVisitor</code> interface that provides enough base functionality
 * to be useful. Visitors that are only interested in VSTNodes of a particular
 * type can then subclass this class and override the methods corresponding
 * to the types they are interested in.
 *
 * @author Ben L. Titzer
 */
public class VSTDepthFirstVisitor implements VSTVisitor {

    public void visit(VSTArrayInitializer a) {
        visitExprs(a.list);
    }

    public void visit(VSTAssign a) {
        a.value.accept(this);
        a.target.accept(this);
    }

    public void visit(VSTBinOp b) {
        b.left.accept(this);
        b.right.accept(this);
    }

    public void visit(VSTComparison c) {
        c.left.accept(this);
        c.right.accept(this);
    }

    public void visit(VSTBlock b) {
        visitStmts(b.stmts);
    }

    public void visit(VSTBooleanLiteral b) {
    }

    public void visit(VSTBreakStmt b) {
    }

    public void visit(VSTCharLiteral c) {
    }

    public void visit(VSTClassDecl d) {
        visitFieldDecls(d.fields);
        visitMethodDecls(d.methods);
        visitIfNonNull(d.declaredConstr);
    }

    public void visit(VSTComponentDecl d) {
        visitFieldDecls(d.fields);
        visitMethodDecls(d.methods);
        visitIfNonNull(d.declaredConstr);
    }

    public void visit(VSTCompoundAssign c) {
        c.value.accept(this);
        c.target.accept(this);
    }

    public void visit(VSTContinueStmt c) {
    }

    public void visit(VSTDoWhileStmt s) {
        s.body.accept(this);
        s.cond.accept(this);
    }

    public void visit(VSTIndexExpr a) {
        a.array.accept(this);
        a.index.accept(this);
    }

    public void visit(VSTEmptyStmt e) {
    }

    public void visit(VSTExprStmt s) {
        s.expr.accept(this);
    }

    public void visit(VSTFieldDecl d) {
        visitModifiers(d.modifiers);
        visitIfNonNull(d.init);
    }

    public void visit(VSTMemberExpr a) {
        a.expr.accept(this);
    }

    public void visit(VSTForStmt s) {
        if (s.init != null) visitExprs(s.init);
        visitIfNonNull(s.cond);
        if (s.update != null) visitExprs(s.update);
        s.body.accept(this);
    }

    public void visit(VSTIfStmt s) {
        s.cond.accept(this);
        visitIfNonNull(s.trueStmt);
        visitIfNonNull(s.falseStmt);
    }

    public void visit(VSTTypeQueryExpr e) {
        e.expr.accept(this);
    }

    public void visit(VSTTypeCastExpr e) {
        e.expr.accept(this);
    }

    public void visit(VSTIntLiteral i) {
    }

    public void visit(VSTRawLiteral i) {
    }

    public void visitFieldDecls(List<VSTFieldDecl> l) {
        for (VSTFieldDecl e : l) e.accept(this);
    }

    public void visitMethodDecls(List<VSTMethodDecl> l) {
        for (VSTMethodDecl e : l) e.accept(this);
    }

    public void visitExprs(List<VSTExpr> l) {
        for (VSTExpr e : l) e.accept(this);
    }

    public void visitParamDecls(List<VSTParamDecl> l) {
        for (VSTParamDecl m : l) m.accept(this);
    }

    private void visitModifiers(List<VSTModifier> l) {
        for (VSTModifier m : l) m.accept(this);
    }

    public void visitStmts(List<VSTStmt> l) {
        for (VSTStmt m : l) m.accept(this);
    }

    private void visitTypeDecls(List<VSTTypeDecl> l) {
        for (VSTTypeDecl m : l) m.accept(this);
    }

    public void visit(VSTMethodDecl d) {
        visitIfNonNull(d.block);
    }

    public void visit(VSTConstructorDecl d) {
        visitIfNonNull(d.supclause);
        visitIfNonNull(d.block);
    }

    public void visit(VSTParamDecl p) {
    }

    public void visit(VSTModule m) {
        visitTypeDecls(m.decls);
    }

    public void visit(VSTNewArrayExpr e) {
        visitExprs(e.list);
    }

    public void visit(VSTNewObjectExpr e) {
        if (e.params != null) visitExprs(e.params);
    }

    public void visit(VSTNullLiteral l) {
    }

    public void visit(VSTReturnStmt r) {
        visitIfNonNull(r.expr);
    }

    public void visit(VSTAppExpr a) {
        a.func.accept(this);
        visitExprs(a.params);
    }

    public void visit(VSTStringLiteral l) {
    }

    public void visit(VSTSwitchCase c) {
        if (c.list != null) visitExprs(c.list);
        c.stmt.accept(this);
    }

    public void visit(VSTSwitchStmt s) {
        s.value.accept(this);
        visitSwitchCases(s.cases);
    }

    public void visitSwitchCases(List<VSTSwitchCase> l) {
        for (VSTSwitchCase c : l) c.accept(this);
    }

    public void visit(VSTTernaryExpr e) {
        e.cond.accept(this);
        e.trueExpr.accept(this);
        e.falseExpr.accept(this);
    }

    public void visit(VSTThisLiteral l) {
    }

    public void visit(VSTSuperClause c) {
        visitExprs(c.params);
    }

    public void visit(VSTTypeRef r) {
    }

    public void visit(VSTUnaryOp u) {
        u.expr.accept(this);
    }

    public void visit(VSTPreOp p) {
        p.expr.accept(this);
    }

    public void visit(VSTPostOp p) {
        p.expr.accept(this);
    }

    public void visit(VSTLocalVarDecl d) {
        visitIfNonNull(d.init);
    }

    public void visit(VSTVarUse u) {
    }

    public void visit(VSTWhileStmt s) {
        s.cond.accept(this);
        s.body.accept(this);
    }

    protected void visitIfNonNull(VSTNode n) {
        if (n != null) n.accept(this);
    }
}
